// $Id: Cursor.cpp,v 1.4 2009/02/05 11:11:49 avalassi Exp $

// Include files
#include <iostream>
#include "CoralServerBase/InternalErrorException.h"
#include "CoralServerBase/NotImplemented.h"

// Local include files
#include "ConnectionProperties.h"
#include "Cursor.h"
#include "SessionProperties.h"
#include "logger.h"

// Namespace
using namespace coral::CoralAccess;

//-----------------------------------------------------------------------------

Cursor::Cursor( const SessionProperties& sessionProperties,
                const coral::Token cursorID,
                const coral::QueryDefinition& queryDef,
                coral::AttributeList& rowBuffer,
                size_t maxRows )
  : m_sessionProperties( sessionProperties )
  , m_cursorID( cursorID )
  , m_queryDef( queryDef )
  , m_rowBuffer( rowBuffer )
  , m_maxRows( maxRows )
  , m_rows() 
  , m_nRows( 0 )
{
  logger << "Create Cursor" << endlog;

  // Fetch ALL rows in one go (maxRows == 0) in the ctor.
  if ( maxRows == 0 )
  {
    if ( cursorID > 0 )
      throw InternalErrorException( "PANIC! CursorID should be 0 if row cache size is 0!", 
                                    "Cursor::Cursor", 
                                    "coral::CoralAccess" );
    m_rows = facade().fetchAllRows( m_sessionProperties.sessionID(), 
                                    m_queryDef,
                                    &m_rowBuffer );
  }

  // Fetch N rows at a time (maxRows > 0).
  // Fetch the first bulk in the ctor and the other bulks in next().
  else
  {
    if ( cursorID == 0 )
      throw InternalErrorException( "PANIC! CursorID should be > 0 if row cache size is > 0!", 
                                    "Cursor::Cursor", 
                                    "coral::CoralAccess" );
    m_rows = facade().fetchRows( m_sessionProperties.sessionID(), 
                                 m_cursorID, 
                                 m_maxRows );  
  }

}

//-----------------------------------------------------------------------------

Cursor::~Cursor()
{
  logger << "Delete Cursor" << endlog;
  if ( m_cursorID > 0 ) facade().releaseCursor( m_cursorID );
}

//-----------------------------------------------------------------------------

bool Cursor::next()
{
  // --- There are more rows in the current bulk
  if ( m_rows->next() ) 
  {
    logger << "Get the next row from the local cache" << endlog;
    m_nRows++;
    if ( m_maxRows > 0 && m_nRows > m_maxRows )
      throw InternalErrorException( "PANIC! More than maxRows in the current bulk (next=true)",
                                    "Cursor::next", 
                                    "coral::CoralAccess" );
    return true;
  }

  // --- There are no more rows in the current bulk
  else
  {

    // All rows were fetched in one go.
    // There is nothing else to fetch.
    if ( m_maxRows == 0 )
    {
      logger << "End of loop over all rows" << endlog;
      return false;
    }

    // Rows are fetched in bulks.
    // The current bulk has < maxRows: there is nothing else to fetch.
    else if ( m_nRows < m_maxRows )
    {
      logger << "End of loop over the last bulk of rows" << endlog;
      return false;
    }

    // Rows are fetched in bulks.
    // The current bulk has maxRows: there are more bulks to fetch.
    else if ( m_nRows == m_maxRows )
    {
      logger << "End of loop over the current bulk of rows: fetch a new one" << endlog;
      m_nRows = 0;
      m_rows = facade().fetchRows( m_sessionProperties.sessionID(), 
                                   m_cursorID, 
                                   m_maxRows );  
      // Iterate over the new bulk.
      return this->next();
    }

    // Rows are fetched in bulks.
    // The current bulk has > maxRows: this is not possible (internal error)!
    else
    {
      throw InternalErrorException( "PANIC! More than maxRows in the current bulk (next=false)",
                                    "Cursor::next", 
                                    "coral::CoralAccess" );
    }
  }
}

//----------------------------------------------------------------------------- 
     
const coral::AttributeList& Cursor::currentRow() const
{
  return m_rows->currentRow();
}

//-----------------------------------------------------------------------------

void Cursor::close()
{
  throw NotImplemented("Cursor::close");
}

//-----------------------------------------------------------------------------

const coral::ICoralFacade& Cursor::facade() const
{
  return m_sessionProperties.connectionProperties().facade();
}

//-----------------------------------------------------------------------------
