// $Id: Schema.cpp,v 1.6 2009/02/05 10:20:15 avalassi Exp $

// Include files
#include <iostream>
#include "CoralServerBase/InternalErrorException.h"
#include "CoralServerBase/NotImplemented.h"
#include "RelationalAccess/SchemaException.h"

// Local include files
#include "ConnectionProperties.h"
#include "Query.h"
#include "Schema.h"
#include "Table.h"
#include "SessionProperties.h"
#include "logger.h"

// Namespace
using namespace coral::CoralAccess;

//-----------------------------------------------------------------------------

Schema::Schema( const SessionProperties& sessionProperties,
                const std::string& schemaName )
  : m_sessionProperties( sessionProperties )
  , m_schemaName( schemaName )
  , m_tables()
  , m_tableListFresh( false )
{
  logger << "Create Schema" << endlog;  
}

//-----------------------------------------------------------------------------

Schema::~Schema()
{
  logger << "Delete Schema" << endlog;  
  for ( std::map< std::string, Table* >::iterator 
          table = m_tables.begin(); table != m_tables.end(); table++ )
  {
    if ( table->second ) delete table->second;
    table->second = 0;
  }
}

//-----------------------------------------------------------------------------

std::string Schema::schemaName() const
{
  return m_schemaName;
}

//-----------------------------------------------------------------------------

std::set<std::string> Schema::listTables() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw TransactionNotActiveException( "coral::Access", "Schema::listTables" );
  return facade().listTables( m_sessionProperties.sessionID(), m_schemaName );
}

//-----------------------------------------------------------------------------

bool Schema::existsTable( const std::string& tableName ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw TransactionNotActiveException( "coral::Access", "Schema::existsTable" );
  i_readTableList();
  return m_tables.find( tableName ) != m_tables.end();
}

//-----------------------------------------------------------------------------

void Schema::dropTable( const std::string& /*tableName*/ )
{
  throw NotImplemented("Schema::dropTable");
}

//-----------------------------------------------------------------------------

void Schema::dropIfExistsTable( const std::string& /*tableName*/ )
{
  throw NotImplemented("Schema::dropIfExistsTable");
}

//-----------------------------------------------------------------------------

coral::ITable& Schema::createTable( const coral::ITableDescription& /*description*/ )
{
  throw NotImplemented("Schema::createTable");
}

//-----------------------------------------------------------------------------

coral::ITable& Schema::tableHandle( const std::string& tableName )
{
  if ( !existsTable( tableName ) )
    throw coral::TableNotExistingException( "coral::CoralAccess", tableName );
  if ( !m_tables[ tableName ] )
  {
    const coral::TableDescription desc =
      facade().tableDescription( m_sessionProperties.sessionID(), m_schemaName, tableName );
    m_tables[ tableName ] = new Table( m_sessionProperties, m_schemaName, desc );
  }  
  return *( m_tables[ tableName ] );
}

//-----------------------------------------------------------------------------

void Schema::truncateTable( const std::string& /*tableName*/ )
{
  throw NotImplemented("Schema::truncateTable");
}

//-----------------------------------------------------------------------------

void Schema::callProcedure( const std::string& /*procedureName*/,
                            const coral::AttributeList& /*inputArguments*/ )
{
  throw NotImplemented("Schema::callProcedure");
}

//-----------------------------------------------------------------------------

coral::IQuery* Schema::newQuery() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw TransactionNotActiveException( "coral::Access", "Schema::newQuery" );
  return new Query( m_sessionProperties, m_schemaName );
}

//-----------------------------------------------------------------------------

coral::IViewFactory* Schema::viewFactory()
{
  throw NotImplemented("Schema::viewFactory");
}

//-----------------------------------------------------------------------------

bool Schema::existsView( const std::string& /*viewName*/ ) const
{
  throw NotImplemented("Schema::existsView");
}

//-----------------------------------------------------------------------------

void Schema::dropView( const std::string& /*viewName*/ )
{
  throw NotImplemented("Schema::dropView");
}

//-----------------------------------------------------------------------------

void Schema::dropIfExistsView( const std::string& /*viewName*/ )
{
  throw NotImplemented("Schema::dropIfExistsView");
}

//-----------------------------------------------------------------------------

std::set<std::string> Schema::listViews() const
{
  throw NotImplemented("Schema::listViews");
}

//-----------------------------------------------------------------------------

coral::IView& Schema::viewHandle( const std::string& /*viewName*/ )
{
  throw NotImplemented("Schema::viewHandle");
}

//-----------------------------------------------------------------------------

const coral::ICoralFacade& Schema::facade() const
{
  return m_sessionProperties.connectionProperties().facade();
}

//-----------------------------------------------------------------------------

void Schema::i_readTableList() const
{
  if ( m_tableListFresh )
  { 
    logger << "Table list has already been read from the data dictionary" << endlog;  
    return;
  }  
  else
  {
    logger << "Read table list from the data dictionary" << endlog;  
    const std::set<std::string> tableList =
      facade().listTables( m_sessionProperties.sessionID(), m_schemaName );
    for ( std::set<std::string>::const_iterator 
            table = tableList.begin(); table != tableList.end(); table++ )
    {
      const std::string& name = *table;
      m_tables[ name ] = 0;
    }
    /*
    const TableDescriptionMap tableMap = 
      facade().tableDescriptionMap( m_sessionProperties.sessionID(), m_schemaName );
    for ( TableDescriptionMap::const_iterator 
            table = tableMap.begin(); table != tableMap.end(); table++ )
    {
      const std::string& name = table->first;
      const TableDescription& desc = table->second;
      if ( desc.name() != name )
        throw InternalErrorException( "PANIC! Table name mismatch for '"+name+"'", 
                                      "Schema::Schema", 
                                      "coral::CoralAccess" );     
      m_tables[ name ] = new Table( m_sessionProperties, m_schemaName, desc );
    }
    */
    m_tableListFresh = true;
  }
}

//-----------------------------------------------------------------------------
