// $Id: CoralServerFacade.h,v 1.5 2009/02/05 11:11:49 avalassi Exp $
#ifndef CORALSERVER_CORALSERVERFACADE_H
#define CORALSERVER_CORALSERVERFACADE_H 1

// Include files
#include "CoralKernel/IHandle.h"
#include "CoralServerBase/ICoralFacade.h"
#include "RelationalAccess/IConnectionService.h"

namespace coral 
{

  namespace CoralServer 
  {

    // Forward declaration
    class IObjectStoreMgr;

    /** @class CoralServerFacade
     *
     *  @author Andrea Valassi
     *  @date   2007-11-26 (python)
     *  @date   2008-01-22 (C++)
     */

    class CoralServerFacade : virtual public ICoralFacade
    { 
 
    public:
      
      /// Standard constructor. The facade creates its own ConnectionService.
      CoralServerFacade();

      /// Destructor.
      virtual ~CoralServerFacade();

      /// Get the sessionID for a new connection.
      /// The facade gets fromProxy=true if the request comes via a proxy.
      /// The client gets back fromProxy=true if the reply comes via a proxy.
      Token connect( const std::string& dbUrl,
                     const coral::AccessMode mode,
                     bool& fromProxy ) const;

      /// Release a session
      void releaseSession( Token sessionID ) const;

      /// Get the properties of a remote session
      const std::vector<std::string> sessionProperties( Token sessionID ) const;

      /// Starts a transaction.
      void startTransaction( Token sessionID, 
                             bool readOnly = false ) const;

      /// Commit a transaction.
      void commitTransaction( Token sessionID ) const;

      /// Get the cursorID for a new query.
      /// If given, the row buffer must be used by the iterator returning
      /// the results (and its specification must be shipped over the network
      /// to be used by the server-side IQuery::defineOutput call). 
      Token executeQuery( Token sessionID,
                          const QueryDefinition& qd,
                          AttributeList* pRowBuffer ) const;

      /// Release a cursor
      void releaseCursor( Token cursorID ) const;

      /// Fetch a new bulk of rows from the cursor.
      /// The number of rows in all bulks but the last must be exactly maxRows;
      /// the number of rows in the last bulk must be between 0 and maxRows-1.
      /// Throws an exception if the last bulk has already been fetched.
      /// Throws an exception if the loop on the previous iterator is not over:
      /// callers who fetched a bulk must loop over it before fetching a new one.
      /// NB Here sessionID is needed ONLY to allow cursor sharing (caching)
      IRowIteratorPtr fetchRows( Token sessionID,
                                 Token cursorID,
                                 size_t maxSize ) const;

      /// Fetch all rows from the cursor.
      /// If given, the row buffer must be used by the iterator returning
      /// the results (and its specification must be shipped over the network
      /// to be used by the server-side IQuery::defineOutput call). 
      IRowIteratorPtr fetchAllRows( Token sessionID,
                                    const QueryDefinition& qd,
                                    AttributeList* pRowBuffer ) const;

      /// Delete rows from a table.
      /// Eventually pass an AttributeList.
      int deleteTableRows( Token sessionID,
                           const std::string& schemaName,
                           const std::string& tableName,
                           const std::string& whereClause,
                           const std::string& whereData ) const;

      /// Prepare a row buffer from the table metadata.
      /// Eventually retrieve an AttributeList.
      /// NB Eventually this will be replaced by other metadata methods.
      const std::string formatRowBufferAsString( Token sessionID,
                                                 const std::string& schemaName,
                                                 const std::string& tableName ) const;

      /// Insert one row.
      /// Eventually pass an AttributeList reference.
      void insertRowAsString( Token sessionID,
                              const std::string& schemaName,
                              const std::string& tableName,
                              const std::string& rowBufferAS ) const;

      /// Get the bulkOperationID for a new bulk insertion.
      /// Eventually pass an AttributeList reference.
      /// NB Is the rowBuffer here really needed or I got it wrong?
      Token bulkInsertAsString( Token sessionID,
                                const std::string& schemaName,
                                const std::string& tableName,
                                const std::string& rowBufferAS,
                                int rowCacheSizeDb ) const;
      
      /// Release a bulk operation
      void releaseBulkOp( Token bulkOpID ) const;

      /// Insert several rows in bulk.
      void processRows( Token bulkOpID,
                        const std::vector<AttributeList>& rows ) const;
      
      /// Flush a bulk operation.
      void flush( Token bulkOpID ) const;

      /// Get the list of tables in a schema.
      const std::set<std::string> listTables( Token sessionID,
                                              const std::string& schemaName ) const;
      
      /// Get the description of a table in a schema.
      const TableDescription tableDescription( Token sessionID,
                                               const std::string& schemaName,
                                               const std::string& tableName ) const;

      /// Get the description of all the tables in a schema.
      const TableDescriptionMap tableDescriptionMap( Token sessionID,
                                                     const std::string& schemaName ) const;

    private:

      /// The CORAL connection service.
      /// [NB Should be mutable if ICoralFacade::connect() is const]
      //ConnectionService m_connSvc;
      IHandle<IConnectionService> m_pConnSvc;

      /// The object store manager (owned by this instance).
      IObjectStoreMgr* m_storeMgr;

    };

  }

}
#endif // CORALSERVER_CORALSERVERFACADE_H
