// $Id: CoralServerFacade.cpp,v 1.18 2009/02/09 16:23:04 avalassi Exp $

// Include files
#include <fstream>
#include <memory>
#include <iostream>
#include <sstream>
#include "CoralKernel/Context.h"
#include "CoralServer/CoralServerFacade.h"
#include "CoralServerBase/InternalErrorException.h"
#include "CoralServerBase/NotImplemented.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/IConnectionServiceConfiguration.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ISessionProperties.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITransaction.h"

// Local include files
#include "ObjectStoreMgr.h"
#include "RowVectorIterator.h"
#include "QueryAndCursor.h"
#include "QueryMgr.h"
#include "SessionAndFlag.h"

// Logger
#define LOGGER_NAME "CoralServer::CoralServerFacade"
#include "logger.h"

// Namespace
using namespace coral::CoralServer;

//-----------------------------------------------------------------------------

CoralServerFacade::CoralServerFacade()
  : m_pConnSvc()
  , m_storeMgr( new ObjectStoreMgr() )
{
  logger << Debug << "Create CoralServerFacade" << endlog;
  // Use the second connection service if it exists (local facade mode)
  // Use the default connection service otherwise 
  // [do not attempt to load it - it must already exist if you are here!]
  m_pConnSvc = Context::instance().query<IConnectionService>( "CORAL/Services/ConnectionService2" );
  if ( !m_pConnSvc.isValid() )
    m_pConnSvc = Context::instance().query<IConnectionService>( "CORAL/Services/ConnectionService" );
  if ( !m_pConnSvc.isValid() )
    throw Exception( "Could not retrieve the connection service", 
                     "CoralServerFacade::CoralServerFacade", 
                     "coral::CoralServer" );    
}

//-----------------------------------------------------------------------------

CoralServerFacade::~CoralServerFacade()
{
  logger << Debug << "Delete CoralServerFacade..." << endlog;
  m_pConnSvc = IHandle<IConnectionService>();
  if ( m_storeMgr ) delete m_storeMgr;
  m_storeMgr = 0;
  logger << Debug << "Deleted CoralServerFacade... DONE" << endlog;
}

//-----------------------------------------------------------------------------

coral::Token CoralServerFacade::connect( const std::string& dbUrl,
                                         const coral::AccessMode mode,
                                         bool& fromProxy ) const
{
  logger << Debug << "CoralServerFacade::connect( " 
         << dbUrl << ", " << mode << ", " << fromProxy << " )" << endlog;
  //fromProxy = true; // HACK to test proxy behaviour
  IConnectionService& connSvc = const_cast<IConnectionService&>( *m_pConnSvc );
  ISessionProxy* session = connSvc.connect( dbUrl, mode );
  logger << Debug << "Got a new session" << endlog;
  if ( fromProxy ) session->transaction().start( true ); // RO PROXY
  SessionAndFlag* saf = new SessionAndFlag( session, fromProxy );
  return m_storeMgr->registerObject( saf, 0 );
}

//-----------------------------------------------------------------------------

void CoralServerFacade::releaseSession( coral::Token sessionID ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  if ( saf->fromProxy() ) saf->session()->transaction().commit(); // RO PROXY
  m_storeMgr->releaseObject( sessionID );
}

//-----------------------------------------------------------------------------

const std::vector<std::string> 
CoralServerFacade::sessionProperties( coral::Token sessionID ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  std::vector<std::string> properties;
  properties.push_back( session->properties().flavorName() );
  properties.push_back( session->properties().serverVersion() );
  return properties;
}

//-----------------------------------------------------------------------------

void CoralServerFacade::startTransaction( coral::Token sessionID, 
                                          bool readOnly ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  if ( saf->fromProxy() ) 
    throw InternalErrorException( "PANIC! Proxy did not drop startTransaction message", 
                                  "CoralServerFacade::startTransaction", 
                                  "coral::CoralServer" );
  ISessionProxy* session = saf->session();
  session->transaction().start( readOnly );
}

//-----------------------------------------------------------------------------

void CoralServerFacade::commitTransaction( coral::Token sessionID ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  if ( saf->fromProxy() ) 
    throw InternalErrorException( "PANIC! Proxy did not drop commitTransaction message", 
                                  "CoralServerFacade::commitTransaction", 
                                  "coral::CoralServer" );
  ISessionProxy* session = saf->session();
  session->transaction().commit();
}

//-----------------------------------------------------------------------------

coral::Token 
CoralServerFacade::executeQuery( coral::Token sessionID,
                                 const coral::QueryDefinition& qd,
                                 coral::AttributeList* pRowBuffer ) const
{
  //logger << Always << "CoralServerFacade::executeQuery show row buffer..." << endlog;
  //if ( !pRowBuffer ) logger << Always << "NULL" << endlog;
  //else logger << Always << *pRowBuffer << endlog;
  //logger << Always << "CoralServerFacade::executeQuery show row buffer... DONE" << endlog;
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  IQuery* query = QueryMgr::newQuery( pRowBuffer, *session, qd );
  ICursor& cursor = query->execute();
  QueryAndCursor* qac = new QueryAndCursor( query, cursor, pRowBuffer );
  return m_storeMgr->registerObject( qac, sessionID );
}

//-----------------------------------------------------------------------------

void CoralServerFacade::releaseCursor( coral::Token cursorID ) const
{
  m_storeMgr->releaseObject( cursorID );
}

//-----------------------------------------------------------------------------

coral::IRowIteratorPtr
CoralServerFacade::fetchRows( coral::Token sessionID,
                              coral::Token cursorID,
                              size_t maxSize ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  QueryAndCursor* qac = m_storeMgr->findObject<QueryAndCursor>( cursorID );
  ICursor& cursor = qac->cursor();
  if ( maxSize == 0 ) throw Exception( "maxSize must be > 0", "CoralServerFacade::fetchRows", "CoralServer" );
  std::vector<AttributeList> rows;
  while ( rows.size() < maxSize && cursor.next() )
    rows.push_back( cursor.currentRow() );
  logger << Debug << "New batch of rows from database cursor: #rows=" << rows.size() << endlog;
  //logger << Always << "CoralServerFacade::fetchRows show row buffer..." << endlog;
  //if ( !qac->buffer() ) logger << Always << "NULL" << endlog;
  //else logger << Always << *qac->buffer() << endlog;
  //logger << Always << "CoralServerFacade::fetchRows show row buffer... DONE" << endlog;
  return IRowIteratorPtr( new RowVectorIterator( qac->buffer(), rows ) );
}

//-----------------------------------------------------------------------------

coral::IRowIteratorPtr
CoralServerFacade::fetchAllRows( coral::Token sessionID,
                                 const coral::QueryDefinition& qd,
                                 coral::AttributeList* pRowBuffer ) const
{
  //logger << Always << "CoralServerFacade::fetchAllRows show row buffer..." << endlog;
  //if ( !pRowBuffer ) logger << Always << "NULL" << endlog;
  //else logger << Always << *pRowBuffer << endlog;
  //logger << Always << "CoralServerFacade::fetchAllRows show row buffer... DONE" << endlog;
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  std::auto_ptr<IQuery> query( QueryMgr::newQuery( pRowBuffer, *session, qd ) );
  ICursor& cursor = query->execute();
  std::vector<AttributeList> rows;
  while ( cursor.next() ) rows.push_back( cursor.currentRow() );
  std::cout << "Fetched all rows from database cursor: #rows=" << rows.size() << std::endl;
  return IRowIteratorPtr( new RowVectorIterator( pRowBuffer, rows ) );
}

//-----------------------------------------------------------------------------

int CoralServerFacade::deleteTableRows( coral::Token sessionID,
                                        const std::string& schemaName,
                                        const std::string& tableName,
                                        const std::string& whereClause,
                                        const std::string& whereData ) const
{
  throw NotImplemented("CoralServerFacade::deleteTableRows");
}

//-----------------------------------------------------------------------------

const std::string 
CoralServerFacade::formatRowBufferAsString( coral::Token sessionID,
                                            const std::string& schemaName,
                                            const std::string& tableName ) const
{
  throw NotImplemented("CoralServerFacade::formatRowBufferAsString");
}

//-----------------------------------------------------------------------------

void CoralServerFacade::insertRowAsString( coral::Token sessionID,
                                           const std::string& schemaName,
                                           const std::string& tableName,
                                           const std::string& rowBufferAS ) const
{
  throw NotImplemented("CoralServerFacade::insertRowAsString");
}

//-----------------------------------------------------------------------------

coral::Token 
CoralServerFacade::bulkInsertAsString( coral::Token sessionID,
                                       const std::string& schemaName,
                                       const std::string& tableName,
                                       const std::string& rowBufferAS,
                                       int rowCacheSizeDb ) const
{
  throw NotImplemented("CoralServerFacade::bulkInsertAsString");
}

//-----------------------------------------------------------------------------

void CoralServerFacade::releaseBulkOp( coral::Token bulkOpID ) const
{
  m_storeMgr->releaseObject( bulkOpID );
}

//-----------------------------------------------------------------------------

void CoralServerFacade::processRows( coral::Token bulkOpID,
                                     const std::vector<coral::AttributeList>& rows ) const
{
  throw NotImplemented("CoralServerFacade::processRows");
}

//-----------------------------------------------------------------------------

void CoralServerFacade::flush( coral::Token bulkOpID ) const
{
  throw NotImplemented("CoralServerFacade::flush");
}

//-----------------------------------------------------------------------------

const std::set<std::string> 
CoralServerFacade::listTables( Token sessionID,
                               const std::string& schemaName ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  if ( schemaName == "" ) return session->nominalSchema().listTables();
  else return session->schema( schemaName ).listTables();
}

//-----------------------------------------------------------------------------

const coral::TableDescription
CoralServerFacade::tableDescription( Token sessionID,
                                     const std::string& schemaName,
                                     const std::string& tableName ) const
{
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  if ( schemaName == "" ) return session->nominalSchema().tableHandle( tableName ).description();
  else return session->schema( schemaName ).tableHandle( tableName ).description();
}

//-----------------------------------------------------------------------------

const coral::TableDescriptionMap
CoralServerFacade::tableDescriptionMap( Token sessionID,
                                        const std::string& schemaName ) const
{
  throw NotImplemented("CoralServerFacade::tableDescriptionMap (is this really needed?)");
  /*
  SessionAndFlag* saf = m_storeMgr->findObject<SessionAndFlag>( sessionID );
  ISessionProxy* session = saf->session();
  ISchema& schema = ( schemaName == "" ) ? session->nominalSchema() : session->schema( schemaName );  
  const std::set<std::string> tableNames = schema.listTables();
  TableDescriptionMap tableMap;
  for ( std::set<std::string>::const_iterator 
          table = tableNames.begin(); table != tableNames.end(); table++ )
    tableMap[ *table ] = schema.tableHandle( *table ).description();
  return tableMap;
  */
}

//-----------------------------------------------------------------------------
