// $Id: CoralServerFacadeService.h,v 1.4 2009/02/05 11:11:49 avalassi Exp $
#ifndef CORALSERVER_CORALSERVERFACADESERVICE_H
#define CORALSERVER_CORALSERVERFACADESERVICE_H 1

// Include files
#include "CoralKernel/Service.h"
#include "CoralServerBase/ICoralFacade.h"
#include "CoralServer/CoralServerFacade.h"

namespace coral 
{

  namespace CoralServer 
  {

    /** @class CoralServerFacadeService
     *
     *  Wrapper around a CoralServerFacade.
     *  This is a plugin that can be loaded at runtime.
     *
     *  @author Andrea Valassi
     *  @date   2009-01-30
     */

    class CoralServerFacadeService : public Service,
                                     virtual public ICoralFacade
    { 
 
    public:
      
      /// Service constructor from the service name.
      CoralServerFacadeService( const std::string& key );

      /// Destructor.
      virtual ~CoralServerFacadeService();

      /// Get the sessionID for a new connection.
      /// The facade gets fromProxy=true if the request comes via a proxy.
      /// The client gets back fromProxy=true if the reply comes via a proxy.
      Token connect( const std::string& dbUrl,
                     const AccessMode mode,
                     bool& fromProxy ) const
      {
        return m_facade.connect( dbUrl, mode, fromProxy );
      }
    
      /// Release a session
      void releaseSession( Token sessionID ) const
      {
        return m_facade.releaseSession( sessionID );
      }
    
      /// Get the properties of a remote session
      const std::vector<std::string> sessionProperties( Token sessionID ) const
      {
        return m_facade.sessionProperties( sessionID );
      }
    
      /// Starts a transaction.
      void startTransaction( Token sessionID, 
                             bool readOnly ) const
      {
        return m_facade.startTransaction( sessionID, readOnly );
      }
    
      /// Commit a transaction.
      void commitTransaction( Token sessionID ) const
      {
        return m_facade.commitTransaction( sessionID );
      }
    
      /// Get the cursorID for a new query.
      /// If given, the row buffer must be used by the iterator returning
      /// the results (and its specification must be shipped over the network
      /// to be used by the server-side IQuery::defineOutput call). 
      Token executeQuery( Token sessionID,
                          const QueryDefinition& qd,
                          AttributeList* pRowBuffer ) const
      {
        return m_facade.executeQuery( sessionID, qd, pRowBuffer );
      }
      
      /// Release a cursor
      void releaseCursor( Token cursorID ) const
      {
        return m_facade.releaseCursor( cursorID );
      }
    
      /// Fetch a new bulk of rows from the cursor.
      /// The number of rows in all bulks but the last must be exactly maxRows;
      /// the number of rows in the last bulk must be between 0 and maxRows-1.
      /// Throws an exception if the last bulk has already been fetched.
      /// Throws an exception if the loop on the previous iterator is not over:
      /// callers who fetched a bulk must loop over it before fetching a new one.
      /// NB Here sessionID is needed ONLY to allow cursor sharing (caching)
      IRowIteratorPtr fetchRows( Token sessionID,
                                 Token cursorID,
                                 size_t maxRows ) const
      {
        return m_facade.fetchRows( sessionID, cursorID, maxRows );
      }
    
      /// Fetch all rows from a new query.
      /// If given, the row buffer must be used by the iterator returning
      /// the results (and its specification must be shipped over the network
      /// to be used by the server-side IQuery::defineOutput call). 
      IRowIteratorPtr fetchAllRows( Token sessionID,
                                    const QueryDefinition& qd,
                                    AttributeList* pRowBuffer ) const
      {
        return m_facade.fetchAllRows( sessionID, qd, pRowBuffer );
      }
    
      /// Delete rows from a table.
      /// Eventually pass an AttributeList.
      int deleteTableRows( Token sessionID,
                           const std::string& schemaName,
                           const std::string& tableName,
                           const std::string& whereClause,
                           const std::string& whereData ) const
      {
        return m_facade.deleteTableRows( sessionID, schemaName, tableName, whereClause, whereData );
      }
 
      /// Prepare a row buffer from the table metadata.
      /// Eventually retrieve an AttributeList.
      /// NB Eventually this will be replaced by other metadata methods.
      const std::string formatRowBufferAsString( Token sessionID,
                                                 const std::string& schemaName,
                                                 const std::string& tableName ) const
      {
        return m_facade.formatRowBufferAsString( sessionID, schemaName, tableName );
      }
    
      /// Insert one row.
      /// Eventually pass an AttributeList reference.
      void insertRowAsString( Token sessionID,
                              const std::string& schemaName,
                              const std::string& tableName,
                              const std::string& rowBufferAS ) const
      {
        return m_facade.insertRowAsString( sessionID, schemaName, tableName, rowBufferAS );
      }
    
      /// Get the bulkOperationID for a new bulk insertion.
      /// Eventually pass an AttributeList reference.
      /// NB Is the rowBuffer here really needed or I got it wrong?
      Token bulkInsertAsString( Token sessionID,
                                const std::string& schemaName,
                                const std::string& tableName,
                                const std::string& rowBufferAS,
                                int rowCacheSizeDb ) const
      {
        return m_facade.bulkInsertAsString( sessionID, schemaName, tableName, rowBufferAS, rowCacheSizeDb );
      }
    
      /// Release a bulk operation
      void releaseBulkOp( Token bulkOpID ) const
      {
        return m_facade.releaseBulkOp( bulkOpID );
      }
    
      /// Insert several rows in bulk.
      void processRows( Token bulkOpID,
                        const std::vector<AttributeList>& rowsAS ) const
      {
        return m_facade.processRows( bulkOpID, rowsAS );
      }
    
      /// Flush a bulk operation.
      void flush( Token bulkOpID ) const
      {
        return m_facade.flush( bulkOpID );
      }

      /// Get the list of tables in a schema.
      const std::set<std::string> listTables( Token sessionID,
                                              const std::string& schemaName ) const
      {
        return m_facade.listTables( sessionID, schemaName );
      }
      
      /// Get the description of a table in a schema.
      const TableDescription tableDescription( Token sessionID,
                                               const std::string& schemaName,
                                               const std::string& tableName ) const
      {
        return m_facade.tableDescription( sessionID, schemaName, tableName );
      }

      /// Get the description of all the tables in a schema.
      const TableDescriptionMap tableDescriptionMap( Token sessionID,
                                                     const std::string& schemaName ) const
      {
        return m_facade.tableDescriptionMap( sessionID, schemaName );
      }

    private:

      /// The CORAL server facade.
      CoralServerFacade m_facade;

    };

  }

}
#endif // CORALSERVER_CORALSERVERFACADESERVICE_H
