// $Id: ObjectStoreMgr.cpp,v 1.3 2009/02/02 23:17:43 avalassi Exp $

// Include files
#include <sstream>
#include "RelationalAccess/IBulkOperation.h"

// Local include files
#include "QueryAndCursor.h"
#include "ObjectStoreMgr.h"
#include "SessionAndFlag.h"

// Logger
#define LOGGER_NAME "CoralServer::ObjectStoreMgr"
#include "logger.h"

// Namespace
using namespace coral::CoralServer;

//-----------------------------------------------------------------------------

ObjectStoreMgr::ObjectStoreMgr()
  : m_guid( 0 )
{
  logger << Debug << "Create ObjectStoreMgr" << endlog;
}

//-----------------------------------------------------------------------------

ObjectStoreMgr::~ObjectStoreMgr()
{
  logger << Debug << "Delete ObjectStoreMgr" << endlog;
  std::map< Token, ObjectAndType >::iterator objIt;
  for ( objIt = m_dict.begin(); objIt != m_dict.end(); objIt++ )
  {
    Token ID = objIt->first;
    logger << Warning << "Deleting unreleased object #" << ID << endlog;
    deleteObject( objIt->second, false );
  }
}

//-----------------------------------------------------------------------------

coral::Token ObjectStoreMgr::registerObject( ObjectAndType obj,
                                             coral::Token parentID )
{
  if ( parentID != 0 && m_dict.find( parentID ) == m_dict.end() ) 
    throw Exception( "Cannot register object: parentID is unknown",
                     "ObjectStoreMgr::registerObject", 
                     "coral::CoralServer" );
  // Register the object
  m_guid++;
  m_dict[m_guid] = obj;
  // Update the list of children of its parent
  if ( parentID != 0 )
  {
    m_parent[m_guid] = parentID;
    m_children[parentID].insert(m_guid);
  }  
  return m_guid;
}

//-----------------------------------------------------------------------------

void ObjectStoreMgr::releaseObject( coral::Token ID )
{
  i_releaseObject( ID, true );
}

//-----------------------------------------------------------------------------

void ObjectStoreMgr::i_releaseObject( coral::Token ID, bool erase )
{
  // Unregister the object from the list of children of its parent
  if ( m_parent.find( ID ) != m_parent.end() ) 
  {
    m_children[ m_parent[ ID ] ].erase( ID );
    m_parent.erase( ID );
  }
  // Release and delete all children
  else if ( m_children.find( ID ) != m_children.end() )
  {
    std::set<coral::Token> children = m_children[ ID ]; // copy
    std::set<coral::Token>::iterator child;
    for ( child = children.begin(); child != children.end(); child++ )
    {
      logger << Warning << "Deleting unreleased child #" << *child << " of parent #" << ID << endlog;
      i_releaseObject( *child, false ); // do not erase... the user may release it later!
    }
  }
  // Delete the object
  ObjectAndType obj = findObject( ID );
  if ( erase ) m_dict.erase( ID );
  else 
  {
    m_dict[ID].first = 0;
  }
  bool throwIfNull = !erase; // do not throw on user calls (erase=true)
  deleteObject( obj, throwIfNull );
}

//-----------------------------------------------------------------------------

IObjectStoreMgr::ObjectAndType ObjectStoreMgr::findObject( coral::Token ID )
{
  if ( m_dict.find( ID ) != m_dict.end() ) 
  {
    return m_dict[ID];
  }
  else
  {
    std::stringstream msg;
    msg << "No object found with ID=" << ID;
    throw Exception( msg.str(), "ObjectStoreMgr::findObject", "coral::CoralServer" );
  }
}

//-----------------------------------------------------------------------------

void ObjectStoreMgr::deleteObject( ObjectAndType obj, bool throwIfNull )
{
  void* pObj = obj.first;
  if ( !pObj )
  {
    if ( !throwIfNull ) return;
    throw Exception( "PANIC! Attempt to delete null object",
                     "ObjectStoreMgr::deleteObject",
                     "coral::CoralServer" );
  }  
  const std::type_info& type = *(obj.second);
  if ( type == typeid( IBulkOperation ) )
    deleteObject<IBulkOperation>( pObj );
  else if ( type == typeid( QueryAndCursor ) )
    deleteObject<QueryAndCursor>( pObj );
  else if ( type == typeid( SessionAndFlag ) )
    deleteObject<SessionAndFlag>( pObj );
  else
    throw Exception( "PANIC! Unsupported type!",
                     "ObjectStoreMgr::deleteObject",
                     "coral::CoralServer" );
}

//-----------------------------------------------------------------------------

bool ObjectStoreMgr::isSupportedType( const std::type_info& type )
{
  if ( type == typeid( IBulkOperation ) ||
       type == typeid( QueryAndCursor ) ||
       type == typeid( SessionAndFlag ) )
    return true;
  else
    return false;
}

//-----------------------------------------------------------------------------

