// $Id: ICoralFacade.h,v 1.7 2009/02/10 15:10:35 avalassi Exp $
#ifndef CORALSERVERBASE_ICORALFACADE_H
#define CORALSERVERBASE_ICORALFACADE_H 1

// Include files
#include <set>
#include <string>
#include <vector>
#include "RelationalAccess/AccessMode.h"
#include "RelationalAccess/TableDescription.h"
#include "CoralServerBase/QueryDefinition.h"
#include "CoralServerBase/IRowIterator.h"

namespace coral 
{

  /// ObjectID for sessionID, cursorID, bulkOpID.
  typedef unsigned int Token;
  
  /// Row iterator pointer.
  typedef std::auto_ptr<IRowIterator> IRowIteratorPtr;

  /// Table description map.
  typedef std::map< std::string, TableDescription > TableDescriptionMap;

  /** @class ICoralFacade
   *
   *  Interface to a facade to the CORAL relational functionalities.
   *  The methods of this interface represent all the messages exchanges
   *  required between the CORAL server and its clients over the network.
   *
   *  @author Andrea Valassi and Alexander Kalkhof
   *  @date   2007-12-04
   */
  
  class ICoralFacade
  { 
    
  public:
    
    /// Destructor.
    virtual ~ICoralFacade(){}
    
    /// Get the sessionID for a new connection.
    /// The facade gets fromProxy=true if the request comes via a proxy.
    /// The client gets back fromProxy=true if the reply comes via a proxy.
    virtual Token connect( const std::string& dbUrl,
                           const AccessMode mode,
                           bool& fromProxy ) const = 0;
    
    /// Release a session
    virtual void releaseSession( Token sessionID ) const = 0;
    
    /// Get the properties of a remote session
    virtual const std::vector<std::string> sessionProperties( Token sessionID ) const = 0;
    
    /// Starts a transaction.
    virtual void startTransaction( Token sessionID, 
                                   bool readOnly ) const = 0;
    
    /// Commit a transaction.
    virtual void commitTransaction( Token sessionID ) const = 0;
    
    /// Fetch rows from a new query (a few at a time).
    /// This method is associated to a server-side open cursor (PULL model):
    /// the cursor is released when the loop ends or the iterator is deleted.
    /// This method must prefetch data into the internal cache of the iterator:
    /// the cache size is given in #rows (MB) if cacheSizeInMB is false (true).
    /// If given, the row buffer must be used by the iterator returning
    /// the results (and its specification must be shipped over the network
    /// to be used by the server-side IQuery::defineOutput call). 
    virtual IRowIteratorPtr fetchRows( Token sessionID,
                                       const QueryDefinition& qd,
                                       AttributeList* pRowBuffer,
                                       size_t cacheSize,
                                       bool cacheSizeInMB ) const = 0;
    
    /// Fetch all rows from a new query.
    /// This method is NOT associated to server-side open cursors (PUSH model).
    /// If given, the row buffer must be used by the iterator returning
    /// the results (and its specification must be shipped over the network
    /// to be used by the server-side IQuery::defineOutput call).
    virtual IRowIteratorPtr fetchAllRows( Token sessionID,
                                          const QueryDefinition& qd,
                                          AttributeList* pRowBuffer ) const = 0;
    
    /// Delete rows from a table.
    /// Eventually pass an AttributeList.
    virtual int deleteTableRows( Token sessionID,
                                 const std::string& schemaName,
                                 const std::string& tableName,
                                 const std::string& whereClause,
                                 const std::string& whereData ) const = 0;
    
    /// Prepare a row buffer from the table metadata.
    /// Eventually retrieve an AttributeList.
    /// NB Eventually this will be replaced by other metadata methods.
    virtual const std::string formatRowBufferAsString( Token sessionID,
                                                       const std::string& schemaName,
                                                       const std::string& tableName ) const = 0;
    
    /// Insert one row.
    /// Eventually pass an AttributeList reference.
    virtual void insertRowAsString( Token sessionID,
                                    const std::string& schemaName,
                                    const std::string& tableName,
                                    const std::string& rowBufferAS ) const = 0;
    
    /// Get the bulkOperationID for a new bulk insertion.
    /// Eventually pass an AttributeList reference.
    /// NB Is the rowBuffer here really needed or I got it wrong?
    virtual Token bulkInsertAsString( Token sessionID,
                                      const std::string& schemaName,
                                      const std::string& tableName,
                                      const std::string& rowBufferAS,
                                      int rowCacheSizeDb ) const = 0;
    
    /// Release a bulk operation
    virtual void releaseBulkOp( Token bulkOpID ) const = 0;
    
    /// Insert several rows in bulk.
    virtual void processRows( Token bulkOpID,
                              const std::vector<AttributeList>& rowsAS ) const = 0;
    
    /// Flush a bulk operation.
    virtual void flush( Token bulkOpID ) const = 0;

    /// Get the list of tables in a schema.
    virtual const std::set<std::string> listTables( Token sessionID,
                                                    const std::string& schemaName ) const = 0;

    /// Get the description of a table in a schema.
    virtual const TableDescription tableDescription( Token sessionID,
                                                     const std::string& schemaName,
                                                     const std::string& tableName ) const = 0;

    /// Get the description of all the tables in a schema.
    virtual const TableDescriptionMap tableDescriptionMap( Token sessionID,
                                                           const std::string& schemaName ) const = 0;

  };
  
}
#endif // CORALSERVERBASE_ICORALFACADE_H
