// $Id: CoralServerUnitTest.h,v 1.1 2009/01/21 11:46:37 avalassi Exp $
#ifndef COMMON_CORALSERVERUNITTEST_H
#define COMMON_CORALSERVERUNITTEST_H 1

// Include files
#include <stdexcept>
#include <iostream>

// Local include files
#include "CppUnit_headers.h"

namespace coral 
{

  //--------------------------------------------------------------------------

  /** @class ProgressListener
   *
   *  Simple TestListener printing one line per test in the standard output.
   *  
   *  Based on  CppUnit::BriefTestProgressListener (copy and paste) 
   *  using std::cout instead of std::cerr.
   *
   *  @author Marco Clemencic
   *  @date   2006-11-13
   */

  class ProgressListener : public CppUnit::TestListener
  {

  public:

    /// Default constructor.
    ProgressListener(): m_lastTestFailed(false) {}
    
    /// Destructor.
    virtual ~ProgressListener() {}
    
    void startTest( CppUnit::Test *test )
    {
      std::cout << test->getName();
      std::cout.flush();      
      m_lastTestFailed = false;
    }

    void addFailure( const CppUnit::TestFailure &failure )
    {
      std::cout << " : " << (failure.isError() ? "error" : "assertion");
      m_lastTestFailed  = true;
    }
    
    void endTest( CppUnit::Test * /*test*/ )
    {
      if ( !m_lastTestFailed )
        std::cout  <<  " : OK";
      std::cout << std::endl;
    }
    
  private:

    bool m_lastTestFailed;

  };


  //---------------------------------------------------------------------------

  /** @class CoralServerlUnitTest
   *
   *  Base class for COOL tests that does not use a database
   *  (use CoolDBUnitTest for tests involving databases).
   *  Copied and reused for CORAL SERVER unit tests.
   *
   *  @author Marco Clemencic
   *  @date   2006-05-09
   */
  class CoralServerUnitTest: public CppUnit::TestFixture {

  public:

    /// Standard constructor
    CoralServerUnitTest(){}

    /// Destructor
    virtual ~CoralServerUnitTest(){}

    /// Program main
    static int Main( int argc, char* argv[] );

  };

  int CoralServerUnitTest::Main( int argc, char* argv[] )
  {
    // Retrieve test path from command line first argument.
    // Default to "" which resolve to the top level suite.
    std::string testPath =
      (argc > 1) ? std::string(argv[1]) : std::string("");

    // Add a listener that colllects test result
    //CppUnit::TestResultCollector result;
    //controller.addListener( &result );

    /// Get the top level suite from the registry
    CppUnit::Test *suite =
      CppUnit::TestFactoryRegistry::getRegistry().makeTest();

    /// Adds the test to the list of test to run
    // CppUnit::TestRunner runner;
    CppUnit::TextTestRunner runner;
    runner.addTest( suite );

    // Change the default outputter to a compiler error format outputter
    // uncomment the following line if you need a compiler outputter.
    runner.setOutputter( new CppUnit::CompilerOutputter( &runner.result(), 
                                                         std::cout ) );

    // Change the default outputter to a xml error format outputter
    // uncomment the following line if you need a xml outputter.
    //runner.setOutputter( new CppUnit::XmlOutputter( &runner.result(),
    //                                                    std::cout ) );

    runner.eventManager().addListener( new coral::ProgressListener() );

    //CppUnit::TestResultCollector *collector = 
    //  new CppUnit::TestResultCollector();
    //runner.eventManager().addListener(collector);

    bool wasSuccessful = false;

    try 
    {
      wasSuccessful = runner.run( testPath, false, true, false );
    }

    // Test path not resolved
    catch ( std::invalid_argument &e )
    {
      std::cout  << std::endl << "ERROR: " << e.what() << std::endl;
      return 0;
    }

    // Should never happen?
    catch ( std::exception& e )
    {
      std::cout  << std::endl
                 << "UNEXPECTED STD EXCEPTION CAUGHT in CoralServerUnitTest: "
                 << e.what() << std::endl;
      return 0;
    }

    // Should never happen?
    catch ( ... )
    {
      std::cout  << std::endl
                 << "UNKNOWN EXCEPTION CAUGHT in CoralServerUnitTest" << std::endl;
      return 0;
    }

    // Return error code 1 if the one of tests failed.
    // Print a message on standard error if something failed (for QMTest)
    if ( ! wasSuccessful ) std::cerr << "Error: CppUnit Failures" << std::endl;
    int retcode = wasSuccessful ? 0 : 1;
    // Uncomment the next line if you want to integrate CppUnit with Oval
    std::cout << "[OVAL] Cppunit-result =" << retcode << std::endl;
    return retcode;

  }

  //---------------------------------------------------------------------------

}

#define CORALSERVERTEST_MAIN(CLASS) int main( int argc, char* argv[] ) \
  { return coral::CLASS::Main( argc, argv ); }

#endif // COMMON_CORALSERVERUNITTEST_H
