// $Id: PacketSocket.cpp,v 1.8 2009/02/10 15:07:45 mwache Exp $

#include <iostream>

// Include files
#include "CoralSockets/GenericSocketException.h"
#include "PacketSocket.h"

#include <sys/time.h> // for gettimeof day debug

//#define DEBUG( out ) do { timeval val; gettimeofday( &val, 0); std::cerr << "["<< val.tv_usec << "]" << out << std::endl; } while (0)

#ifndef DEBUG
#define DEBUG( out )
#endif

// Namespace
using namespace coral::CoralSockets;


//---------------------------------------------------------------------------

PacketSocket::PacketSocket( ISocketPtr socket) 
  : m_socket( socket )
{
}

//---------------------------------------------------------------------------

void PacketSocket::sendPacket( const PacketSLAC& Message) {
    // - cork, uncork?
    // - avoid deep copies
  const CTLPacketHeader &ctlHeader( Message.getHeader() );
  const ByteBuffer &payload( Message.getPayload() );

  DEBUG("dumping header ver:" << ctlHeader.version() 
      << " status:"  << ctlHeader.status()
      << " size:" << ctlHeader.packetSize()
      << " requestID:" << ctlHeader.requestID()
      << " clientID:" << ctlHeader.clientID()
      << " segmentNumber:" << ctlHeader.segmentNumber()
      << " moreSegments: " << ctlHeader.moreSegments()
      << std::endl
      );

  if ( ctlHeader.packetSize() - CTLPACKET_HEADER_SIZE > 
      PacketSLAC::CTLPACKET_MAX_PAYLOAD_SIZE )
    // we don't want to send too much in one go
    throw ErrorInHeaderException(ErrorInHeaderException::ErrPacketTooLarge,
        "packet size is too large!",
        "PacketSocket::sendPacket()");


  m_socket->writeAll( ctlHeader.data(), CTLPACKET_HEADER_SIZE );
  m_socket->writeAll( payload.data(), payload.usedSize() );
}

//---------------------------------------------------------------------------

PacketPtr PacketSocket::receivePacket( ) {

  // todo: 
  // -this can be optimized quite a lot, since we know all the expected sizes:
  //  first read the header, get the packet size, and then read the rest
  // -also we should avoid the deep copies all the time

  // receive the header
  unsigned char header[ CTLPACKET_HEADER_SIZE ];
  DEBUG("reading header" << std::endl);
  m_socket->readAll( (unsigned char*)header, CTLPACKET_HEADER_SIZE );

  // this is a bit stupid, but I'll leave it for now
  // we need a CTLHeader to know the size
  CTLPacketHeader ctlHeader( (unsigned char*)header, CTLPACKET_HEADER_SIZE);

  DEBUG("dumping header ver:" << ctlHeader.version() 
      << " status:"  << ctlHeader.status()
      << " size:" << ctlHeader.packetSize()
      << " requestID:" << ctlHeader.requestID()
      << " clientID:" << ctlHeader.clientID()
      << " segmentNumber:" << ctlHeader.segmentNumber()
      << " moreSegments: " << ctlHeader.moreSegments()
      << std::endl);

  // first check: version
  if ( ctlHeader.version() != CTLPACKET_CURRENT_VERSION )
    throw ErrorInHeaderException(ErrorInHeaderException::ErrWrongVersion,
        "received packet with wrong version number!",
        "PacketSocket::receivePacket()");

  // second check: status
  if ( ctlHeader.status() != CTLOK ) {
    if ( ctlHeader.status() == CTLWrongMagicWord )
      throw PacketStatusNotOkException("Received a packet with status WrongMagicWord",
          "PacketSocket::receivePacket()");
    else if ( ctlHeader.status() == CTLWrongVersion )
      throw PacketStatusNotOkException("Received a packet with status WrongVersion",
          "PacketSocket::receivePacket()");
    else if ( ctlHeader.status() == CTLWrongChecksum )
      throw PacketStatusNotOkException("Received a packet with status WrongChecksum",
          "PacketSocket::receivePacket()");
  };

  // check the size
  if ( ctlHeader.packetSize()< CTLPACKET_HEADER_SIZE ) 
    // at least on byte payload
    throw ErrorInHeaderException(ErrorInHeaderException::ErrPacketTooSmall,
        "packet size in received header is too small!",
        "PacketSocket::receivePacket()");

  if ( ctlHeader.packetSize() - CTLPACKET_HEADER_SIZE > 
      PacketSLAC::CTLPACKET_MAX_PAYLOAD_SIZE )
    // we don't want to allocate too much memory in one go
    throw ErrorInHeaderException(ErrorInHeaderException::ErrPacketTooLarge,
        "packet size in received header is too large!",
        "PacketSocket::receivePacket()");

  // receive the payload
  std::auto_ptr<ByteBuffer> payload( new ByteBuffer( 
        ctlHeader.packetSize()-CTLPACKET_HEADER_SIZE ) );
  DEBUG("reading payload" << std::endl);
  m_socket->readAll( payload->data(), payload->freeSize() );
  payload->setUsedSize( payload->freeSize() );

  if ( ctlHeader.payloadChecksum() != CTLPacketHeader::computeChecksum(
        payload->data(), payload->usedSize() ) )
    throw ErrorInHeaderException(ErrorInHeaderException::ErrWrongChecksum,
        "payload checksum in received header is wrong!",
        "PacketSocket::receivePacket()");

  // construct PacketPtr
  PacketPtr ret( new PacketSLAC( ctlHeader, payload ) );

  return ret;
}

//---------------------------------------------------------------------------

bool PacketSocket::isOpen() {
  return m_socket->isOpen();
}

//---------------------------------------------------------------------------

void PacketSocket::close() {
  return m_socket->close();
}

//---------------------------------------------------------------------------

int PacketSocket::getFd() {
  return m_socket->getFd();
}

