// $Id: ReplyManager.cpp,v 1.5 2009/02/11 11:31:57 mwache Exp $

// Include files
#include "CoralSockets/GenericSocketException.h"

#include "ReplyManager.h"


// Namespace
using namespace coral;
using namespace coral::CoralSockets;


#include <sys/time.h> // for gettimeof day debug

//#define DEBUG( out ) do { timeval val; gettimeofday( &val, 0); std::cerr << "["<< val.tv_usec << "]" << out << std::endl; } while (0)

#ifndef DEBUG
#define DEBUG( out )
#endif


unsigned int ReplySlot::timeout=30*60; // half an hour

//-----------------------------------------------------------------------------

void ReplySlot::appendReply( std::auto_ptr<ByteBuffer> reply, int segmentNr,
    bool lastReply )
{
  boost::shared_ptr<ByteBuffer> buffer(reply.get());
  reply.release();

  boost::mutex::scoped_lock lock( m_mutex);
  if ( !m_used )
    throw GenericSocketException("Slot is not used!","ReplySlot::appendReply");
  if ( m_status==ALL_REPLIES )
    throw GenericSocketException("Slot has already got all replies!",
        "ReplySlot::appendReply");
  if ( m_segmentNr != segmentNr )
    throw SegmentationErrorException("multi segment reply, but segment no is"
        " not continous!","ReplySlot::appendReply");


  m_replies.push( buffer );
  m_segmentNr++;
  if ( lastReply )
    m_status = ALL_REPLIES;
  m_cond.notify_all();
}

//-----------------------------------------------------------------------------

boost::shared_ptr<ByteBuffer> ReplySlot::nextReply( bool &lastReply )
{
  boost::mutex::scoped_lock lock( m_mutex );

  // return 0 if the last reply has already been returned
  if ( m_status == ALL_REPLIES && m_replies.empty() )
    return boost::shared_ptr<ByteBuffer>( (ByteBuffer*)0 );

  unsigned int wait_time=0;
  while ( m_replies.empty() && m_status == OK ) {
    // it looks like on Mac OS signals can get lost?
    // so better use a time out..
    // todo: handle errors (connection lost etc...) 
    boost::xtime xt;
    xtime_get(&xt, boost::TIME_UTC);
    xt.sec += 20;
    if ( !m_cond.timed_wait(lock, xt) ) {
      DEBUG( "wait for reply timed out..."<< m_requestID << std::endl);
      wait_time+=20;
      if (!m_replies.empty() )
        std::cerr << "wait for reply timed out but reply is there .." << m_requestID << std:: endl;
    };
    if ( wait_time > timeout )
      setTimedOut();
  }

  if ( m_status == TIME_OUT )
    throw GenericSocketException("reply timed out","ReplySlot::nextReply");
  if ( m_status == CLOSED )
    throw GenericSocketException("connection closed", "ReplySlot::nextReply");

  if ( m_replies.empty() )
    throw GenericSocketException("Panic! replies empty before pop!",
        "ReplySlot::nextReply");

  boost::shared_ptr<ByteBuffer> buffer( m_replies.front() );
  m_replies.pop();
  lastReply = m_replies.empty() && (m_status==ALL_REPLIES);
  return buffer;
}


//-----------------------------------------------------------------------------

void ReplySlot::freeSlot() 
{
  boost::mutex::scoped_lock lock( m_mutex );
  DEBUG("freeSlot() called" << std::endl);

  while ( !m_replies.empty() )
    m_replies.pop();
  
  m_used=false;
  m_status=OK;
}

//-----------------------------------------------------------------------------

bool ReplySlot::takeSlot( int requestID ) 
{
  boost::mutex::scoped_lock lock( m_mutex );
  DEBUG("takeSlot( " << requestID << " ) called" << std::endl );

  if ( m_used )
    return false;

  m_requestID=requestID;
  m_used=true;
  m_status=OK;
  m_segmentNr=0;
  return true;
}

//-----------------------------------------------------------------------------

void ReplySlot::setTimedOut()
{
  // private method, expects m_mutex to be locked!
  if (m_status != OK)
    throw GenericSocketException("Panic! m_status!=OK", 
        "ReplySlot::setTimedOut");

  m_status=TIME_OUT;
}

//-----------------------------------------------------------------------------

void ReplySlot::setConnectionClosed()
{
  boost::mutex::scoped_lock lock( m_mutex );

  m_status=CLOSED;
  m_cond.notify_all();
}

//-----------------------------------------------------------------------------

bool SocketReplyIterator::next()
{
  m_currentBuf = m_slot.nextReply( m_lastReply );
  return m_currentBuf.get() != 0;
}

//-----------------------------------------------------------------------------

const ByteBuffer& SocketReplyIterator::currentBuffer() const
{
  if ( m_currentBuf.get() == 0) 
    throw GenericSocketException("called currentReply() before the first "
        "next() or after the last reply", "SocketReplyIterator::currentReply");

  return *m_currentBuf;
}

//-----------------------------------------------------------------------------

ReplySlot& ReplyManager::requestSlot( int requestID ) 
{
  int slot = (requestID + m_requestOffset) % m_sleepersSize;
  int tries=0;
  int maxTries=m_sleepersSize;
  while ( m_sleepers[ slot ].used() != 0 && tries< maxTries ) {
    tries++;
    std::cerr << "skipping slot " << slot << " for id " << requestID << std::endl;
    nextSlot( slot );
  }

  if ( !m_sleepers[ slot ].takeSlot( requestID ) )
    throw GenericSocketException("PANIC! internal error could not take slot!",
        "ReplyManager::requestSlot");

  if ( tries>=maxTries )
    throw GenericSocketException( "requestSlot max tries" );
  m_requestOffset=(m_requestOffset+tries)%m_sleepersSize;
  return m_sleepers[ slot ];
}

//-----------------------------------------------------------------------------

ReplySlot& ReplyManager::findSlot( int requestID ) {
  int slot = (requestID + m_answerOffset ) % m_sleepersSize;
  int tries=0;
  int maxTries=m_sleepersSize;
  if ( m_sleepers[slot].requestID() == requestID )
    return m_sleepers[ slot ];

  if ( m_sleepers[slot].requestID() > requestID ) {
    while ( m_sleepers[slot].requestID() != requestID && tries< maxTries ) {
      tries++;
      std::cerr << "prev answer skipping slot " << slot 
        <<" id " << m_sleepers[slot].requestID() << " searching " << requestID << std::endl;
      prevSlot(slot);
    }
  } else {
    while ( m_sleepers[slot].requestID() != requestID && tries< maxTries ) {
      tries++;
      std::cerr << "answer skipping slot " << slot
        <<" id " << m_sleepers[slot].requestID() << " searching " << requestID << std::endl;
      nextSlot(slot);
    }
    m_answerOffset=(m_answerOffset+tries)%m_sleepersSize;
  };
  if ( m_sleepers[ slot ].requestID() != requestID )
    throw GenericSocketException( " didn't find correct slot!",
        "ReplyManager::findSlot()");
  if ( tries>=maxTries )
    throw GenericSocketException( "replySlot max tries" );
  return m_sleepers[ slot ];
}

//-----------------------------------------------------------------------------

void ReplyManager::close() 
{
  for ( int i=0; i < m_sleepersSize; i++) 
    m_sleepers[ i ].setConnectionClosed();
}
