// $Id: SocketRequestHandler.cpp,v 1.6 2009/02/11 11:33:45 mwache Exp $

// Include files
#include <iostream>

#include "SocketRequestHandler.h"

// Local include files
#include "PacketSocket.h"
#include "ReplyManager.h"

#include <sys/time.h> // for gettimeof day debug

//#define DEBUG( out ) do { timeval val; gettimeofday( &val, 0); std::cerr << "["<< val.tv_usec << "]" << out << std::endl; } while (0)

#ifndef DEBUG
#define DEBUG( out )
#endif


// Namespace
using namespace coral::CoralSockets;
using namespace coral;

//-----------------------------------------------------------------------------

namespace coral {

  namespace CoralSockets {


    /// a class which does the receiving of packets in an own thread
    class SocketRequestHandler::receiveThread {
      public:
        receiveThread( const PacketSocketPtr& socket, ReplyManager& rplMngr ) :
          m_Socket( socket ),
          m_rplMngr( rplMngr)
        {
          DEBUG("socketRequestHandler::receiveThread constructor");
        };

        virtual ~receiveThread()
        {
          DEBUG("socketRequestHandler::receiveThread destructor");
        };

        /// the main loop of the thread
        void operator()();

      private:
        /// the socket which is used to receive data
        const PacketSocketPtr& m_Socket;

        /// storage container for the waiting requests
        ReplyManager& m_rplMngr;                
    };

  }
} // namespace CORAL & CoralSockets

//-----------------------------------------------------------------------------

// the main loop of the receiver thread
void SocketRequestHandler::receiveThread::operator()() {
  DEBUG("SocketRequestHandler::receiveThread main loop started");
  int debug=1;

  while ( true ) 
  { 
    try
    {
      PacketPtr packet( m_Socket->receivePacket() );

      const int requestID = packet->getHeader().requestID();
      if (debug)
        std::cout << "Got reply to request " << requestID << std::endl;

      // find sleeper slot
      ReplySlot &slot = m_rplMngr.findSlot( requestID );

      // put reply and wake up the sleeper!
      slot.appendReply( packet->getPayloadPointer(), 
          packet->getHeader().segmentNumber(), 
          !packet->getHeader().moreSegments() /* isLast */ );

    }
    catch( SocketClosedException& e ) {
      std::cout << "The socket has been closed. Stopping threads." << std::endl;
      break;
    }
    catch( std::exception& e ) {
      std::cerr << "ERROR! Standard C++ exception: '" << e.what() << "'" << std::endl;
      break;
    }
    catch( ... ) {
      std::cerr << "ERROR! Unknown exception caught" << std::endl;
      break;
    }

  }

  m_rplMngr.close();

  // todo: tear down: wake up all the sleepers and notify of the error condition
  DEBUG("SocketRequestHandler::receiveThread main loop ended");
}

//-----------------------------------------------------------------------------

IByteBufferIteratorPtr SocketRequestHandler::replyToRequest( 
    const ByteBuffer& requestBuffer )
{
  ReplySlot &slot = send( requestBuffer );

  return IByteBufferIteratorPtr( new SocketReplyIterator( slot ) );
}      

//-----------------------------------------------------------------------------

SocketRequestHandler::SocketRequestHandler( PacketSocketPtr socket, 
                                            bool debug )
  : IRequestHandler()
  , m_debug( debug )
  , m_Socket( socket )
  , m_requestID( 0 )
  , m_rplMngr( new ReplyManager() )
{
  DEBUG("SocketRequestHandler constructor connection:"<< m_Socket->desc());

  receiveThread receiver( m_Socket, *m_rplMngr );
  m_receiveThread = std::auto_ptr<boost::thread>(new boost::thread( receiver ));
}

//-----------------------------------------------------------------------------

SocketRequestHandler::~SocketRequestHandler()
{
  DEBUG("Destructor SocketRequestHandler started");
  m_Socket->close();
  // wait until the receive thread has terminated
  m_receiveThread->join();
  DEBUG("Destructor SocketRequestHandler finished");
}

//-----------------------------------------------------------------------------

ReplySlot& SocketRequestHandler::send( const ByteBuffer& requestMsg )
{
  boost::mutex::scoped_lock lock( m_mutexForSend );
  m_requestID++;
  ReplySlot& slot=m_rplMngr->requestSlot( m_requestID );

  PacketSLAC packet(  m_requestID, 0 /*segment number*/, 0 /*clientID*/,
     false /* more segments*/, requestMsg );

  /*
  std::cerr<<"packet header ver:" << packet.getHeader().version() 
      << " status:"  << packet.getHeader().status()
      << " size:" << packet.getHeader().packetSize()
      << " requestID:" << packet.getHeader().requestID()
      << " clientID:" << packet.getHeader().clientID()
      << " segmentNumber:" << packet.getHeader().segmentNumber()
      << " moreSegments: " << packet.getHeader().moreSegments()
      << std::endl;
  */   

  try {
    m_Socket->sendPacket( packet );
  }
  catch( std::exception& e )
  {
    std::cerr << "ERROR! Standard C++ exception: '" << e.what() << "'" << std::endl;
    throw;
  }

  catch( ... ) 
  {
    std::cerr << "ERROR! Unknown exception caught" << std::endl;
    throw;
 }

  return slot;
}

//-----------------------------------------------------------------------------

