// $Id: SocketThread.cpp,v 1.15 2009/02/11 11:33:45 mwache Exp $

// Include files
#include "CoralServerBase/IRequestHandler.h"
#include "CoralServerBase/CTLMagicWordException.h"
#include "CoralSockets/GenericSocketException.h"

#include "SocketThread.h"

// Namespace
using namespace coral::CoralSockets;


#include <sys/time.h> // for gettimeof day debug

//#define DEBUG( out ) do { timeval val; gettimeofday( &val, 0); std::cerr << "["<< val.tv_usec << "]" << out << std::endl; } while (0)

#ifndef DEBUG
#define DEBUG( out )
#endif

//-----------------------------------------------------------------------------

/// class for handling requests
class SocketThread::HandlerThread : public Runnable {

public:
  virtual ~HandlerThread() {}

  HandlerThread( SocketThread* sThread,
                 const std::string& desc,
                 bool debug )
    : m_sThread( sThread ) 
    , m_desc( desc)
    , m_debug( debug )
  {}

  /// main loop of the handler
  void operator()();

  /// 
  void endThread() 
  {};

private:
  /// the "owner" of this handler thread
  SocketThread* m_sThread;

  /// description of the connection, for error messages
  const std::string m_desc;
  const bool m_debug;
};

//-----------------------------------------------------------------------------

void SocketThread::HandlerThread::operator()()
{
  DEBUG( "( " << m_desc << " ) "
            << "Enter handler thread " << this << " for client socket" 
            << std::endl);
  coral::IRequestHandler *handler( &m_sThread->getRequestHandler() );

  while ( m_sThread->isActive() ) 
  {
    try 
    {
      PacketPtr packet1( m_sThread->getNewRequest() );
      if ( packet1.get() == 0 ){
        // no new packet -> the thread should end
        continue;
      } 

      const int requestID = packet1->getHeader().requestID();
      const int clientID = packet1->getHeader().clientID();
      const ByteBuffer& request = packet1->getPayload();
      DEBUG( "( " << m_desc << " ) "
                    << "Request #" << requestID << ":" << std::endl);

      std::auto_ptr<IByteBufferIterator> replyIt( 
                        handler->replyToRequest( request ) );

      uint32_t segmentID=0;
      while ( replyIt->next() ) {
        DEBUG("( " << m_desc << " ) sending "
                    << "Reply #" << requestID << " SegementID " 
                    << segmentID << std::endl);

        PacketSLAC packet( requestID, clientID, segmentID,
            !replyIt->isLast() /* moreSegments */,
            replyIt->currentBuffer() );

        m_sThread->send( packet );
        segmentID++;
      };
      DEBUG("( " << m_desc << " ) "
                    << "Reply #" << requestID << " finished" << std::endl);
 
    }
    catch ( std::exception& e )
    {
      std::cout << "( " << m_desc << " ) "
                << "EXCEPTION CAUGHT: " << e.what() << std::endl;
    }
    catch ( ... )
    {
      std::cout << "( " << m_desc << " ) "
                << "UNKNOWN EXCEPTION CAUGHT" << std::endl;
    }
  }
  DEBUG( "( " << m_desc << " ) "
            << "Exit handler thread " << this << " for client socket #" 
            << "??" << std::endl);
}

//-----------------------------------------------------------------------------

void SocketThread::send( const PacketSLAC& packet ) 
{
  boost::mutex::scoped_lock lock( m_mutexForSend );
  m_cSocket->sendPacket( packet );
}

//-----------------------------------------------------------------------------

PacketPtr SocketThread::getNewRequest()
{
  boost::mutex::scoped_lock lock( m_mutexForRequests );
  while ( m_requests.empty() && m_isActive )
    m_condForRequests.wait( lock );

  if ( !m_isActive ) 
    return PacketPtr( (PacketSLAC*)0 );

  PacketPtr packet = m_requests.front();
  m_requests.pop();
  return packet;
}

//-----------------------------------------------------------------------------

void SocketThread::operator()()
{
  DEBUG( "( " << m_cSocket->desc() <<" ) "
            << "Enter thread for client socket" << std::endl);
  for ( int i=0; i<m_nHThreads; i++ )
    m_thrManager.addThread( new HandlerThread( this, m_cSocket->desc(), 
          m_debug ) );

  try 
  {
    while ( m_isActive ) 
    {
      PacketPtr packet( m_cSocket->receivePacket() );

      if ( packet->getHeader().moreSegments() )
        throw GenericSocketException("received multi-buffer request, but "
           "they are not yet supported!","SocketThread::operator()()");
      if ( packet->getHeader().segmentNumber() != 0 )
        throw GenericSocketException("received packet with segment number != 0",
            "SocketThread::operator()()");

      DEBUG("( " << m_cSocket->desc() << " ) got packet request #"
          << packet1->getHeader().requestID() );
      // queue packet for processing
      {
        boost::mutex::scoped_lock lock( m_mutexForRequests );
        m_requests.push( packet );
        m_condForRequests.notify_one( );
      }
    }
  }
  catch (SocketClosedException & e) {
    std::cout << "The connection '" << m_cSocket->desc() 
      <<"' has been closed. Stopping threads."<<std::endl;
  }
  catch ( CTLMagicWordException &e) {
    std::cout << "Received packet with errors in the header '"
      << e.what() << "'" << std::endl;

    // send error msg packet back
    PacketSLAC errPacket( CTLWrongMagicWord );
    send( errPacket );
  }
  catch( ErrorInHeaderException & e) {
    std::cout << "Received packet with errors in the header '"
      << e.what() << "'" << std::endl;

    // send error msg packet back
    if ( e.getErrorCode() == ErrorInHeaderException::ErrWrongVersion ) {
      PacketSLAC errPacket( CTLWrongVersion );
      send( errPacket );
    } else if ( e.getErrorCode() == ErrorInHeaderException::ErrWrongChecksum ) {
      PacketSLAC errPacket( CTLWrongChecksum );
      send( errPacket );
    } else {
      // send back a general checksum error
      PacketSLAC errPacket( CTLWrongChecksum );
      send( errPacket );
    }
  }
  catch ( std::exception& e )
  {
    std::cout << "( " << m_cSocket->desc() << " ) "
              << "EXCEPTION CAUGHT: " << e.what() << std::endl;
  }
  catch ( ... )
  {
    std::cout << "( " << m_cSocket->desc() << " ) "
              << "UNKNOWN EXCEPTION CAUGHT" << std::endl;
  }
  DEBUG( "( " << m_cSocket->desc() << " ) "
          << "Waiting for the threads to end" << std::endl);
  m_isActive = false;
  m_condForRequests.notify_all( );
  m_thrManager.joinAll();

  DEBUG( "( " << m_cSocket->desc() << " ) "
          << "Close the client socket" << std::endl);
  m_cSocket->close();
 
  //m_server->deleteThreadForSocket( m_cSocket->getFd() );
  DEBUG( "( " << m_cSocket->desc() << " ) "
            << "Exit thread for client socket" << std::endl);
}


