// $Id: TcpSocket.cpp,v 1.5 2009/02/10 15:06:40 mwache Exp $

// Include files
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <netinet/tcp.h> /* for TCP_NODELAY */ 
#include <fcntl.h>
#include <netdb.h>
#include <poll.h>

#include <sys/time.h> // for gettimeof day debug

#include <iostream>
#include <sstream>

#include "CoralSockets/GenericSocketException.h"

#include "TcpSocket.h"

// Namespace
using namespace coral::CoralSockets;


//#define DEBUG( out ) do { timeval val; gettimeofday( &val, 0); std::cerr << "["<< val.tv_usec << "]" << out << std::endl; } while (0)

#ifndef DEBUG
#define DEBUG( out )
#endif

TcpSocket::TcpSocket( int fd, const std::string & desc ) 
  : m_fd( fd )
  , m_desc( desc )
{
  configureSocket();
}

//---------------------------------------------------------------------------
TcpSocket::~TcpSocket() {
  if ( isOpen() ) {
    DEBUG("Closing socket in destructor");
    close();
  };
}

//---------------------------------------------------------------------------
const std::string& TcpSocket::desc() const {
  return m_desc;
}

//---------------------------------------------------------------------------

void TcpSocket::setSockOpt(int level, int option, int value) {
  if ( setsockopt( m_fd, level, option, &value, sizeof( value ) ) < 0 )
    throw GenericSocketException("Error setting socket option.",
          "TcpSocket::setSockOpt");
}

//---------------------------------------------------------------------------

void TcpSocket::setFdFlags(int getCMD, int setCMD, int flags) {
  int oflags = fcntl( m_fd, getCMD, 0 );
  
  if (oflags < 0 ) 
    throw GenericSocketException("Error getting fd flags.", 
        "TcpSocket::setFdFlags");

  if ( fcntl( m_fd, setCMD, oflags | flags ) < 0 )
    throw GenericSocketException("Error setting fd flags.",
        "TcpSocket::setFdFlags");
}

//-------------------------------------------------------------------------

void TcpSocket::configureSocket() {
  // fast reuse enable
  setSockOpt( SOL_SOCKET, SO_REUSEADDR, 1 );
  // keepalive enable
  setSockOpt( SOL_SOCKET, SO_KEEPALIVE, 1 );
#ifdef __APPLE__
  // don't send broken pipe signal
  setSockOpt( SOL_SOCKET, SO_NOSIGPIPE, 1 ); 
  // For Linux this is done on per message basis ( see read_write() )
#endif// __APPLE__
  // close on exec
  setFdFlags( F_GETFD, F_SETFD, FD_CLOEXEC );
  // non-blocking mode
  setFdFlags( F_GETFL, F_SETFL, O_NONBLOCK );

  // no delay (for debugging)
  setSockOpt( IPPROTO_TCP, TCP_NODELAY, 1 );

//   int value = 1;
//  ::setsockopt( m_cSocket, IPPROTO_TCP, TCP_NODELAY, &value, sizeof( value ));

}

//-------------------------------------------------------------------------

void TcpSocket::fillAddr( const std::string& hostName, 
                          int port, sockaddr_in& addr) 
{
  memset(&addr, 0, sizeof(addr));  // Zero out address structure
  addr.sin_family = AF_INET;       // Internet address
  hostent* host;  // Resolve name
  if ( ( host = gethostbyname(hostName.c_str()) ) == NULL ) {
    throw GenericSocketException( "Failed to resolve"
       " name (gethostbyname( '" + hostName + "'))",
        "TcpSocket::fillAddr" );
  }
  addr.sin_addr.s_addr = *( (unsigned long*)host->h_addr_list[0] );
  addr.sin_port = htons(port);     // Assign port in network byte order
}

//--------------------------------------------------------------------------

TcpSocketPtr TcpSocket::connectTo( const std::string& host, unsigned int port) 
{
  sockaddr_in addr;

  fillAddr( host, port, addr);
  int fd = socket( PF_INET, SOCK_STREAM, 0 );

  if ( fd < 0 )
    throw GenericSocketException("Error creating socket.",
        "TcpSocket::connectTo");

  if ( ::connect( fd, (sockaddr*)&addr, sizeof( addr ) ) < 0 ) {
    std::stringstream str;
    str << "Connection failed: " << strerror( errno );
#ifdef WIN32
    ::closesocket( fd );
#else
    ::close( fd );
#endif
    throw GenericSocketException( str.str(), "TcpSocket::connectTo" );
  };

  sockaddr_in laddr;
  memset(&laddr, 0, sizeof( laddr ));
  socklen_t llen=sizeof( laddr );
  if ( getsockname( fd, (sockaddr*)&laddr, &llen ) < 0 )
    std::cerr << "Could not get local port!" << std::endl;
 
  std::stringstream desc;
  desc << "connectTo: "<< host <<":"<< port << " from port "
    << ntohs(laddr.sin_port);

  return TcpSocketPtr( new TcpSocket( fd, desc.str() ) );
}
//----------------------------------------------------------------------

TcpSocketPtr TcpSocket::listenOn( const std::string& ip, unsigned int port) {
  sockaddr_in addr;
  fillAddr( ip, port, addr);
  
  int fd = socket( PF_INET, SOCK_STREAM, 0 );
  
  if ( fd < 0 )
   throw GenericSocketException("Error creating socket.",
       "TcpSocket::listenOn");
  
  memset( addr.sin_zero,'\0', sizeof( addr.sin_zero ) );

  if ( bind( fd, (struct sockaddr*) &addr, sizeof( addr ) ) < 0 ) {
    std::stringstream str;
    str << "Error binding socket: " << strerror( errno );
#ifdef WIN32
    ::closesocket( fd );
#else
    ::close( fd );
#endif
    throw GenericSocketException( str.str(), "TcpSocket::listenOn");
   };

  int backlog=100;
  if ( listen( fd, backlog ) < 0 ) {
    std::stringstream str;
    str << "Listen on socket failed: " << strerror( errno );
#ifdef WIN32
    ::closesocket( fd );
#else
    ::close( fd );
#endif
    throw GenericSocketException( str.str(), "TcpSocket::listenOn");
   };
  std::stringstream desc;
  desc << "listenOn: "<< ip <<":"<< port;

  return TcpSocketPtr( new TcpSocket( fd, desc.str() ) );
}

//----------------------------------------------------------------------

TcpSocketPtr TcpSocket::accept() {
  sockaddr_in cAddrStruct;
#ifdef WIN32
  int cAddrStructSize = sizeof(cAddrStruct);
#else
  unsigned int cAddrStructSize = sizeof(cAddrStruct);
#endif
  int cSocket;

  cSocket = ::accept( m_fd, (struct sockaddr*) &cAddrStruct, 
      &cAddrStructSize );

  if ( cSocket < 0 ) 
    throw GenericSocketException( "Client socket accept() failed" );

  // A client has succesfully connected

  std::string cHost = inet_ntoa( cAddrStruct.sin_addr );
  int cPort = ntohs( cAddrStruct.sin_port );

  std::cout << "New client connection from " 
      << "( " << cHost << ", " << cPort << " )" << std::endl;

  std::stringstream desc;
  desc << "accept from " << cHost << ":" << cPort;
  return TcpSocketPtr( new TcpSocket( cSocket, desc.str() ) );
} 

//----------------------------------------------------------------------

void TcpSocket::readAll( unsigned char *buf, size_t len ) const {
  size_t res = read_write( P_READ, buf, len, -1 );
  if ( res != len )
    throw GenericSocketException("Panic! readAll() could not read all bytes!");
}

void TcpSocket::writeAll( const unsigned char *buf, size_t len ) const {
  size_t res = read_write( P_WRITE, const_cast< unsigned char*> (buf), len, -1 );
  if ( res != len )
    throw GenericSocketException("Panic! writedAll() could not write all bytes!");
}

//----------------------------------------------------------------------

size_t TcpSocket::read( unsigned char *buf, size_t len, int timeout ) const {
  return read_write( P_READ, buf, len, timeout );
}

size_t TcpSocket::write( const unsigned char *buf, size_t len, 
                int timeout ) const {
  return read_write( P_WRITE, const_cast< unsigned char*> (buf), len, timeout );
}

//----------------------------------------------------------------------

size_t TcpSocket::read_write( PMode mode, 
    unsigned char *buf, size_t len, int timeout ) const 
{
  if ( mode != P_READ && mode != P_WRITE )
    throw GenericSocketException("Panic! Wrong mode in read_write!");

  const char *label = mode==P_READ ? "TcpSocket::read" : "TcpSocket::write";

  DEBUG( label << ":" << m_desc << " enter read write timeout " << timeout);

  if (m_fd < 0 )
        throw SocketClosedException( label );
#ifndef __APPLE__
  static const int flags = MSG_NOSIGNAL;
#else
  static const int flags = 0; 
  // flag SO_NOSIGPIPE gives stranges results on Mac Os
#endif

  unsigned char *p=buf;
  size_t read=0;
  while ( len > read) {
    DEBUG( label << ":" << m_desc << " len " << len << " read " << read <<" p " << (int)p );

    if  ( timeout != 0 ) {
      // wait until we can read...
      if (timeout < 0 )
        // no timeout set, we sleep until data is available,
        // but don't return unless we got all data
        poll( mode, 200 );
      else if ( !poll( mode, timeout ) ) {
        DEBUG( label << ": " << m_desc << " timeout" );
        // timeout occured, return
        break;
      };
    }

    // has the socket been closed?
    if (m_fd < 0 )
      throw SocketClosedException( label );

    ssize_t nResult=0;
    if ( mode == P_READ)
      nResult = ::recv( m_fd, p, len-read, flags);
    else 
      nResult = ::send( m_fd, p, len-read, flags);

    if ( nResult < 0 ) {
      DEBUG( label << ":" << m_desc << " nResult < 0 " << errno << " " << strerror(errno) );
      // something is wrong, find out what
      if ( errno == EINTR )
        // try again on interrupt
        continue;
      if ( errno == EAGAIN ) {
        DEBUG( label << ":" << m_desc <<" EAGAIN this should not happen?"  );
        // no data available
        // as long as we poll, this shoud not happen... 
        if (timeout == 0)
                break;
        continue;
      } 
      if ( errno == EPIPE || errno == ENOTCONN || errno == ECONNRESET ){
        DEBUG( label << ":" << m_desc <<"Error : " << strerror( errno ) );
        // connection closed
        throw SocketClosedException( label );
      };

      // hmm, seems to be something serious
      throw GenericSocketException( strerror(errno) , label);
    };

    if ( nResult == 0 ) {
      // connection lost
      // in case we already read something return what we have
      if (read!=0)
        break;
      DEBUG( label << ":" << m_desc << " nrecv return 0 " );
      throw SocketClosedException( label );
    };
  
    read+=nResult;
    p+=nResult;
  };
  
  DEBUG( label << ":" << m_desc << " leave read write return:" << read );
  return (size_t) read;
}

/*
size_t TcpSocket::read( char *buf, size_t len, int timeout ) const {
  if (m_fd < 0 )
        throw SocketClosedException( "TcpSocket::read" );
#ifndef __APPLE__
  static const int flags = MSG_NOSIGNAL;
#endif
  static const int flags = 0; 
  // flag SO_NOSIGPIPE gives stranges results on Mac Os


  char *p=buf;
  size_t read=0;
  while ( len > read) {
    std::cerr << m_desc << " len " << len << "read " << read <<"p " << p << std::endl;
    // wait until we can read...
    if (timeout <= 0 )
      poll( P_READ, 200 );
      // no timeout set, we sleep, but don't return unless we got all data
    else if ( !poll( P_READ, timeout ) )
      // timeout occured, return
      break;

    ssize_t nRead = ::recv( m_fd, p, len, flags);

    if ( nRead < 0 ) {
      std::cerr << m_desc << " nRead < 0 " << errno << " " << strerror(errno) << std::endl;
      // something is wrong, find out what
      if ( errno == EINTR )
        // try again on interrupt
        continue;
      if ( errno == EAGAIN )
        // no data available
        continue;
      if ( errno == EPIPE || errno == ENOTCONN || errno == ECONNRESET ){
        std::cerr <<"Error : " << strerror( errno ) << std::endl;
        // connection closed
        throw SocketClosedException( "TcpSocket::read" );
      };

      // hmm, seems to be something serious
      throw GenericSocketException( strerror(errno) , "TcpSocket::read");
    };

    if ( nRead == 0 ) {
      // connection lost
      std::cerr << m_desc << " nrecv return 0 " << std::endl;
      throw SocketClosedException( "TcpSocket::read" );
    };
  
    read+=nRead;
    p+=nRead;
  };
  
  return (size_t) read;
}

//------------------------------------------------------------------------

size_t TcpSocket::write( const char *buf, size_t len, int timeout ) const {
  if (m_fd < 0 )
        throw SocketClosedException( "TcpSocket::write" );
#ifndef __APPLE__
  static const int flags = MSG_NOSIGNAL;
#else
  static const int flags = 0;
#endif

  const char *p=buf;
  size_t written=0;
  while ( len > written) {
    //std::cerr << "len " << len << "written " << written <<"p " << p << std::endl;
    // wait until we can write...
    if (timeout <= 0 )
      poll( P_WRITE, 200 );
      // no timeout set, we sleep, but don't return unless we got all data
    else if ( !poll( P_WRITE, timeout ) )
      // timeout occured, return
      break;

    ssize_t nWritten = ::send( m_fd, p, len-written, flags);

    if ( nWritten < 0 ) {
      // something is wrong, find out what
      if ( errno == EINTR )
        // try again on interrupt
        continue;
      if ( errno == EAGAIN )
        // try again
        continue;
      if ( errno == EPIPE || errno == ENOTCONN || errno == ECONNRESET )
        // connection closed
        throw SocketClosedException( "TcpSocket::write" );


      // hmm, seems to be something serious
      throw GenericSocketException( strerror( errno ) , "TcpSocket::write");
    };

    if ( nWritten == 0 ) {
      // connection lost
      throw GenericSocketException( "Connection lost", "TcpSocket::write");
    };

    written+=nWritten;
    p+=nWritten;
  };
  return (size_t)written;
}
*/
//--------------------------------------------------------------------------

bool TcpSocket::poll( PMode mode, int timeout ) const {
  struct pollfd fds;
  // has socket been closed?
  if (m_fd < 0 )
    throw SocketClosedException( "TcpSocket::poll" );

  fds.fd = m_fd;

  if (mode == P_READ )
    fds.events = POLLIN;
  else if ( mode == P_WRITE )
    fds.events = POLLOUT;
  else if ( mode == P_READ_WRITE )
    fds.events = POLLOUT | POLLIN;

  if ( ::poll( &fds, 1, timeout) == 0 )
    return false;
  // FIXME what about errors?
  return true;
}

//--------------------------------------------------------------------------

void TcpSocket::close() {
  if ( m_fd < 0 )
    // already closed
    return;

  if ( ::shutdown( m_fd, SHUT_RDWR ) < 0 ) {
    // FIXME log something? already closed? 
  };
#ifdef WIN32
  ::closesocket( m_fd );
#else
  ::close( m_fd );
#endif
  m_fd = -1;
}

//---------------------------------------------------------------------------
        
bool TcpSocket::isOpen() const {
  return m_fd >0;
}

