// $Id: ThreadManager.h,v 1.3 2009/02/09 12:27:55 mwache Exp $
#ifndef CORALSOCKETS_THREADMANAGER_H 
#define CORALSOCKETS_THREADMANAGER_H 1

#include <map>
#include <string>
#include <boost/shared_ptr.hpp>
#include "CoralSockets/BoostThread_headers.h"
#include "CoralSockets/GenericSocketException.h"

namespace coral {

  namespace CoralSockets {

    class ThreadWrapper; // forward decleration

    class Runnable {
      public:
        /// signal the thread that it should end
        virtual void endThread() = 0;

        /// this method is the main method of the thread
        virtual void operator()()=0;

        // return a description of the thread
        virtual const std::string desc() const
        { return std::string("thread with no description"); };

        // destructor
        virtual ~Runnable()
        {};
    };

    class ThreadManager {
      typedef boost::thread* ThreadPtr;

      public:
      ThreadManager()
      {};

      virtual ~ThreadManager(); 

      /// takes the ownership of runClass
      void addThread( Runnable *runClass ); 

      /// returns true if the runClass is in the map
      bool existsThread( Runnable *runClass );

      /// called by teh ThreadWrapper class after the 
      /// thread is finished. runClass will be removed
      /// from the maps during the next call to housekeeping()
      void markDoneThread( Runnable *runClass ); 

      /// check which threads are done and remove them from the
      /// maps
      void housekeeping();

      /// waits until all threads are finished
      void joinAll();

      private:
      /// copy constructor is private
      ThreadManager(const ThreadManager& );
      /// assignment operator as well
      ThreadManager& operator=(const ThreadManager& );

      /// mutex protecting m_threads
      boost::mutex m_threadsMutex;

      /// a map of the running threads
      std::map<Runnable*, ThreadPtr > m_threads;

      /// which threads are done
      std::map<Runnable*, bool > m_threadsDone;

      /// the thread group
      boost::thread_group m_group;

    };

    // a wrapper to unsure that the threads a deregistered when they have
    // finished
    class ThreadWrapper {
      public:
        ThreadWrapper( ThreadManager * manager, Runnable* runClass)
          : m_manager( manager )
            , m_runClass( runClass )
        {};

        virtual ~ThreadWrapper()
        {};

        void operator()() {
          m_runClass->operator()();

          // avoid race conditions by making sure this thread has been
          // added before we remove it from the map
          while ( !m_manager->existsThread( m_runClass ) )
            usleep(10000);

          m_manager->markDoneThread( m_runClass );
        };

      private:
        ThreadManager *m_manager;
        Runnable* m_runClass;
    };



  }
}


#endif // CORALSOCKETS_THREADMANAGER_H
