// $Id: test_SocketRequestHandler.cpp,v 1.2 2009/02/11 11:33:46 mwache Exp $

#include <iostream>

#include "../../../CoralServerBase/tests/Common/CoralServerUnitTest.h"

// Local include files 
#include "../../CoralSockets/BoostThread_headers.h"
#include "../../CoralSockets/GenericSocketException.h"
#include "../../src/RingBufferSocket.h"
#include "../../src/TcpSocket.h"
#include "../../src/SocketRequestHandler.h"

// Namespace
using namespace coral::CoralSockets;
using namespace coral;

typedef std::auto_ptr<ByteBuffer> ByteBufferPtr;

// port used for testing
const int testPort = 50007;


std::auto_ptr<RingBufferPipes> pipes;
TcpSocketPtr listenSocket;

namespace coral 
{

  class SocketRequestHandlerTest : public CoralServerUnitTest 
  {

    CPPUNIT_TEST_SUITE( SocketRequestHandlerTest );
    CPPUNIT_TEST( test_Basic );
    CPPUNIT_TEST( test_Close );
    CPPUNIT_TEST( test_WrongVersionPacket );
    CPPUNIT_TEST( test_WrongChecksumPacket );
    CPPUNIT_TEST( test_WrongMagicWordPacket );
    CPPUNIT_TEST_SUITE_END();

    public:

    void setUp() {}

    void tearDown() {
    }

    // ---------------------------------------------------
    // helpers
    void createSockets( ISocketPtr &src, ISocketPtr &dst)
    {
#if 1
      pipes = std::auto_ptr<RingBufferPipes>(
          new RingBufferPipes( 10000, "test pipes") );

      src=pipes->getSrc();
      dst=pipes->getDst();
#else
      listenSocket = TcpSocketPtr( (TcpSocket*)0);

      while ( listenSocket.get() == 0 )
      {
        try {
          listenSocket = TcpSocket::listenOn( "localhost", testPort );
        }
        catch (GenericSocketException &e) {
          std::string expMsg="Error binding socket";
          if (expMsg == std::string( e.what(), expMsg.size() ) ) {
            std::cout << "Caught exeption: '" << e.what() 
              << "' sleeping."<< std::endl;
            sleep( 10 );
            continue;
          }
          throw;
        }
      };


      src = TcpSocket::connectTo( "localhost", testPort );
      dst = listenSocket->accept();
#endif
    }

    static ByteBufferPtr createBuffer( int num ) {
      ByteBufferPtr buf( new ByteBuffer( sizeof( int ) ) );
      *((int*)buf->data()) = num;
      buf->setUsedSize( sizeof( int ) );
      return buf;
    };

    static int getBufferNum( ByteBufferPtr buf ) {
      return *(int*)buf->data();
    };

    static int getBufferNum( const ByteBuffer& buf ) {
      return *(int*)buf.data();
    };

    ByteBufferPtr  string2ByteBuffer( std::string str ) {
      ByteBufferPtr  result( new ByteBuffer( str.length() ) );
      const char * cstr=str.c_str();

      for (unsigned int i=0; i<str.length(); i++ )
        *(result->data()+i) = (unsigned char)*(cstr+i);

      result->setUsedSize( str.length() );
      return result;
    }

    std::string byteBuffer2String( const ByteBuffer& buf ) {
      return std::string( (char*)buf.data(), buf.usedSize() );
    };

    // ------------------------------------------------------
    void test_Basic() 
    {
      ISocketPtr src;
      ISocketPtr dst;

      // test if the normal case works as it should

      createSockets( src, dst );
      PacketSocket sSocket( src );
      PacketSocketPtr dSocket( new PacketSocket( dst ) );

      SocketRequestHandler handler( dSocket );

      ByteBufferPtr buf;

      buf = createBuffer( 10 );

      // requst a reply
      IByteBufferIteratorPtr it = handler.replyToRequest( *buf );

      // read the request from the socket
      PacketPtr packet;
      packet=sSocket.receivePacket();

      // check that this is really our packet
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 1",
          10, getBufferNum( packet->getPayload() ) );

      // send back a reply
      buf = createBuffer( 20 );
      PacketSLAC reply( packet->getHeader().requestID(), 0 /*clientID*/,
          0 /*segment number*/, false /* more segments*/, *buf );
      sSocket.sendPacket( reply );

      // check if the reply arrived
      CPPUNIT_ASSERT_EQUAL_MESSAGE("next 1", true,
          it->next() );
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 2",
          20, getBufferNum( it->currentBuffer() ) );
      CPPUNIT_ASSERT_EQUAL_MESSAGE("isLast 1", true,
          it->isLast() );

      CPPUNIT_ASSERT_EQUAL_MESSAGE("next 2", false,
          it->next() );

    }

    // ------------------------------------------------------
    void test_Close()
    {
      ISocketPtr src;
      ISocketPtr dst;

      // close this side of the socket before the reply arrived

      createSockets( src, dst );
      PacketSocket sSocket( src );
      PacketSocketPtr dSocket( new PacketSocket( dst ) );

      SocketRequestHandler handler( dSocket );

      ByteBufferPtr buf;

      buf = createBuffer( 10 );

      // requst a reply
      IByteBufferIteratorPtr it = handler.replyToRequest( *buf );

      // read the request from the socket
      PacketPtr packet;
      packet=sSocket.receivePacket();

      // check that this is really our packet
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 1",
          10, getBufferNum( packet->getPayload() ) );

      // instead of replying, close the socket
      sSocket.close();

      try {
        it->next();
        CPPUNIT_FAIL("connection closed exception expected");
      }
      catch( GenericSocketException & e) {
        std::string expMsg="connection closed";
        CPPUNIT_ASSERT_EQUAL_MESSAGE( "exception message",
            expMsg, std::string( e.what(), expMsg.size() ) );
        // cut CORAL info
      }
   }

   // ------------------------------------------------------

    void test_WrongVersionPacket()
    {
      ISocketPtr src;
      ISocketPtr dst;

      // send "wrong version" packet as reply

      createSockets( src, dst );
      PacketSocket sSocket( src );
      PacketSocketPtr dSocket( new PacketSocket( dst ) );

      SocketRequestHandler handler( dSocket );

      ByteBufferPtr buf;

      buf = createBuffer( 10 );

      // requst a reply
      IByteBufferIteratorPtr it = handler.replyToRequest( *buf );

      // read the request from the socket
      PacketPtr packet;
      packet=sSocket.receivePacket();

      // check that this is really our packet
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 1",
          10, getBufferNum( packet->getPayload() ) );

      // send back a reply with status "wrong version"
      buf = createBuffer( 20 );

      CTLPacketHeader header(
          CTLWrongVersion,
          CTLPACKET_HEADER_SIZE+buf->usedSize() /*packet size*/,
          0, /*request id*/
          0, /* client id*/
          0, /* segment number */
          false, /* more segments */
          CTLPacketHeader::computeChecksum( buf->data(), buf->usedSize() )+40
          );

      src->writeAll( header.data(), CTLPACKET_HEADER_SIZE );
      src->writeAll( buf->data(), buf->usedSize() );

      try {
        it->next();
        CPPUNIT_FAIL("connection closed exception expected");
      }
      catch( GenericSocketException & e) {
        std::string expMsg="connection closed";
        CPPUNIT_ASSERT_EQUAL_MESSAGE( "exception message",
            expMsg, std::string( e.what(), expMsg.size() ) );
        // cut CORAL info
      }    
    }

    // ------------------------------------------------------
    void test_WrongChecksumPacket()
    {
      ISocketPtr src;
      ISocketPtr dst;

      // send "wrong checksum" packet as reply

      createSockets( src, dst );
      PacketSocket sSocket( src );
      PacketSocketPtr dSocket( new PacketSocket( dst ) );

      SocketRequestHandler handler( dSocket );

      ByteBufferPtr buf;

      buf = createBuffer( 10 );

      // requst a reply
      IByteBufferIteratorPtr it = handler.replyToRequest( *buf );

      // read the request from the socket
      PacketPtr packet;
      packet=sSocket.receivePacket();

      // check that this is really our packet
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 1",
          10, getBufferNum( packet->getPayload() ) );

      // send back a reply with status "wrong version"
      buf = createBuffer( 20 );

      CTLPacketHeader header(
          CTLWrongChecksum,
          CTLPACKET_HEADER_SIZE+buf->usedSize() /*packet size*/,
          0, /*request id*/
          0, /* client id*/
          0, /* segment number */
          false, /* more segments */
          CTLPacketHeader::computeChecksum( buf->data(), buf->usedSize() )+40
          );

      src->writeAll( header.data(), CTLPACKET_HEADER_SIZE );
      src->writeAll( buf->data(), buf->usedSize() );

      try {
        it->next();
        CPPUNIT_FAIL("connection closed exception expected");
      }
      catch( GenericSocketException & e) {
        std::string expMsg="connection closed";
        CPPUNIT_ASSERT_EQUAL_MESSAGE( "exception message",
            expMsg, std::string( e.what(), expMsg.size() ) );
        // cut CORAL info
      }    
    }

    // ------------------------------------------------------

    void test_WrongMagicWordPacket()
    {
      ISocketPtr src;
      ISocketPtr dst;

      // send "wrong checksum" packet as reply

      createSockets( src, dst );
      PacketSocket sSocket( src );
      PacketSocketPtr dSocket( new PacketSocket( dst ) );

      SocketRequestHandler handler( dSocket );

      ByteBufferPtr buf;

      buf = createBuffer( 10 );

      // requst a reply
      IByteBufferIteratorPtr it = handler.replyToRequest( *buf );

      // read the request from the socket
      PacketPtr packet;
      packet=sSocket.receivePacket();

      // check that this is really our packet
      CPPUNIT_ASSERT_EQUAL_MESSAGE("buffer num 1",
          10, getBufferNum( packet->getPayload() ) );

      // send back a reply with status "wrong version"
      buf = createBuffer( 20 );

      CTLPacketHeader header(
          CTLWrongMagicWord,
          CTLPACKET_HEADER_SIZE+buf->usedSize() /*packet size*/,
          0, /*request id*/
          0, /* client id*/
          0, /* segment number */
          false, /* more segments */
          CTLPacketHeader::computeChecksum( buf->data(), buf->usedSize() )+40
          );

      src->writeAll( header.data(), CTLPACKET_HEADER_SIZE );
      src->writeAll( buf->data(), buf->usedSize() );

      try {
        it->next();
        CPPUNIT_FAIL("connection closed exception expected");
      }
      catch( GenericSocketException & e) {
        std::string expMsg="connection closed";
        CPPUNIT_ASSERT_EQUAL_MESSAGE( "exception message",
            expMsg, std::string( e.what(), expMsg.size() ) );
        // cut CORAL info
      }    
    } 
  };

  CPPUNIT_TEST_SUITE_REGISTRATION( SocketRequestHandlerTest );
  
}

CORALSERVERTEST_MAIN( SocketRequestHandlerTest )
