#include "CoralStubs/ClientStub.h"

//CoralStubs includes
#include "StreamBuffer.h"
#include "StaticBufferPool.h"
#include "RowIteratorAll.h"
#include "RowIteratorFetch.h"
#include "CALProtocol.h"
#include "ClientUtils.h"
#include "Exceptions.h"
#include "CppTypes.h"
#include "CppMutex.h"
#include "CppMutexFct.h"

//CoralServerBase includes
#include "CoralServerBase/IRequestHandler.h"

//Coral includes
#include "CoralBase/Exception.h"

//CoralExceptions
#include "RelationalAccess/ConnectionServiceException.h"
#include "RelationalAccess/AuthenticationServiceException.h"
#include "RelationalAccess/RelationalServiceException.h"
#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/MonitoringException.h"
#include "RelationalAccess/WebCacheControlException.h"
#include "RelationalAccess/LookupServiceException.h"
#include "CoralBase/AttributeException.h"
#include "CoralBase/AttributeListException.h"

//C++ includes
#include <iostream>
#include <sstream>

namespace coral { namespace CoralStubs {

ClientStub::ClientStub(IRequestHandler& requestHandler)
  : m_requestHandler( requestHandler )
{
}

ClientStub::~ClientStub()
{
}

coral::Token
ClientStub::connect( const std::string& dbUrl, const coral::AccessMode mode, bool& fromProxy ) const
{
  coral::CALOpcode opcode = mode == coral::ReadOnly ? coral::CALOpcodes::ConnectRO : coral::CALOpcodes::ConnectRW;
  
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( opcode );
      
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(dbUrl);
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  
  const coral::ByteBuffer& replyBuffer = prepareReplyMessageFP(*(reply.get()), opcode, fromProxy);
  
  //create a streambuffer for in
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);
  
  uint32_t mytoken;
  
  stream_in.read(mytoken);
  ///FIXME output only for debug
//  std::cout << "Could establish connection to database: SessionID " << mytoken << std::endl;
  
  return mytoken;
}

void
ClientStub::releaseSession( coral::Token sessionID ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::Disconnect );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  prepareReplyMessage( *(reply.get()), coral::CALOpcodes::Disconnect );
  ///FIXME output only for debug
//  std::cout << "Connection to database dropped" << std::endl;
}

void
ClientStub::startTransaction( coral::Token sessionID, bool readOnly ) const
{
  coral::CALOpcode opcode = readOnly ? coral::CALOpcodes::StartTransactionRO : coral::CALOpcodes::StartTransactionRW;
  
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( opcode);
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  prepareReplyMessage( *(reply.get()), opcode );
  ///FIXME output only for debug
//  std::cout << "Transaction started" << std::endl;
}

void
ClientStub::commitTransaction( coral::Token sessionID ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::CommitTransaction );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  prepareReplyMessage( *(reply.get()), coral::CALOpcodes::CommitTransaction );
  ///FIXME output only for debug
//  std::cout << "Transaction commited" << std::endl;
}

const std::vector<std::string>
ClientStub::sessionProperties( Token sessionID ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::SessionProperties );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  
  stream_out->print();
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::SessionProperties );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);
  //create a new set
  std::vector<std::string> ret;
  //fill the set
  stream_in.read(ret);
    //debug
  stream_in.print();
  //return the set
  return ret; 
}

IRowIteratorPtr
ClientStub::fetchRows( Token sessionID, const QueryDefinition& qd, AttributeList* pRowBuffer, size_t cacheSize, bool cacheSizeInMB ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::FetchRows );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  
  stream_out->write(qd);
  
  stream_out->write(cacheSizeInMB);
  
  stream_out->write((uint32_t)cacheSize);
  //check if we have an empty AttributeList
  if(pRowBuffer == NULL){
    stream_out->write(false);
  }else{
    stream_out->write(true);
    stream_out->write(*pRowBuffer);
  }  
  //stream_out->print();
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::FetchRows );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);

  uint32_t cursorID;
  
  stream_in.read(cursorID);
  //create the iterator  
  RowIteratorFetch* tfetch = new RowIteratorFetch(m_requestHandler, cursorID, cacheSize, pRowBuffer);
  //return the iterator packed into a smart pointer
  return std::auto_ptr<IRowIterator>(tfetch);
}

IRowIteratorPtr
ClientStub::fetchAllRows( Token sessionID, const QueryDefinition& qd, AttributeList* rowBuffer ) const
{
  RowIteratorAll* tall; 
  
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::FetchAllRows );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  //check if we have an empty AttributeList
  if(rowBuffer == NULL){
  
    tall = new RowIteratorAll;
    
    stream_out->write(false);
  }else{
  
    tall = new RowIteratorAll(rowBuffer);
    
    stream_out->write(true);
    
    stream_out->write(*rowBuffer);
  }  
  stream_out->write(qd);
  
///  stream_out->print();
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::FetchAllRows );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);

  
  uint32_t size;
  stream_in.read(size);
  for(uint32_t i = 0; i < size; i++ ) stream_in.read(tall->addRow());
  
///  stream_in.print();

  return std::auto_ptr<IRowIterator>(tall);
}

const std::set<std::string>
ClientStub::listTables( Token sessionID, const std::string& schemaName ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::ListTables );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  stream_out->write(schemaName);
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::ListTables );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);
  //create a new set
  std::set<std::string> myset;
  //fill the set
  stream_in.read(myset);
  //return the set
  return myset;
}

const TableDescription
ClientStub::tableDescription( Token sessionID, const std::string& schemaName, const std::string& tableName ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::TableDescription );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  stream_out->write(schemaName);
  stream_out->write(tableName);
  //debug
  stream_out->print();
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::TableDescription );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);
  //create a new set
  TableDescription td;
  
  stream_in.read(td);
  //debug
  stream_in.print();
    
  return td;
}

const TableDescriptionMap
ClientStub::tableDescriptionMap( Token sessionID, const std::string& schemaName ) const
{
  coral::ByteBuffer* requestBuffer = prepareRequestMessage( coral::CALOpcodes::TableDescriptionMap );
  //create the stream out buffer
  WStreamBuffer* stream_out = new WStreamBuffer(*requestBuffer);
  //encode here
  stream_out->write(sessionID);
  stream_out->write(schemaName);
  //debug
  stream_out->print();
  //delete the stream buffer and set the buffer position = stream position 
  delete stream_out;
  //call the request handler and keep the reply interator
  coral::IByteBufferIteratorPtr reply = m_requestHandler.replyToRequest( *requestBuffer );
  //release the request buffer
  delete requestBuffer;
  //get the reply
  const coral::ByteBuffer& replyBuffer = prepareReplyMessage( *(reply.get()), coral::CALOpcodes::TableDescriptionMap );
  //create the stream in buffer
  RStreamBuffer stream_in(replyBuffer, CALPACKET_HEADER_SIZE);
  //create a new set
  TableDescriptionMap tdm;
  
  stream_in.read(tdm);
  //debug
  stream_in.print();
  
  return tdm;
}

int
ClientStub::deleteTableRows( Token /*sessionID*/, const std::string& /*schemaName*/, const std::string& /*tableName*/, const std::string& /*whereClause*/, const std::string& /*whereData*/ ) const
{
  throw StubsException("ClientStub::deleteTableRows, This function is not available");
}

const std::string
ClientStub::formatRowBufferAsString( Token /*sessionID*/, const std::string& /*schemaName*/, const std::string& /*tableName*/ ) const
{
  throw StubsException("ClientStub::formatRowBufferAsString, This function is not available");
}

void
ClientStub::insertRowAsString( Token /*sessionID*/, const std::string& /*schemaName*/, const std::string& /*tableName*/, const std::string& /*rowBufferAS*/ ) const
{
  throw StubsException("ClientStub::insertRowAsString, This function is not available");
}

Token
ClientStub::bulkInsertAsString( Token /*sessionID*/, const std::string& /*schemaName*/, const std::string& /*tableName*/, const std::string& /*rowBufferAS*/, int /*rowCacheSizeDb*/ ) const
{
  throw StubsException("ClientStub::bulkInsertAsString, This function is not available");
}

void
ClientStub::releaseBulkOp( Token /*bulkOpID*/ ) const
{
  throw StubsException("ClientStub::releaseBulkOp, This function is not available");
}

void
ClientStub::processRows( Token /*bulkOpID*/, const std::vector<coral::AttributeList>& /*rowsAS*/ ) const
{
  throw StubsException("ClientStub::processRows, This function is not available");
}

void
ClientStub::flush( Token /*bulkOpID*/ ) const
{
  throw StubsException("ClientStub::flush, This function is not available");
}

                              
} }
