#include "StreamBuffer.h"

#include "CoralServerBase/hexstring.h"

#include "CoralBase/Blob.h"
#include "CoralBase/Date.h"
#include "CoralBase/TimeStamp.h"

#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IColumn.h"

#include "Exceptions.h"

#include <iostream>

#include <sstream>
#define LOGGER_NAME "CoralStubs::StreamBuffer"
#include "logger.h"

#define STRINGVECTORSIZE uint16_t

namespace coral { namespace CoralStubs {

  WStreamBuffer::WStreamBuffer(coral::ByteBuffer& bytebuffer)
    : m_bytebuffer( bytebuffer )
    , m_write_pos( bytebuffer.usedSize() )
  {
  }
  
  WStreamBuffer::~WStreamBuffer()
  {
    m_bytebuffer.setUsedSize(m_write_pos);
  }
  
  void
  WStreamBuffer::flush()
  {
    m_bytebuffer.setUsedSize(m_write_pos);
  }

  template<>
  void
  WStreamBuffer::write(unsigned char const& data)
  {
    //dereference the address and write the data
    *(m_bytebuffer.data() + m_write_pos) = data;
    //increase the write position
    m_write_pos++;
  }

  template<>
  void
  WStreamBuffer::write(bool const& data)
  {
    //dereference the address and write the data
    *(m_bytebuffer.data() + m_write_pos) = (unsigned char)data;
    //increase the write position
    m_write_pos++;
  }

  template<>
  void
  WStreamBuffer::write(uint16_t const& data)
  {
    //dereference the address, cast to unsigned int with byte depth of 4
    //overrride the 4 chars with the unsigned int
    *((uint16_t*)(m_bytebuffer.data() + m_write_pos)) = data;
    //increase the write position
    m_write_pos = m_write_pos + 2;  
  }

  template<>
  void
  WStreamBuffer::write(uint32_t const& data)
  {
    //dereference the address, cast to unsigned int with byte depth of 4
    //overrride the 4 chars with the unsigned int
    *((uint32_t*)(m_bytebuffer.data() + m_write_pos)) = data;
    //increase the write position
    m_write_pos = m_write_pos + 4;
  }

  template<>
  void
  WStreamBuffer::write(uint64_t const& data)
  {
    //dereference the address, cast to unsigned int with byte depth of 4
    //overrride the 4 chars with the unsigned int
    *((uint64_t*)(m_bytebuffer.data() + m_write_pos)) = data;
    //increase the write position
    m_write_pos = m_write_pos + 8;
  }

  template<>
  void
  WStreamBuffer::write(const std::string& data)
  {
    //
    STRINGVECTORSIZE size = data.size();
    write(size);
    unsigned char* pos = m_bytebuffer.data() + m_write_pos;
    STRINGVECTORSIZE i;
    for(i = 0; i < size; i++, pos++ ){
      *pos = data[i];
    }
    m_write_pos = m_write_pos + i;
  }

  template<>
  void
  WStreamBuffer::write(const std::vector<std::string>& data)
  {
    STRINGVECTORSIZE size = data.size();
    write(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      write(data[i]);
    }
  }

  template<>
  void
  WStreamBuffer::write(const std::set<std::string>& data)
  {
    STRINGVECTORSIZE size = data.size();
    write(size);
    std::set<std::string>::const_iterator i;
    for(i = data.begin(); i != data.end(); i++ ){
      write(*i);
    }
  }
  
  template<>
  void
  WStreamBuffer::write(const std::vector< std::pair<std::string, std::string> >& data)
  {
    STRINGVECTORSIZE size = data.size();
    write(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      write(data[i].first);
      write(data[i].second);
    }
  }
  
  template<>
  void
  WStreamBuffer::write(const IQueryDefinition::SetOperation& data)
  {
    if(data == IQueryDefinition::Union)
      write((unsigned char)0x01);
    else if(data == IQueryDefinition::Minus)
      write((unsigned char)0x02);
    else if(data == IQueryDefinition::Intersect)
      write((unsigned char)0x03);
    else
      write((unsigned char)0x00);
  }
  
  template<>
  void WStreamBuffer::write(const TableDescription& data)
  {
    //write the table name
    write(data.name());
    //write the table type
    write(data.type());
    //write the table space name
    write(data.tableSpaceName());
    //write the columns
    STRINGVECTORSIZE columns = data.numberOfColumns();
    write(columns);
    for(STRINGVECTORSIZE i = 0; i < columns; i++ ){
      const IColumn& column = data.columnDescription( i );
      write(column.name());
      write(column.type());
      write((uint32_t)column.size());
      write(column.isSizeFixed());
      write(column.tableSpaceName());
    }
    //write unique constraints
    STRINGVECTORSIZE constraints = data.numberOfUniqueConstraints();
    write(constraints);
    for(STRINGVECTORSIZE i = 0; i < constraints; i++ ){
      const IUniqueConstraint& conststr = data.uniqueConstraint( i );
      write(conststr.columnNames());
      write(conststr.name());
      write(conststr.tableSpaceName());
    }    
    //write primary keys
    bool isprikey = data.hasPrimaryKey();
    write(isprikey);
    if(isprikey){
      const IPrimaryKey& prikey = data.primaryKey();
      write(prikey.columnNames());
      write(prikey.tableSpaceName());
    }
    //write foreign keys
    STRINGVECTORSIZE foreignkeys = data.numberOfForeignKeys();
    write(foreignkeys);
    for(STRINGVECTORSIZE i = 0; i < foreignkeys; i++ ){
      const IForeignKey& forkey = data.foreignKey( i );
      write(forkey.columnNames());
      write(forkey.name());
      write(forkey.referencedTableName());
      write(forkey.referencedColumnNames());    
    }
    //write indexes
    STRINGVECTORSIZE indexes = data.numberOfIndices();
    write(indexes);    
    for(STRINGVECTORSIZE i = 0; i < indexes; i++ ){
      const IIndex& ind = data.index( i );
      write(ind.columnNames());
      write(ind.name());
      write(ind.isUnique());
      write(ind.tableSpaceName());    
    }
  }
  
  template<>
  void
  WStreamBuffer::write(const std::map< std::string, TableDescription >& data)
  {
    STRINGVECTORSIZE size = data.size();
    write(size);
    std::map< std::string, TableDescription >::const_iterator i;
    for(i = data.begin(); i != data.end(); i++ ){
      write(i->first);
      write(i->second);
    }
  }

  template<>
  void
  WStreamBuffer::write(const QueryDefinition& data)
  {
    write(data.getSchemaName());
    //default values    
    write(data.getDistinct());
    write(data.getOutputList());
    write(data.getTableList());
    write(data.getCondition());
    write(data.getConditionData());
    write(data.getGroupBy());
    write(data.getOrderList());
    //special values    
    const std::map< std::string, QueryDefinition >& map = data.getSubQueries();
    write((STRINGVECTORSIZE)map.size());
    std::map< std::string, QueryDefinition >::const_iterator i;
    for(i = map.begin(); i != map.end(); i++ )
    {
      write(i->first);
      write(i->second);
    }
    
    const std::pair<int,int>* limits = data.getRowLimitAndOffset();
    if(limits == NULL){
      write(false);
    }else{
      write(true);
      write((uint32_t)limits->first);
      write((uint32_t)limits->second);
    }
    
    const std::pair<IQueryDefinition::SetOperation, QueryDefinition>* operations = data.getSetOperation();
    if(operations == NULL){
      write(false);
    }else{
      write(true);
      write(operations->first);
      write(operations->second);
    }
  }
  
  template<class T> inline
  void
  WStreamBuffer::writeAttribute(const coral::CALOpcode& opcode, const bool isnull, const Attribute& attr)
  {
    write(opcode);
    if(!isnull) write( *( (const T*)attr.addressOfData() ) );
  }
 
  template<>
  void
  WStreamBuffer::write(const AttributeList& data)
  {
    STRINGVECTORSIZE size = data.size();
    write(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      const Attribute& attr = data[i];
      const AttributeSpecification& spec = attr.specification();
      write(spec.name());
      const std::type_info& type = spec.type();
      bool isnull = attr.isNull();
      write(isnull);

      if      ( type == typeid(bool) ) writeAttribute<bool>(0x01, isnull, attr);
      else if ( type == typeid(char) ) writeAttribute<unsigned char>(0x02, isnull, attr);
      else if ( type == typeid(unsigned char) ) writeAttribute<unsigned char>(0x03, isnull, attr);
      else if ( type == typeid(signed char) ) writeAttribute<unsigned char>(0x04, isnull, attr);
      else if ( type == typeid(short) ) writeAttribute<uint16_t>(0x05, isnull, attr);
      else if ( type == typeid(unsigned short) ) writeAttribute<uint16_t>(0x06, isnull, attr);
      else if ( type == typeid(int) ) writeAttribute<uint32_t>(0x07, isnull, attr);
      else if ( type == typeid(unsigned int) ) writeAttribute<uint32_t>(0x08, isnull, attr);
      else if ( type == typeid(long) ) writeAttribute<uint64_t>(0x09, isnull, attr);
      else if ( type == typeid(unsigned long) ) writeAttribute<uint64_t>(0x0a, isnull, attr);
      else if ( type == typeid(long long) ) writeAttribute<uint64_t>(0x0b, isnull, attr);
      else if ( type == typeid(unsigned long long) ) writeAttribute<uint64_t>(0x0c, isnull, attr);
      else if ( type == typeid(float) ) writeAttribute<uint32_t>(0x0d, isnull, attr);
      else if ( type == typeid(double) ) writeAttribute<uint64_t>(0x0e, isnull, attr);
      else if ( type == typeid(long double) ) writeAttribute<uint64_t>(0x0f, isnull, attr);
      else if ( type == typeid(std::string) ) writeAttribute<std::string>(0x10, isnull, attr);
      else if ( type == typeid(coral::Blob) ) writeAttribute<coral::Blob>(0x11, isnull, attr);
      else if ( type == typeid(coral::Date) ) writeAttribute<coral::Date>(0x12, isnull, attr);
      else if ( type == typeid(coral::TimeStamp) ) writeAttribute<coral::TimeStamp>(0x13, isnull, attr);
      else throw StreamBufferException("Not supported attribute type", "WStreamBuffer::write<const AttributeList&>");
    }
  }
  
  void
  WStreamBuffer::write(IRowIterator& data)
  {
    size_t wpos = m_write_pos;
    uint32_t size = 0;
    write(size);
    while(data.next())
    {
      write(data.currentRow());
      size++;
    }
    //override the size position
    *((uint32_t*)(m_bytebuffer.data() + wpos)) = size;    
  }
  
  template<>
  void
  WStreamBuffer::write(const coral::Blob& data)
  {
    const void* startpos = data.startingAddress();
    uint64_t size = data.size();
    write(size);
    ::memcpy(m_bytebuffer.data() + m_write_pos, startpos, size );
    m_write_pos = m_write_pos + size;
  }
    
  template<>
  void
  WStreamBuffer::write(const coral::Date& data)
  {
    write((uint16_t)data.year());
    write((uint16_t)data.month());
    write((uint16_t)data.day());
  }
    
  template<>
  void
  WStreamBuffer::write(const coral::TimeStamp& data)
  {
    write((uint16_t)data.year());
    write((uint16_t)data.month());
    write((uint16_t)data.day());
    write((uint16_t)data.hour());
    write((uint16_t)data.minute());
    write((uint16_t)data.second());
    write((uint64_t)data.nanosecond());  
  }

  void
  WStreamBuffer::print()
  {
    std::stringstream msg;

    for(size_t i = 0; i < m_write_pos; i++ ){
      char c = *(m_bytebuffer.data() + i);

      if((c > 126) || (c < 32))
        msg << "[#" << hexstring(&c,1).substr(2,2) << "]";
      else
        msg << "[ " << c << " ]";
      
      if(!((i + 1) % 20)) msg << "\n";

    }
    logger << "StreamBuffer::write::pos " << m_write_pos << "\n"
           << msg.str() << endlog;
  }


  RStreamBuffer::RStreamBuffer(const coral::ByteBuffer& bytebuffer, size_t offset)
    : m_bytebuffer( bytebuffer )
    , m_read_pos( offset )
  {
  }
  
  RStreamBuffer::~RStreamBuffer()
  {
  }
  
  template<>
  void
  RStreamBuffer::read(unsigned char& data)
  {
    data = *(m_bytebuffer.data() + m_read_pos);
    m_read_pos++;
  }

  template<>
  void
  RStreamBuffer::read(bool& data)
  {
    data = *(m_bytebuffer.data() + m_read_pos);
    m_read_pos++;
  }
  
  template<>
  void RStreamBuffer::read(uint16_t& data)
  {
    data = *((uint16_t*)(m_bytebuffer.data() + m_read_pos));
    m_read_pos = m_read_pos + 2;
  }

  template<>
  void
  RStreamBuffer::read(uint32_t& data)
  {
    data = *((uint32_t*)(m_bytebuffer.data() + m_read_pos));
    m_read_pos = m_read_pos + 4;
  }

  template<>
  void
  RStreamBuffer::read(uint64_t& data)
  {
    data = *((uint64_t*)(m_bytebuffer.data() + m_read_pos));
    m_read_pos = m_read_pos + 8;
  }

  template<>
  void
  RStreamBuffer::read(std::string& data)
  {
    data.clear();
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      unsigned char c;
      read(c);
      data += c;
    }
  }

  template<>
  void
  RStreamBuffer::read(std::vector<std::string>& data)
  {
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      std::string s;
      read(s);
      data.push_back(s);
    }
  }

  template<>
  void
  RStreamBuffer::read(std::set<std::string>& data)
  {
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      std::string s;
      read(s);
      data.insert(s);
    }
  }

  template<>
  void
  RStreamBuffer::read(std::vector< std::pair<std::string, std::string> >& data)
  {
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      std::string s1;
      read(s1);
      std::string s2;
      read(s2);
      data.push_back(std::pair<std::string, std::string>(s1, s2));
    }
  }
    
  template<>
  void RStreamBuffer::read(TableDescription& data)
  {
    //read the table name
    std::string tableName;
    read(tableName);
    data.setName(tableName);
    //read the table type
    std::string tableType;
    read(tableType);
    data.setType(tableType);
    //read the table space name
    std::string tableSpaceName;
    read(tableSpaceName);
    data.setTableSpaceName(tableSpaceName);
    //get the columns
    STRINGVECTORSIZE columns;
    read(columns);
    for(STRINGVECTORSIZE i = 0; i < columns; i++ ){
      std::string cname;
      read(cname);
      std::string ctype;
      read(ctype);
      uint32_t csize;
      read(csize);
      bool cfixedsize;
      read(cfixedsize);
      std::string cspacename;
      read(cspacename);
      data.insertColumn( cname, ctype, csize, cfixedsize, cspacename );
    }
    //set the unique constraints
    STRINGVECTORSIZE constraints;
    read(constraints);
    for(STRINGVECTORSIZE i = 0; i < constraints; i++ ){
      std::vector<std::string> cnames;
      read(cnames);
      std::string cname;
      read(cname);
      std::string cspacename;
      read(cspacename);
      data.setUniqueConstraint( cnames, cname, true, cspacename );
    }    
    //set primary key
    bool isprikey;
    read(isprikey);
    if(isprikey){
      std::vector<std::string> cnames;
      read(cnames);
      std::string cspacename;
      read(cspacename);
      data.setPrimaryKey( cnames, cspacename );
    }    
    //set foreign keys
    STRINGVECTORSIZE foreignkeys;
    read(foreignkeys);
    for(STRINGVECTORSIZE i = 0; i < foreignkeys; i++ ){
      std::vector<std::string> cnames;
      read(cnames);      
      std::string cname;
      read(cname);
      std::string crtablename;
      read(crtablename);
      std::vector<std::string> crcolumns;
      read(crcolumns);
      data.createForeignKey( cname, cnames, crtablename, crcolumns );
    }    
    //set indexes
    STRINGVECTORSIZE indexes;
    read(indexes);
    for(STRINGVECTORSIZE i = 0; i < indexes; i++ ){
      std::vector<std::string> cnames;
      read(cnames);      
      std::string cname;
      read(cname);
      bool cisunique;
      read(cisunique);
      std::string cspacename;
      read(cspacename);
      data.createIndex( cname, cnames, cisunique, cspacename );
    }    
  }
  
  template<>
  void RStreamBuffer::read(std::map< std::string, TableDescription >& data)
  {
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      std::string s;
      read(s);
      TableDescription td;
      read(td);
      data.insert(std::pair< std::string, TableDescription >(s, td));
    }    
  }
  
  template<>
  void RStreamBuffer::read(QueryDefinition& data)
  {
    //set some iterators
    std::vector< std::pair<std::string,std::string> >::iterator i;
    std::vector< std::string >::iterator j;
    
    std::string schemaname;
    read(schemaname);
    data.setSchemaName( schemaname );
    //default values
    bool distinct;
    read(distinct);
    if(distinct) data.setDistinct();
    
    std::vector< std::pair<std::string,std::string> > outputlist;
    read(outputlist);
    for(i = outputlist.begin(); i != outputlist.end(); i++ ) data.addToOutputList(i->first, i->second);

    std::vector< std::pair<std::string,std::string> > tablelist;
    read(tablelist);
    for(i = tablelist.begin(); i != tablelist.end(); i++ ) data.addToTableList(i->first, i->second);
    
    std::string condition;
    read(condition);
    
    AttributeList conditionData;
    read(conditionData);
    data.setCondition( condition, conditionData );
    
    std::string groupby;
    read(groupby);
    data.groupBy(groupby);
    
    std::vector< std::string > orderlist;
    read(orderlist);
    for(j = orderlist.begin(); j != orderlist.end(); j++ ) data.addToOrderList(*j);
    
    //special values    
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE k = 0; k < size; k++ )
    {
      std::string alias;
      read(alias);
      QueryDefinition& qd = dynamic_cast<QueryDefinition&>(data.defineSubQuery( alias ));
      read(qd);
    }
    
    bool limits;
    read(limits);
    if(limits){
      uint32_t l01;
      read(l01);
      uint32_t l02;
      read(l02);
      data.limitReturnedRows( l01, l02 );
    }
    
    bool operations;
    read(operations);
    if(operations){
      IQueryDefinition::SetOperation setopt;
      read(setopt);
      QueryDefinition& qd = dynamic_cast<QueryDefinition&>(data.applySetOperation( setopt ));
      read(qd);
    }
  }
    
  template<>
  void RStreamBuffer::read(IQueryDefinition::SetOperation& data)
  {
    unsigned char res = 0x00;
    if(res == 0x01)
      data = IQueryDefinition::Union;
    else if(res == 0x02)
      data = IQueryDefinition::Minus;
    else if(res == 0x03)
      data = IQueryDefinition::Intersect;
  }
  
  template<class T> inline
  void
  RStreamBuffer::readAttribute(AttributeList& data, const bool isnull, const std::string& name, const std::type_info& type)
  {
    data.extend( name, type);
    Attribute& attr = data[data.size() - 1];
    
    if(isnull) attr.setNull( true );
    else read( *( (T*)attr.addressOfData() ) );
  }

  template<>
  void RStreamBuffer::read(AttributeList& data)
  {
    STRINGVECTORSIZE size;
    read(size);
    for(STRINGVECTORSIZE i = 0; i < size; i++ ){
      //get the attribute name
      std::string name;
      read(name);
      //get is null
      bool isnull;
      read(isnull);
      //get the type of attribute
      coral::CALOpcode topcode;
      read(topcode);
      //check all opcodes and do the transformation
      if      ( topcode == 0x01 ) readAttribute<bool>(data, isnull, name, typeid(bool));
      else if ( topcode == 0x02 ) readAttribute<unsigned char>(data, isnull, name, typeid(char));
      else if ( topcode == 0x03 ) readAttribute<unsigned char>(data, isnull, name, typeid(unsigned char));
      else if ( topcode == 0x04 ) readAttribute<unsigned char>(data, isnull, name, typeid(signed char));
      else if ( topcode == 0x05 ) readAttribute<uint16_t>(data, isnull, name, typeid(short));
      else if ( topcode == 0x06 ) readAttribute<uint16_t>(data, isnull, name, typeid(unsigned short));
      else if ( topcode == 0x07 ) readAttribute<uint32_t>(data, isnull, name, typeid(int));
      else if ( topcode == 0x08 ) readAttribute<uint32_t>(data, isnull, name, typeid(unsigned int));
      else if ( topcode == 0x09 ) readAttribute<uint64_t>(data, isnull, name, typeid(long));
      else if ( topcode == 0x0a ) readAttribute<uint64_t>(data, isnull, name, typeid(unsigned long));
      else if ( topcode == 0x0b ) readAttribute<uint64_t>(data, isnull, name, typeid(long long));
      else if ( topcode == 0x0c ) readAttribute<uint64_t>(data, isnull, name, typeid(unsigned long long));
      else if ( topcode == 0x0d ) readAttribute<uint32_t>(data, isnull, name, typeid(float));
      else if ( topcode == 0x0e ) readAttribute<uint64_t>(data, isnull, name, typeid(double));
      else if ( topcode == 0x0f ) readAttribute<uint64_t>(data, isnull, name, typeid(long double));
      else if ( topcode == 0x10 ) readAttribute<std::string>(data, isnull, name, typeid(std::string));
      else if ( topcode == 0x11 ) readAttribute<coral::Blob>(data, isnull, name, typeid(coral::Blob));
      else if ( topcode == 0x12 ) readAttribute<coral::Date>(data, isnull, name, typeid(coral::Date));
      else if ( topcode == 0x13 ) readAttribute<coral::TimeStamp>(data, isnull, name, typeid(coral::TimeStamp));
      else throw StreamBufferException("Not supported attribute opcode", "RStreamBuffer::read<AttributeList&>");
    }
    //std::stringstream msg;
    //data.toOutputStream( msg );
    //std::cout << "RStreamBuffer::read(AL&): " << msg.str() << std::endl;
  }
    
  template<>
  void
  RStreamBuffer::read(coral::Blob& data)
  {
    uint64_t size;
    read(size);
    data.resize( size );
    unsigned char* p = static_cast<unsigned char*>( data.startingAddress() );
    ::memcpy(p, m_bytebuffer.data() + m_read_pos, size );
    m_read_pos = m_read_pos + size;
  }
    
  template<>
  void
  RStreamBuffer::read(coral::Date& data)
  {
    uint16_t year;
    read(year);
    uint16_t month;
    read(month);
    uint16_t day;
    read(day);
    data = coral::Date(year, month, day);
  }
    
  template<>
  void
  RStreamBuffer::read(coral::TimeStamp& data)
  {
    uint16_t year;
    read(year);
    uint16_t month;
    read(month);
    uint16_t day;
    read(day);
    uint16_t hour;
    read(hour);
    uint16_t minute;
    read(minute);
    uint16_t second;
    read(second);
    uint64_t nsecond;
    read(nsecond);
    data = coral::TimeStamp( year, month,  day, hour, minute, second, nsecond);
  }

  void
  RStreamBuffer::print()
  {
    std::stringstream msg;

    for(size_t i = 0; i < m_read_pos; i++ ){
    
      char c = *(m_bytebuffer.data() + i);
      
      if((c > 126) || (c < 32))
        msg << "[#" << hexstring(&c,1).substr(2,2) << "]";
      else
        msg << "[ " << c << " ]";
      
      if(!((i + 1) % 20)) msg << "\n";
    }
    logger << "StreamBuffer::read::pos " << m_read_pos << "\n"
           << msg.str() << endlog;
  }


} }
