#ifndef CONNECTIONSERVICE_CONNECTIONPARAMS_H
#define CONNECTIONSERVICE_CONNECTIONPARAMS_H

#include "RelationalAccess/AccessMode.h"
#include <map>
#include <string>

namespace coral {

  namespace ConnectionService {

    /// encapsulation of the parameters describing the connection request 
    class ConnectionParams {

      public:

      /// constructor 
      ConnectionParams( const std::string& connectionString );
      
      /// constructor 
      ConnectionParams( const std::string& connectionString,
                        const std::string& serviceName,
                        const std::string& technologyName,
                        const std::string& schemaName,
                        const std::string& userName,
                        const std::string& password );

      /// destructor
      virtual ~ConnectionParams();

      /// copy constructor
      ConnectionParams( const ConnectionParams& rhs);
      
      /// assignment operator
      ConnectionParams& operator=(const ConnectionParams& rhs);

      /// setter for db parameters
      void setDbParams(const std::string& serviceName,
                       const std::string& technologyName,
                       const std::string& schemaName);

      /// setter for authentication parameters
      void setAuthenticationParams(const std::string& userName,
                                   const std::string& password );
      
      /// returns the user associated to the connection request
      const std::string& userName() const;

      /// returns the password associated to the connection request
      const std::string& password() const;

       /// returns the connection string associated to the connection request
      const std::string& connectionString() const;

       /// returns the service connection string associated to the connection request
      const std::string& serviceKey() const;

      /// returns the service name associated to the connection request
      const std::string& serviceName() const;
      
       /// returns the technology name associated to the connection request
      const std::string& technologyName() const;

      /// returns the schema name associated to the connection request
      const std::string& schemaName() const;

      /// returns the reference to the service parameter map
      std::map<std::string,std::string>& serviceParameters();
      
      /// returns the reference to the service parameter map
      const std::map<std::string,std::string>& serviceParameters() const;
      
      private:

      /// the stored connectionString
      std::string m_connectionString;

      /// the stored service name
      std::string m_serviceName;

      /// the stored service name
      std::string m_serviceKey;

      /// the stored technology name
      std::string m_technologyName;

      /// the stored schema name
      std::string m_schemaName;

      /// the username
      std::string m_userName;
      
      /// the password
      std::string m_password;

      /// the service parameter map
      std::map<std::string, std::string> m_serviceParameters;

    };
    
  }
  
}

inline
coral::ConnectionService::ConnectionParams::ConnectionParams( const std::string& connectionString ):
  m_connectionString( connectionString ),
  m_serviceName( "" ),
  m_serviceKey( "" ),
  m_technologyName( "" ),
  m_schemaName( "" ),
  m_userName( "" ),
  m_password( "" ),
  m_serviceParameters(){
}

inline
coral::ConnectionService::ConnectionParams::ConnectionParams( const std::string& connectionString,
                                                              const std::string& serviceName,
                                                              const std::string& technologyName,
                                                              const std::string& schemaName,
                                                              const std::string& userName,
                                                              const std::string& password):
  m_connectionString( connectionString ),
  m_serviceName( serviceName ),
  m_serviceKey( technologyName ),
  m_technologyName( technologyName ),
  m_schemaName( schemaName ),
  m_userName( userName ),
  m_password( password ),
  m_serviceParameters(){
  m_serviceKey.append(":").append(m_serviceName);
}

/// destructor
inline
coral::ConnectionService::ConnectionParams::~ConnectionParams(){}

/// copy constructor
inline
coral::ConnectionService::ConnectionParams::ConnectionParams( const ConnectionParams& rhs):
  m_connectionString( rhs.m_connectionString ),
  m_serviceName(  rhs.m_serviceName ),
  m_serviceKey( rhs.m_serviceKey ),
  m_technologyName(  rhs.m_technologyName ),
  m_schemaName(  rhs.m_schemaName ),
  m_userName( rhs.m_userName ),
  m_password( rhs.m_password ),
  m_serviceParameters( rhs.m_serviceParameters ){}

/// assignment operator
inline
coral::ConnectionService::ConnectionParams&
coral::ConnectionService::ConnectionParams::operator=(const ConnectionParams& rhs){
  m_connectionString = rhs.m_connectionString;
  m_serviceName = rhs.m_serviceName;
  m_serviceKey = rhs.m_serviceKey;
  m_technologyName = rhs.m_technologyName;
  m_schemaName = rhs.m_schemaName;
  m_userName = rhs.m_userName;
  m_password = rhs.m_password;
  m_serviceParameters = rhs.m_serviceParameters;
  return *this;
}

inline
void
coral::ConnectionService::ConnectionParams::setDbParams(const std::string& serviceName,
                                                        const std::string& technologyName,
                                                        const std::string& schemaName){
  m_serviceName = serviceName;
  m_serviceKey = technologyName;
  m_serviceKey.append(":").append(m_serviceName);
  m_technologyName = technologyName;
  m_schemaName = schemaName;
}

inline
void
coral::ConnectionService::ConnectionParams::setAuthenticationParams(const std::string& userName,
                                                                    const std::string& password ){
  m_userName = userName;
  m_password = password;
}

/// returns the user associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::userName() const {
  return m_userName;
}

/// returns the user password associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::password() const {
  return m_password;
}

/// returns the service connection name associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::connectionString() const {
  return m_connectionString;
}

/// returns the service connection string associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::serviceName() const{
  return m_serviceName;
}

/// returns the service connection string associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::serviceKey() const{
  return m_serviceKey;
}

/// returns the service connection string associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::technologyName() const{
  return m_technologyName;
}

/// returns the schema name associated to the connection request
inline
const std::string&
coral::ConnectionService::ConnectionParams::schemaName() const {
  return m_schemaName;
}

/// returns the reference to the service parameter map
inline
std::map<std::string,std::string>&
coral::ConnectionService::ConnectionParams::serviceParameters(){
  return m_serviceParameters;
}

/// returns the reference to the service parameter map
inline
const std::map<std::string,std::string>&
coral::ConnectionService::ConnectionParams::serviceParameters() const {
  return m_serviceParameters;
}

#endif
