/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>

#include "CoralMessaging/Tcp/TcpSocketFactory.h"
#include "CoralMessaging/Insulation/SessionException.h"

#include "Domain.h"
#include "Connection.h"
#include "Error.h"
#include "Logger.h"
#include "ConnectionStub.h"
#include "TcpConnectionStubFactory.h"

START_CORALACCESS_NAMESPACE

Connection::Connection(Domain& domain, const std::string& url, const uint32_t clientID)
	: coral::MonitorObject(),
	  _domain(domain),
	  _serverVersion(false, ""),
	  _clientID(clientID),
	  _stubFactory(new TcpConnectionStubFactory),
	  _isSessionOpen(false)
{
	typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
	// Separator between the host name and the port
	boost::char_separator<char> sep(":");
	tokenizer tk(url, sep);
	std::vector<std::string> tokens(tk.begin(), tk.end());

	if (tokens.size() != 2)
		EXCEPTION_THROW(Err_Custom(std::string("Mal-formed server URL: ") + url));

	_hostName = *tokens.begin();
	const std::string& portStr = *(tokens.begin() + 1);

	try {
		_port = boost::lexical_cast<unsigned int>(portStr);
	}
	catch (boost::bad_lexical_cast&) {
		EXCEPTION_THROW(Err_Custom(std::string("Malformed server port: ") + portStr));
	}

	_serverName = _hostName + ":" + boost::lexical_cast<std::string>(_port);
}

//-----------------------------------------------------------------------------

Connection::~Connection()
{
MONITOR_START_CRITICAL
	// EMPTY
MONITOR_END_CRITICAL
}

//-----------------------------------------------------------------------------

void Connection::connect()
{
MONITOR_START_CRITICAL
	if (_stub.get())
		LOGGER_NEWLOG(coral::Warning) << "A connection is already present for \"" << _hostName << ":" << _port << "\"" << commit;
	else
		_stub.reset(_stubFactory->create(_domain.name(), _hostName, _port, _clientID));
MONITOR_END_CRITICAL
}

//-----------------------------------------------------------------------------

bool Connection::isConnected(bool probePhysicalConnection)
{
MONITOR_START_CRITICAL
	return _isConnected(probePhysicalConnection);
MONITOR_END_CRITICAL
}

//-----------------------------------------------------------------------------

void Connection::disconnect()
{
MONITOR_START_CRITICAL
	_serverVersion.first = false;

	if (_stub.get())
		_stub.reset(NULL);
MONITOR_END_CRITICAL
}

//-----------------------------------------------------------------------------

std::string Connection::serverVersion() const
{
MONITOR_START_CRITICAL
	if (!_serverVersion.first)
		throw ServerException(_domain.name(), "Unknown server version, session not started yet", __FUNCTION__);

 	return _serverVersion.second;
MONITOR_END_CRITICAL
}

//-----------------------------------------------------------------------------

const std::string& Connection::serverName() const
{
MONITOR_START_CRITICAL
	return _serverName;
MONITOR_END_CRITICAL
}

coral::ISession* Connection::newSession(const std::string& dbUri, coral::AccessMode mode) const
{
MONITOR_START_CRITICAL
	if (!_isConnected())
		throw ConnectionNotActiveException(_domain.name(), __FUNCTION__);

	if (_isSessionOpen)
		EXCEPTION_THROW(Err_Custom("CoralAccess cannot allow multiple session for the same connection."));

	_isSessionOpen = true;
	return new Session(const_cast<Connection&>(*this), dbUri, mode);
MONITOR_END_CRITICAL
}

IConnectionStub& Connection::getStub()
{
MONITOR_START_CRITICAL
	CORALACCESS_DEBUG_ASSERT(_stub.get());
	return *_stub;
MONITOR_END_CRITICAL
}

Domain& Connection::domain()
{
	return _domain;
}

ITypeConverter& Connection::typeConverter()
{
	return _typeConverter;
}

void Connection::_setServerVersion(const std::string& serverVersion)
{
MONITOR_START_CRITICAL
	_serverVersion.first = true;
	_serverVersion.second = serverVersion;
MONITOR_END_CRITICAL
}

bool Connection::_isConnected(bool probePhysicalConnection) const
{
	if (_stub.get())
		return probePhysicalConnection ? _stub->connected() : true;

	return false;
}

void Connection::_sessionClosed()
{
MONITOR_START_CRITICAL
	_isSessionOpen = false;
MONITOR_END_CRITICAL
}

END_CORALACCESS_NAMESPACE
