/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <boost/foreach.hpp>

#include "CoralMessaging/Insulation/Exceptions.h"
#include "CoralMessaging/CAL/MsgObj/QueryRepresentation.h"
#include "CoralMessaging/CAL/MsgObj/AttributeListRepresentation.h"

#include "Error.h"
#include "Session.h"
#include "Domain.h"
#include "Connection.h"
#include "Pointers.h"
#include "QueryDefinition.h"

START_CORALACCESS_NAMESPACE

using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::QueryRepresentationTypes;
using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::AttributeListRepresentation;

QueryDefinition::QueryDefinition(Session& session)
	: MonitorObject(),
	  _session(&session),
	  _isExecuted(false)
{
	// EMPTY
}

//replace
void QueryDefinition::setDistinct()
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::setDistinct repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type;
	_eqr.replaceMethodCall(repr);
MONITOR_END_CRITICAL
}

//insert
void QueryDefinition::addToOutputList(const std::string& expression, std::string alias)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::addToOutputList repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(expression, alias);
	_eqr.insertMethodCall(repr);
MONITOR_END_CRITICAL
}

// replace
void QueryDefinition::setCondition(const std::string& condition, const AttributeList& inputData)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::setCondition repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(condition, AttributeListRepresentation(inputData, true));
	_eqr.replaceMethodCall(repr);
MONITOR_END_CRITICAL
}

// replace
void QueryDefinition::groupBy(const std::string& expression)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::groupBy repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(expression);
	_eqr.replaceMethodCall(repr);
MONITOR_END_CRITICAL
}

// insert
void QueryDefinition::addToOrderList(const std::string& expression)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::addToOrderList repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(expression);
	_eqr.insertMethodCall(repr);
MONITOR_END_CRITICAL
}

// replace
void QueryDefinition::limitReturnedRows(int maxRows, int offset)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	typedef QueryRepresentationTypes::limitReturnedRows repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(maxRows, offset);
	_eqr.replaceMethodCall(repr);
MONITOR_END_CRITICAL
}

// replace
IQueryDefinition& QueryDefinition::applySetOperation(SetOperation operationType)
{
MONITOR_START_CRITICAL
	_checkIfExecuted();
	_setQuery.first = operationType;
	_setQuery.second.reset(_getSimilar());
	return *_setQuery.second;
MONITOR_END_CRITICAL
}

bool QueryDefinition::operator == (const QueryDefinition& x) const
{
MONITOR_START_CRITICAL
	return this == &x ? true : _eqr == x._eqr;
MONITOR_END_CRITICAL
}

IQueryDefinition& QueryDefinition::_defineSubQuery(const std::string& alias, QueryDefinition* qd)
{
	DEBUG_ASSERT_PTR(qd);
	_subQueries.push_back(std::make_pair(alias, qd));
	return *qd;
}

void QueryDefinition::_addToTableList(const std::string& tableName, std::string alias)
{
	typedef QueryRepresentationTypes::addToTableList repr_type;
	repr_type repr;
	repr.ptr = new repr_type::type(tableName, alias);
	_eqr.insertMethodCall(repr);
}

void QueryDefinition::_executeQueryDefinition()
{
	// Add the subqueries
	typedef _SubQueriesType::iterator ItrType;
	ItrType end = _subQueries.end();

	for (ItrType qd = _subQueries.begin(); qd != _subQueries.end(); ++qd) {
		// Execute recursively the subqueries
		DEBUG_ASSERT_PTR(qd->second);
		qd->second->_executeQueryDefinition();
		typedef QueryRepresentationTypes::defineSubQuery repr_type;
		repr_type repr;
		repr.ptr = new repr_type::type(qd->first, qd->second->_eqr);
		_eqr.insertMethodCall(repr);
	}
	// Add the set operation
	if (_setQuery.second.get()) {
		typedef QueryRepresentationTypes::applySetOperation repr_type;
		repr_type repr;
		repr.ptr = new repr_type::type(static_cast<uint8_t>(_setQuery.first), _setQuery.second->_eqr);
		_eqr.replaceMethodCall(repr);
	}
}

void QueryDefinition::_checkIfExecuted() const
{
	DEBUG_ASSERT_PTR(_session);

	if (_isExecuted)
		throw coral::QueryExecutedException(_session->connection().domain().name(), __FUNCTION__);
}

void QueryDefinition::_setExecuted()
{
	_isExecuted = true;
}

// -------------- SingleTableQueryDefinition -----------------------------------

SingleTableQueryDefinition::SingleTableQueryDefinition(Session& session, const std::string& tableName)
	: QueryDefinition(session),
	  _tableName(tableName)
{
	_addToTableList(tableName);
}

//insert
IQueryDefinition& SingleTableQueryDefinition::defineSubQuery(const std::string& alias)
{
MONITOR_START_CRITICAL
	return _defineSubQuery(alias, new SingleTableQueryDefinition(*_session, _tableName));
MONITOR_END_CRITICAL
}

void SingleTableQueryDefinition::addToTableList(const std::string&, std::string)
{
MONITOR_START_CRITICAL
	DEBUG_ASSERT_PTR(_session);
    throw coral::QueryException(_session->connection().domain().name(),
                                "Method cannot be invoked on queries on a single table",
                                __FUNCTION__);
MONITOR_END_CRITICAL
}

QueryDefinition* SingleTableQueryDefinition::_getSimilar() const
{
	DEBUG_ASSERT_PTR(_session);
	return new SingleTableQueryDefinition(*_session, _tableName);
}

// ------------------------ MultiTableQueryDefinition --------------------------

MultiTableQueryDefinition::MultiTableQueryDefinition(Session& session)
	: QueryDefinition(session)
{
	// EMPTY
}

//insert
IQueryDefinition& MultiTableQueryDefinition::defineSubQuery(const std::string& alias)
{
MONITOR_START_CRITICAL
	DEBUG_ASSERT_PTR(_session);
	return _defineSubQuery(alias, new MultiTableQueryDefinition(*_session));
MONITOR_END_CRITICAL
}

void MultiTableQueryDefinition::addToTableList(const std::string& tableName, std::string alias)
{
MONITOR_START_CRITICAL
	_addToTableList(tableName, alias);
MONITOR_END_CRITICAL
}

QueryDefinition* MultiTableQueryDefinition::_getSimilar() const
{
	DEBUG_ASSERT_PTR(_session);
	return new MultiTableQueryDefinition(*_session);
}

END_CORALACCESS_NAMESPACE
