/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include "NotImplemented.h"
#include "Session.h"
#include "ViewFactory.h"
#include "Connection.h"
#include "Domain.h"
#include "Schema.h"

START_CORALACCESS_NAMESPACE

Schema::Schema(Session& session, const std::string& schemaName)
  : coral::MonitorObject(),
    _session(session),
    _schemaProperties(schemaName),
    _sb(session),
    _tableListFresh(false),
    _viewListFresh(false)
{
	// EMPTY
}

Schema::~Schema()
{
MONITOR_START_CRITICAL
	// EMPTY
MONITOR_END_CRITICAL
}

std::string Schema::schemaName() const
{
MONITOR_START_CRITICAL
	return _schemaProperties.schemaName();
MONITOR_END_CRITICAL
}

Schema::NameListType Schema::listTables() const
{
MONITOR_START_CRITICAL
	return _listTables();
MONITOR_END_CRITICAL
}

void Schema::setTableListStale()
{
MONITOR_START_CRITICAL
  	_tableListFresh = false;
MONITOR_END_CRITICAL
}

bool Schema::existsTable(const std::string& tableName)  const
{
MONITOR_START_CRITICAL
	return _existsTable(tableName);
MONITOR_END_CRITICAL
}

void Schema::dropTable(const std::string& tableName)
{
MONITOR_START_CRITICAL
	_throwIfTableNotExists(tableName);
	_dropExistingTable(tableName);
MONITOR_END_CRITICAL
}

void Schema::dropIfExistsTable(const std::string& tableName)
{
MONITOR_START_CRITICAL
	if (_existsTable(tableName)) _dropExistingTable(tableName);
MONITOR_END_CRITICAL
}

coral::ITable& Schema::createTable(const coral::ITableDescription& description)
{
MONITOR_START_CRITICAL
	// Get the name of the table
	const std::string tableName = description.name();
	_throwIfTableExists(tableName);
	_sb.createTable(description);
    // This call makes sure the table handle registry gets in sync
    return _tableHandle(tableName);
MONITOR_END_CRITICAL
}


coral::ITable& Schema::tableHandle(const std::string& tableName)
{
MONITOR_START_CRITICAL
	return _tableHandle(tableName);
MONITOR_END_CRITICAL
}

void Schema::truncateTable(const std::string& tableName)
{
MONITOR_START_CRITICAL
	_throwIfTableNotExists(tableName);
	_sb.truncateTable(tableName);
MONITOR_END_CRITICAL
}

void Schema::callProcedure(const std::string&,	const coral::AttributeList& /*inputArguments*/)
{
	// TODO
	throw NotImplemented(__FUNCTION__);
}

bool Schema::existsView(const std::string& viewName) const
{
MONITOR_START_CRITICAL
	return _existsView(viewName);
MONITOR_END_CRITICAL
}

void Schema::dropView(const std::string& viewName)
{
MONITOR_START_CRITICAL
	_throwIfViewNotExists(viewName);
	_dropExistingView(viewName);
MONITOR_END_CRITICAL
}

void Schema::dropIfExistsView( const std::string& viewName )
{
MONITOR_START_CRITICAL
  if (this->_existsView(viewName)) this->_dropExistingView(viewName);
MONITOR_END_CRITICAL
}

Schema::NameListType Schema::listViews() const
{
MONITOR_START_CRITICAL
	return this->_listViews();
MONITOR_END_CRITICAL
}

coral::IView& Schema::viewHandle(const std::string&)
{
MONITOR_START_CRITICAL
	// TODO
	throw NotImplemented(__FUNCTION__);
	// I know...
	coral::IView* p = 0;
	return *p;
MONITOR_END_CRITICAL
}

IQuery* Schema::newQuery() const
{
MONITOR_START_CRITICAL
	 return new MultiTableQuery(_session, _schemaProperties.schemaName());
MONITOR_END_CRITICAL
}

IViewFactory* Schema::viewFactory()
{
MONITOR_START_CRITICAL
	return new ViewFactory(_session);
MONITOR_END_CRITICAL
}

void Schema::_dropExistingTable(const std::string& tableName)
{
	DEBUG_ASSERT(_schemaProperties.isTableExists(tableName));
	_sb.dropTable(tableName);
    // Update schema registry
	_schemaProperties.eraseTable(tableName);
}

Schema::NameListType Schema::_listTables() const
{
    _tableListFresh = true;
    NameListType list = _sb.listTables();
    _schemaProperties.refreshTableNameList(list);
	return list;
}

bool Schema::_existsTable(const std::string& tableName) const
{
	if (!_tableListFresh)
		_listTables();

	return _schemaProperties.isTableExists(tableName);
}

Schema::NameListType Schema::_listViews() const
{
    _viewListFresh = true;
    NameListType viewList = _sb.listViews();
    _schemaProperties.refreshViewNameList(viewList);
	return viewList;
}

coral::ITable& Schema::_tableHandle(const std::string& tableName)
{
	_throwIfTableNotExists(tableName);
	// Go to the schema properties to pick up the table handle from the cache
	Table* table = _schemaProperties.findTable(tableName);
	return table == NULL ? *_schemaProperties.newTable(tableName, _session) : *table;
}

/// Check if a table with this name already exists
void Schema::_throwIfTableExists(const std::string& tableName)
{
	if (this->_existsTable(tableName))
		throw coral::TableAlreadyExistingException(_session.connection().domain().name(), (_schemaProperties.schemaName() +"." + tableName));
}

/// Check if a table with this name already exists
void Schema::_throwIfTableNotExists(const std::string& tableName)
{
	if (!this->_existsTable(tableName))
	    throw coral::TableNotExistingException(_session.connection().domain().name(), (_schemaProperties.schemaName() +"." + tableName), __FUNCTION__);
}

/// Check if a view with this name already exists
void Schema::_throwIfViewNotExists(const std::string& viewName)
{
	if (!_existsView(viewName))
		throw coral::ViewNotExistingException(_session.connection().domain().name(), viewName, __FUNCTION__);
}

void Schema::_dropExistingView(const std::string& viewName)
{
	DEBUG_ASSERT(_schemaProperties.isViewExists(viewName));
  	_sb.dropView(viewName);
  	_schemaProperties.eraseView(viewName);
}

bool Schema::_existsView(const std::string& viewName) const
{
	if (!_viewListFresh)
		_listViews(); // Update the list of views

	return _schemaProperties.isViewExists(viewName);
}

END_CORALACCESS_NAMESPACE
