/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include "CoralCommon/URIParser.h"
#include "CoralMessaging/CAL/CALMessages.h"
#include "CoralMessaging/Insulation/SessionException.h"

#include "Logger.h"
#include "Schema.h"
#include "Transaction.h"
#include "NotImplemented.h"
#include "Connection.h"
#include "Domain.h"
#include "Session.h"

START_CORALACCESS_NAMESPACE

Session::Session(Connection& connection, const std::string& dbUri, coral::AccessMode mode)
	: coral::IDevSession(connection),
	  coral::MonitorObject(),
	  _connection(connection),
	  _mode(mode),
	  _isUserSessionActive(false),
	  _transaction(connection),
	  _dbUri(dbUri),
    m_remoteTechnologyName( "Unknown technology for DB server connected via CORAL server" ), // TODO
    m_remoteServerVersion( "Unknown version for DB server connected via CORAL server" ) // TODO
{
	coral::URIParser parser;
	parser.setURI(dbUri);
	_schema.reset(new Schema(const_cast<Session&>(*this), parser.databaseOrSchemaName()));
}

Session::~Session()
{
MONITOR_START_CRITICAL
	_endUserSession();
MONITOR_END_CRITICAL
}

coral::IMonitoring& Session::monitoring()
{
	return _monitorController;
}

std::string Session::nominalSchemaNameForConnection(const std::string& connectionName) const
{
	coral::URIParser parser;
	parser.setURI(connectionName);
	return parser.databaseOrSchemaName();
}

bool Session::isUserSessionActive() const
{
MONITOR_START_CRITICAL
	return _isUserSessionActive;
MONITOR_END_CRITICAL
}

coral::ITransaction& Session::transaction()
{
	return _transaction;
}

coral::ISchema& Session::nominalSchema()
{
MONITOR_START_CRITICAL
	return _nominalSchema();
MONITOR_END_CRITICAL
}

coral::ISchema& Session::schema(const std::string& schemaName)
{
MONITOR_START_CRITICAL
	if (!_connection.isConnected())
		throw coral::ConnectionNotActiveException(_connection.domain().name());

	if (schemaName == _schema->schemaName())
		return this->_nominalSchema();

	Schema* schema = _schreg.find(schemaName);

	if (!schema) {
		// The call below throws coral::InvalidSchemaNameException if the schema does not exist
		if (_sl.schemaExists(const_cast<Session&>(*this), schemaName)) {
			schema = new Schema(const_cast<Session&>(*this), schemaName);
			_schreg.add(schemaName, schema);
		}
	}

	return *schema;
MONITOR_END_CRITICAL
}

Connection& Session::connection()
{
	return _connection;
}

void Session::startUserSession(const std::string& userName, const std::string& password)
{
MONITOR_START_CRITICAL
	LOGGER_NEWLOG(coral::Info) << "Starting user session...";
	coral::CoralMessaging::CAL::MsgObj::Connection connection(_mode == coral::ReadOnly, _dbUri, userName, password);
	char buffer[255];
	gethostname(buffer, sizeof(buffer));
	coral::CoralMessaging::CAL::MsgObj::MonitoringInformation monitor(getpid(), buffer);
	IMessagePtr reply = _connection.getStub().sendRecv(coral::CoralMessaging::CAL::Connect(connection, monitor), coral::CoralProtocol::Opcode::E_ConnectReply);
	theLogger() << "successfull." << commit;
	_isUserSessionActive = true;
	using coral::CoralMessaging::CAL::ConnectReply;
	ConnectReply& r = dynamic_cast<ConnectReply&>(*reply);
	_connection._setServerVersion(r.serverVersion());
MONITOR_END_CRITICAL
}

void Session::endUserSession()
{
MONITOR_START_CRITICAL
	_endUserSession();
MONITOR_END_CRITICAL
}

Schema* Session::_SchemaRegistry::find(const std::string& schemaName) const
{
	_ContainerType::const_iterator sch = _cont.find(schemaName);
	return sch == _cont.end() ? NULL : const_cast<Schema*>(sch->second);
}

void Session::_SchemaRegistry::add(const std::string& schemaName, Schema* sch)
{
	// cast is necessary, as the insert method's key parameter is a non-const reference
	_cont.insert(const_cast<std::string&>(schemaName), sch);
}

void Session::_endUserSession()
{
	if (_isUserSessionActive) {
		// Send close session...
		LOGGER_NEWLOG(coral::Info) << "Closing user session..." << commit;
		_connection.getStub().sendRecv(coral::CoralMessaging::CAL::CloseSession(), coral::CoralProtocol::Opcode::E_OK);
		_isUserSessionActive = false;
		_connection._sessionClosed();
	}
}

coral::ISchema& Session::_nominalSchema()
{
	if (!_isUserSessionActive)
		throw coral::ConnectionNotActiveException(_connection.domain().name());

	return *_schema;
}

std::string 
coral::CoralAccess::Session::remoteTechnologyName() const
{
  return m_remoteTechnologyName;
}

std::string 
coral::CoralAccess::Session::remoteServerVersion() const
{
  return m_remoteServerVersion;
}

END_CORALACCESS_NAMESPACE
