/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include "base.h"
#include <boost/foreach.hpp>

#include "CoralMessaging/Utils/Error.h"
#include "CoralMessaging/CAL/CALMessages.h"

#include "Connection.h"
#include "ConnectionStub.h"
#include "Session.h"
#include "TableSchemaLoader.h"

START_CORALACCESS_NAMESPACE

TableSchemaLoader::TableSchemaLoader(Session& s, const std::string& schemaName)
	: _session(s),
	  _schemaName(schemaName)
{
	// EMPTY
}

TableSchemaLoader::~TableSchemaLoader()
{
	// EMPTY
}

void TableSchemaLoader::refreshTableDescription(coral::TableDescription& description)
{
	coral::CoralMessaging::CAL::MsgObj::TableSchemaRead obj(_schemaName, description.name());
	coral::CoralMessaging::CAL::TableSchemaRead request(obj);
	IMessagePtr reply = _session.connection().getStub().sendRecv(request, coral::CoralProtocol::Opcode::E_TableSchemaReadReply);

	if(!reply.get())
		EXCEPTION_THROW(coral::CoralMessaging::Err_Custom("Table Schema Read failed"));

	coral::CoralMessaging::CAL::TableSchemaReadReply& msg = dynamic_cast<coral::CoralMessaging::CAL::TableSchemaReadReply&>(*reply);
	// Insert the new column metadata into description
	BOOST_FOREACH(coral::CoralMessaging::CAL::MsgObj::ColumnDescription i, msg.columnDescription()) {
		description.insertColumn(i.columnName(), i.cppType(), i.length(), i.fixedSize(), i.tableSpaceName());

		if (i.notNull())
			description.setNotNullConstraint(i.columnName());
	}
	// Insert the index description
	BOOST_FOREACH(coral::CoralMessaging::CAL::MsgObj::IndexDescription i, msg.indexDescription())
		description.createIndex(i.indexName(), i.columnNames(), i.isUnique(), i.tableSpaceName());
	// Fill foreign key descriptions
	BOOST_FOREACH(coral::CoralMessaging::CAL::MsgObj::ForeignKeyDescription i, msg.foreignKeyDescription())
		description.createForeignKey(i.keyName(), i.columnNames(), i.referencedTableName(), i.referencedColumnNames());
	// Fill unique constraints
	BOOST_FOREACH(coral::CoralMessaging::CAL::MsgObj::UniqueConstraintDescription i, msg.uniqueConstraintDescription())
		description.setUniqueConstraint(i.columnNames(), i.constraintName(), true, i.tableSpaceName());
	// Fill the primary keys
	if (!msg.primaryKeyDescription().columnNames().empty())
		description.setPrimaryKey(msg.primaryKeyDescription().columnNames(), msg.primaryKeyDescription().tableSpaceName());

	description.setType(msg.type());
	description.setTableSpaceName(msg.tableSpaceName());
}

END_CORALACCESS_NAMESPACE
