/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include "RelationalAccess/ISessionProperties.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ConnectionService.h"

#include "../../src/Connection.h"
#include "../../src/Domain.h"
#include "../../src/Session.h"
#include "../../src/Error.h"

#include "DomainHandle.h"
#include "ConnectionTest.h"

START_CORALACCESS_NAMESPACE

void ConnectionTest::testSuccessfullConnection()
{
	DomainHandle dh;
	Domain& domain = dh.get();
	Pointer<Connection>::Scoped connection(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	connection->resetFactory<ConnectionTestConnectionFactory>();
	_testSuccessfullConnection(*connection);

	if (!TestData::WithServer)
		return;

	Pointer<Connection>::Scoped connection_ws(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	_testSuccessfullConnection(*connection_ws);

	// Test with the connection service
	coral::ConnectionService connService;
	Pointer<coral::ISessionProxy>::Scoped session(connService.connect(TestData::CoralServiceName, coral::ReadOnly));
	CPPUNIT_ASSERT(session.get());
	CPPUNIT_ASSERT_EQUAL(TestData::ServerVersion, session->properties().serverVersion());
	CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, session->nominalSchema().schemaName());
}

void ConnectionTest::_testSuccessfullConnection(Connection& connection)
{
	CPPUNIT_ASSERT(!connection.isConnected());
	bool catched = false;

	try {
		connection.serverVersion();
	} catch (ServerException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	connection.typeConverter(); // should not throw
	CPPUNIT_ASSERT_EQUAL(TestData::CoralServerUrl, connection.serverName());
	catched = false;

	try {
		connection.newSession(TestData::ExistingDbUri, coral::ReadOnly);
	} catch (ConnectionNotActiveException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	connection.disconnect();
	CPPUNIT_ASSERT(!connection.isConnected());
	// OK, connect now...
	connection.connect();
	CPPUNIT_ASSERT(connection.isConnected());
	catched = false;
	// Session not started, so the server version must be unknown...
	try {
		connection.serverVersion();
	} catch (ServerException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	connection.getStub(); // should not crash
	// Check disconnect
	connection.disconnect();
	CPPUNIT_ASSERT(!connection.isConnected());
	connection.connect();
	CPPUNIT_ASSERT(connection.isConnected());
	// Ok, the session now...
	Pointer<ISession>::Scoped session(connection.newSession(TestData::ExistingDbUri, coral::ReadOnly));
	CPPUNIT_ASSERT(session.get());
	session->startUserSession("", ""); // Message exchange 1
	CPPUNIT_ASSERT_EQUAL(TestData::ServerVersion, connection.serverVersion());
	CPPUNIT_ASSERT_EQUAL(TestData::CoralServerUrl, connection.serverName());

	// Cannot open a second session...
	catched = false;

	try {
		connection.newSession(TestData::ExistingSchemaName, coral::ReadOnly);
	} catch (Err_Custom&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	session->endUserSession();// Message exchange 2 (closing session)

	// OK, now we can start another session as the previous one is closed
	session.reset(connection.newSession(TestData::ExistingDbUri, coral::ReadOnly));
	CPPUNIT_ASSERT(session.get());
	session->startUserSession("", ""); // Message exchange 3
	session->endUserSession();// Message exchange 4 (closing session)

	connection.disconnect();
	CPPUNIT_ASSERT(!connection.isConnected());
	catched = false;

	try {
		connection.newSession(TestData::ExistingDbUri, coral::ReadOnly);
	} catch (ConnectionNotActiveException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	catched = false;
	// Session not started, so the server version must be unknown...
	try {
		connection.serverVersion();
	} catch (ServerException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
}

IConnectionStub* ConnectionTestConnectionFactory::create(const std::string&, const std::string&, uint32_t, const uint32_t) const
{
	return new ConnectionTestConnectionStub;
}

int ConnectionTestConnectionStub::_msgExchange = 1;

IMessagePtr ConnectionTestConnectionStub::sendRecv(const IMessage& m, uint8_t opcode)
{
	using coral::CoralMessaging::CAL::Connect;
	using coral::CoralMessaging::CAL::ConnectReply;
	using coral::CoralMessaging::CAL::OK;

	switch (_msgExchange++) {
	case 1:
	case 3:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_Connect), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ConnectReply), static_cast<int>(opcode));
		return IMessagePtr(new ConnectReply(TestData::ServerVersion));
	case 2:
	case 4:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_CloseSession), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		return IMessagePtr(new OK);
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}


END_CORALACCESS_NAMESPACE
