/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include <boost/assign/std/vector.hpp>

#include "CoralMessaging/Insulation/AttributeList.h"
#include "CoralMessaging/CAL/CALMessages.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProperties.h"
#include "RelationalAccess/ISessionProxy.h"
#include "CoralBase/Blob.h"

#include "../../src/Connection.h"
#include "../../src/Domain.h"
#include "../../src/Session.h"
#include "../../src/Error.h"
#include "../../src/Query.h"

#include "OracleSessionCreator.h"
#include "LocalSession.h"
#include "QueryTest.h"

START_CORALACCESS_NAMESPACE

void QueryTest::SetupQuery_msg_2(Query& q)
{
	q.setForUpdate();
}

#define MSG_SET_testQueryConstruction 1
#define MSG_SET_testQueryEquality 2
#define MSG_SET_testSimpleQuery 3

void QueryTest::testQueryConstruction()
{
	LocalSession<QueryTestConnectionFactory<MSG_SET_testQueryConstruction> > ls;
	Session& session = ls.session();
	MultiTableQuery q1(session, TestData::ExistingSchemaName);
	q1.execute(); 	// Msg exchange 1
	// Query with method calls
	MultiTableQuery q2(session, TestData::ExistingSchemaName);
	SetupQuery_msg_2(q2);	// Msg exchange 2
	q2.execute();
	// Msg Exchange 3 (close session)
}

void QueryTest::testQueryEquality()
{
	LocalSession<QueryTestConnectionFactory<MSG_SET_testQueryEquality> > ls;
	Session& session = ls.session();
	MultiTableQuery q1(session, TestData::ExistingSchemaName);
	SetupQuery_msg_2(q1);
	MultiTableQuery q2(session, TestData::ExistingSchemaName);
	SetupQuery_msg_2(q2);
	CPPUNIT_ASSERT(q1 == q2);
}

void QueryTest::testSimpleQueryWithServer()
{
	if (!TestData::WithServer)
		return;

	std::vector<AttributeList> fixture = _fillData(TestData::DbServiceName);
	::sleep(1);
	OracleSessionCreator osc(coral::ReadOnly);
	Session& ora_session = osc.session();
	_testSimpleQuery(ora_session, fixture);
}

void QueryTest::testStressWithQuery()
{
	if (!TestData::WithServer)
		return;

	std::vector<AttributeList> fixture = _fillData(TestData::DbServiceName);
	::sleep(1);
	OracleSessionCreator osc(coral::ReadOnly);
	Session& ora_session = osc.session();

	for (int i = 0; i < _QueryNumber(); ++i)
		_testSimpleQuery(ora_session, fixture);
}

void QueryTest::_testSimpleQuery(Session& session, const std::vector<AttributeList>& fixture)
{
	coral::ISchema& schema = session.nominalSchema();
	Pointer<coral::IQuery>::Scoped query(schema.tableHandle(_TableName()).newQuery());
	coral::ICursor& cursor = query->execute();

	for (int i = 0; i < _RowNum(); ++i) {
		CPPUNIT_ASSERT(cursor.next());
		const coral::AttributeList& currentRow = cursor.currentRow();
		CPPUNIT_ASSERT_EQUAL(fixture[i].size(), currentRow.size());

		for (size_t s = 0; s < currentRow.size(); ++s)
			CPPUNIT_ASSERT_EQUAL(fixture[i][s], currentRow[s]);

		CPPUNIT_ASSERT_EQUAL(fixture[i], currentRow);
		const coral::Attribute& blobAttribute = currentRow["data2"];

		if ( ! blobAttribute.isNull()) {
			const coral::Blob& blob = blobAttribute.data<coral::Blob>();
			long int blobSize = ( i + 1 ) * 1000;
			CPPUNIT_ASSERT_EQUAL(blobSize, blob.size());
			const unsigned char* p = static_cast<const unsigned char*>( blob.startingAddress() );

			for (int j = 0; j < blobSize; ++j, ++p)
				CPPUNIT_ASSERT_EQUAL(*p, static_cast<unsigned char>(j%256));
		}
	}
}

std::vector<AttributeList> QueryTest::_fillData(const std::string& serviceName)
{
	coral::ConnectionService connService;
	Pointer<coral::ISessionProxy>::Scoped session(connService.connect(serviceName, coral::Update));
	CPPUNIT_ASSERT(session.get());
	session->transaction().start();
	coral::ISchema& schema = session->nominalSchema();
	schema.dropIfExistsTable(_TableName());

	coral::TableDescription description;
	description.setName(_TableName());
	description.insertColumn("ID", AttributeSpecification::typeNameForId(typeid(int)));
	description.setPrimaryKey("ID");
	description.insertColumn("x", coral::AttributeSpecification::typeNameForId(typeid(float)));
	description.setNotNullConstraint( "x" );
	description.insertColumn("Y", coral::AttributeSpecification::typeNameForId(typeid(double)));
	description.insertColumn("Z", coral::AttributeSpecification::typeNameForId(typeid(double)));
	description.insertColumn( "data1", coral::AttributeSpecification::typeNameForId(typeid(std::string)), 100, false);
	description.insertColumn( "data2", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)));

	coral::ITable& table = schema.createTable( description );
	coral::AttributeList rowBuffer;
	table.dataEditor().rowBuffer(rowBuffer);
	std::vector<AttributeList> result;

	for (int i = 0; i < _RowNum(); ++i) {
		rowBuffer["ID"].data<int>() = i + 10;
		rowBuffer["x"].data<float>() = i + 0.1 * i;
		rowBuffer["Y"].setNull(false);
		rowBuffer["Y"].data<double>() = i + 0.01 * i;
		rowBuffer["Y"].setNull(i%2);
		rowBuffer["Z"].data<double>() = i + 0.001 * i;
		rowBuffer["data1"].data<std::string>() = "Data";

		rowBuffer["data2"].setNull(false);
		coral::Blob& blob = rowBuffer["data2"].data<coral::Blob>();
		int blobSize = ( i + 1 ) * 1000;
		blob.resize( blobSize );
		unsigned char* p = static_cast<unsigned char*>( blob.startingAddress() );

		for ( int j = 0; j < blobSize; ++j, ++p )
			*p = j%256;

		rowBuffer["data2"].setNull(i%2);
		table.dataEditor().insertRow( rowBuffer );
		result.push_back(rowBuffer);
	}

	session->transaction().commit();
	return result;
}

template<int MSG_SET>
IConnectionStub* QueryTestConnectionFactory<MSG_SET>::create(const std::string&, const std::string&, uint32_t, const uint32_t) const
{
	return new QueryTestConnectionStub<MSG_SET>();
}

template<int MSG_SET>
int QueryTestConnectionStub<MSG_SET>::_msgExchange = 1;

using CORALMESSAGING_NAMESPACE_NAME::CAL::Connect;
using CORALMESSAGING_NAMESPACE_NAME::CAL::ConnectReply;
using CORALMESSAGING_NAMESPACE_NAME::CAL::ExecuteQuery;
using CORALMESSAGING_NAMESPACE_NAME::CAL::ExecuteQueryReply;
using CORALMESSAGING_NAMESPACE_NAME::CAL::OK;
using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::QueryRepresentation;

template<>
IMessagePtr QueryTestConnectionStub<MSG_SET_testQueryConstruction>::sendRecv(const IMessage& m, uint8_t opcode)
{
	switch (_msgExchange++) {
	case 1: {
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ExecuteQuery), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ExecuteQueryReply), static_cast<int>(opcode));
		const ExecuteQuery& qMsg = dynamic_cast<const ExecuteQuery&>(m);
		CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, qMsg.query().schemaName());
		return IMessagePtr(new ExecuteQueryReply);
	}
	case 2: {
		LocalSession<QueryTestConnectionFactory<MSG_SET_testQueryConstruction> > ls;
		Session& session = ls.session();
		MultiTableQuery q1(session, TestData::ExistingSchemaName);
		const ExecuteQuery& qMsg = dynamic_cast<const ExecuteQuery&>(m);
		const_cast<QueryRepresentation&>(qMsg.query()).replay(q1);
		MultiTableQuery q2(session, TestData::ExistingSchemaName);
		QueryTest::SetupQuery_msg_2(q2);
		CPPUNIT_ASSERT(q1 == q2);
		return IMessagePtr(new ExecuteQueryReply);
	}
	case 3:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_CloseSession), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		return IMessagePtr(new OK);
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}

template<>
IMessagePtr QueryTestConnectionStub<MSG_SET_testQueryEquality>::sendRecv(const IMessage& m, uint8_t opcode)
{
	switch (_msgExchange++) {
	case 1: {
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ExecuteQuery), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ExecuteQueryReply), static_cast<int>(opcode));
		const ExecuteQuery& qMsg = dynamic_cast<const ExecuteQuery&>(m);
		CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, qMsg.query().schemaName());
		return IMessagePtr(new ExecuteQueryReply);
	}
	case 2: {
		LocalSession<QueryTestConnectionFactory<MSG_SET_testQueryEquality> > ls;
		Session& session = ls.session();
		MultiTableQuery q1(session, TestData::ExistingSchemaName);
		const ExecuteQuery& qMsg = dynamic_cast<const ExecuteQuery&>(m);
		const_cast<QueryRepresentation&>(qMsg.query()).replay(q1);
		MultiTableQuery q2(session, TestData::ExistingSchemaName);
		QueryTest::SetupQuery_msg_2(q2);
		CPPUNIT_ASSERT(q1 == q2);
		return IMessagePtr(new ExecuteQueryReply);
	}
	case 3:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_CloseSession), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		return IMessagePtr(new OK);
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}

END_CORALACCESS_NAMESPACE
