/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <iterator>

#include <boost/assign/std/set.hpp>
#include <boost/assign/list_of.hpp>

#include "CoralMessaging/CAL/CALMessages.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "CoralBase/AttributeSpecification.h"

#include "../../src/SchemaLoader.h"
#include "../../src/Domain.h"
#include "../../src/Connection.h"
#include "../../src/Pointers.h"
#include "../../src/SchemaEditor.h"

#include "OracleSessionCreator.h"
#include "DomainHandle.h"
#include "SchemaEditorTest.h"

START_CORALACCESS_NAMESPACE

std::set<std::string>& SchemaEditorTest::TableList()
{
    using namespace boost;
    using namespace boost::assign;

	static std::set<std::string> lst = list_of("T_1")("T_2")("T_3").to_container(lst);
	return lst;
}

void SchemaEditorTest::testListTablesWithoutServer()
{
	// Test without any server
	DomainHandle dh;
	Domain& domain = dh.get();
	Pointer<Connection>::Scoped connection(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	CPPUNIT_ASSERT(connection.get());
	connection->resetFactory<SchemaEditorTestConnectionFactory>();
	connection->connect();
	Pointer<Session>::Scoped session(dynamic_cast<Session*>(connection->newSession(TestData::ExistingDbUri, coral::ReadOnly)));
	CPPUNIT_ASSERT(session.get());
	SchemaEditor se(*session);
	// Check the empty list
	CPPUNIT_ASSERT(se.listTables().empty()); // Message exchange 1
	// Check non-empty list
	CPPUNIT_ASSERT(TableList() == se.listTables()); // Message exchange 2
}

void SchemaEditorTest::testListTablesWithServer()
{
	if (!_withServer())
		return;

	_setupTables(TestData::DbServiceName);
	OracleSessionCreator osc(coral::ReadOnly);
	Session& session = osc.session();
	SchemaEditor se(session);
	std::set<std::string> res(se.listTables());
	// The resulting list of tables must include the newly created tables. We determine it by calculating the
	// intersection of the two sets.
	std::set<std::string> intr;
	std::set_intersection(res.begin(), res.end(), TableList().begin(), TableList().end(), std::inserter(intr, intr.end()));
	CPPUNIT_ASSERT(TableList() == intr);
}

void SchemaEditorTest::_setupTables(const std::string& serviceName)
{
	coral::ConnectionService connService;
	Pointer<coral::ISessionProxy>::Scoped session(connService.connect(serviceName, coral::Update));
	CPPUNIT_ASSERT(session.get());
	session->transaction().start(false);
	coral::ISchema& schema = session->nominalSchema();

	for (std::set<std::string>::const_iterator i = TableList().begin(); i != TableList().end(); ++i) {
		schema.dropIfExistsTable(*i);
		coral::TableDescription description;
		description.setName(*i);
		description.insertColumn("ID", coral::AttributeSpecification::typeNameForId(typeid(int)));
		schema.createTable(description);
	}

	session->transaction().commit();
}

bool SchemaEditorTest::_withServer()
{
	return TestData::WithServer;
}

IConnectionStub* SchemaEditorTestConnectionFactory::create(const std::string&, const std::string&, uint32_t, const uint32_t) const
{
	return new SchemaEditorTestConnectionStub;
}

int SchemaEditorTestConnectionStub::_msgExchange = 1;

IMessagePtr SchemaEditorTestConnectionStub::sendRecv(const IMessage& m, uint8_t opcode)
{
	CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ListTables), static_cast<int>(m.opcode()));
	CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ListTablesReply), static_cast<int>(opcode));
	using coral::CoralMessaging::CAL::ListTables;
	using coral::CoralMessaging::CAL::ListTablesReply;

	switch (_msgExchange++) {
	case 1:
		return IMessagePtr(new ListTablesReply);
	case 2:
		return IMessagePtr(new ListTablesReply(std::vector<std::string>(
				SchemaEditorTest::TableList().begin(), SchemaEditorTest::TableList().end())));
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}

END_CORALACCESS_NAMESPACE
