/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include "CoralMessaging/Insulation/SchemaException.h"
#include "CoralMessaging/Insulation/SessionException.h"
#include "CoralMessaging/CAL/CALMessages.h"

#include "../../src/IConnectionStubFactory.h"
#include "../../src/IConnectionStub.h"

#include "DomainHandle.h"
#include "SessionTest.h"

START_CORALACCESS_NAMESPACE

void SessionTest::testSuccessfullConnection()
{
	DomainHandle dh;
	Domain& domain = dh.get();
	Pointer<Connection>::Scoped connection(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	connection->resetFactory<SessionTestConnectionFactory>();
	_testSuccessfullConnection(*connection);

	if (!TestData::WithServer)
		return;

	Pointer<Connection>::Scoped connection_ws(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	_testSuccessfullConnection(*connection_ws);
}

void SessionTest::_testSuccessfullConnection(Connection& connection)
{
	connection.connect();
	Pointer<Session>::Scoped session(dynamic_cast<Session*>(connection.newSession(TestData::ExistingDbUri, coral::ReadOnly)));
	CPPUNIT_ASSERT(session.get());
	session->connection(); // should not throw
	session->monitoring(); // should not throw
	session->transaction(); // should not throw
	CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, session->nominalSchemaNameForConnection(TestData::ExistingDbUri));

	bool catched = false;

	try {
		session->nominalSchema().schemaName();
	} catch (ConnectionNotActiveException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	catched = false;

	try {
		session->schema(TestData::ExistingSchemaName);
	} catch (ConnectionNotActiveException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
	CPPUNIT_ASSERT(!session->isUserSessionActive());

	session->startUserSession(UserName(), Password()); // Message exchange 1
	CPPUNIT_ASSERT(session->isUserSessionActive());

	CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, session->nominalSchema().schemaName());
	// Should NOT send a message (the nominal schema)
	CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, session->schema(TestData::ExistingSchemaName).schemaName());
	// Should send a message and cache it - Message exchange 2
	CPPUNIT_ASSERT_EQUAL(TestData::OtherExistingSchemaName, session->schema(TestData::OtherExistingSchemaName).schemaName());
	// Should not send a message (schema from cache)
	CPPUNIT_ASSERT_EQUAL(TestData::OtherExistingSchemaName, session->schema(TestData::OtherExistingSchemaName).schemaName());

	// Not existing schema name
	catched = false;

	try {
		session->schema(TestData::NotExistingSchemaName); // Message exchange 3
	} catch (InvalidSchemaNameException&) {
		catched = true;
	}

	// TODO - with server, it dows not work due to Oracle bug
	if (!TestData::WithServer)
		CPPUNIT_ASSERT(catched);

	session->endUserSession(); // Message exchange 4
	CPPUNIT_ASSERT(!session->isUserSessionActive());
	catched = false;

	try {
		session->schema(TestData::ExistingSchemaName);
	} catch (ConnectionNotActiveException&) {
		catched = true;
	}

	CPPUNIT_ASSERT(catched);
}


IConnectionStub* SessionTestConnectionFactory::create(const std::string&, const std::string&, uint32_t, const uint32_t) const
{
	return new SessionTestConnectionStub;
}

using coral::CoralProtocol::Opcode;
using namespace coral::CoralMessaging;

int SessionTestConnectionStub::_msgExchange = 0;

IMessagePtr SessionTestConnectionStub::sendRecv(const IMessage& m, uint8_t opcode)
{
	switch (++_msgExchange) {
	case 1: {
		const CAL::Connect& msg = dynamic_cast<const CAL::Connect&>(m);
		CPPUNIT_ASSERT_EQUAL(static_cast<uint32_t>(Opcode::E_Connect), static_cast<uint32_t>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(SessionTest::UserName(), msg.connection().username());
		CPPUNIT_ASSERT_EQUAL(SessionTest::Password(), msg.connection().password());
		CPPUNIT_ASSERT_EQUAL(true, msg.connection().accessMode());
		CPPUNIT_ASSERT_EQUAL(TestData::ExistingDbUri, msg.connection().dbUrl());
		CPPUNIT_ASSERT_EQUAL(static_cast<uint32_t>(Opcode::E_ConnectReply), static_cast<uint32_t>(opcode));
		return IMessagePtr(new CAL::ConnectReply(TestData::ServerVersion));
	}
	case 2: {
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_TableSchemaExists), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		const CAL::TableSchemaExists& msg = dynamic_cast<const CAL::TableSchemaExists&>(m);
		CPPUNIT_ASSERT_EQUAL(TestData::OtherExistingSchemaName, msg.schemaName());
		return IMessagePtr(new CAL::OK);
	}
	case 3: {
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_TableSchemaExists), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		const CAL::TableSchemaExists& msg = dynamic_cast<const CAL::TableSchemaExists&>(m);
		CPPUNIT_ASSERT_EQUAL(TestData::NotExistingSchemaName, msg.schemaName());
		throw InvalidSchemaNameException();
	}
	case 4:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_CloseSession), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		return IMessagePtr(new CAL::OK);
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}

END_CORALACCESS_NAMESPACE
