/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include "CoralMessaging/CAL/CALMessages.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ConnectionService.h"

#include "../../src/TableSchemaLoader.h"
#include "../../src/Domain.h"
#include "../../src/Connection.h"
#include "../../src/Pointers.h"
#include "../../src/Session.h"

#include "DomainHandle.h"
#include "OracleSessionCreator.h"
#include "TableSchemaLoaderTest.h"

START_CORALACCESS_NAMESPACE

void TableSchemaLoaderTest::testSchemaExists()
{
	DomainHandle dh;
	Domain& domain = dh.get();
	Pointer<Connection>::Scoped connection(dynamic_cast<Connection*>(domain.newConnection(TestData::CoralServerUrl)));
	connection->resetFactory<TableSchemaLoaderTestConnectionFactory>();
	connection->connect();
	Pointer<Session>::Scoped session(dynamic_cast<Session*>(connection->newSession(TestData::ExistingDbUri, coral::ReadOnly)));
	session->startUserSession("", ""); // Message exchange 1
	TableSchemaLoader tsl(*session, TestData::ExistingSchemaName);
	coral::TableDescription desc1;
	desc1.setName(TableName());
	tsl.refreshTableDescription(desc1); // Message exchange 2
	_testSchemaExists(desc1);
	session->endUserSession(); // message exchange 3 (CloseSession)

	if (!TestData::WithServer)
		return;

	coral::ConnectionService connService;
	Pointer<coral::ISessionProxy>::Scoped ora_session_upd(connService.connect(TestData::DbServiceName, coral::Update));
	ora_session_upd->transaction().start();
	coral::ISchema& schema = ora_session_upd->nominalSchema();
	schema.dropIfExistsTable(TableName());
	schema.dropIfExistsTable(FKDesc().referencedTableName());
	coral::TableDescription ref_desc;
	ref_desc.setName(FKDesc().referencedTableName());
	ref_desc.insertColumn(FKDesc().referencedColumnNames()[0], coral::AttributeSpecification::typeNameForId(typeid(int)));
	ref_desc.setPrimaryKey(FKDesc().referencedColumnNames()[0]);
	schema.createTable(ref_desc);
	schema.createTable(desc1);
	ora_session_upd->transaction().commit();

	OracleSessionCreator osc(coral::ReadOnly);
	Session& ora_session = osc.session();
	ora_session.transaction().start();
	TableSchemaLoader ora_tsl(ora_session, TestData::ExistingSchemaName);
	coral::TableDescription desc2;
	desc2.setName(TableName());
	ora_tsl.refreshTableDescription(desc2);
	_testSchemaExists(desc2);
	ora_session.transaction().commit();
}

void TableSchemaLoaderTest::_testSchemaExists(coral::TableDescription& desc)
{
	CPPUNIT_ASSERT_EQUAL(TableName(), desc.name());
	CPPUNIT_ASSERT_EQUAL(Type(), desc.type());
	CPPUNIT_ASSERT_EQUAL(TableSpaceName(), desc.tableSpaceName());

	CPPUNIT_ASSERT_EQUAL(2, desc.numberOfColumns());
	const IColumn& colDesc = desc.columnDescription(0);
	CPPUNIT_ASSERT_EQUAL(ColDesc_1().columnName(), colDesc.name());
	CPPUNIT_ASSERT_EQUAL(ColDesc_1().cppType(), colDesc.type());
	CPPUNIT_ASSERT_EQUAL(static_cast<unsigned long long>(ColDesc_1().length()), static_cast<unsigned long long>(colDesc.size()));
	CPPUNIT_ASSERT_EQUAL(ColDesc_1().fixedSize(), colDesc.isSizeFixed());
	CPPUNIT_ASSERT_EQUAL(ColDesc_1().notNull(), colDesc.isNotNull());
	//CPPUNIT_ASSERT_EQUAL(ColDesc_1().tableSpaceName(), colDesc.tableSpaceName());

	CPPUNIT_ASSERT(desc.hasPrimaryKey());
	const IPrimaryKey& pk = desc.primaryKey();
	CPPUNIT_ASSERT(pk.columnNames() == std::vector<std::string>(1, ColDesc_1().columnName()));
	CPPUNIT_ASSERT_EQUAL(TableSpaceName(), pk.tableSpaceName());

	CPPUNIT_ASSERT_EQUAL(0, desc.numberOfIndices());
	//const IIndex& index = desc.index(0);
	//CPPUNIT_ASSERT_EQUAL(IndexDesc().indexName(), index.name());
	//CPPUNIT_ASSERT(IndexDesc().columnNames() == index.columnNames());
	//CPPUNIT_ASSERT_EQUAL(IndexDesc().isUnique(), index.isUnique());
	//CPPUNIT_ASSERT_EQUAL(TableSpaceName(), index.tableSpaceName());

	CPPUNIT_ASSERT_EQUAL(1, desc.numberOfForeignKeys());
	const IForeignKey& fk = desc.foreignKey(0);
	CPPUNIT_ASSERT_EQUAL(FKDesc().keyName(), fk.name());
	CPPUNIT_ASSERT(FKDesc().columnNames() == fk.columnNames());
	CPPUNIT_ASSERT_EQUAL(FKDesc().referencedTableName(), fk.referencedTableName());
	CPPUNIT_ASSERT(FKDesc().referencedColumnNames() == fk.referencedColumnNames());
}

IConnectionStub* TableSchemaLoaderTestConnectionFactory::create(const std::string&, const std::string&, uint32_t, const uint32_t) const
{
	return new TableSchemaLoaderTestConnectionStub;
}

int TableSchemaLoaderTestConnectionStub::_msgExchange = 1;

IMessagePtr TableSchemaLoaderTestConnectionStub::sendRecv(const IMessage& m, uint8_t opcode)
{
	using coral::CoralMessaging::CAL::TableSchemaRead;
	using coral::CoralMessaging::CAL::TableSchemaReadReply;
	using coral::CoralMessaging::CAL::OK;
	using coral::CoralMessaging::CAL::ConnectReply;

	switch (_msgExchange++) {
	case 1:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_Connect), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_ConnectReply), static_cast<int>(opcode));
		return IMessagePtr(new ConnectReply(TestData::ServerVersion));
	case 2: {
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_TableSchemaRead), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_TableSchemaReadReply), static_cast<int>(opcode));
		const TableSchemaRead& msg = dynamic_cast<const TableSchemaRead&>(m);
		CPPUNIT_ASSERT_EQUAL(TestData::ExistingSchemaName, msg.schema().schemaName());
		CPPUNIT_ASSERT_EQUAL(TableSchemaLoaderTest::TableName(), msg.schema().tableName());

		std::vector<MsgObj::ColumnDescription> columnDescription;
		columnDescription.push_back(TableSchemaLoaderTest::ColDesc_1());
		columnDescription.push_back(TableSchemaLoaderTest::ColDesc_2());
		//std::vector<MsgObj::IndexDescription> indexDescription(1, TableSchemaLoaderTest::IndexDesc());
		std::vector<MsgObj::IndexDescription> indexDescription;
		std::vector<MsgObj::ForeignKeyDescription> foreignKeyDescription(1, TableSchemaLoaderTest::FKDesc());
		std::vector<MsgObj::UniqueConstraintDescription> ucDescription(1, TableSchemaLoaderTest::UCDesc());

		return IMessagePtr(new TableSchemaReadReply(columnDescription, indexDescription, foreignKeyDescription, ucDescription, TableSchemaLoaderTest::PKDesc(), TableSchemaLoaderTest::Type(), TableSchemaLoaderTest::TableSpaceName()));
	}
	case 3:
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_CloseSession), static_cast<int>(m.opcode()));
		CPPUNIT_ASSERT_EQUAL(static_cast<int>(coral::CoralProtocol::Opcode::E_OK), static_cast<int>(opcode));
		return IMessagePtr(new OK);
	default:
		CPPUNIT_ASSERT(0);
	}

	return IMessagePtr();
}

END_CORALACCESS_NAMESPACE
