#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/AttributeListException.h"
#include "AttributeData.h"
#include "AttributeDataFactory.h"
#include <ostream>

coral::Attribute::Attribute( const coral::AttributeSpecification& specification ):
  m_specification( &specification ),
  m_data( coral::AttributeDataFactory::factory().create( m_specification->type() ) )
{}


coral::Attribute::~Attribute()
{
  if ( m_data ) m_data->release();
}

coral::Attribute&
coral::Attribute::operator=( const coral::Attribute& rhs )
{
  if ( !( rhs.m_specification->type() == this->m_specification->type() ) )
    throw coral::AttributeListException( "Cannot a \"" + rhs.m_specification->typeName() +
                                         "\" to a \"" + this->m_specification->typeName() + "\"" );
  this->fastCopy( rhs );
  return *this;
}


void
coral::Attribute::fastCopy( const Attribute& rhs )
{
  this->setValueFromAddress( rhs.addressOfData() );
  this->setNull( rhs.isNull() );
}


void
coral::Attribute::setValueFromAddress( const void* externalAddress )
{
  m_data->setValueFromAddress( externalAddress );
}

void
coral::Attribute::copyValueToAddress( void* externalAddress ) const
{
  m_data->copyValueToAddress( externalAddress );
}

void*
coral::Attribute::addressOfData()
{
  return m_data->addressOfData();
}

const void*
coral::Attribute::addressOfData() const
{
  return m_data->addressOfData();
}


void
coral::Attribute::shareData( const Attribute& sourceAttribute )
{
  if ( !( sourceAttribute.m_specification->type() == this->m_specification->type() ) )
    throw coral::AttributeListException( "Cannot share a \"" + sourceAttribute.m_specification->typeName() +
                                         "\" with a \"" + this->m_specification->typeName() + "\"" );
  m_data->shareData( *( sourceAttribute.m_data ) );
}


void
coral::Attribute::bindUnsafely( const void* externalAddress )
{
  m_data->bind( const_cast< void* >( externalAddress ) );
}

void
coral::Attribute::bindUnsafely( void* externalAddress )
{
  m_data->bind( externalAddress );
}


void
coral::Attribute::bindVariable( const std::type_info& type,
                                const void* externalAddress )
{
  if ( ! ( this->m_specification->type() == type ) ) {
    throw coral::AttributeListException( "Attempted binding with incompatible types" );
  }

  this->bindUnsafely( externalAddress );
}

void
coral::Attribute::bindVariable( const std::type_info& type,
                                void* externalAddress )
{
  if ( ! ( this->m_specification->type() == type ) )  {
    throw coral::AttributeListException( "Attempted binding with incompatible types" );
  }

  this->bindUnsafely( externalAddress );
}


void
coral::Attribute::setNull( bool isVariableNull )
{
  m_data->setNull( isVariableNull );
}


bool
coral::Attribute::isNull() const
{
  return m_data->isNull();
}


int
coral::Attribute::size() const
{
  return m_data->size();
}



std::ostream&
coral::Attribute::toOutputStream( std::ostream& os,
                                  bool valueOnly) const
{
  if ( ! valueOnly )
    os << m_specification->name() << " (" << m_specification->typeName() << ") : ";
  m_data->toOutputStream( os );
  return os;
}

void
coral::Attribute::setValue( const std::type_info& type,
                            const void* externalAddress )
{
  if ( ! ( this->m_specification->type() == type ) ) {
    throw coral::AttributeListException( "Attempted setting the value from a wrong type" );
  }
  this->setValueFromAddress( externalAddress );
}


bool
coral::Attribute::operator==( const Attribute& rhs ) const
{
	if (this->m_data->isNull() || rhs.m_data->isNull())
		return this->m_data->isNull() == rhs.m_data->isNull();

	return this->m_specification->type() == rhs.m_specification->type() &&
         *( this->m_data ) == *( rhs.m_data );
}

bool
coral::Attribute::operator!=( const Attribute& rhs ) const
{
	return !(*this == rhs);
}
