/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#ifndef CORALBASE_ATTRIBUTETYPETOOLS_H_
#define CORALBASE_ATTRIBUTETYPETOOLS_H_

#include <string>

namespace coral {

/// List of supported Attribute types. This is the core of the Attribute type operations.
/// Code generators take it as input, and generete code for each tpye in several places.
#define CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_1 (bool)(char)(unsigned char)(short)\
	(unsigned short)(int)(unsigned int)(long)(unsigned long)(long long)(unsigned long long)

#define CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_2 \
	(float)(double)(long double)(std::string)(coral::Blob)(coral::Date)(coral::TimeStamp)

/// We need two list, as due to compiler limitations, a type list length cannot exceed
/// a limit (normally it is 17).
#define CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST \
	CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_1 \
	CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_2

/// Type manipulations for the supported attribute types
struct AttributeTypeTools
{
public:
	// -------------------------------------------------------------------------
	// C++ name (string) of the supported types
	// -------------------------------------------------------------------------
	
	/** Give a string representation for the fully qualified C++ type name. */
	template<typename T> static const std::string& typeName()
	{
		// In case of unhandled type Attribute (that is not in  CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST)
		// generate a compile time error.
		BOOST_MPL_ASSERT_MSG((boost::mpl::bool_<false>::value, UNSUPPORTED_CORAL_ATTRIBUTE_TYPE, (bool)));
		// Just to please the compiler. This line will never be executed (as this function will never be
		// instantiated).
		return std::string();
	}
	
	// Generate specializations for each supported type
	
	// The input to the code generator
	#define CODE_GENERATOR(r, data, TYPENAME) \
		template<> static const std::string& typename<TYPENAME>() \
		{ \
			const std::string s(BOOST_PP_STRINGIZE(TYPENAME)); \
			return s; \
		}
	
	// Code generation
	// Iterates through the type list and creates a template specialization for each type.
	BOOST_PP_SEQ_FOR_EACH(CODE_GENERATOR, ~, CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST);
	
	// -------------------------------------------------------------------------
	// A C++ typelist of all supported types
	// -------------------------------------------------------------------------
	#define CREATE_LIST_ELEMENT(r, data, i, elem) elem BOOST_PP_COMMA_IF(BOOST_PP_LESS(i, data))
	#define CREATE_LIST(lst) BOOST_PP_SEQ_FOR_EACH_I(CREATE_LIST_ELEMENT, BOOST_PP_SEQ_SIZE(lst), lst)
	
	// we need two lists due to compiler limitations
	typedef boost::mpl::list<CREATE_LIST(CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_1)> typelist_1;
	typedef boost::mpl::list<CREATE_LIST(CORALBASE_ATTRIBUTETYPETOOLS_TYPELIST_2)> typelist_2;
	
	#undef CREATE_LIST_ELEMENT
	#undef CREATE_LIST
};

} // namespace coral

#endif /*CORALBASE_ATTRIBUTETYPETOOLS_H_*/
