#include <boost/foreach.hpp>

#include "CoralKernel/Context.h"
#include "CoralKernel/ILoadableComponent.h"

#include "PluginManager.h"
#include "PropertyManager.h"

coral::Context&
coral::Context::instance( coral::IPluginManager* pluginManager )
{
  static coral::Context theInstance( pluginManager );

  return theInstance;
}


bool
coral::Context::existsComponent( const std::string& componentName )
{
  boost::mutex::scoped_lock lock( m_mutex );
  if ( m_components.find( componentName ) != m_components.end() )
    return true;
  else
    return false;
}


void
coral::Context::loadComponent( const std::string& componentName, coral::IPluginManager* pluginManager )
{
  /// Check if the component has already been loaded.
  if ( this->existsComponent( componentName ) ) return;

  /// Use the provided plugin manager or the default implementation
  coral::IPluginManager* thePluginManager = pluginManager;

  if ( pluginManager == 0 )
  {
    boost::mutex::scoped_lock lock( m_mutex );
    if ( ! m_pluginManager )
    {
      m_nativePluginManager = new coral::PluginManager();
      m_pluginManager = m_nativePluginManager;
    }

    thePluginManager = m_pluginManager;
  }

  ILoadableComponent* component = thePluginManager->newComponent( componentName );

  if ( component )
  {
    boost::mutex::scoped_lock lock( m_mutex );
    m_components.insert( std::make_pair( componentName, component ) );
  }
}

coral::IPropertyManager& coral::Context::PropertyManager()
{
	return *m_propertyManager;
}

std::set<std::string>
coral::Context::loadedComponents() const
{
	std::set<std::string> comlist;

	BOOST_FOREACH (m_components_type::value_type v, m_components)
		comlist.insert(v.first);

	return comlist;
}

std::set<std::string>
coral::Context::knownComponents() const
{
	std::set<std::string> comlist = m_pluginManager->knownPlugins();
	return comlist;
}

coral::Context::Context( coral::IPluginManager* pluginManager ):
  m_mutex(),
  m_components(),
  m_pluginManager( 0 ),
  m_propertyManager(new coral::PropertyManager)
{
  if ( pluginManager == 0 )
  {
    boost::mutex::scoped_lock lock( m_mutex );
    if ( ! m_pluginManager )
    {
      m_nativePluginManager = new coral::PluginManager();
      m_pluginManager = m_nativePluginManager;
    }
  }
  else
    m_pluginManager = pluginManager;
}

coral::Context::~Context()
{
  for ( std::map<std::string, coral::ILoadableComponent*>::iterator iComponent = m_components.begin();
        iComponent != m_components.end(); ++iComponent )
    iComponent->second->removeReference();

  if ( m_nativePluginManager )
    delete m_nativePluginManager;
}
