/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALMESSAGING_CAL_MSGOBJ_ATTRIBUTETYPETOOLS_H_
#define CORALMESSAGING_CAL_MSGOBJ_ATTRIBUTETYPETOOLS_H_

#include <boost/mpl/assert.hpp>
#include <boost/mpl/list.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/has_key.hpp>
#include <boost/mpl/if.hpp>
#include <boost/mpl/at.hpp>
#include <boost/mpl/map.hpp>
#include <boost/preprocessor/arithmetic/inc.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/for_each_i.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/seq/size.hpp>
#include <boost/preprocessor/comparison/less.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/arithmetic/dec.hpp>

#include "CoralMessaging/Insulation/AttributeSpecification.h"
#include "CoralBase/AttributeListSpecification.h"

#include "CoralMessaging/CAL/CppType_ProtocolID_Map.h"
#include "CoralMessaging/CAL/MsgObj/Objects.h"

#include <string>
#include <typeinfo>

namespace coral
{
	class TimeStamp;
	class Blob;
	class Date;
}

/// List of supported Attribute types. This is the core of the Attribute type operations.
/// Code generators take it as input, and generate code for each tpye in several places.
#define CORALBASE_ATTRIBUTE_TYPELIST_1 (bool)(char)(unsigned char)(short)\
	(unsigned short)(int)(unsigned int)(long)(unsigned long)

#define CORALBASE_ATTRIBUTE_TYPELIST_2 (long long)(unsigned long long)\
	(float)(double)(long double)(std::string)(coral::Blob)(coral::Date)(coral::TimeStamp)

/// We need two list, as due to compiler limitations, a type list length cannot exceed
/// a limit (normally it is 17).
#define CORALBASE_ATTRIBUTE_TYPELIST \
	CORALBASE_ATTRIBUTE_TYPELIST_1 \
	CORALBASE_ATTRIBUTE_TYPELIST_2

START_CORALMESSAGING_NAMESPACE

namespace CAL {

namespace MsgObj {

/// List of supported Attribute types.
#define CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_1 CORALBASE_ATTRIBUTE_TYPELIST_1
#define CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_2 CORALBASE_ATTRIBUTE_TYPELIST_2
#define CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST   CORALBASE_ATTRIBUTE_TYPELIST

/// Type manipulations for the supported attribute types
class AttributeTypeTools
{
public:
	// -------------------------------------------------------------------------
	// Protocol ID of the supported types
	// -------------------------------------------------------------------------

	/** Give a string representation for the fully qualified C++ type name. */
	template<typename T> struct ProtocolIDType
	{
		//BOOST_MPL_ASSERT_MSG(::boost::mpl::bool_<false>::value, UNSUPPORTED_CORAL_ATTRIBUTE_TYPE, (bool));
	};

	// -------------------------------------------------------------------------
	// Protocol ID from a typeinfo
	// -------------------------------------------------------------------------
	static const uint32_t ProtocolID(const std::type_info& ti);

	// -------------------------------------------------------------------------
	// A C++ typelist of all supported types
	// -------------------------------------------------------------------------
	#define CREATE_LIST_ELEMENT(r, data, i, elem) elem BOOST_PP_COMMA_IF(BOOST_PP_LESS(i, data))
	#define CREATE_LIST(lst) BOOST_PP_SEQ_FOR_EACH_I(CREATE_LIST_ELEMENT, BOOST_PP_DEC(BOOST_PP_SEQ_SIZE(lst)), lst)

	// we need two lists due to compiler limitations
    typedef ::boost::mpl::list<CREATE_LIST(CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_1)> typelist_1;
    typedef ::boost::mpl::list<CREATE_LIST(CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_2)> typelist_2;

	#undef CREATE_LIST_ELEMENT
	#undef CREATE_LIST

	// -------------------------------------------------------------------------
	// A type map that maps a C++ type to its protocol representation type
	// -------------------------------------------------------------------------

private:
	// Identity function: for types requiring no conversion. Those types can be
	// handled directly by the MessageStream{Writer, Reader} classes.
	template <typename FROM, typename TO>
	struct _CppProtocolTypeConverter
	{
		static TO convert(FROM from)
		{
			return static_cast<TO>(from);
		}
	};

	template <typename FROM, typename TO>
	struct _ProtocolCppTypeConverter
	{
		static TO convert(FROM from)
		{
			return static_cast<TO>(from);
		}
	};

	template<class CppType, class ReprType> struct _MapElement
	{
		typedef _ProtocolCppTypeConverter<CppType, ReprType> pct;
		typedef _CppProtocolTypeConverter<CppType, ReprType> cpt;
        typedef ::boost::mpl::pair<pct, cpt> innerPair;
        typedef ::boost::mpl::pair<ReprType, innerPair> value;
		// Metafunction return value
        typedef ::boost::mpl::pair<CppType, value> type;
	};

	// The type map: C++ type -> message object type
    typedef ::boost::mpl::map <
		_MapElement<char, uint8_t>::type,
		_MapElement<unsigned char, uint8_t>::type,
		_MapElement<short, uint16_t>::type,
		_MapElement<unsigned short, uint16_t>::type,
		_MapElement<int, uint32_t>::type,
		_MapElement<unsigned int, uint32_t>::type,
		_MapElement<long, uint64_t>::type,
		_MapElement<unsigned long, uint64_t>::type,
		_MapElement<long long, uint64_t>::type,
		_MapElement<unsigned long long, uint64_t>::type,
		_MapElement<float, uint32_t>::type,
		_MapElement<double, uint64_t>::type,
		_MapElement<long double, MsgObj::LongDoubleRepresentation>::type,
		_MapElement<coral::TimeStamp, MsgObj::TimeStamp>::type,
		_MapElement<coral::Date, MsgObj::Date>::type,
		_MapElement<coral::Blob, std::string>::type
	> _CppProtocolTypeMap;

	// The message object selector metafunction
	template <typename TYPE> struct _MsgObjConverterSelector
	{
		// Try to find the type in the map
		typedef typename ::boost::mpl::has_key<_CppProtocolTypeMap, TYPE>::type FoundInMap;

		// If the element found...
		typedef typename ::boost::mpl::if_<FoundInMap,
			// return the corresponding message object type
			typename ::boost::mpl::at<_CppProtocolTypeMap, TYPE>::type,
			// else return the type itself
			typename _MapElement<TYPE, TYPE>::value
		>::type type;
	};

public:
	// The message object selector metafunction
	template <typename TYPE> struct GetMsgObjConverter
	{
		typedef typename _MsgObjConverterSelector<TYPE>::type::second::second type;
	};

	template <typename TYPE> struct GetMsgObjBackConverter
	{
		typedef typename _MsgObjConverterSelector<TYPE>::type::second::first type;
	};

	template <typename TYPE> struct GetMsgObjType
	{
		typedef typename _MsgObjConverterSelector<TYPE>::type::first type;
	};

private:
	struct _TypeMap : public  CppType_ProtocolID_Map
	{
		_TypeMap();
	};

	static _TypeMap _typeMap;
	static const uint32_t _UnsupportedTypeID = 0;
};

// Generate ProtocolID specializations for each supported type

// The input to the code generator
#define CODE_GENERATOR(r, data, i, TYPENAME) \
	template<> struct AttributeTypeTools::ProtocolIDType<TYPENAME> \
	{ \
		static const uint32_t id() {return BOOST_PP_INC(i);} \
	};

// Code generation
// Iterates through the type list and creates a template specialization for each type.
BOOST_PP_SEQ_FOR_EACH_I(CODE_GENERATOR, ~, CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST);
#undef CODE_GENERATOR

// Specializations: Convert the coral type to the message object
template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<coral::TimeStamp, MsgObj::TimeStamp>
{
	static MsgObj::TimeStamp convert(const coral::TimeStamp& from);
};

template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<coral::Date, MsgObj::Date>
{
	static MsgObj::Date convert(const coral::Date& from);
};

template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<coral::Blob, std::string>
{
	static std::string convert(const coral::Blob& from);
};

template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<float, uint32_t>
{
	static uint32_t convert(const float& from);
};

template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<double, uint64_t>
{
	static uint64_t convert(const double& from);
};

template<>
struct AttributeTypeTools::_CppProtocolTypeConverter<long double, MsgObj::LongDoubleRepresentation>
{
	static MsgObj::LongDoubleRepresentation convert(const long double& from);
};


// Specializations: Convert the message object to C++ type
template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<coral::TimeStamp, MsgObj::TimeStamp>
{
	static coral::TimeStamp convert(const MsgObj::TimeStamp& from);
};

template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<coral::Date, MsgObj::Date>
{
	static coral::Date convert(const MsgObj::Date& from);
};

template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<coral::Blob, std::string>
{
	static coral::Blob convert(const std::string& from);
};

template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<float, uint32_t>
{
	static float convert(const uint32_t from);
};

template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<double, uint64_t>
{
	static double convert(const uint64_t from);
};

template<>
struct AttributeTypeTools::_ProtocolCppTypeConverter<long double, MsgObj::LongDoubleRepresentation>
{
	static long double convert(const MsgObj::LongDoubleRepresentation& from);
};

} // namespace MsgObj

} // namespace CAL

END_CORALMESSAGING_NAMESPACE

#endif /* #ifndef CORALMESSAGING_CAL_MSGOBJ_ATTRIBUTETYPETOOLS_H_*/
