/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALMESSAGING_CAL_MSGOBJ_PARAMETERCODECS_H_
#define CORALMESSAGING_CAL_MSGOBJ_PARAMETERCODECS_H_

#include <boost/call_traits.hpp>

#include "CoralProtocol/arch.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/IQueryDefinition.h"

#include "CoralMessaging/Stream/DataStringStream.h"
#include "CoralMessaging/Stream/MessageStreamReader.h"
#include "CoralMessaging/Stream/MessageStreamWriter.h"

#include "IMessageObjectImpl.h"
#include "ProtocolCoralTypeConverter.h"

START_CORALMESSAGING_NAMESPACE

namespace CAL {

namespace MsgObj {

class Parameter_0 : public IMessageObjectImpl<Parameter_0>
{
public:
	Parameter_0()
		: IMessageObjectImpl<Parameter_0>(this)
	{}

	Parameter_0(const Parameter_0&)
		: IMessageObjectImpl<Parameter_0>(this)
	{}

	virtual bool decode(const std::string&) {return true;}

	/// Decode the object into a string and return the string. It should always be successfull.
	virtual IMessageObject::EncodedType encode() const {return IMessageObject::EncodedType(new std::string());}

	virtual std::ostream& log(std::ostream& ost) const
	{
		ost << "Par 0";
		return ost;
	}

	Parameter_0& operator = (const Parameter_0&) {return *this;}

	bool operator == (const Parameter_0&) const {return true;}

	bool operator < (const IMessageObject&) const {return true;}

	template<class QUERY_TYPE, class METHOD>
	void call(QUERY_TYPE& q, METHOD m)
	{
		(q.*m)();
	}
};

template <class T1> class Parameter_1 : public IMessageObjectImpl<Parameter_1<T1> >
{
private:
	typedef typename Pointer<Stream::DataStringStream>::Shared _StrType;
	typedef typename boost::call_traits<T1>::param_type _P1Type;

public:
	Parameter_1()
		: IMessageObjectImpl<Parameter_1<T1> >(this)
	{}

	explicit Parameter_1(_P1Type p1)
		: IMessageObjectImpl<Parameter_1<T1> >(this),
		  _p1(p1)
	{}

	Parameter_1(const Parameter_1& x)
		: IMessageObjectImpl<Parameter_1<T1> >(this),
		  _p1(x._p1)
	{}

	typename boost::call_traits<T1>::param_type p1() const {return _p1;}

	virtual bool decode(const std::string& raw)
	{
		_StrType str(new typename _StrType::element_type(raw));
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<_StrType> msr(str, CORALPROTOCOL_NAMESPACE_NAME::_system_endian);
		msr.read(_p1);
		return !msr.fail();
	}

	/// Decode the object into a string and return the string. It should always be successfull.
	virtual IMessageObject::EncodedType encode() const
	{
		_StrType str(new typename _StrType::element_type);
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<_StrType> mbw(str);
		mbw.write(_p1);
		mbw.flush();
		DEBUG_ASSERT(!str->fail());
		return IMessageObject::EncodedType(new std::string(str->str().str()));
	}

	Parameter_1& operator = (const Parameter_1& x)
	{
		if (&x != this)
			_p1 = x._p1;

		return *this;
	}

	bool operator == (const Parameter_1& x) const {return _p1 == x._p1;}

	bool operator < (const IMessageObject& x) const
	{
		return _p1 < dynamic_cast<const Parameter_1&>(x)._p1;
	}

	template<class QUERY_TYPE, class METHOD>
	void call(QUERY_TYPE& q, METHOD m)
	{
		(q.*m)(ProtocolCoralTypeConverter::GetTypeConverter<T1>::type::convert(_p1));
	}

	virtual std::ostream& log(std::ostream& ost) const
	{
		ost << "Par 1 <" << _p1 << ">";
		return ost;
	}

private:
	T1 _p1;
};

template <class T1, class T2>
class Parameter_2 : public IMessageObjectImpl<Parameter_2<T1, T2> >
{
private:
	typedef typename Pointer<Stream::DataStringStream>::Shared _StrType;
	typedef typename boost::call_traits<T1>::param_type _P1Type;
	typedef typename boost::call_traits<T2>::param_type _P2Type;

public:
	Parameter_2()
		: IMessageObjectImpl<Parameter_2<T1, T2> >(this)
	{}

	Parameter_2(_P1Type p1, _P2Type p2)
		: IMessageObjectImpl<Parameter_2<T1, T2> >(this),
		  _p1(p1), _p2(p2)
	{}

	Parameter_2(const Parameter_2& x)
		: IMessageObjectImpl<Parameter_2<T1, T2> >(this),
		  _p1(x._p1),
		  _p2(x._p2)
	{}

	typename boost::call_traits<T1>::param_type p1() const {return _p1;}

	typename boost::call_traits<T2>::param_type p2() const {return _p2;}

	virtual bool decode(const std::string& raw)
	{
		_StrType str(new typename _StrType::element_type(raw));
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<_StrType> msr(str, CORALPROTOCOL_NAMESPACE_NAME::_system_endian);
		msr.read(_p1);
		msr.read(_p2);
		return !msr.fail();
	}

	/// Decode the object into a string and return the string. It should always be successfull.
	virtual IMessageObject::EncodedType encode() const
	{
		_StrType str(new typename _StrType::element_type);
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<_StrType> mbw(str);
		mbw.write(_p1);
		mbw.write(_p2);
		mbw.flush();
		DEBUG_ASSERT(!str->fail());
		return IMessageObject::EncodedType(new std::string(str->str().str()));
	}

	bool operator == (const Parameter_2& x) const {return _p1 == x._p1 && _p2 == x._p2;}

	bool operator < (const IMessageObject& x) const
	{
		const Parameter_2& cx = dynamic_cast<const Parameter_2&>(x);
		return _p1 < cx._p1 || _p2 < cx._p2;
	}

	Parameter_2& operator = (const Parameter_2& p)
	{
		if (&p != this) {
			_p1 = p._p1;
			_p2 = p._p2;
		}

		return *this;
	}

	template<class QUERY_TYPE, class METHOD>
	void call(QUERY_TYPE& q, METHOD m)
	{
		(q.*m)(ProtocolCoralTypeConverter::GetTypeConverter<T1>::type::convert(_p1),
			   ProtocolCoralTypeConverter::GetTypeConverter<T2>::type::convert(_p2));
	}

	typedef IQueryDefinition& (IQueryDefinition::*defineSubQuery_type)(const std::string&);

	template<class QUERY_TYPE>
	void call(QUERY_TYPE& q, defineSubQuery_type m)
	{
		IQueryDefinition& qd = (q.*m)(_p1);
		_p2.replay(qd);
	}

	typedef IQueryDefinition& (IQueryDefinition::*applySetOperation_type)(IQueryDefinition::SetOperation);

	template<class QUERY_TYPE>
	void call(QUERY_TYPE& q, applySetOperation_type m)
	{
		IQueryDefinition& qd = (q.*m)(static_cast<IQueryDefinition::SetOperation>(_p1));
		_p2.replay(qd);
	}

	virtual std::ostream& log(std::ostream& ost) const
	{
		ost << "Par 2 <" << _p1 << "> ; <" << _p2 << ">";
		return ost;
	}

private:
	T1 _p1;
	T2 _p2;
};

} // namespace CAL

} // namespace MsgObj

END_CORALMESSAGING_NAMESPACE

#endif /* CORALMESSAGING_CAL_MSGOBJ_PARAMETERCODECS_H_ */
