/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CTLERRORCODE_H_
#define CTLERRORCODE_H_

#include "CoralMessaging/base.h"

START_CORALMESSAGING_NAMESPACE

namespace CTL {

/// We use the Boost preprocessor metaprogramming library to generate some repeated
/// code and code with repeating pattern. The labels here represent the the CTL
/// messages. The order of the labels are in increasing CTL message code, the
/// first label os zero. We assume that there is no gap in the message codes - 
/// if we fix the error codes in that way, it greatly simplifies coding the
/// repeats.
#define ERRCODE_LABELS (OK)(VersionMismatch)(WrongOpcode)(SegmentationError)\
(NotConnected)(SyntaxError)(OpcodeMismatch)(GeneralError)

/// Fetch the highest error code from the ERRCODE_LABELS. This will be the terminator
/// of the meta-sequences.
#define LAST_ERRCODE BOOST_PP_CAT(E_, BOOST_PP_SEQ_HEAD(BOOST_PP_SEQ_REVERSE(ERRCODE_LABELS)))

/** Represents a CTL error (message) code.
 * 
 * The class encodes/decodes the CTL error codes.
 */
class ErrorCode
{
public:
	// Generate enum values for all the CTL error codes. The labels are given by 
	// ERRCODE_LABELS, and prefixed by E_. So, the generated code will look like 
	//
	// enum {E_OK = 0, E_VersionMismatch = 1, ...., E_Invalid = 0xff}
	#define ENUM_GEN(r, data, i, elem) BOOST_PP_CAT(E_, elem) = i,
	
	enum { BOOST_PP_SEQ_FOR_EACH_I(ENUM_GEN, ~, ERRCODE_LABELS)
		   E_Invalid = 0xFF
	};

	#undef ENUM_GEN
	
	/// Constructor
	explicit ErrorCode(const uint8_t errcode = E_Invalid) : _errcode(errcode) {}
	
	/// Check if the stored value represents a valid CLT error code. 
	///
	/// This method is called when we construct the messgae from the stream, and we need to check
	/// if the error code in the message is valid or not.
	bool valid() const {return _errcode <= LAST_ERRCODE;}
	
	/// Convert the error code into a byte.
	operator uint8_t () const {return _errcode;}
	
private:
	uint8_t _errcode; /**< Value of the error code (may be invalid...) */
};

/// Error code class generator. The code is given by the template parameter. The
/// function checks in compile time if the value is valid or not.
template <int ERRCODE> inline ErrorCode ErrorCode_gen()
{
	BOOST_STATIC_ASSERT((ERRCODE <= ErrorCode::LAST_ERRCODE));
	return ErrorCode(ERRCODE);
}

#undef LAST_ERRCODE

} // namespace CTL

END_CORALMESSAGING_NAMESPACE

#endif /*CTLERRORCODE_H_*/
