/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CTLMESSAGEBUFFER_H_
#define CTLMESSAGEBUFFER_H_

#include <boost/statechart/event.hpp>
#include <boost/statechart/state_machine.hpp>
#include <boost/statechart/simple_state.hpp>
#include <boost/statechart/transition.hpp>
#include <boost/statechart/custom_reaction.hpp>
#include <boost/mpl/list.hpp>

#include "CoralMessaging/CoralSocket.h"
#include "CoralMessaging/Pointers.h"
#include "CTLPacket.h"
#include "CTLMessages.h"

#include <deque>

START_CORALMESSAGING_NAMESPACE

namespace sc = boost::statechart;

namespace CTL {

namespace MessageBufferSM {

typedef Pointer<Packet>::Shared PacketPtr; /**< Pointer to a Packet */
typedef std::deque<PacketPtr> ListOfPackets; /**< List of valid packets that constitute the message */

/** Event representing arrival of a new packet.
 *
 * The event stores the packet and the sending CORAL socket.
 * This event may be sent only in StEmptyMessage and StReceivingPackets states.
 */
struct EvAddPacket : public sc::event<EvAddPacket> {
	/// Constructor
	EvAddPacket(const PacketPtr& p, /**< The new packet */
			    CoralInputSocket& csock /**< The sender of the message */) : packet(p), csock(csock) {}

	/// Copy constructor
	EvAddPacket(const EvAddPacket& x) : sc::event<EvAddPacket>(), packet(x.packet), csock(x.csock) {}

	PacketPtr packet; /**< The new packet */
	CoralInputSocket csock; /**< The sender of the packet */
};

/** Event representing reading out a message.
 *
 * The event informs the state machine where the raw message data should be copied.
 * The event handler will do the copy.
 * The message data may be readed out in chunks. When all the message data is read out,
 * the state machine changes to StEmptyMessage state.
 *
 * This message may be sent only in StReadingOut state.
 */
struct EvReadData : public sc::event<EvReadData> {
	/// Constructor
	EvReadData(char* d, /**< The place where the message data should be copied */
			   size_t l, /**< The maximal length of required data */
			   size_t& r /**< The actually copied data */) : data(d), len(l), readed(r) {}

	/// Copy constructor
	EvReadData(const EvReadData& x) : sc::event<EvReadData>(), data(x.data), len(x.len), readed(x.readed) {}

	char* data; /**< The place where the message data should be copied */
	size_t len; /**< The maximal length of required data */
	size_t& readed; /**< The actually copied data */
};

/** Polymorfic base of the states, required by the state machine. */
struct StateOperations
{
	virtual ~StateOperations() {}

	virtual bool received() const = 0;

	virtual CoralProtocol::Opcode opcode() const = 0;

	virtual uint8_t firstByte() const = 0;
};

// Preliminary declaration
class StEmptyMessage;

/**The MessageBuffer state machine.
 *
 * See the state machine design description.
 * Starts in StEmptyMessage state. Terminated state means failure.
 */
class MessageBufferStateMachine : public sc::state_machine<MessageBufferStateMachine, StEmptyMessage>
{
public:
	/// The individual states store the relevant data here. This is the point where the
	/// series of segments are distilled into the message - the "assembly line".
	/// All the state operations add
	/// information here that is extracted at that particular state and reused by the
	/// other states. This structure is just a collection of objects with no internal
	/// behaviour, and those objects are manipulated by all the classes in this compilation
	/// unit.
	struct SharedData {
		/// Constructor
		SharedData() : packetList(new ListOfPackets),  messageLength(0) {}

		CoralProtocol::Opcode opcode; /**< The opcode of the message */
		CoralProtocol::SequenceNumber lastSeqNum; /**< The last valid sequence number */
		Pointer<ListOfPackets>::Shared packetList; /**< The list of valid packets constituting the message */
		uint8_t firstByte; /**< The very first byte of the message. Necessary to be able to determine the CTL message type. */
		long long messageLength; /**< The full length of the message body (calculated by the individual packet segments) */
	};

	/// True if the message is received fully (all the segments). Delegates the call to the actual state
	bool received() const;

	/// The opcode of the message. Error if called before this information is known. Delegates the call to the actual state.
	CoralProtocol::Opcode opcode() const;

	/// The first byte of the message. Error if called before this information is known.
	uint8_t firstByte() const;

	/// Return the message data to handle by the states.
	SharedData& data() {return _data;}

	/// Return the message data to use by the states.
	const SharedData& const_data() const {return _data;}

private:
	SharedData _data;
};

/// Operations that have different behaviours in different states. The user communicates
/// with the state machine via these methods.
///
/// The default behaviour is software error (assertion), meaning that a given method
/// cannot be called from that particular state. The states must override those methods
/// that have meaningful meaning in that state.
///
/// The same applies for event handlers: the default event handler is software failure.
/// Sending an event to a state that should not handle that event signals logic error
/// somewhere else, so we cannot simply ignore it. Let's crash the code visibly.
template <class DERIVED>
struct BaseState : public StateOperations,
                   public sc::simple_state<DERIVED, MessageBufferStateMachine>
{
	/// List of handled events
    typedef ::boost::mpl::list< sc::custom_reaction<EvAddPacket>, sc::custom_reaction<EvReadData> > reactions;

	/// True if the message is received fully (all the segments)
	virtual bool received() const
	{
		return false;
	}

	/// The opcode of the message. Error if called before this information is known.
	virtual CoralProtocol::Opcode opcode() const
	{
		FORCE_CRASH_HERE
		return CoralProtocol::Opcode();
	}

	/// The first byte of the message. Error if called before this information is known.
	virtual uint8_t firstByte() const
	{
		FORCE_CRASH_HERE
		return uint8_t();
	}

	/// Handle the EvReadData message
	virtual sc::result react(const EvReadData&)
	{
		FORCE_CRASH_HERE
		return this->terminate();
	}

	/// Handle the EvAddPacket message
	virtual sc::result react(const EvAddPacket&)
	{
		FORCE_CRASH_HERE
		return this->terminate();
	}
};

/// When the message is assembled, then we can return meaningful information in the states
/// coming after message assembly.
template <class DERIVED>
struct BaseStateWhenAssembled: public BaseState<DERIVED>
{
	using BaseState<DERIVED>::react;

	// From StateOperations
	virtual bool received() const
	{
		return true;
	}

	virtual CoralProtocol::Opcode opcode() const
	{
		return this->outermost_context().const_data().opcode;
	}

	virtual uint8_t firstByte() const
	{
		return this->outermost_context().const_data().firstByte;
	}
};

/** State representing the final state of the state machine: the message is fully received
 * and totally read out.
 *
 * See the state machine design document.
 */
struct StFinished : public BaseStateWhenAssembled<StFinished>
{

};

/** State class. The message has been arrived fully, the reconstructed (unsegmented)
 * message payload is being read out.
 *
 * See the state machine design document.
 *
 * The user copies the message payload into a continous buffer, so the segmentation
 * information is lost (teh message is reconstructed here). It iterates through the
 * packets, and returns their payload. The state changes into "finished" if the payload
 * of all the packets is copied.
 */
class StReadingOut : public BaseStateWhenAssembled<StReadingOut>
{
public:
	using BaseStateWhenAssembled<StReadingOut>::react;

	/// Constructor.
	StReadingOut();

	/// From StateOperations
	virtual sc::result react(const EvReadData& er);

private:
	PacketPtr _actual; /**< The actual packet of the message whose payload is readed out. */
	size_t _cursor; /**< The position of the end of the last read in the actual packet payload */
};

/** State class. The message has been arrived fully.
 *
 * See the state machine design document.
 *
 * It is an internal (transitional) state, performs some housekeeping after
 * the mesage reception, and turns into "Reading out" state. It makes the value of the
 * first byte available.
 */
struct StReceived : public BaseState<StReceived>
{
	/// List of handled events
	typedef sc::custom_reaction<EvAddPacket> reactions;

	/// Handler of EvAddPacket. This was the last packet, so turn into "reading out" state.
	sc::result react(const EvAddPacket&);
};

/** State class. Receives packets. Stay in this state until the last packet has been arrived.
 *
 * See the state machine design document.
 */
struct StReceivingPackets : public BaseState<StReceivingPackets>
{
	using BaseState<StReceivingPackets>::react;

	/// Handler of EvAddPacket. Check the segmantation, and the packet to the list. send back
	/// protocol error in case of segmentation error.
	virtual sc::result react(const EvAddPacket& er);

	// Here we have opcode information.
	virtual CoralProtocol::Opcode opcode() const;
};

/** State class representing an empty message.
 *
 * This is the starting state of the state machine. It waits for the first packet.
 *
 * See the state machine design document.
 */
struct StEmptyMessage : public BaseState<StEmptyMessage>
{
	using BaseState<StEmptyMessage>::react;

	/// Handler of EvAddPacket.
	virtual sc::result react(const EvAddPacket& er);
};

} // namespace MessageBufferSM

/** Class responsible for reconstructing a message from the message segments (packets).
 *
 * The class does not receive data from the line, it is the responsibility of the PacketReceiver.
 * The MessageBuffer takes the received, valid packets and checks if they have valid
 * segmantation. If so, it creates a list of valid segments. The message is received when
 * all the segments arrived with valid segmentation.
 *
 * Not all the methods of the class may be called in every state of the state machine.
 * In wrong method calls, assertion is triggered.
 */
class MessageBuffer
{
public:
	/// Constructor
	MessageBuffer();

	/// Destructor
	virtual ~MessageBuffer() {};

	/// Add a new packet to the message
	bool addPacket(Pointer<Packet>::Shared& packet, /**< The packet to be added */
			       CoralInputSocket& csock /**< The sender of the packet */);

	/// Return the first byte of the message payload
	uint8_t firstByte() const;

	/// Copy a given amount of data from the recronstructed message to a buffer
	/// (read out the message).
	///
	/// Subsequent calls will continue where the actual call finished the copy.
	/// The segmentation information is lost here, we see the whole message payload
	/// as a continous buffer.
	/// Returns the actually copied number of bytes.
	size_t read(char* data, /**< The destination of the copy */
			    size_t len /**< Amount of data to be copied */);

	/// Return true if failure occured during the message reconstruction
	bool fail() const;

	/// Return the opcode of the message
	CoralProtocol::Opcode opcode() const;

private:
	MessageBufferSM::MessageBufferStateMachine _sm; /**< The message receiver state machine */
};

} // namespace CTL

END_CORALMESSAGING_NAMESPACE

#endif /*CTLMESSAGEBUFFER_H_*/
