/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CTLPACKETREADER_H_
#define CTLPACKETREADER_H_

#include <string>

#include <boost/statechart/event.hpp>
#include <boost/statechart/state_machine.hpp>
#include <boost/statechart/simple_state.hpp>
#include <boost/statechart/state.hpp>
#include <boost/statechart/transition.hpp>
#include <boost/statechart/custom_reaction.hpp>

#include "CoralMessaging/Stream/DataStream.h"
#include "CoralMessaging/Utils/Logger.h"
#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Timeout.h"

#include "CTLPacket.h"

namespace sc = boost::statechart;

START_CORALMESSAGING_NAMESPACE

namespace CTL {

namespace PacketReaderSM {

/** Get the next packet from the line.
 * 
 * The event stores a timeout value: this is the max. time while we wait for a packet.
 * The packet can returned only if it has been arrived fully.
 */ 
class EvPickPacket : public sc::event<EvPickPacket> 
{
public:
	/// Constructor
	explicit EvPickPacket(const Timeout& to /**< Max. time of waiting for packet arrival */) : _to(to.clone()) {}
	
	/// Copy constructor.
	EvPickPacket(const EvPickPacket& x) : sc::event<EvPickPacket>(), _to(x._to->clone()) {}
	
	/// The timeout stored in the event.
	const Timeout& timeout() const {return *_to;}
	
private:
	Pointer<Timeout>::Scoped _to; /**< Max, wait time for a packet */
};

/** Polymorfic base of the states, required by the state machine. */
struct StateOperations
{
	/// Destructor
	virtual ~StateOperations() {}
	
	virtual Pointer<Packet>::Shared get() const = 0; 
};

// The nitial state
class StObtainingHeader;

/**The PacketReader state machine.
 *
 * See the state machine design description.
 * Starts in StObtainingHeader state. Terminated state means failure.
 */
class PacketReaderStateMachine : public sc::state_machine<PacketReaderStateMachine, StObtainingHeader>
{
public:
	/// The individual states store the relevant data here. This is the point where the
	/// packet is received and built - the "assembly line". All the state operations add
	/// information here that is extracted at that particular state and reused by the
	/// other states. This structure is just a collection of objects with no internal
	/// behaviour, and those objects are manipulated by all the classes in this compilation
	/// unit.
	struct SharedData {
		typedef Pointer<ICoralIOHandler>::Shared IOHandler;
		
		/// Constructor
		SharedData(IOHandler& s) : dataStr(s) {}
		
		IOHandler dataStr; /**< The physical stream where the packet is read from */
		Pointer<Packet>::Shared packet; /**< The packet is constructed here */
	};
	
	/// Constructor
	PacketReaderStateMachine(Pointer<ICoralIOHandler>::Shared& s) : _data(s) {}
	
	/// Get the packet. The packet is provided only in the right states, defer the execution to the actual state.
	/// Zero pointer indicates that we have not received a full packet.
	Pointer<Packet>::Shared get() const; 
	
	/// Return the message data to handle by the states.
	SharedData& data() {return _data;}
	
	/// Return the message data to use by the states.
	const SharedData& const_data() const {return _data;}
	
private: 
	SharedData _data;
};

/// Operations that have different behaviours in different states. The user communicates
/// with the state machine via these methods.
///
/// The default behaviour is software error (assertion), meaning that a given method
/// cannot be called from that particular state. The states must override those methods
/// that have meaningful meaning in that state.
///
/// The same applies for event handlers: the default event handler is software failure.
/// Sending an event to a state that should not handle that event signals logic error 
/// somewhere else, so we cannot simply ignore it. Let's crash the code visibly.
template <class DERIVED>
struct BaseState : public StateOperations,
                   public sc::simple_state<DERIVED, PacketReaderStateMachine>
{
	/// List of events to be handled.
	typedef sc::custom_reaction<EvPickPacket> reactions;
};

/** State class. A full and valid packet has been received. 
 * 
 * See the state maching documentation.
 * */
struct StPacketReceived : public BaseState<StPacketReceived> 
{
	/// Handle the EvPickPacket event.
	sc::result react(const EvPickPacket& er);
	
	virtual Pointer<Packet>::Shared get() const;
};

/** Base of the states of StObtainingPacket. */
template<class State>
class ObtainingData : public BaseState<State>
{
public:			
	/// Handle the EvPickPacket event - the common logic of receiving the header and the payload..
	template<class DataStreamType>
	sc::result react_global(const EvPickPacket& er, DataStreamType& str) 
	{
		// this is the do / obtainHeaderData action
		str.read(er.timeout());
		if (str.isFail()) {
			LOGGER_NEWLOG(coral::Debug) << "Stream failure, terminating" << commit;
			return this->terminate();
		}
		if (str.isReceiving()) return this->discard_event();
		// We got the required data, now handle the situation in the particular state
		if (str.isFinished()) return _handleDataReception(str.data(), er);
		DEBUG_ASSERT(1);
		return this->discard_event(); 
	}
	
	virtual Pointer<Packet>::Shared get() const
	{
		return Pointer<Packet>::Shared();
	}
	
protected:
	/// When we got the required data, handle the situation in the particular state
	virtual sc::result _handleDataReception(const std::string& data, const EvPickPacket& er) = 0;
};

/** State of obtaining the payload of a packet. */
class StObtainingPayload : public ObtainingData<StObtainingPayload>
{
public:
	/// Handle the EvPickPacket event
	sc::result react(const EvPickPacket& er);
	
private:
	virtual sc::result _handleDataReception(const std::string&, const EvPickPacket&);
	
	typedef Stream::DataStreamConcepts::ExternalBuf _StreamType; /** The payload of the packet is written explicitely */
	Pointer<_StreamType>::Scoped _str; /**< The specific DataStream object bound to the physical line */
};

/** State of obtaining the header of a packet. */
class StObtainingHeader : public ObtainingData<StObtainingHeader>
{
public:
	/// Handle the EvPickPacket event
	sc::result react(const EvPickPacket& er);
	
private:
	virtual sc::result _handleDataReception(const std::string& data, const EvPickPacket& er);
	
	typedef Stream::DataStreamConcepts::InternalBuf _StreamType; /**< we maintain an internal data reception buffer,
	                                                                  and construct the packet header of the content */ 
	Pointer<_StreamType>::Scoped _str; /**< The specific DataStream object bound to the physical line */
};

} // namespace PacketReaderSM

/** Class responsible for reading a packet from the physical data source.
 * 
 * Internally, the class employs a state machine with states like receiving header,
 * receiving payload, etc. The pickPacket method attempts to get a packet from the line.
 * In a given timeout, it either returns a full, valid packet or returns a zero
 * pointer. Zero pointer indicates that the packet cannot be received fully, or there is
 * no data coming. In case of communication error, exception should be thrown by the
 * communication line handlers.
 * 
 * The object can be used to obtain packets repeatedly. After receiving a packet, the next
 * pickPacket call will wait for the next packet. 
 */
class PacketReader : public ::boost::noncopyable
{
public:
	/// Constructor
	PacketReader(Pointer<ICoralIOHandler>::Shared s /**< Handler of the physical data line */);
	
	/// Destructor
	virtual ~PacketReader() {};
	
	/// Attempt to read a packet from the line. Return a pointer to the packet if in the given
	/// timeout, we were able to read a full, valid packet, else zero pointer.
	Pointer<Packet>::Shared pickPacket(const Timeout& to /**< Timeout */); 
	
	/// Return true if the packet reception is terminated for any reason.
	bool terminated() const;
	
private:
	PacketReaderSM::PacketReaderStateMachine _sm;
};

} // namespace CTL

END_CORALMESSAGING_NAMESPACE

#endif /*CTLPACKETREADER_H_*/
