/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSOCKET_H_
#define CORALSOCKET_H_

#include "CoralProtocol/PacketHeader.h"
#include "CoralMessaging/CoralSocket/CoralInputSocket.h"

START_CORALMESSAGING_NAMESPACE

class ICoralIOHandler;
class IMessage;

/** class CoralSocket
 * 
 * Descriptor of the origin/target of a Coral message. It encapsulates:
 * 
 * - The stream where the comes from/should be sent to
 * - The client ID
 * - The reques ID
 * 
 * The CoralSocket remains is assigned to the (physical line - client - request) triplet
 * until the object lifetime. 
 * 
 * The three pieces of information identifies unambigouosly the origin/target of the message
 * and the request itself.
 * 
 * When we send a message, then the CoralSocket is responsible for the proper segmentation
 * of the messages. Therefore, the socket needs to know the information that would
 * go into the packet headers in advance: the message opcode and the message cacheability.
 * Using this information, the CoralSocket is able to produce automatically all the packet headers
 * for the message segments.
 * 
 * The CoralSocket implements the messgage segmentation part of the CORAL protocol.
 */
class CoralSocket : public CoralInputSocket
{
public:
	/// Constructor. Requires explicitely the values in the packet header.
	CoralSocket(Pointer<ICoralIOHandler>::Shared str, /**< The physical handler of the data */ 
			    uint32_t clientID, /**< The client ID of the sender */
				uint32_t requestID, /**< The request ID of the message */
				uint8_t opcode, /**< The opcode ID io the message */
				bool cacheable /**< true if the message is cacheable else false */ );

	/// Constructor. Extract the values that go to the packet header from a given message.
	CoralSocket(Pointer<ICoralIOHandler>::Shared str, /**< The physical handler of the data */ 
			    uint32_t clientID, /**< The client ID of the sender */
			    uint32_t requestID, /**< The request ID of the message */ 
			    const IMessage& msg/**< The opcode and the cacheability are extracted from here */ );
	
	/// Constructor. Extract the values that go to the packet header from a given packet header and a message.
	CoralSocket(Pointer<ICoralIOHandler>::Shared str, /**< The physical handler of the data */
			    const CoralProtocol::PacketHeader& phead, /**< The request and client ID-s are extracted from here */ 
			    const IMessage& msg /**< The opcode and the cacheability are extracted from here */);

	/// Constructor. Extract the values that go to the packet header from a given CoralInputSocket header and a message.
	CoralSocket(const CoralInputSocket& csock, /**< The request and client ID-s are extracted from here */
				const IMessage& msg /**< The opcode and the cacheability are extracted from here */);
	
	/// Destructor
	virtual ~CoralSocket() {};
	
	/// Write the raw data into a CoralSocket. The data is in the buffer, and the given bytes of
	/// data is written. The socket attempts to write all the data.
	void write(const char* data, size_t size);
	
	/// True if the socket is in failure state (the last communication was unsuccessfull). It is an
	/// unrecoverable state.
	bool fail() const {return _fail;}
	
	/// Send the actually accumulated data as a packet and start a new packet. Return true in case of success. 
	bool flush(bool isFinal = true /**< Indicate if the actually sent packet is the final one (true). True is the default value. */); 

private:
	uint8_t _opcode; /**< The opcode that are written in the packet headers */
	bool _cacheable; /**< Cacheable flag is set by this variable (true: cacheable flag is set) */
	CoralProtocol::SequenceNumber _seqNum; /** The sequence numbers are tracked here, increaded after sending a packet */
	std::string _data; /**< The data buffer. Data is accumulated here, and sent after flush, when the content is cleared */
	bool _fail; /**< Failure state of the object. Nonrecoverable: once it is set, no more communication will be performed in this CoralSocket. */
};

END_CORALMESSAGING_NAMESPACE

#endif /*CORALSOCKET_H_*/
