/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef ICORALIOHANDLER_H_
#define ICORALIOHANDLER_H_

#include "CoralMessaging/Timeout.h"
#include "CoralMessaging/Utils/RWLockable.h"

START_CORALMESSAGING_NAMESPACE

/** Class represents a communication endpoint (where raw byte data is readed and written).
 * 
 * It may be a string stream, a TCP socket, a file, etc., encapsulated behind the interface.
 * The class must maintain failura state: in case of communication error, the fail()
 * function must return false.
 * 
 * The interface is not really "clean", as it inherits from the RWLockable class. The class
 * is an utility class providing scoped read/write locks. It means that all the objects
 * implementing ICoralIOHandler are read/write lockable by scoped locks.
 */
class ICoralIOHandler : public RWLockable
{
public:
	/** destructor */
	virtual ~ICoralIOHandler() {};
	
	/** Read raw data into the buffer. 
	 * 
	 * The buffer must be big enough to hold the given amount of data. The read fails
	 * if there is less data than the required amount.
	 */
	virtual size_t read(char* buf, /**< The buffer where the data is read in */ 
					    size_t len /**< The required bytes of data to be read */) = 0;
	
	/** Read a given amount of raw data into the buffer. Return true if all the data was
	 * received successfully in the given timeout else false. */
	virtual bool readAll(char* buf, /**< The buffer where the data is read in */
			             size_t len, /**< The required bytes of data to be read */
			             const Timeout& to  /** Try to read until this timeout */) = 0;
	
	/** Attempt to write the given amount of raw data to the communication endpoint. Return the amount of data
	 * actually written. */
	virtual size_t write(const char* buf, /**< The buffer where the data is written from */ 
			             size_t len /**< The required bytes of data to be written */) = 0;
	
	/** Write a given amount of raw data to the communication endpoint. Return true if all the data was
	 * written successfully in the given timeout else false. */
	virtual bool writeAll(const char* buf, /**< The buffer where the data is written from */ 
			              size_t len, /**< The required bytes of data to be written */
			              const Timeout& to /** Try to write until this timeout */) = 0;
	
	/** Returns fail if communication error occured during any of the reads, else true. */
	virtual bool fail() const = 0;
	
	/** Waits until data is ready or timeout occured. Return true if data is ready, false if 
	 * timeout occured.
	 * 
	 * We can use the default wait time. The behavour (blocking or polling) depends on
	 * the default behaviour of Timeout. */
	virtual bool waitForData(const Timeout& to) = 0;
	
	virtual bool isClosed() const = 0;
};

END_CORALMESSAGING_NAMESPACE

#endif /* ICORALIOHANDLER_H_ */
