/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALMESSAGING_IMESSAGEOBJECT_H_
#define CORALMESSAGING_IMESSAGEOBJECT_H_

#include "CoralMessaging/Pointers.h"
#include <string>
#include <iosfwd>

START_CORALMESSAGING_NAMESPACE

/** Interface to a message element (object).
 *
 * If a message field is a composite object, it is encoded/decoded into arrays when
 * it is written to / read from a stream. This interface gives access to the encoding/decoding
 * functionalities.
 */
class IMessageObject : private boost::noncopyable
{
public:
	typedef Pointer<std::string>::Shared EncodedType;

	/// destructor
	virtual ~IMessageObject() {}

	/// Will be called by the CAL protocol layer. Encoded a string to the object.
	/// Return true if the string representation is syntactically correct.
	virtual bool decode(const std::string& raw /**< The raw string (Array) field that holds the encoded object */) = 0;

	/// Decode the object into a string and return the string. It should always be successfull.
	virtual EncodedType encode() const = 0;

	/// Write the textual representation of the content into a (log) stream. Default: write nothing.
	virtual std::ostream& log(std::ostream& ost) const
	{
		ost << "<not implemented>";
		return ost;
	}

	virtual bool operator < (const IMessageObject& x) const = 0;

	virtual IMessageObject* clone() const = 0;

	virtual bool isEqual(const IMessageObject& x) const = 0;
};

struct IMessageObjectComparator : std::binary_function<IMessageObject, IMessageObject, bool>
{
	bool operator() (const IMessageObject& lhs, const IMessageObject& rhs) const
	{
		return lhs.isEqual(rhs);
	}
};

inline std::ostream& operator << (std::ostream& ost, const IMessageObject& mo)
{
	return mo.log(ost);
}

END_CORALMESSAGING_NAMESPACE

#endif /* CORALMESSAGING_IMESSAGEOBJECT_H_ */
