// -*- C++ -*-
// $Id: Attribute.h,v 1.3 2008/10/16 03:34:07 molnarzs Exp $

#ifdef CORAL_SEALED

#ifndef CORALBASE_ATTRIBUTE_H
#define CORALBASE_ATTRIBUTE_H

#include "AttributeException.h"
#include "CoralBase/AttributeSpecification.h"

#include <typeinfo>
#include <iosfwd>

namespace coral {

  // forward declarations
  class AttributeData;

  /**
     @class Attribute Attribute.h CoralBase/Attribute.h

     An attribute class which holds the data and a reference to its specification (type, name)
  */
  class Attribute {
    public:
      /// Assignment operator. Copies data.
      Attribute& operator=( const Attribute& rhs );

      /// Copies data without performing check on the type
      void fastCopy( const Attribute& rhs );

      /// Returns the specification of the attribute.
      const AttributeSpecification& specification() const;

      /// Returns a reference to the actual data.
      template<typename T> T& data();
      template<typename T> const T& data() const;

      /// Sets the value from an external variable. Performs type checking
      template<typename T> void setValue( const T& value );

      /// Binds safely to an external variable.
      template<typename T> void bind( const T& externalVariable );
      template<typename T> void bind( T& externalVariable );

      /// Copies data from an external address. By definition this is an unsafe operation.
      void setValueFromAddress( const void* externalAddress );

      /// Copies data to an external address. By definition this is an unsafe operation.
      void copyValueToAddress( void* externalAddress ) const;

      /// Returns the addess of the actual data.
      void* addressOfData();
      const void* addressOfData() const;

      /// Shares the data with another Attribute.
      void shareData( const Attribute& sourceAttribute );

      /// Binds to an external address (unsafe operation).
      void bindUnsafely( const void* externalAddress );
      void bindUnsafely( void* externalAddress );

      /// Sets the NULL-ness of the variable
      void setNull( bool isVariableNull = true );

      /// Retrieves the NULL-ness of the variable
      bool isNull() const;

      /// Returns the size of the variable
      int size() const;

      /// Equal operator compares data and type. Not the name!
      bool operator==( const Attribute& rhs ) const;

      /// Comparison operator compares data and type. Not the name!
      bool operator!=( const Attribute& rhs ) const;

      /// Dumps the attribute contents
      std::ostream& toOutputStream( std::ostream& os, bool valueOnly = false ) const;

    private:
      /// Binds safely to an external variable.
      void bindVariable( const std::type_info& type, const void* externalAddress );
      void bindVariable( const std::type_info& type, void* externalAddress );

      /// Sets the value safely from an external source
      void setValue( const std::type_info& type, const void* externalAddress );

    private:
      /// Only the AttributeList can create Attributes
      friend class AttributeList;

      /// Constructor
      explicit Attribute( const AttributeSpecification& specification );

      /// Destructor
      ~Attribute();

      /// No copy constructor.
      Attribute( const Attribute& rhs );

    private:
      /// The specification
      const AttributeSpecification* m_specification;

      /// The data
      AttributeData* m_data;
  };


  // Inline methods
  inline const AttributeSpecification&
  Attribute::specification() const
  {
    return *m_specification;
  }

  
  template<typename T>
  const T& Attribute::data() const
  {
    if( this->specification().type() != typeid(T))
      throw coral::AttributeException( "Attempt to assign attribute \"" + this->specification().name() + "\" of type " + this->specification().typeName() + " with " + coral::AttributeSpecification::typeNameForId(typeid(T)) );
    
    return *( static_cast< const T* >( this->addressOfData() ) );
  }


  template<typename T>
  T& Attribute::data()
  {
    if( this->specification().type() != typeid(T))
      throw coral::AttributeException( "Attempt to assign attribute \"" + this->specification().name() + "\" of type " + this->specification().typeName() + " with " + coral::AttributeSpecification::typeNameForId(typeid(T)) );

    return *( static_cast< T* >( this->addressOfData() ) );
  }


  template<typename T>
  void Attribute::bind( const T& externalVariable )
  {
    this->bindVariable( typeid(T), &externalVariable );
  }


  template<typename T>
  void Attribute::bind( T& externalVariable )
  {
    this->bindVariable( typeid(T), &externalVariable );
  }

  
  template<typename T>
  void Attribute::setValue( const T& value )
  {
    this->setValue( typeid(T), &value );
  }
}


inline std::ostream& operator << (std::ostream& os, const coral::Attribute& x)
{
    x.toOutputStream(os);
    return os;
}


#endif // CORALBASE_ATTRIBUTE_H

#else // #ifdef CORAL_SEALED

#include "CoralBase/Attribute.h"

#endif // #ifdef CORAL_SEALED

