// -*-C++-*-
#ifdef CORAL_SEALED

#ifndef CORAL_CORALBASE_TIMESTAMP_H
#define CORAL_CORALBASE_TIMESTAMP_H

#include <sstream>
#include "SealBase/Time.h"

namespace coral
{
  /**
     @class TimeStamp TimeStamp.h CoralBase/TimeStamp.h

     A class defining the TIMESTAMP type implemented on the seal::Time class
   */
  class TimeStamp
  {
  public:
	typedef signed long long int ValueType;
	
    /// Default constructor
    TimeStamp();

    /// Constructor
    TimeStamp( int year,
               int month,
               int day,
               int hour,
               int minute,
               int second,
               long nanosecond );

    /// Constructor from a seal time
    explicit TimeStamp( const seal::Time& sealTime );

    /// Destructor
    ~TimeStamp();

    /// Copy constructor
    TimeStamp( const TimeStamp& rhs );

    /// Assignment operator
    TimeStamp& operator=( const TimeStamp& rhs );

    /// Returns the year
    int year() const;

    /// Returns the month [1-12]
    int month() const;

    /// Returns the day [1-31]
    int day() const;

    /// Returns the hour [0-23]
    int hour() const;

    /// Returns the minute [0-59]
    int minute() const;

    /// Returns the second [0-59]
    int second() const;

    /// Returns the nanosecond [0-999999999]
    long nanosecond() const;

    /// Returns the underlying seal::Time object
    const seal::Time& time() const;

    /// Equal operator
    bool operator==( const TimeStamp& rhs ) const;

    /// Comparison operator
    bool operator!=( const TimeStamp& rhs ) const;

    std::string toString() const
    {
      std::stringstream ss;
      ss << day()  << "/" << month()  << "/" << year()   << " "
         << hour() << ":" << minute() << ":" << second() << "." << nanosecond();
      return ss.str();
    }
    
    /// The number of nanoseconds from epoch 01/01/1970 UTC, normally should fit into 64bit signed integer, depends on the BOOST installation
    ValueType total_nanoseconds() const
    {
	   return 0;
    }
    
  public:
    static TimeStamp now();

  private:
    /// The actual time object
    seal::Time m_time;
  };
}




/// Inline methods
inline
coral::TimeStamp::TimeStamp():
  m_time()
{}

inline
coral::TimeStamp::TimeStamp( int year,
                             int month,
                             int day,
                             int hour,
                             int minute,
                             int second,
                             long nanosecond):
  m_time( year, month-1, day, hour, minute, second, nanosecond )
{}

inline
coral::TimeStamp::TimeStamp( const seal::Time& sealTime ):
  m_time( sealTime )
{}

inline
coral::TimeStamp::~TimeStamp()
{}

inline
coral::TimeStamp::TimeStamp( const coral::TimeStamp& rhs ):
  m_time( rhs.m_time )
{}

inline coral::TimeStamp&
coral::TimeStamp::operator=( const coral::TimeStamp& rhs )
{
  m_time = rhs.m_time;
  return *this;
}

inline int
coral::TimeStamp::year() const
{
  return m_time.year( true );
}

inline int
coral::TimeStamp::month() const
{
  return m_time.month( true ) + 1;
}

inline int
coral::TimeStamp::day() const
{
  return m_time.day( true );
}

inline int
coral::TimeStamp::hour() const
{
  return m_time.hour( true );
}

inline int
coral::TimeStamp::minute() const
{
  return m_time.minute( true );
}

inline int
coral::TimeStamp::second() const
{
  return m_time.second( true );
}

inline long
coral::TimeStamp::nanosecond() const
{
  return m_time.nsecond();
}

inline const seal::Time&
coral::TimeStamp::time() const
{
  return m_time;
}

inline bool
coral::TimeStamp::operator==( const coral::TimeStamp& rhs ) const
{
  return ( this->m_time == rhs.m_time );
}

inline bool
coral::TimeStamp::operator!=( const coral::TimeStamp& rhs ) const
{
  return ( this->m_time != rhs.m_time );
}

#endif

#else // #ifdef CORAL_SEALED

#include "CoralBase/TimeStamp.h"

#endif // #ifdef CORAL_SEALED
