/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

/* This is the class generator for the messgae objects. The generated message objects are able to:
 *
 *   - be written (encoded) to a message stream
 *   - be read from a stream (decoded), respecting the sender endianness
 *
 * All the messages must be composed from these message objects plus the simple types:
 * array, byte, word.
 *
 * The class generator uses the Boost preprocessor metaprogramming library. To understand the
 * code, google for "goost preprocessor", and read the documentation.
 */
#ifndef PAR_NUM

#ifndef CORALMESSAGING_MESSAGE_CLASSGENERATOR_H_
#define CORALMESSAGING_MESSAGE_CLASSGENERATOR_H_

#include "CoralProtocol/arch.h"

#include "CoralMessaging/Stream/MessageStreamWriter.h"
#include "CoralMessaging/Stream/MessageStreamReader.h"

#include <iosfwd>
#include <iterator>
#include <vector>

START_CORALMESSAGING_NAMESPACE

/// Converts the argument into a string literal
#define TO_STRING(arg) #arg

/** Generate code for different message object classes. Paramaters:
 *
 * CLASSNAME: the name of the generated class (no namespace)
 * TYPEn: types of the fields that compose the object
 *
 * The suffix in the macro name corresponds to the number of composing objects.
 *
 * Before generating a class, the following macros must be defined:
 *
 * CLASSNAME: the name of the class
 * CLASSNAME_STR: String representetion of the classname (goes to the logs)
 * PAR_NUM: number of the parameters (the composing objects)
 * PAR_NAMEn: method name of the n-th parameter
 *
 * Example: we generate a class called "Example", with one uint32_t parameter called
 * "method":
 *
 * #define CLASSNAME Example
 * #define CLASSNAME_STR "My Example class"
 * #define PAR_NUM 1
 * #define PAR_NAME0 method
 * CLASSGENERATOR_1(Example, uint32_t)
 *
 * The skeleton of the generated class will be:
 *
 * class Example {
 * public:
 * 	   Example(const uint32_t& par0 = uint32_t()) : _par0(par0) {}
 *     // generated methods, destructor, I/O functions, etc.
 *
 *     const uint32_t& method() const {return _par0;}
 *
 *     const std::string& name() const {static std::string s(CLASSNAME_STR); return s;}
 *
 * private:
 * 	   uint32_t _par0;
 * };
 */

#define CLASSGENERATOR_0(CLASSNAME) \
	typedef gen::CLASSNAME::CLASSNAME CLASSNAME

#define CLASSGENERATOR_1(CLASSNAME, TYPE1) \
	typedef gen::CLASSNAME::CLASSNAME<TYPE1> CLASSNAME

#define CLASSGENERATOR_2(CLASSNAME, TYPE1, TYPE2) \
	typedef gen::CLASSNAME::CLASSNAME<TYPE1, TYPE2> CLASSNAME

#define CLASSGENERATOR_3(CLASSNAME, TYPE1, TYPE2, TYPE3) \
	typedef gen::CLASSNAME::CLASSNAME<TYPE1, TYPE2, TYPE3> CLASSNAME

#define CLASSGENERATOR_4(CLASSNAME, TYPE1, TYPE2, TYPE3, TYPE4) \
	typedef gen::CLASSNAME::CLASSNAME<TYPE1, TYPE2, TYPE3, TYPE4> CLASSNAME

#define CLASSGENERATOR_5(CLASSNAME, TYPE1, TYPE2, TYPE3, TYPE4, TYPE5) \
	typedef gen::CLASSNAME::CLASSNAME<TYPE1, TYPE2, TYPE3, TYPE4, TYPE5> CLASSNAME

END_CORALMESSAGING_NAMESPACE

#endif /*#ifndef CORALMESSAGING_MESSAGE_CLASSGENERATOR_H_ */

#else // #ifndef PAR_NUM

#include <boost/preprocessor/logical/or.hpp>
#include <boost/preprocessor/logical/and.hpp>
#include <boost/preprocessor/control/iif.hpp>
#include <boost/preprocessor/repetition.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/comparison/less.hpp>
#include <boost/call_traits.hpp>

namespace gen { namespace CLASSNAME {

#if PAR_NUM
#  define TEMPLATE_EXPRESSION template < BOOST_PP_ENUM_PARAMS(PAR_NUM, class T) >
#  define TEMPLATE_TYPE_LIST < BOOST_PP_ENUM_PARAMS(PAR_NUM, T) >
#else
#  define TEMPLATE_EXPRESSION
#  define TEMPLATE_TYPE_LIST
#endif

TEMPLATE_EXPRESSION class CLASSNAME
{
public:
	#define GENERATOR_CONSTRUCTOR_INIT(z, n, unused) _par##n(par##n)
	#define GENERATOR_PARAM(z, n, unused) typename boost::call_traits<T##n>::param_type par##n = T##n()

	explicit CLASSNAME(BOOST_PP_ENUM(PAR_NUM, GENERATOR_PARAM, @) BOOST_PP_COMMA_IF(PAR_NUM) coral::CoralProtocol::t_endianness e = coral::CoralProtocol::_system_endian)
		: _valid(true), _endian(e)
		  BOOST_PP_COMMA_IF(PAR_NUM) BOOST_PP_ENUM(PAR_NUM, GENERATOR_CONSTRUCTOR_INIT, @)
	{
		// EMPTY
	}

	#undef 	GENERATOR_CONSTRUCTOR_INIT
	#undef 	GENERATOR_PARAM

	CLASSNAME(const CLASSNAME& x)
		: _valid(x._valid),
		  _endian(x._endian) BOOST_PP_COMMA_IF(PAR_NUM)
		  #define CODE_GEN(z, n, size) _par##n(x._par##n) BOOST_PP_COMMA_IF(BOOST_PP_LESS(n, size))
		  BOOST_PP_REPEAT(PAR_NUM, CODE_GEN, BOOST_PP_DEC(PAR_NUM))
		  #undef CODE_GEN
	{
		// EMPTY
	}

	virtual ~CLASSNAME() {};

	bool valid() const {return _valid;}

	const std::string& name() const
	{
		static std::string s(CLASSNAME_STR);
		return s;
	}

	template<class INPUT>
	bool read(INPUT in)
	{
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<INPUT> msr(in, _endian);
		#define GENERATOR_READ(z, n, unused) msr.read(_par##n);
		BOOST_PP_REPEAT(PAR_NUM, GENERATOR_READ, @)
		#undef GENERATOR_READ
		return (_valid = !msr.fail());
	}

	template<class OUTPUT>
	bool write(OUTPUT out) const
	{
		typename CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<OUTPUT> mbw(out);
		#define GENERATOR_APPEND(z, n, unused) mbw.write(_par##n);
		BOOST_PP_REPEAT(PAR_NUM, GENERATOR_APPEND, @)
		#undef GENERATOR_APPEND
		mbw.flush();
		return !out->fail();
	}

	bool operator == (const CLASSNAME& x) const
	{
		if (this == &x) return true;
		#define GENERATOR_COMPARE(z, n, unused) if (_par##n != x._par##n) return false;
		BOOST_PP_REPEAT(PAR_NUM, GENERATOR_COMPARE, @)
		#undef GENERATOR_COMPARE
		return true;
	}

	bool operator != (const CLASSNAME& x) const {return !(*this == x);}

	CLASSNAME& operator = (const CLASSNAME& rhs)
	{
		if (this != &rhs) {
			_valid = rhs._valid;
			_endian = rhs._endian;

			#define CODE_GEN(z, n, size) _par##n = rhs._par##n;
			BOOST_PP_REPEAT(PAR_NUM, CODE_GEN, @);
			#undef CODE_GEN
		}

		return *this;
	}

private:
	bool _valid;
	coral::CoralProtocol::t_endianness _endian;

	#define GENERATOR_MEMBERDEF(z, n, unused) \
	private: \
		T##n _par##n; \
	public: \
		typename boost::call_traits<T##n>::param_type PAR_NAME##n() const {return _par##n;}

	BOOST_PP_REPEAT(PAR_NUM, GENERATOR_MEMBERDEF, @)
	#undef GENERATOR_MEMBERDEF
};

template<class T> inline std::ostream& operator << (std::ostream& os, const std::vector<T>& x)
{
	std::copy(x.begin(), x.end(), std::ostream_iterator<const T>(os, " ; "));
	return os;
}

TEMPLATE_EXPRESSION inline std::ostream& operator <<
   (std::ostream& os, const CLASSNAME TEMPLATE_TYPE_LIST& BOOST_PP_IF(PAR_NUM, x, ))
{
	#define GENERATOR_OUT(z, n, unused) os << std::hex << "Field " << n << ": " << x.PAR_NAME##n() << " ; ";
	BOOST_PP_REPEAT(PAR_NUM, GENERATOR_OUT, @)
	#undef GENERATOR_OUT
	return os;
}

#undef TEMPLATE_EXPRESSION
#undef TEMPLATE_TYPE_LIST

}} // namespace gen

#undef PAR_NUM

#ifdef PAR_NAME0
#  undef PAR_NAME0
#endif

#ifdef PAR_NAME1
#  undef PAR_NAME1
#endif

#ifdef PAR_NAME2
#  undef PAR_NAME2
#endif

#ifdef PAR_NAME3
#  undef PAR_NAME3
#endif

#ifdef PAR_NAME4
#  undef PAR_NAME4
#endif

#ifdef PAR_NAME5
#  undef PAR_NAME5
#endif

#ifdef PAR_NAME6
#  undef PAR_NAME6
#endif

#ifdef PAR_NAME7
#  undef PAR_NAME7
#endif

#undef CLASSNAME
#undef CLASSNAME_STR

#endif // #ifndef PAR_NUM

