/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef MESSAGERECEIVER_H_
#define MESSAGERECEIVER_H_

#include <boost/ptr_container/ptr_map.hpp>

#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Timeout.h"
#include "CoralMessaging/CTL/CTLPacketReader.h"
#include "CoralMessaging/CTL/CTLMessageBuffer.h"
#include "CoralMessaging/CoralSocket.h"

START_CORALMESSAGING_NAMESPACE

class ICoralIOHandler;
class IMessage;

/** Class implementing the CORAL protocol.
 * 
 * Reads messages from the communication line, takes care of segmentation, opcodes, 
 * error handling. Only valid and syntactically correct messages are forwarded to the
 * client/server message handlers.
 */ 
class MessageReceiver
{
public:
	/// Constructor
	explicit MessageReceiver(Pointer<ICoralIOHandler>::Shared& str /**< The physical line */);
	
	/// Destructor
	virtual ~MessageReceiver() {};
	
	/** Receive a message.
	 * 
	 * Wait for a valid message until it arrives or timeout occurs. Only valid
	 * messages arrive: unknown, wrongly segmented, syntactically errorneous
	 * messages are discarded by the underlying protocol, and the required actions are
	 * taken.
	 * 
	 * The method returns true if a new message arrived, false if timeout occured.
	 */
	bool receive(const Timeout& to /**< Wait for a message until this time */);

	/** Return the origin of the last received message. Do not call it if there is no
	 * message arrived yet, it will crash the code */
	const Pointer<CoralInputSocket>::Shared from() const;
	
	/** Return the last received message. Do not call it if there is no
	 * message arrived yet, it will crash the code. */
	const Pointer<IMessage>::Shared message() const;

private:
	Pointer<ICoralIOHandler>::Shared _stream; /**< The physical line */
	CTL::PacketReader _preader; /** Read the individual packets from the line */
	
	/// Maps the origin of the message to a message buffer. The individial packets
	/// are demultiplexed by their origin (CoralSocket), and collected in a corresponding
	/// messgage buffer.
	typedef boost::ptr_map<Pointer<CoralInputSocket>::Shared, CTL::MessageBuffer> IncomingMessages;
	
	IncomingMessages _incoming; /**< the origon -> message map */
	Pointer<CoralInputSocket>::Shared _lastFrom; /**< store the origin of the last received message */
	Pointer<IMessage>::Shared _lastMessage; /**< store the last received message */
};

END_CORALMESSAGING_NAMESPACE

#endif /*MESSAGERECEIVER_H_*/
