/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#ifndef DATASTREAM_H_
#define DATASTREAM_H_

#include <string>
#include <boost/mpl/if.hpp>

#include "CoralMessaging/Stream/DataStreamLogic.h"

START_CORALMESSAGING_NAMESPACE

class Timeout;

namespace Stream {

/** Data stream maintaining external buffer */
class DataStreamExternal
{
public:
	/// Constructor
	DataStreamExternal(std::string& buf /**< Reference to the extarnal buffer */) 
		: _buf(buf) {};

protected:
	std::string& _buf; /**< Reference to the extarnal buffer */
};

/** Data stream maintaining interbal buffer */
class DataStreamInternal
{
public:
	/// Constructor 
	DataStreamInternal(size_t len /**< Length of the internal buffer */) 
		: _buf(len, 0) {};

protected:
	std::string _buf; /**< The internal buffer */
};


/** Read data from the physical stream
 * 
 * The class may have external or internal buffer. In case of external buffer, the
 * user of the class must ensure that the buffer is prepared, and exists until
 * the end of the DataStream object scope. In case of internal buffer, the buffer
 * is prepared and maintained by the object.
 * 
 * The buffer nature is a compile time decision, according to the the template parameter 
 * of the class.
 * 
 * The two constructors of the class cannot be used in the same time, one can be used
 * only extarnal, the other only internal buffer case. Mixing up the two constructors generates
 * compile time error.
 */
template <bool CreateOwnBuffer /**< true: internal buffer, false: external buffer */>
class DataStream : public boost::mpl::if_c<CreateOwnBuffer, DataStreamInternal, DataStreamExternal>::type
{
public:
	typedef typename boost::mpl::if_c<CreateOwnBuffer, DataStreamInternal, DataStreamExternal>::type BaseType;
	using BaseType::_buf;

	/// Constructor for internal buffer case
	DataStream(size_t len, /**< Length of the internal buffer */
			   typename DataStreamLogic::DataStream& src /**< The physical stream */)
		: BaseType(len), _logic(_buf, src) {}
	
	/// Constructor for external buffer case
	DataStream(std::string& buf, /**< The external buffer */
	           typename DataStreamLogic::DataStream& src /**< The physical stream */) 
		: BaseType(buf), _logic(_buf, src) {}
	
	/// Destructor
	virtual ~DataStream() {}
	
	/// Read the data. Return the read state after the read attempt.
	void read(const Timeout& to) 
	{
		_logic.read(to);
	}

	/// Return the size of the received data so far.
	size_t size() const 
	{ 
		return _logic.size(); 
	}
	
	/** Return data buffer (either external or internal).
	 * 
	 * Use this method only when the read is finished. Using it earlier does not
	 * make sense, so in debug mode it will generate a runtime error in that case.
	 */
	const std::string& data() const 
	{ 
		// Runtime error if state is not finished 
		DEBUG_ASSERT(_logic.state() == DataStreamLogic::Finished);
		return _buf; 
	}
	
	/// Return true if data read failed due to communication error.
	bool isFail() const 
	{
		return _logic.state() == DataStreamLogic::Failed;
	}
	
	/// Return true if data read finished (all the required data has been read).
	bool isFinished() const 
	{
		return _logic.state() == DataStreamLogic::Finished;
	}
	
	/// Return true if data has not been arrived fully (timeout occured in read)
	bool isReceiving() const 
	{
		return _logic.state() == DataStreamLogic::Receiving;
	}
	
private:
	DataStreamLogic _logic; /**< The physical data stream reading logic */
};

/// Type collection for the users of DataStream class
struct DataStreamConcepts
{
	typedef DataStream<false> ExternalBuf; /**< Type of DataStream with external buffer */
	typedef DataStream<true> InternalBuf; /**< Type of DataStream with internal buffer */
	typedef DataStreamLogic::DataStream DataStream; /**< Type of the underlaying physical stream class */
};

} // namespace Stream

END_CORALMESSAGING_NAMESPACE

#endif /*DATASTREAM_H_*/
