/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#ifndef DATASTREAMLOGIC_H_
#define DATASTREAMLOGIC_H_

#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Timeout.h"

START_CORALMESSAGING_NAMESPACE

class ICoralIOHandler;

namespace Stream {

/** Class encapsulating the data receiving logic required by the messaging system.
 * 
 * The communication endpoints are wrapped into this class. The behaviour it provides:
 * 
 * - Reads a required amount of data only. 
 * - Respects timeout on read
 * - The state of the data read can be queried.
 */
class DataStreamLogic 
{
public:
	/// Encapsulate the type of the underlying comm. endpoint. It must follow the pointer semantics.
	typedef Pointer<ICoralIOHandler>::Shared DataStream;
	
	/// The possible states of the reading
	typedef enum State {Finished, /**< All the required data has been arrived, no timeout occured */
	                    Failed, /**< Unrecoverable communication error */
	                    Receiving /**< The whole requested data has not been arrived in the given timeout */};
	
	/// Constructor
	DataStreamLogic(std::string& buf, /**< Read the data into this buffer. The required data amount is the length of the buffer. */
	                DataStream& str /**< Read data from this (physical) stream */); 
	
	/// Destructor
	virtual ~DataStreamLogic() {}
	
	/// Return the size of the received data so far.
	size_t size() const;
	
	/// Return the actual state of the data read.
	State state() const;
	
	/// Read the data. Return the read state after the read attempt.
	State read(const Timeout& to);
	
protected:
	std::string& _buf; /**< Read data into this buffer */
	size_t _cursor; /**< The size of actually read data so far */
	DataStream _str; /**< Read data from this stream */
	mutable State _state; /**< The actual state of the read */
};

} // namespace Stream

END_CORALMESSAGING_NAMESPACE

#endif /*DATASTREAMLOGIC_H_*/
