/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#ifndef MESSAGESTREAMREADER_H_
#define MESSAGESTREAMREADER_H_

#include <vector>
#include <string>
#include <boost/mpl/if.hpp>
#include <boost/type_traits/is_pointer.hpp>
#include <boost/type_traits/remove_pointer.hpp>

#include "CoralProtocol/utils.h"
#include "CoralProtocol/arch.h"
#include "CoralProtocol/SequenceNumber.h"

#include "CoralMessaging/Message/FixedValue.h"
#include "CoralMessaging/Message/Array.h"
#include "CoralMessaging/Stream/MessageStreamReader.h"
#include "CoralMessaging/IMessageObject.h"
#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Utils/ctl.h"

START_CORALMESSAGING_NAMESPACE

namespace Stream {

/** Read the objects that constitues the messages from a stream.
 *
 * The stream is modelled by the @see ICommunicationEndpoint abstract interface.
 * The class takes care of the endianness and the differenct codec protocols.
 *
 * The MessageStreamReader objects maintain failure state. Failure occurs in case of:
 *
 * - communication error
 * - less data than expected
 * - syntax error in the received data (accorting to the codec protocols)
 *
 * If the object turns into failure state, it cannot be reverted to good state, and all the readed
 * data should be handled as corrupt data.
 */
template <class INPUT>
class MessageStreamReader : public boost::noncopyable
{
public:
	/** Constructor */
	explicit MessageStreamReader(
			INPUT& in, /**< The communication endpoint (stream) */
			coral::CoralProtocol::t_endianness e  = CORALPROTOCOL_NAMESPACE_NAME::_system_endian /**< The endianness of the message stream */
	)
		: _in(in), _endian(e), _fail(false)
    {
    	DEBUG_ASSERT_PTR(_in);
    }

	/** Destructor */
	virtual ~MessageStreamReader() {};

	/** Read a word */
	void read(uint32_t& s)
	{
		uint32_t res;
		_in->read(reinterpret_cast<char*>(&res), sizeof(uint32_t));
		_fail = _in->fail();

		if (!_fail) {
			s = (_endian == CoralProtocol::_system_endian ? res : CoralProtocol::Utils::swap_word(res));
		}
	}

	void read(uint64_t& s)
	{
		uint64_t res;
		_in->read(reinterpret_cast<char*>(&res), sizeof(uint64_t));
		_fail = _in->fail();

		if (!_fail) {
			s = (_endian == CoralProtocol::_system_endian ? res : CoralProtocol::Utils::swap_word(res));
		}
	}

	/** Read a byte */
	void read(uint16_t& s)
	{
		uint16_t res;
		_in->read(reinterpret_cast<char*>(&res), sizeof(uint16_t));
		_fail = _in->fail();

		if (!_fail) {
			s = (_endian == CoralProtocol::_system_endian ? res : CoralProtocol::Utils::swap_word(res));
		}
	}

	/** Read a byte */
	void read(uint8_t& s)
	{
		uint8_t res;
		_in->read(reinterpret_cast<char*>(&res), sizeof(uint8_t));
		_fail = _in->fail();

		if (!_fail) {
			s = res;
		}
	}

	/** Read a byte */
	void read(bool& s)
	{
		uint8_t res;
		read(res);

		if (!_fail)
		{
			if (res <= 1)
				s = static_cast<bool>(res);
			else
				_fail = true;
		}
	}

	/** Read a string (string is stored internally as an array)*/
	void read(std::string& s)
	{
		Message::Array a;
		read(a);
		s.resize(0);
		if (!_fail) s.append(a.data());
	}

	/** Read an Array */
	void read(Message::Array& s)
	{
		uint32_t size;
		read(size);

		if (!_fail) {
			std::string data(size, 0);
			_in->read(const_cast<char*>(data.data()), size);
			_fail = _in->fail();

			if (!_fail) {
				s = Message::Array(data);
			}
		}
	}

	template<class T>
	void read(std::vector<T>& s)
	{
		_readVector(s);
	}

	/** Read a message object */
	void read(IMessageObject& mo /**< It will contain the encoded version of the query description */)
	{
		std::string qStr;
		read(qStr);

		if (!_fail) _fail = !mo.decode(qStr);
	}

	/** Read SequenceNumber object */
	void read(coral::CoralProtocol::SequenceNumber& x)
	{
		_readByteObjects(x);
	}

	/** Read a FixedByte (with value check - the value must equal to the template parameter) */
	template <uint8_t VALUE> void read(Message::FixedByte<VALUE>& val)
	{
		uint8_t v;
		read(v);

		if (!_fail) _fail = !val.isValid(v);
	}

	/** Return false if failure occurd during the read operations */
	bool fail() const {return _fail;}

private:
	/// Read objects that are represented on a byte
	template<class T>
	void _readByteObjects(T& x)
	{
		uint8_t v;
		read(v);

		if (!_fail) _fail = !x.setValue(v);
	}

	template <class T>
	struct _SimpleReader
	{
		static void read(T& x, MessageStreamReader& mr) {mr.read(x);}
	};

	template <class T>
	struct _PtrReader
	{
		static void read(T& x, MessageStreamReader& mr)
		{
			x = new typename boost::remove_pointer<T>::type;
			mr.read(*x);
		}
	};

	template<class T, template <class> class Vector>
	void _readVector(Vector<T>& s)
	{
		DEBUG_ASSERT(s.size() == 0);
		uint32_t rowNum;
		read(rowNum);

		if (!_fail) {
			s.resize(rowNum);
			typedef typename Vector<T>::iterator ItrT;
			ItrT end = s.end();

			for (ItrT i = s.begin(); i != end; ++i) {
				typedef typename boost::mpl::if_<boost::is_pointer<typename Vector<T>::value_type>, _PtrReader<T>, _SimpleReader<T> >::type Reader;
				Reader::read(*i, *this);
				if (_fail) break;
			}
		}
	}

	INPUT _in;
	coral::CoralProtocol::t_endianness _endian;
	bool _fail;
};

} // namespace Stream

END_CORALMESSAGING_NAMESPACE

#endif /*MESSAGESTREAMREADER_H_*/
