/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#ifndef MESSAGESTREAMWRITER_H_
#define MESSAGESTREAMWRITER_H_

#include <vector>
#include <string>
#include <boost/foreach.hpp>
#include <boost/mpl/if.hpp>
#include <boost/type_traits/is_pointer.hpp>
#include <boost/type_traits/remove_pointer.hpp>

#include "CoralProtocol/arch.h"
#include "CoralProtocol/SequenceNumber.h"
#include "CoralMessaging/Message/Array.h"
#include "CoralMessaging/IMessageObject.h"
#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Stream/MessageStreamWriter.h"

START_CORALMESSAGING_NAMESPACE

namespace Stream {

/** Write the objects that constitue a message into an output stream.
 *
 * The stream is always written in system endianness, and respects the differenct codec protocols.
 *
 * The MessageStreamWriter objects maintain failure state. Failure occurs in case of:
 *
 * - communication error
 *
 * If the object turns into failure state, it cannot be reverted to good state.
 */
template<class OUTPUT>
class MessageStreamWriter
{
public:
	/// Constructor
	MessageStreamWriter(OUTPUT& out /**< Output stream */)
		: _out(out)
	{
		DEBUG_ASSERT_PTR(_out);
	}

	/// Destructor
	virtual ~MessageStreamWriter()
	{
		// EMPTY
	}

	/// Write a string into the CoralSocket. Return true if all the data is written successfully.
	bool write(const std::string& data)
	{
		Message::Array a(data);
		return write(a);
	}

	/// Write a long word into the CoralSocket. Return true if all the data is written successfully.
	bool write(const uint64_t data)
	{
		_out->write(reinterpret_cast<const char*>(&data), sizeof(uint64_t));
		return !_out->fail();
	}

	/// Write a word into the CoralSocket. Return true if all the data is written successfully.
	bool write(const uint32_t data)
	{
		_out->write(reinterpret_cast<const char*>(&data), sizeof(uint32_t));
		return !_out->fail();
	}

	/// Write a short word into the CoralSocket. Return true if all the data is written successfully.
	bool write(const uint16_t data)
	{
		_out->write(reinterpret_cast<const char*>(&data), sizeof(uint16_t));
		return !_out->fail();
	}

	/// Write a byte into the CoralSocket. Return true if all the data is written successfully.
	bool write(const uint8_t data)
	{
		_out->write(reinterpret_cast<const char*>(&data), sizeof(uint8_t));
		return !_out->fail();
	}

	/// Write a bool into the CoralSocket. Return true if all the data is written successfully.
	bool write(const bool data)
	{
		return write(uint8_t(data ? 1 : 0));
	}

	/// Write a Sequence Number into the CoralSocket. Return true if all the data is written successfully.
	bool write(const coral::CoralProtocol::SequenceNumber& acc)
	{
		return write(acc.value());
	}

	/// Write an Array into the CoralSocket. Return true if all the data is written successfully.
	bool write(const Message::Array& data)
	{
		uint32_t len = data.size();
		write(len);
		_out->write(data.data().c_str(), len);
		return !_out->fail();
	}

	/// Write a query representation into the CoralSocket. Return true if all the data is written successfully.
	bool write(const IMessageObject& data)
	{
		return write(*data.encode());
	}

	template<class T>
	bool write(const std::vector<T>& data)
	{
		return _writeVector(data);
	}

	/// Send the actually accumulated data as a packet and start a new packet. Return true in case of success.
	bool flush(bool isFinal = true /**< Indicate if the actually sent packet is the final one (true). True is the default value. */)
	{
		return _out->flush(isFinal);
	}

private:
	template <class T>
	struct _SimpleWriter
	{
		static bool write(T& x, MessageStreamWriter& mw) {return mw.write(x);}
	};

	template <class T>
	struct _PtrWriter
	{
		static bool write(T& x, MessageStreamWriter& mw) {return mw.write(*x);}
	};

	/// Write a vector of strings into the CoralSocket. Return true if all the data is written successfully.
	template<class T, template <class> class Vector>
	bool _writeVector(const Vector<T>& data)
	{
		uint32_t len = data.size();
		write(len);
		typedef typename boost::mpl::if_<boost::is_pointer<typename Vector<T>::value_type>, _PtrWriter<T>, _SimpleWriter<T> >::type Writer;
		BOOST_FOREACH(T i, data) Writer::write(i, *this);
		return !_out->fail();
	}

	OUTPUT _out; /**< The output stream where the packets must be sent into. */
};

} // namespace Stream

END_CORALMESSAGING_NAMESPACE

#endif /*MESSAGESTREAMWRITER_H_*/
