/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef TCPSOCKET_H_
#define TCPSOCKET_H_

#include "CoralMessaging/Timeout.h"
#include "CoralMessaging/ICoralIOHandler.h"

START_CORALMESSAGING_NAMESPACE

namespace Tcp {

/** A wrapper class around the TCP socket implementation of the system.
 *
 * The class collects and implements the project-wide common functionality related
 * to TCP socket read/write. The class is an ICommunicationEndpoint, represents
 * the connection between the project and the physical I/O.
 * The class is thread safe, and read/write lockable. The lock is important as we
 * always send/receive a whole packet in one go. We may send/receive multiple packets
 * initiated from multiple threads in the same time, so the line must be locked until a full packet
 * has been received/sent.
 *
 * The class also provide the common configuration of the TCP socket.
 *
 * A TcpSocket object must be initialized with a system socket descriptor - meaning an open socket.
 * Socket opening should be controlled by the @see TcpSocketFactory class. This solution is
 * chosen as different sockets (active, passive) must be opened differently, but they are
 * used in the same way.
 */
class TcpSocket : public ICoralIOHandler
{
public:
	/// Constructor
	TcpSocket(int sock /**< An open socket descriptor */);

	/// Destructor. It closes the system socket as well.
	virtual ~TcpSocket();

	/// Return teh socket file descriptor.
	int getFd() const {return _sock;};

	/// Return true if the socket is closed.
	bool isClosed() const;

	/// Close the socket.
	void close();

	// From ICommunicationEndpoint
	virtual bool fail() const;

	virtual size_t read(char* buf, size_t len);

	virtual bool readAll(char* buf, size_t len, const Timeout& to);

	virtual size_t write(const char* buf, size_t len);

	virtual bool writeAll(const char* buf, size_t len, const Timeout& to);

	virtual bool waitForData(const Timeout& to);

private:
	/// Common execution and error handling skeleton for sending and receiving data
	template<class OP, class BUF>
	size_t _readwrite(BUF buf, /** The read or write buffer */
			          size_t len, /** Size of the buffer */
			          OP op, /** ::send or ::recv */
			          const char* label /** Goes into the log, indicates the message direction ("sent" or "received") */);

	/// Common execution  skeleton for sending and receiving all data
	template<class OP, class BUF>
	bool _readwriteAll(BUF buf, /** The read or write buffer */
		               size_t len,  /** Size of the buffer */
		               const Timeout& to, /** Timeout of the operation */
		               OP op /** read or write */);

	/// Configure the socket
	void _configSocket();

	/// Set an fd flag of the underlying socket
	void _setFdFlag(int getCmd, int setCmd, long f);

	/// Set a socket option of the underlying socket
	void _setSocketOption(int optname);

	int _sock; /**< The socket file descriptor */
	bool _fail; /**< Failure state of the socket */
};

} // namespace Tcp

END_CORALMESSAGING_NAMESPACE

#endif /*TCPSOCKET_H_*/
