/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALMESSAGING_TIMEOUT_H_
#define CORALMESSAGING_TIMEOUT_H_

#include "CoralBricks/TimeTraits.h"
#include "CoralMessaging/base.h"

#include <limits>
#include <sys/time.h>

START_CORALMESSAGING_NAMESPACE

/** A type representing timeout.
 * 
 * The timeout may be specified in nanosecond granularity. The time is given relative to 
 * the actual time. so, if we specify Timeout(5, 0, 0), then timeout orrurs 5 seond after 
 * the construction. The timeout may be actualized: it means that the countdown restarts from now. 
 */
class Timeout
{
public:
	/** Constructor. 
	 * 
	 * The microsec and nanosec parts are additive: 1200 usec will represent 1 sec 200 usec, and 1 sec
	 * will be added to the second part. The same attribute has the nanosecond parameter.
	 * 
	 * The constructor also actualizes the timeout, and the countdown starts immediately.
	 * */
	explicit Timeout(int secs = 0 /**< second part */, 
		        	 int usecs = 0 /**< microsecond part */, 
			         int nsecs = 0 /**< nanosecond part */) : _s(secs), _us(usecs), _ns(nsecs) {actualize();};
			
	/** Destructor */
	virtual ~Timeout() {}
	
	/** Assignment operator.
	 * 
	 * Copies the second, nanosecond, microsecond values and actualizes the timeout
	 * (the countdown starts immediately).
	 */
	Timeout& operator = (const Timeout& x);
	
	/** boost::xtime representation of the timeout.
	 * 
	 * It is the absolute time when the timeout should occur. The Boost library requires this
	 * representation. 
	 */
	coral::TimeTraits::TIME getXtime() const { return _xt; }
	
	/** Return the second part of the timeout, as given in the constructor */
	int sec() const { return _s; }
	
	/** Return the microsecond part of the timeout, as given in the constructor */
	int usec() const { return _us; }
	
	/** Return the nanosecond part of the timeout, as given in the constructor */
	int nsec() const { return _ns; }
	
	/** Actualize the timeout (countdown restarts from now). Return itself. */
	Timeout& actualize();
	
	/** Return true if timeout occured else false. */
	virtual bool occured() const;
	
	virtual timeval* tval() const;
	
	virtual Timeout* clone() const; 
	
protected:
	/** Copy constructor
	 * 
	 * Copies the second, nanosecond, microsecond values and actualizes the timeout
	 * (the countdown starts immediately).
	 */
	Timeout(const Timeout& x) : _s(x._s), _us(x._us), _ns(x._ns) {};
	
private:
    int _s; /**< second part of the timeout */ 
    int _us; /**< microsecond part of the timeout */
    int _ns; /**< nanosecond part of the timeout */
    TimeTraits::TIME _xt; /**< absolut time of timeout (the @see actualize() method sets it) */
    mutable timeval _tval;
};

/** Class representing infinite timeout (timeout never occurs...) */
class InfiniteTimeout : public Timeout
{
public:
	InfiniteTimeout() : Timeout() {}
	virtual ~InfiniteTimeout() {}
	
	virtual bool occured() const;
	
	virtual timeval* tval() const;
	
	virtual Timeout* clone() const;
	
private:
	InfiniteTimeout(const InfiniteTimeout& x) : Timeout(x) {}
};

END_CORALMESSAGING_NAMESPACE

#endif /*CORALMESSAGING_TIMEOUT_H_*/
