/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include "CoralMessaging/base.h"

#include <boost/preprocessor/seq/for_each.hpp>

#include "CoralMessaging/CAL/MsgObj/AttributeTypeTools.h"
#include "CoralMessaging/CAL/MsgObj/AttributeListRepresentation.h"
#include "CoralMessaging/Stream/MessageStreamReader.h"
#include "CoralMessaging/Stream/MessageStreamWriter.h"

START_CORALMESSAGING_NAMESPACE

namespace CAL {

namespace MsgObj {
   
AttributeListRepresentation::_GetDataStringTypeMap AttributeListRepresentation::_getDataStringTypeMap;
AttributeListRepresentation::_ListExtenders AttributeListRepresentation::_listExtenders;
AttributeListRepresentation::_DataDecoderMap AttributeListRepresentation::_dataDecoderMap;

AttributeListRepresentation::AttributeListRepresentation()
	: IMessageObjectImpl<AttributeListRepresentation>(this),
	  _isShare(false)
{
	// EMPTY
}

AttributeListRepresentation::AttributeListRepresentation(const AttributeList& attList, const bool isShare)
	: IMessageObjectImpl<AttributeListRepresentation>(this),
	  _attList(new AttributeList(attList)),
	  _isShare(isShare)
{
	_share(attList);
}

AttributeListRepresentation::AttributeListRepresentation(const AttributeListRepresentation& x)
	: IMessageObjectImpl<AttributeListRepresentation>(this),
	  _attList(x._attList.get() ? new AttributeList(x.attributeList()) : NULL),
	  _isShare(x._isShare)
{
	if (_attList.get()) _share(x.attributeList());
}

bool AttributeListRepresentation::decode(const std::string& raw)
{
	std::vector<AttributeRepresentation> attList;
	_StrType str(new _StrType::element_type(raw));
	CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<_StrType> msr(str, CORALPROTOCOL_NAMESPACE_NAME::_system_endian);
	msr.read(attList);

	if (msr.fail())
		return false;

	_attList.reset(new AttributeList);

    for (std::vector<AttributeRepresentation>::iterator attr = attList.begin(); attr != attList.end(); ++attr) {
		_ListExtenderMethodType method = _listExtenders(attr->type());
		((*_attList).*method)(attr->attributeName());
		Attribute& a = (*_attList)[attr->attributeName()];
		a.setNull(attr->isNull());
		_dataDecoderMap(attr->type())(a, attr->data());
	}

	return true;
}

IMessageObject::EncodedType AttributeListRepresentation::encode() const
{
	DEBUG_ASSERT_PTR(_attList);
	std::vector<AttributeRepresentation> attList;
	AttributeList::const_iterator end = _attList->end();

	for (AttributeList::const_iterator i = _attList->begin(); i != end; ++i)
		attList.push_back(_encodeAttribute(*i));

	_StrType str(new _StrType::element_type);
	CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<_StrType> mbw(str);
	mbw.write(attList);
	mbw.flush();
	DEBUG_ASSERT(!str->fail());
	return IMessageObject::EncodedType(new std::string(str->str().str()));
}

bool AttributeListRepresentation::operator < (const IMessageObject&) const
{
	return true;
}

std::ostream& AttributeListRepresentation::log(std::ostream& ost) const
{
	if (!_attList.get())
		ost << "<empty list>";
	else
		_attList->toOutputStream(ost);

	return ost;
}

AttributeListRepresentation& AttributeListRepresentation::operator = (const AttributeListRepresentation& x)
{
	if (&x != this && x._attList.get())
		_attList.reset(x._attList.get() ? new AttributeList(x.attributeList()) : NULL);

	_isShare = x._isShare;
	if (_attList.get()) _share(x.attributeList());
	return *this;
}

bool AttributeListRepresentation::operator == (const AttributeListRepresentation& x) const
{
	if (_attList.get() == x._attList.get()) return true;

	if (_attList.get() == NULL && x._attList.get() != NULL) return false;

	if (_attList.get() != NULL && x._attList.get() == NULL) return false;

	return *_attList == *x._attList;
}

const AttributeList& AttributeListRepresentation::attributeList() const
{
	DEBUG_ASSERT_PTR(_attList);
	return *_attList;
}

const AttributeRepresentation& AttributeListRepresentation::_encodeAttribute(const coral::Attribute& att) const
{
	static Pointer<AttributeRepresentation>::Scoped attRepr;
	attRepr.reset(new AttributeRepresentation(att.specification().name(),
			                                  AttributeTypeTools::ProtocolID(att.specification().type()),
			                                  _getDataStringTypeMap(att.specification().type())(att),
			                                  att.isNull()));
	return *attRepr;
}

void AttributeListRepresentation::_share(const AttributeList& attList)
{
	if (_isShare) {
	   const size_t numberOfVariables = attList.size();

	   for (size_t i = 0; i < numberOfVariables; ++i)
		   (*_attList)[i].shareData(attList[i]);
	}
}

template<typename T>
std::string AttributeListRepresentation::_getSpecificDataString(const coral::Attribute& att)
{
	static _StrType str(new _StrType::element_type);
	static CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<_StrType> mbw(str);

	str->reset();
	const T* data = static_cast<const T*>(att.addressOfData());
	DEBUG_ASSERT_PTR(data);
	typename AttributeTypeTools::GetMsgObjConverter<T>::type converter;
	mbw.write(converter.convert(*data));
	mbw.flush();
	DEBUG_ASSERT(!str->fail());
	return str->str().str();
}

template<typename T>
void AttributeListRepresentation::_decodeData(Attribute& a, const std::string& data)
{
	_StrType str(new _StrType::element_type(data));
	CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<_StrType> msr(str, CORALPROTOCOL_NAMESPACE_NAME::_system_endian);
	// Read the type representation from the data
	typename AttributeTypeTools::GetMsgObjType<T>::type msgObj;
	msr.read(msgObj);
	typedef typename AttributeTypeTools::GetMsgObjBackConverter<T>::type converter;
	T converted(converter::convert(msgObj));
	DEBUG_ASSERT(!str->fail());
	a.setValueFromAddress(static_cast<void*>(&converted));
}

AttributeListRepresentation::_GetDataStringTypeMap::_GetDataStringTypeMap()
{
	#define CODE_GENERATOR(r, data, TYPENAME) \
		_map[&typeid(TYPENAME)] = &AttributeListRepresentation::_getSpecificDataString<TYPENAME>;

	BOOST_PP_SEQ_FOR_EACH(CODE_GENERATOR, ~, CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST);
	#undef CODE_GENERATOR
}

AttributeListRepresentation::_ListExtenders::_ListExtenders()
{
	#define CODE_GENERATOR(r, data, TYPENAME) \
		_map[AttributeTypeTools::ProtocolIDType<TYPENAME>::id()] = &::coral::AttributeList::extend<TYPENAME>;

	BOOST_PP_SEQ_FOR_EACH(CODE_GENERATOR, ~, CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST);
	#undef CODE_GENERATOR
}

AttributeListRepresentation::_DataDecoderMap::_DataDecoderMap()
{
	#define CODE_GENERATOR(r, data, TYPENAME) \
		_map[AttributeTypeTools::ProtocolIDType<TYPENAME>::id()] = &AttributeListRepresentation::_decodeData<TYPENAME>;

	BOOST_PP_SEQ_FOR_EACH(CODE_GENERATOR, ~, CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST);
	#undef CODE_GENERATOR
}

} // namespace MsgObj

} // namespace CAL

END_CORALMESSAGING_NAMESPACE
