/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <ieee754.h>
#include <boost/mpl/map.hpp>
#include <boost/mpl/for_each.hpp>

#include "CoralMessaging/CAL/MsgObj/AttributeTypeTools.h"

START_CORALMESSAGING_NAMESPACE

namespace CAL {

namespace MsgObj {

AttributeTypeTools::_TypeMap AttributeTypeTools::_typeMap;

const uint32_t AttributeTypeTools::ProtocolID(const std::type_info& ti)
{
	return _typeMap(ti);
}

AttributeTypeTools::_TypeMap::_TypeMap()
	: CppType_ProtocolID_Map()
{
	#define CODE_GENERATOR(r, LIST_LENGTH, i, TYPENAME) \
		ElemGen<BOOST_PP_INC(i), TYPENAME>::type BOOST_PP_COMMA_IF(BOOST_PP_LESS(BOOST_PP_INC(i), LIST_LENGTH))

	typedef ::boost::mpl::map <
		BOOST_PP_SEQ_FOR_EACH_I(CODE_GENERATOR, BOOST_PP_SEQ_SIZE(CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_1), CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_1)
	> creatormap_1;

    typedef ::boost::mpl::map <
		BOOST_PP_SEQ_FOR_EACH_I(CODE_GENERATOR, BOOST_PP_SEQ_SIZE(CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_2), CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_2)
	> creatormap_2;

	#undef CODE_GENERATOR

	CppType_ProtocolID_Map::Insert ins1(*this);
	::boost::mpl::for_each<creatormap_1>(ins1);
	CppType_ProtocolID_Map::Insert ins2(*this, BOOST_PP_SEQ_SIZE(CORALMESSAGING_ATTRIBUTETYPETOOLS_TYPELIST_1));
	::boost::mpl::for_each<creatormap_2>(ins2);
}

MsgObj::TimeStamp AttributeTypeTools::_CppProtocolTypeConverter<coral::TimeStamp, MsgObj::TimeStamp>::convert(const coral::TimeStamp& from)
{
	return MsgObj::TimeStamp(from.year(), from.month(), from.day(), from.hour(), from.minute(), from.second(), from.nanosecond());
}


MsgObj::Date AttributeTypeTools::_CppProtocolTypeConverter<coral::Date, MsgObj::Date>::convert(const coral::Date& from)
{
	return MsgObj::Date(from.year(), from.month(), from.day());
}


std::string AttributeTypeTools::_CppProtocolTypeConverter<coral::Blob, std::string>::convert(const coral::Blob& from)
{
	std::string ret;
	ret.append(static_cast<const char*>(from.startingAddress()), from.size());
	return ret;
}

uint32_t AttributeTypeTools::_CppProtocolTypeConverter<float, uint32_t>::convert(const float& from)
{
	ieee754_float f;
	f.f = from;
	uint32_t ret = 0;
	ret |= f.ieee.mantissa;
	ret |= f.ieee.exponent << 23;
	ret |= f.ieee.negative << (23 + 8);
	return ret;
}

uint64_t AttributeTypeTools::_CppProtocolTypeConverter<double, uint64_t>::convert(const double& from)
{
	ieee754_double f;
	f.d = from;
	uint64_t ret = 0;
	ret |= f.ieee.mantissa1;
	ret |= static_cast<uint64_t>(f.ieee.mantissa0) << 32;
	ret |= static_cast<uint64_t>(f.ieee.exponent) << (32 + 20);
	ret |= static_cast<uint64_t>(f.ieee.negative) << (32 + 20 + 11);
	return ret;
}

MsgObj::LongDoubleRepresentation AttributeTypeTools::_CppProtocolTypeConverter<long double, MsgObj::LongDoubleRepresentation>::convert(const long double& from)
{
	ieee854_long_double f;
	f.d = from;
	uint64_t part64 = 0;
	part64 |= f.ieee.mantissa1;
	part64 |= static_cast<uint64_t>(f.ieee.mantissa0) << 32;

	uint32_t part32 = 0;
	part32 |= f.ieee.empty;
	part32 |= f.ieee.exponent << 16;
	part32 |= f.ieee.negative << (16 + 15);

	return MsgObj::LongDoubleRepresentation(part64, part32);
}

// Back converters
coral::TimeStamp AttributeTypeTools::_ProtocolCppTypeConverter<coral::TimeStamp, MsgObj::TimeStamp>::convert(const MsgObj::TimeStamp& from)
{
	return coral::TimeStamp(from.year(), from.month(), from.day(), from.hour(), from.minute(), from.second(), from.nanosecond());
}

coral::Date AttributeTypeTools::_ProtocolCppTypeConverter<coral::Date, MsgObj::Date>::convert(const MsgObj::Date& from)
{
	return coral::Date(from.year(), from.month(), from.day());
}

coral::Blob AttributeTypeTools::_ProtocolCppTypeConverter<coral::Blob, std::string>::convert(const std::string& from)
{
	coral::Blob blob(from.size());
	::memcpy(blob.startingAddress(), from.data(), blob.size());
	return blob;
}

float AttributeTypeTools::_ProtocolCppTypeConverter<float, uint32_t>::convert(const uint32_t from)
{
	ieee754_float f;

	f.ieee.mantissa = 2;

	f.ieee.mantissa = 0x7FFFFF & from;
	f.ieee.exponent = 0xFF & (from >> 23);
	f.ieee.negative = 1 & (from >> (23 + 8));
	return f.f;
}

double AttributeTypeTools::_ProtocolCppTypeConverter<double, uint64_t>::convert(const uint64_t from)
{
	ieee754_double f;
	f.ieee.mantissa1 = 0xFFFFFFFF & from;
	f.ieee.mantissa0 = 0xFFFFFFull & (from >> 32);
	f.ieee.exponent = 0x7FFull & (from >> (32 + 20));
	f.ieee.negative = 1ull & (from >> (32 + 20 + 11));
	return f.d;
}

long double AttributeTypeTools::_ProtocolCppTypeConverter<long double, MsgObj::LongDoubleRepresentation>::convert(const MsgObj::LongDoubleRepresentation& from)
{
	ieee854_long_double f;
	f.ieee.mantissa1 = 0xFFFFFFFF & from.part64();
	f.ieee.mantissa0 = 0xFFFFFFFFull & (from.part64() >> 32);
	f.ieee.empty = 0xFFFF & from.part32();
	f.ieee.exponent = 0x7FFF & (from.part32() >> 16);
	f.ieee.negative = 1 & (from.part32() >> (16 + 15));
	return f.d;
}

} // namespace MsgObj

} // namespace CAL

END_CORALMESSAGING_NAMESPACE
