/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include "CoralMessaging/CTL/CTLMessageBuffer.h"

START_CORALMESSAGING_NAMESPACE

namespace CTL {

namespace MessageBufferSM {

// -----------------------------------------------------------------------------
// MessageBufferStateMachine implementation
// -----------------------------------------------------------------------------
bool MessageBufferStateMachine::received() const
{
	return this->terminated() ? false : this->state_cast< const StateOperations& >().received();
}

CoralProtocol::Opcode MessageBufferStateMachine::opcode() const
{
	return this->terminated() ? _data.opcode : this->state_cast< const StateOperations& >().opcode();
}

uint8_t MessageBufferStateMachine::firstByte() const
{
	return this->terminated() ? _data.firstByte : this->state_cast< const StateOperations& >().firstByte();
}

// -----------------------------------------------------------------------------
// StReadingOut implementation
// -----------------------------------------------------------------------------
StReadingOut::StReadingOut()
	: _cursor(0)
{
	// EMPTY
}

sc::result StReadingOut::react(const EvReadData& er)
{
	er.readed = 0;
	Pointer<ListOfPackets>::Shared& plist = this->outermost_context().data().packetList;

	if(!plist.get()) {
		return this->transit<StFinished>();
	}

	size_t x = er.len;
	char* data = er.data;

	while(x != 0) {
		if (_cursor == 0) {
			if (plist->empty()) {
				plist.reset();
				break;
			}
			_actual = plist->front();
			plist->pop_front();
		}

		size_t distanceFromEnd = _actual->payload().size() - _cursor;

		if (x <= distanceFromEnd) {
			memcpy(data, _actual->payload().data() + _cursor, x);
			_cursor += x;
			data += x;
			x = 0;
		} else {
			memcpy(data, _actual->payload().data() + _cursor, distanceFromEnd);
			_cursor = 0;
			data += distanceFromEnd;
			x -= distanceFromEnd;
		}
	}

	er.readed = er.len - x;
	return this->discard_event();
}

// -----------------------------------------------------------------------------
// StReceived implementation
// -----------------------------------------------------------------------------
sc::result StReceived::react(const EvAddPacket&)
{
	DEBUG_ASSERT(this->outermost_context().data().packetList.get());
	MessageBufferStateMachine::SharedData& d = this->outermost_context().data();
	if (d.messageLength > 0) d.firstByte = d.packetList->front()->payload().at(0);
	return this->transit<StReadingOut>();
}

// -----------------------------------------------------------------------------
// StReceivingPackets implementation
// -----------------------------------------------------------------------------
CoralProtocol::Opcode StReceivingPackets::opcode() const
{
	return this->outermost_context().const_data().opcode;
}

sc::result StReceivingPackets::react(const EvAddPacket& er)
{
	DEBUG_ASSERT(er.packet.get());
	Packet& p = *er.packet;
	MessageBufferStateMachine::SharedData& data = this->outermost_context().data();

	if (p.header().opcode() != data.opcode) {
		LOGGER_NEWLOG(coral::Info) << "Opcode mismatch" << commit;
		OpcodeMismatch msg(data.lastSeqNum);
		CoralSocket csock(er.csock, msg);
		msg.write(csock);
		DEBUG(StatisticsCollector::Instance().incDiscardedPacketNumber());
		return this->terminate();
	} else if (!p.header().sequenceNum().isNextOf(data.lastSeqNum)) {
		LOGGER_NEWLOG(coral::Debug) << "Message segmentation error" << commit;
		SegmentationError msg(data.lastSeqNum);
		CoralSocket csock(er.csock, msg);
		DEBUG(StatisticsCollector::Instance().incDiscardedPacketNumber());
		return this->terminate();
	} else if (p.header().payloadSize() > 0) {
		data.packetList->push_back(er.packet);
		data.messageLength += p.header().payloadSize();
	}

	if (p.header().sequenceNum().isLast()) {
		LOGGER_NEWLOG(coral::Info) << "The full message received" << commit;
		this->post_event(er);
		return this->transit<StReceived>();
	} else {
		data.lastSeqNum = p.header().sequenceNum();
		return this->discard_event();
	}
}

// -----------------------------------------------------------------------------
// StEmptyMessage implementation
// -----------------------------------------------------------------------------
sc::result StEmptyMessage::react(const EvAddPacket& er)
{
	DEBUG_ASSERT(er.packet.get());
	CTL::Packet& p = *er.packet;
	MessageBufferStateMachine::SharedData& data = this->outermost_context().data();

	if (!p.header().sequenceNum().isFirst()) {
		LOGGER_NEWLOG(coral::Debug) << "Message segmentation error" << commit;
		SegmentationError_First msg;
		CoralSocket csock(er.csock, msg);
		msg.write(csock);
		DEBUG(StatisticsCollector::Instance().incDiscardedPacketNumber());
		return this->discard_event();
	} else if (!p.header().opcode().valid()) {
		LOGGER_NEWLOG(coral::Info) << "Unsupported opcode" << commit;
		WrongOpcode msg;
		CoralSocket csock(er.csock, msg);
		msg.write(csock);
		DEBUG(StatisticsCollector::Instance().incDiscardedPacketNumber());
		return this->discard_event();
	} else if (p.header().payloadSize() > 0) {
		data.packetList->push_back(er.packet);
		data.messageLength += p.header().payloadSize();
	}

	data.opcode = p.header().opcode();

	if (p.header().sequenceNum().isLast()) {
		// OK, we received a message with a single packet
		LOGGER_NEWLOG(coral::Info) << "The full message received" << commit;
		this->post_event(er);
		return this->transit<StReceived>();
	} else {
		data.lastSeqNum = p.header().sequenceNum();
		return this->transit<StReceivingPackets>();
	}
}

} // namespace MessageBufferSM

// -----------------------------------------------------------------------------
// MessageBuffer implementation
// -----------------------------------------------------------------------------
MessageBuffer::MessageBuffer()
{
	_sm.initiate();
}

bool MessageBuffer::addPacket(Pointer<Packet>::Shared& packet, CoralInputSocket& csock)
{
	_sm.process_event(MessageBufferSM::EvAddPacket(packet, csock));
	return _sm.received();
}

uint8_t MessageBuffer::firstByte() const
{
	DEBUG_ASSERT(_sm.received());
	return _sm.firstByte();
}

size_t MessageBuffer::read(char* data, size_t len)
{
	size_t readed;
	_sm.process_event(MessageBufferSM::EvReadData(data, len, readed));
	return readed;
}

bool MessageBuffer::fail() const
{
	return _sm.terminated();
}

CoralProtocol::Opcode MessageBuffer::opcode() const
{
	return _sm.opcode();
}

} // namespace CTL

END_CORALMESSAGING_NAMESPACE
