/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include "CoralMessaging/Utils/StatisticsCollector.h"
#include "CoralMessaging/CoralSocket.h"
#include "CoralMessaging/CTL/CTLMessages.h"
#include "CoralMessaging/CTL/CTLPacketReader.h"

START_CORALMESSAGING_NAMESPACE

namespace CTL {

namespace PacketReaderSM {

// -----------------------------------------------------------------------------
// PacketReaderStateMachine implementation
// -----------------------------------------------------------------------------
Pointer<Packet>::Shared PacketReaderStateMachine::get() const 
{
	return this->terminated() ? Pointer<Packet>::Shared() : this->state_cast< const StateOperations& >().get();
}

// -----------------------------------------------------------------------------
// StPacketReceived implementation
// -----------------------------------------------------------------------------
sc::result StPacketReceived::react(const EvPickPacket& er) 
{
	this->post_event(er);
	return this->transit<StObtainingHeader>();
}

Pointer<Packet>::Shared StPacketReceived::get() const
{
	return this->outermost_context().const_data().packet;
}

// -----------------------------------------------------------------------------
// StObtainingPayload implementation
// -----------------------------------------------------------------------------
sc::result StObtainingPayload::react(const EvPickPacket& er) 
{
	DEBUG_ASSERT(this->outermost_context().data().packet.get());
	
	_str.reset(new _StreamType(this->outermost_context().data().packet->payload(), 
			                   this->outermost_context().data().dataStr));

	sc::result r = this->react_global(er, *_str);
	return r;
}
	
sc::result StObtainingPayload::_handleDataReception(const std::string&, const EvPickPacket&) 
{ 
	DEBUG(StatisticsCollector::Instance().incRecvPacketNumber());
	return this->transit<StPacketReceived>(); 
}

// -----------------------------------------------------------------------------
// StObtainingHeader implementation
// -----------------------------------------------------------------------------
sc::result StObtainingHeader::react(const EvPickPacket& er) 
{
	_str.reset(new _StreamType(CoralProtocol::PacketHeader::size(), this->outermost_context().data().dataStr));		
	return this->react_global(er, *_str);
}
	
sc::result StObtainingHeader::_handleDataReception(const std::string& data, const EvPickPacket& er) 
{	
	Pointer<CoralProtocol::PacketHeader>::Scoped ph(CoralProtocol::PacketHeader::create(data));
	
	if (ph.get()) {
		if (ph->isVersionMismatch()) {
			LOGGER_NEWLOG(coral::Info) << "Version mismatch" << commit;
			CTL::VersionMismatch msg(ph->version(), "");
			CoralSocket csock(this->outermost_context().data().dataStr, *ph, msg);
			msg.write(csock);
			DEBUG(StatisticsCollector::Instance().incDiscardedPacketNumber());
			return this->terminate();
		}
		// Header is Ok, let's go for the payload
		this->outermost_context().data().packet.reset(new Packet(*ph));
		// Instruct the next state to handle the event
		this->post_event(er);
		return this->transit<StObtainingPayload>();
	}
	// Failure, we got garbage
	LOGGER_NEWLOG(coral::Info) << "Garbage arrived" << commit;
	return this->terminate();
}

} // namespace PacketReaderSM 

// -----------------------------------------------------------------------------
// PacketReader implementation
// -----------------------------------------------------------------------------
PacketReader::PacketReader(Pointer<ICoralIOHandler>::Shared s) 
	: _sm(s) 
{
	_sm.initiate();
}
	
Pointer<Packet>::Shared PacketReader::pickPacket(const Timeout& to) 
{
	_sm.process_event(PacketReaderSM::EvPickPacket(to));
	return _sm.get();
}

bool PacketReader::terminated() const 
{
	return _sm.terminated();
}

} // namespace CTL

END_CORALMESSAGING_NAMESPACE

