/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <typeinfo>

#include <boost/mpl/map.hpp>
#include <boost/mpl/for_each.hpp>
#include <boost/mpl/if.hpp>
#include <boost/mpl/at.hpp>

#include "CoralMessaging/Insulation/Exceptions.h"
#include "CoralMessaging/Utils/ctl.h"
#include "CoralMessaging/CAL/CoralExceptionImpl.h"

START_CORALMESSAGING_NAMESPACE

#define CORAL_EXCEPTION_NUMBER_MAX 43

namespace CAL {

CoralExceptionImpl::_Map CoralExceptionImpl::_map;

void CoralExceptionImpl::checkAndThrow(const IMessage& msg)
{
	// Try to cast to the Exception message. If it is successfull...
	const CORALMESSAGING_NAMESPACE_NAME::CAL::Exception* ex_msg =
		message_cast<const CORALMESSAGING_NAMESPACE_NAME::CAL::Exception>(&msg);

	if (ex_msg == NULL) return;
	// ...get the exception ID. According to the ID, select a specialized template method
	// that creates the right exception type, sets the message string and throws the exception.
	#define CASE_GEN(r, state) \
		case BOOST_PP_TUPLE_ELEM(2, 0, state): \
			_createAndThrow<BOOST_PP_TUPLE_ELEM(2, 0, state)>(*ex_msg);

	switch (ex_msg->id()) {
		CORALMESSAGING_CTL_FOR(0, CORAL_EXCEPTION_NUMBER_MAX, CASE_GEN)
	default:
		// Should not get here
		FORCE_CRASH_HERE;
	}

	#undef CASE_GEN
}

typedef boost::mpl::map
<
	// LookupServiceException-s
	ElemGen<1, coral::LookupServiceException>::type,
	ElemGen<2, coral::InvalidReplicaIdentifierException>::type,
	// MonitoringException-s
	ElemGen<3, coral::MonitoringException>::type,
	// RelationalServiceException-s
	ElemGen<4, coral::RelationalServiceException>::type,
	ElemGen<5, coral::NonExistingDomainException>::type,
	ElemGen<6, coral::InvalidOperationException>::type,
	// SchemaException-s
	ElemGen<7, coral::SchemaException>::type,
	ElemGen<8, coral::TableAlreadyExistingException>::type,
	ElemGen<9, coral::TableNotExistingException>::type,
	ElemGen<10, coral::ViewAlreadyExistingException>::type,
	ElemGen<11, coral::QueryException>::type,
	ElemGen<12, coral::QueryExecutedException>::type,
	ElemGen<13, coral::InvalidUniqueConstraintIdentifierException>::type,
	ElemGen<14, coral::InvalidForeignKeyIdentifierException>::type,
	ElemGen<15, coral::InvalidIndexIdentifierException>::type,
	ElemGen<16, coral::InvalidColumnIndexException>::type,
	ElemGen<17, coral::InvalidColumnNameException>::type,
	ElemGen<18, coral::NoPrimaryKeyException>::type,
	ElemGen<19, coral::ExistingPrimaryKeyException>::type,
	ElemGen<20, coral::ViewNotExistingException>::type
> _creatormap_1;

typedef boost::mpl::map
<
	ElemGen<21, coral::UniqueConstraintAlreadyExistingException>::type,
	ElemGen<22, coral::DataEditorException>::type,
	ElemGen<23, coral::DuplicateEntryInUniqueKeyException>::type,
	// SessionException-s
	ElemGen<24, coral::MaximumNumberOfSessionsException>::type,
	ElemGen<25, coral::StartSessionException>::type,
	ElemGen<26, coral::InvalidOperationInReadOnlyModeException>::type,
	ElemGen<27, coral::InvalidSchemaNameException>::type,
	ElemGen<28, coral::NoAuthenticationServiceException>::type,
	ElemGen<29, coral::AuthenticationFailureException>::type,
	ElemGen<30, coral::MonitoringServiceNotFoundException>::type,
	ElemGen<31, coral::ServerException>::type,
	ElemGen<32, coral::DatabaseNotAccessibleException>::type,
	ElemGen<33, coral::ConnectionException>::type,
	ElemGen<34, coral::ConnectionNotActiveException>::type,
	ElemGen<35, coral::TransactionException>::type,
	ElemGen<36, coral::TransactionNotStartedException>::type,
	ElemGen<37, coral::TransactionNotCommittedException>::type,
	ElemGen<38, coral::TransactionNotActiveException>::type,
	ElemGen<39, coral::TypeConverterException>::type,
	ElemGen<40, coral::SessionException>::type
> _creatormap_2;

typedef boost::mpl::map
<
	ElemGen<41, coral::UnSupportedSqlTypeException>::type,
	ElemGen<42, coral::UnSupportedCppTypeException>::type,
	// WebCacheControlException-s
	ElemGen<43, coral::WebCacheControlException>::type,
	// The general exception
	ElemGen<0, coral::Exception>::type
> _creatormap_3;


template<int ID>
void CoralExceptionImpl::_createAndThrow(const CORALMESSAGING_NAMESPACE_NAME::CAL::Exception& msg)
{
	typedef typename boost::mpl::has_key<_creatormap_1, boost::mpl::int_<ID> >::type isIn_1;
	typedef typename boost::mpl::has_key<_creatormap_2, boost::mpl::int_<ID> >::type isIn_2;
	typedef typename boost::mpl::has_key<_creatormap_3, boost::mpl::int_<ID> >::type isIn_3;
	typedef typename boost::mpl::if_<isIn_1, typename boost::mpl::at<_creatormap_1, typename boost::mpl::int_<ID> >::type,
	                                         typename boost::mpl::if_<isIn_2, typename boost::mpl::at<_creatormap_2, typename boost::mpl::int_<ID> >::type,
	                                                                          typename boost::mpl::at<_creatormap_3, boost::mpl::int_<ID> >::type>::type>::type ex_type;

	ex_type ex;
	ex.setMessage(msg.data());
	throw ex;
}

uint32_t CoralExceptionImpl::getID(const std::type_info& ti)
{
	return _map(ti);
}

CoralExceptionImpl::_Map::_Map()
	: CppType_ProtocolID_Map()
{
	Insert ins(*this);
	boost::mpl::for_each<_creatormap_1>(ins);
	boost::mpl::for_each<_creatormap_2>(ins);
	boost::mpl::for_each<_creatormap_3>(ins);
}

} // namespace CAL

END_CORALMESSAGING_NAMESPACE

