// $Id: MessageStream.cpp,v 1.1 2008/07/30 18:28:31 molnarzs Exp $
//====================================================================
//  Implementation file of the MessageStream object
//--------------------------------------------------------------------
//
//  Package: coral/CoralBase (The CORAL project)
//
//  Description: Generic data persistency for LHC
//
//  @author      M.Frank
//====================================================================
#ifdef CORAL_SEALED

#include "CoralMessaging/Insulation/MessageStream.h"
#include <iostream>

// namespace pool  {
//   void setLogLevel(seal::Msg::Level lvl);
//   seal::Msg::Level getLogLevel();
// }

// POOL namespace
// using namespace pool;

// Preset default message reporter to None
static std::auto_ptr<coral::IMsgReporter> s_msgReporter(new coral::MsgReporter);

/// Default constructor
coral::MsgReporter::MsgReporter()
: m_level( coral::Error )
{
  static bool first = true;
  if ( first )
  {
    first = false;
    if ( getenv( "CORAL_MSGLEVEL" ) ) 
    {
      // Check only the first char of the environment variable
      switch ( *getenv( "CORAL_MSGLEVEL" ) )
      {
      case '0':
      case 'n':
      case 'N': m_level = coral::Nil; break;
        
      case '1':
      case 'v':
      case 'V': m_level = coral::Verbose; break;
        
      case '2':
      case 'd':
      case 'D': m_level = coral::Debug; break;
        
      case '3':
      case 'i':
      case 'I': m_level = coral::Info; break;
        
      case '4':
      case 'w':
      case 'W': m_level = coral::Warning; break;
        
      case '5':
      case 'e':
      case 'E': m_level = coral::Error; break;
        
      case '6':
      case 'f':
      case 'F': m_level = coral::Fatal; break;
        
      case '7':
      case 'a':
      case 'A': m_level = coral::Always; break;
        
      default : break; // keep the default
      }
    }
  }
}

/// Access output level
coral::MsgLevel coral::MsgReporter::outputLevel() const
{
//  return coral::MsgLevel(pool::getLogLevel());
  return m_level;
}

/// Modify output level
void coral::MsgReporter::setOutputLevel(coral::MsgLevel lvl)
{ 
//   pool::setLogLevel((seal::Msg::Level)lvl); 
  m_level = lvl;
} 

/// report mesage to stdout
void coral::MsgReporter::report(int lvl, const std::string& src, const std::string& msg)
{
  if ( lvl < m_level ) return;
  const char* level = "Unknown";
  switch ( lvl )    {
#define SET(x)  case coral::x:  level=#x;  break 
    SET( Nil );
    SET( Verbose );
    SET( Debug );
    SET( Info );
    SET( Warning );
    SET( Error );
    SET( Fatal );
    case coral::Always:
      level="Success";
      break;
    default:
      level="Unknown";
      break;
    break;
#undef SET
  }

  std::cout << src << " " << level << " " << msg << std::endl;
}

/// Install new message reporter object
void coral::MessageStream::installMsgReporter(IMsgReporter* reporter)
{
  if ( reporter )
  {
    if ( s_msgReporter.get() )
    {
      s_msgReporter->release();
      s_msgReporter.release();
    }
    s_msgReporter = std::auto_ptr<coral::IMsgReporter>(reporter);
    return;
  }
  
  s_msgReporter = std::auto_ptr<coral::IMsgReporter>(new coral::MsgReporter);
}

/// Set the verbosity of the default message reporter.
void coral::MessageStream::setMsgVerbosity(coral::MsgLevel lvl)
{
  if ( s_msgReporter.get() ) s_msgReporter->setOutputLevel(lvl);
}

/// Get the verbosity of the default message reporter.
const coral::MsgLevel coral::MessageStream::msgVerbosity()
{
  return s_msgReporter->outputLevel();
}

/// Standard initializing constructor
coral::MessageStream::MessageStream(const std::string& source) 
  : m_act(false), m_source(source), m_currLevel(Nil)
{
  //m_level = coral::MsgLevel(pool::getLogLevel());
  m_level = coral::Verbose; // Accept all messages (fix for bug #36831)
}

/// Propagate message to reporter
void coral::MessageStream::doOutput()
{
//  std::cout << m_source << " " << m_level << " " << m_currLevel << " " 
//            << (int)pool::getLogLevel() << " " << m_stream.str() << std::endl;
  s_msgReporter->report(m_currLevel,m_source,m_stream.str());
  m_stream.str("");
}

/// Print data of type long long int
void coral::MessageStream::prtLongLong(const long long int arg)
{
#ifdef _WIN32
  char buf[128];
  ::sprintf(buf,m_stream.flags()&std::ios::hex ? "%I64x" : "%I64d", arg);
  m_stream << buf;
#else
  m_stream << arg;
#endif
}

/// Print data of type unsigned long long int
void coral::MessageStream::prtLongLong(const unsigned long long int arg)
{
#ifdef _WIN32
  char buf[128];
  ::sprintf(buf,m_stream.flags()&std::ios::hex ? "%I64x" : "%I64d", arg);
  m_stream << buf;
#else
  m_stream << arg;
#endif
}

#endif // #ifdef CORAL_SEALED
