/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <boost/foreach.hpp>
#include <boost/preprocessor/repetition/enum.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/stringize.hpp>

#include "CoralMessaging/CAL/MsgObj/QueryRepresentation.h"
#include "CoralMessaging/Stream/DataStringStream.h"
#include "CoralMessaging/Utils/ctl.h"

START_CORALMESSAGING_NAMESPACE

namespace CAL {

namespace MsgObj {

QueryRepresentation::_MethodStringMap QueryRepresentation::_methodStringMap;

QueryRepresentation::QueryRepresentation()
	: IMessageObjectImpl<QueryRepresentation>(this)
{
	// EMPTY
}

QueryRepresentation::QueryRepresentation(const QueryRepresentation& x)
	: IMessageObjectImpl<QueryRepresentation>(this)
{
	*this = x;
}

QueryRepresentation::~QueryRepresentation()
{
	// EMPTY
}

QueryRepresentation& QueryRepresentation::operator = (const QueryRepresentation& r)
{
	if (&r != this) {
		_methods = r._methods;
		_schemaName = r._schemaName;
	}

	return *this;
}

bool QueryRepresentation::decode(const std::string& raw)
{
	_StrType str(new _StrType::element_type(raw));
	CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamReader<_StrType> msr(str, CORALPROTOCOL_NAMESPACE_NAME::_system_endian);
	uint32_t bits;
	msr.read(bits);
    _methods.clear();

	for (uint32_t i = 0; i < QueryRepresentationTypes::MethodNumber(); ++i) {
		if (((uint32_t) 1 << i) & bits) {

			#define CODE_GEN(r, data, elem) \
				case QueryRepresentationTypes::BOOST_PP_CAT(E_, elem): { \
					typedef QueryRepresentationTypes::elem repr_type; \
					std::vector<repr_type::type*> mData; \
					msr.read(mData); \
                    _MethodCallsType* imoData = new _MethodCallsType; \
                    _methods[i].reset(imoData); \
				    \
					BOOST_FOREACH(repr_type::type* ptr, mData) \
						imoData->push_back(static_cast<IMessageObject*>(ptr)); \
				} \
				break;

			switch (i) {
				BOOST_PP_SEQ_FOR_EACH(CODE_GEN, ~, QRT_ALL_LABELS)
			default:
				// Should not get here
				DEBUG_ASSERT(0);
			};

			#undef CODE_GEN
		}
	}

	msr.read(_schemaName);
	return !msr.fail();
}

IMessageObject::EncodedType QueryRepresentation::encode() const
{
	_StrType str(new _StrType::element_type);
	CORALMESSAGING_NAMESPACE_NAME::Stream::MessageStreamWriter<_StrType> msw(str);
	msw.write(_generateBitmap());

	BOOST_FOREACH(_MethodsType::value_type i, _methods) {
		// No sorting due to expected side effect in OracleAccess
		//i.second->sort();
		std::vector<IMessageObject*> data;
		BOOST_FOREACH(_MethodCallsType::reference x, *(i.second))
            data.push_back(&x);
		msw.write(data);
    }

	msw.write(_schemaName);
	return IMessageObject::EncodedType(new std::string(str->str().str()));
}

std::ostream& QueryRepresentation::log(std::ostream& ost) const
{
	ost << "Schrema name: " << _schemaName << "\n";

	BOOST_FOREACH(_MethodsType::value_type x, _methods) {
		ost << "At slot " << x.first << ": method <" << _methodStringMap(x.first) << ">:";

		BOOST_FOREACH(_MethodCallsType::reference d, *(x.second))
			ost << "\n\t" << d;

		ost << "\n";
	}

	return ost;
}

bool QueryRepresentation::operator < (const IMessageObject& x) const
{
	const QueryRepresentation& casted = dynamic_cast<const QueryRepresentation&>(x);
	return _generateBitmap() < casted._generateBitmap();
}

bool QueryRepresentation::operator == (const QueryRepresentation& x) const
{
	if (this == &x) return true;

	return _schemaName == x._schemaName &&
		_methods.size() == x._methods.size() &&
		std::equal(_methods.begin(), _methods.end(), x._methods.begin(), _MethodsElementComparator());
}

#define CASE_GEN(r, CLASS, method) \
	case BOOST_PP_CAT(QueryRepresentationTypes::E_, method): {\
		BOOST_FOREACH(_MethodCallsType::reference pars, *(i.second)) { \
			typedef QueryRepresentationTypes::method::type ReprType; \
			ReprType& p = dynamic_cast<ReprType&>(pars); \
			p.call(q, &CLASS::method); \
		} \
	} \
	break;

void QueryRepresentation::replay(IQuery& q)
{
	BOOST_FOREACH(_MethodsType::value_type i, _methods) {
		switch (i.first) {
			BOOST_PP_SEQ_FOR_EACH(CASE_GEN, IQuery, QRT_ALL_LABELS)
		default:
			DEBUG_ASSERT(0);
		};
	}
}

void QueryRepresentation::replay(IQueryDefinition& q)
{
	BOOST_FOREACH(_MethodsType::value_type i, _methods) {
		switch (i.first) {
			BOOST_PP_SEQ_FOR_EACH(CASE_GEN, IQueryDefinition, QRT_QUERYDEFINITION_LABELS)
		default:
			// Should not get here
			DEBUG_ASSERT(0);
		};
	}
}

#undef CASE_GEN

void QueryRepresentation::setSchemaName(const std::string& schemaName)
{
	_schemaName = schemaName;
}

const std::string& QueryRepresentation::schemaName() const
{
	return _schemaName;
}

uint32_t QueryRepresentation::_generateBitmap() const
{
	uint32_t bits = 0;

	BOOST_FOREACH(_MethodsType::value_type i, _methods)
		bits |= (1 << i.first);

	return bits;
}

QueryRepresentation::_MethodStringMap::_MethodStringMap()
{
	#define CODE_GENERATOR(r, data, ELEM) \
		_map[QueryRepresentationTypes::BOOST_PP_CAT(E_, ELEM)] = BOOST_PP_STRINGIZE(ELEM);

	BOOST_PP_SEQ_FOR_EACH(CODE_GENERATOR, ~, QRT_ALL_LABELS);
	#undef CODE_GENERATOR
}

IMessageObject* QueryRepresentation::_CloneAllocator::allocate_clone(const IMessageObject& a)
{
	return a.clone();
}

void QueryRepresentation::_CloneAllocator::deallocate_clone(const IMessageObject* a)
{
	delete a;
}

bool QueryRepresentation::_MethodsElementComparator::operator () (
	const QueryRepresentation::_MapElementType& l,
	const QueryRepresentation::_MapElementType& r
)const
{
    if (l.first != r.first) return false;
    if (l.second.get() == r.second.get()) return true;

    return l.second->size() == r.second->size() &&
		   std::equal(l.second->begin(), l.second->end(), r.second->begin(), IMessageObjectComparator());
}

} // namespace MsgObj

} // namespace CAL

END_CORALMESSAGING_NAMESPACE
