/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "CoralMessaging/Insulation/TimeStamp.h"
#include "CoralMessaging/Utils/StatisticsCollector.h"
#include "CoralMessaging/Utils/Error.h"
#include "CoralMessaging/Utils/Logger.h"
#include "CoralMessaging/Tcp/TcpSocketSelector.h"
#include "CoralMessaging/Tcp/TcpSocket.h"

START_CORALMESSAGING_NAMESPACE

namespace Tcp {

TcpSocket::TcpSocket(int sock)
	: _sock(sock), _fail(false)
{
	_configSocket();
	LOGGER_NEWLOG(coral::Info) << "TCP socket " << _sock << " open" << commit;
}

TcpSocket::~TcpSocket()
{
	close();
}

void TcpSocket::close()
{
	if (!isClosed())
	{
		// graceful shutdown
		::shutdown(_sock, SHUT_RDWR);
		::close(_sock);
	}
}

bool TcpSocket::isClosed() const
{
    int r = fcntl(_sock, F_GETFL);
    return !(r >= 0 && (r & O_ACCMODE));
}

bool TcpSocket::fail() const
{
	return _fail;
}

size_t TcpSocket::read(char* buf, size_t len)
{
	TimeStamp::ValueType start = TimeStamp::now().total_nanoseconds();
	size_t res = _readwrite(buf, len, ::recv, "received");
	StatisticsCollector::Instance().addStreamReadTime(TimeStamp::now().total_nanoseconds() - start);
	return res;
}

bool TcpSocket::readAll(char* buf, size_t len, const Timeout& to)
{
	return _readwriteAll(buf, len, to, &TcpSocket::read);
}

size_t TcpSocket::write(const char* buf, size_t len)
{
	TimeStamp::ValueType start = TimeStamp::now().total_nanoseconds();
	size_t res = _readwrite(buf, len, ::send, "sent");
	StatisticsCollector::Instance().addStreamWriteTime(TimeStamp::now().total_nanoseconds() - start);
	return res;
}

bool TcpSocket::writeAll(const char* buf, size_t len, const Timeout& to)
{
	return _readwriteAll(buf, len, to, &TcpSocket::write);
}

bool TcpSocket::waitForData(const Timeout& to)
{
	TcpSocketSelector tss;
	tss.insert(*this);
	return tss.wait(to);
}

template<class OP, class BUF>
size_t TcpSocket::_readwrite(BUF buf, size_t len, OP op, const char* label)
{
#ifdef __APPLE__
	static const int flags = SO_NOSIGPIPE;
#else
	static const int flags = MSG_NOSIGNAL;
#endif

	ssize_t nbytes;
op_restart:
	nbytes = op(_sock, buf, len, flags);

	if (nbytes == -1)
	{
		// No worry about the goto. We need to restart the system call if we got EINTR signal,
		// as it is not an error. This is the cleanest soulution not breaking the logic.
		if (errno == EINTR) goto op_restart;
		// The error handling logic is actually here
		// The socket is in non-blocking mode, and we need to wait...
		if (errno == EAGAIN) return 0;
		// Socket error. Turn to failure mode.
		LOGGER_NEWLOG(coral::Info) << "Socket error: " << addErr << commit;
		_fail = true;
		return 0;
	}

	if (nbytes == 0)
    {
        LOGGER_NEWLOG(coral::Info) << "Connection dropped by the other end." << commit;
        _fail = true;
    }

	if (nbytes > 0)
	{
		LOGGER_NEWLOG(coral::Verbose) << std::dec << nbytes << " bytes " << label << commit;
	}

    return static_cast<size_t>(nbytes);
}

template<class OP, class BUF>
bool TcpSocket::_readwriteAll(BUF buf, size_t len, const Timeout& to, OP op)
{
	size_t nbytes = 0;

	while (nbytes != len)
	{
		nbytes += (this->*op)(buf + nbytes, len - nbytes);
		if (_fail || to.occured()) return false;
	}

	return true;
}

void TcpSocket::_setFdFlag(int getCmd, int setCmd, long f)
{
	int flags = fcntl(_sock, getCmd, 0);
    if (flags == -1) EXCEPTION_THROW(Err_System("fnctl read error."));
    if (fcntl(_sock, setCmd, flags | f) == -1) EXCEPTION_THROW(Err_System("fnctl write error."));
}

void TcpSocket::_setSocketOption(int optname)
{
    static const int yes = 1;

    if (setsockopt(_sock, SOL_SOCKET, optname, &yes, sizeof(yes)) < 0)
    {
        EXCEPTION_THROW(Err_System("Error setting socket option."));
    }
}

void TcpSocket::_configSocket()
{
	// fast reuse enable
    _setSocketOption(SO_REUSEADDR);
    // keepalive enable
    _setSocketOption(SO_KEEPALIVE);
    // close on exec
    _setFdFlag(F_GETFD, F_SETFD, FD_CLOEXEC);
    // non-blocking mode
    _setFdFlag(F_GETFL, F_SETFL, O_NONBLOCK);
}

} // namespace Tcp

END_CORALMESSAGING_NAMESPACE
