/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "CoralMessaging/Utils/Error.h"
#include "CoralMessaging/Utils/Logger.h"
#include "CoralMessaging/Pointers.h"
#include "CoralMessaging/Tcp/TcpSocketFactory.h"
#include "CoralMessaging/Tcp/TcpSocket.h"

START_CORALMESSAGING_NAMESPACE

namespace Tcp {

Pointer<TcpSocket>::Shared TcpSocketFactory::CreateActiveSocket(unsigned int fd)
{
	return Pointer<TcpSocket>::Shared(new TcpSocket(fd));
}

Pointer<TcpSocket>::Shared TcpSocketFactory::CreateListeningSocket(unsigned int port, const std::string& ip)
{
	int sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) EXCEPTION_THROW(Err_System("Error opening socket"));
	
    int yes = 1;
    /* fast reuse enable */
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) < 0)
    	EXCEPTION_THROW(Err_System("Error setting socket option."));   
	
    struct sockaddr_in my_addr; 
    my_addr.sin_family = AF_INET;        // host byte order
    my_addr.sin_port = htons(port);      // short, network byte order

    if (inet_aton(ip.c_str(), &my_addr.sin_addr) == 0)  
    	EXCEPTION_THROW(Err_Custom(std::string("Failed listening on invalid IP ") + ip));

    memset(my_addr.sin_zero, '\0', sizeof my_addr.sin_zero);
    
    if (bind(sock, (struct sockaddr *)&my_addr, sizeof my_addr) < 0 ) 
        EXCEPTION_THROW(Err_System("Error binding socket"));
    
    LOGGER_NEWLOG(coral::Info) << "Network server start listening on port " << port << commit;
    
    if (listen(sock, _BACKLOG) < 0) EXCEPTION_THROW(Err_System("Error listening to socket."));
     
	return Pointer<TcpSocket>::Shared(new TcpSocket(sock));
}

Pointer<TcpSocket>::Shared TcpSocketFactory::CreateConnectingSocket(unsigned int port, const std::string& host)
{
    struct hostent *he;
    struct sockaddr_in their_addr; // connector's address information

    if ((he=gethostbyname(host.c_str())) == NULL) {  // get the host info
    	EXCEPTION_THROW(Err_System(std::string("Failed listening on invalid hostname ") + host));
    }
    
    int sock = socket(PF_INET, SOCK_STREAM, 0);
 
    if (sock == -1) EXCEPTION_THROW(Err_System("Error opening socket"));
 
    their_addr.sin_family = AF_INET;    // host byte order
    their_addr.sin_port = htons(port);  // short, network byte order
    their_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(their_addr.sin_zero, '\0', sizeof their_addr.sin_zero);

    if (connect(sock, (struct sockaddr *)&their_addr, sizeof their_addr) == -1) {
    	EXCEPTION_THROW(Err_System("Cannot connect"));
    }
    	
    return Pointer<TcpSocket>::Shared(new TcpSocket(sock));
}

} // namespace Tcp

END_CORALMESSAGING_NAMESPACE
