/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <boost/assign/std/vector.hpp>

#include "CoralBase/AttributeList.h"
#include "CoralBase/Blob.h"

#include "CoralMessaging/Stream/MessageStreamReader.h"
#include "CoralMessaging/Stream/MessageStreamWriter.h"
#include "CoralMessaging/CAL/MsgObj/AttributeListRepresentation.h"
#include "CoralMessaging/Utils/ctl.h"

#include "AttributeListRepresentationTest.h"

START_CORALMESSAGING_NAMESPACE

using CAL::MsgObj::AttributeListRepresentation;

void AttributeListRepresentationTest::testBlob()
{
	AttributeList al;
	al.extend("blob", typeid(coral::Blob));
	coral::Blob& blob = al[0].data<coral::Blob>();
	std::string s("abcde");
	blob.resize(s.size());
	::memcpy(blob.startingAddress(), s.c_str(), s.size() + 1);
	// Test encode - decode
	AttributeListRepresentation alr(al);
	IMessageObject::EncodedType enc = alr.encode();
	AttributeListRepresentation alr2;
	alr2.decode(*enc);
	CPPUNIT_ASSERT_EQUAL(alr, alr2);
	CPPUNIT_ASSERT(alr.attributeList() == alr2.attributeList());

	// test read-write
	StrType str(new Stream::DataStringStream);
	Stream::MessageStreamWriter<StrType> msw(str);
	CPPUNIT_ASSERT(msw.write(alr));
	Stream::MessageStreamReader<StrType> msr(str);
	AttributeListRepresentation alr3;
	msr.read(alr3);
	CPPUNIT_ASSERT(!msr.fail());
	CPPUNIT_ASSERT_EQUAL(alr, alr3);
	CPPUNIT_ASSERT(alr.attributeList() == alr3.attributeList());

	// Test big blob
	long int bSize = 100000;
	const_cast<AttributeList&>(alr.attributeList())[0].data<coral::Blob>().resize(bSize);
	str->reset();
	CPPUNIT_ASSERT(msw.write(alr));
	msr.read(alr3);
	CPPUNIT_ASSERT(!msr.fail());
	CPPUNIT_ASSERT_EQUAL(bSize, alr3.attributeList()[0].data<coral::Blob>().size());
	CPPUNIT_ASSERT_EQUAL(alr, alr3);
	CPPUNIT_ASSERT(alr.attributeList() == alr3.attributeList());
}

void AttributeListRepresentationTest::testVector()
{
	AttributeList al;
	al.extend("blob", typeid(coral::Blob));
	std::vector<AttributeListRepresentation> v1;
	boost::assign::push_back(v1)(AttributeListRepresentation(al))(AttributeListRepresentation(al))(AttributeListRepresentation(al));

	StrType str(new Stream::DataStringStream);
	Stream::MessageStreamWriter<StrType> msw(str);
	CPPUNIT_ASSERT(msw.write(v1));
	Stream::MessageStreamReader<StrType> msr(str);
	std::vector<AttributeListRepresentation> v2;
	msr.read(v2);
	CPPUNIT_ASSERT(!msr.fail());
	CPPUNIT_ASSERT_EQUAL(v1, v2);
}

void AttributeListRepresentationTest::testSharedAttributeList()
{
	AttributeList al;
	al.extend("id", typeid(float));
	AttributeListRepresentation alr(al, true);
	float& al_val = al[0].data<float>();
	const float& alr_val = alr.attributeList()[0].data<float>();
	al_val = 6;
	CPPUNIT_ASSERT_EQUAL(al_val, alr_val);

	// Share after copy
	AttributeListRepresentation alr2(alr);
	al_val = 8;
	CPPUNIT_ASSERT_EQUAL(float(8), alr2.attributeList()[0].data<float>());

	AttributeListRepresentation alr3;
	alr3 = alr;
	al_val = 10;
	CPPUNIT_ASSERT_EQUAL(float(10), alr2.attributeList()[0].data<float>());
}

END_CORALMESSAGING_NAMESPACE
