/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <algorithm>
#include <boost/bind.hpp>

#include "CoralBricks/ThreadTraits.h"
#include "CoralMessaging/Timeout.h"

#include "CoralIOHandlerTestBase.h"

START_CORALMESSAGING_NAMESPACE

void CoralIOHandlerTestBase::_ThrReadWrite(const CoralIOHandlerTestBase::ThrCommand& cmd)
{
	const std::string& s = cmd.s;
	ICoralIOHandler& ioh = *cmd.ioh;
	
	size_t received = 0;
	char buf[cmd.s.size()];
	Timeout& to(_GetTimeout());
	
	while(received != s.size())
	{
		received += ioh.read(buf + received, s.size() - received);
		CPPUNIT_ASSERT(!to.occured());
	}
	
	CPPUNIT_ASSERT(std::equal(s.begin(), s.end(), buf));
}

void CoralIOHandlerTestBase::_ThrReadWriteAll(const CoralIOHandlerTestBase::ThrCommand& cmd)
{
	ICoralIOHandler& ioh = *cmd.ioh;
	char buf[cmd.s.size()];
	Timeout& to(_GetTimeout());
	CPPUNIT_ASSERT(ioh.readAll(buf, cmd.s.size(), to));
	CPPUNIT_ASSERT(std::equal(cmd.s.begin(), cmd.s.end(), buf));
}

void CoralIOHandlerTestBase::_ThrReadFailure(CoralIOHandlerTestBase::ThrCommand& cmd)
{
	ICoralIOHandler& ioh = *cmd.ioh;
	Timeout& to(_GetWaitTimeout());
	// The strategy: attempt to read data until timeout or failure occur. Timeout mains test failure.
	
	while(!to.occured())
	{
		char buf = 0;
		ioh.read(&buf, 1);
		if (ioh.fail()) break;
	}
	
	CPPUNIT_ASSERT(ioh.fail());
}

void CoralIOHandlerTestBase::_ThrWriteFailure(CoralIOHandlerTestBase::ThrCommand& cmd)
{	
	ICoralIOHandler& ioh = *cmd.ioh;
	Timeout& to(_GetWaitTimeout());
	// The strategy: attempt to write data until timeout or failure occur. Timeout mains test failure.
	
	while(!to.occured())
	{
		const char buf = 0;
		ioh.write(&buf, 1);
		if (ioh.fail()) break;
	}
	
	CPPUNIT_ASSERT(ioh.fail());
}

void CoralIOHandlerTestBase::_ThrWait(ThrCommand& cmd)
{
	ICoralIOHandler& ioh = *cmd.ioh;
	ioh.waitForData(_GetWaitTimeout());
	char buf = 0;
	ioh.write(&buf, 1);
}

const std::string& CoralIOHandlerTestBase::_GenerateLongRandomData()
{
	static std::string s(_DataSize, 0);
	std::generate(s.begin(), s.end(), std::rand);
	return s;
}

void CoralIOHandlerTestBase::_testReadWrite(ICoralIOHandler& sender, ICoralIOHandler& receiver)
{
	// Create long random data to write
	const std::string& s = _GenerateLongRandomData();
	// Create a thread that consumes and compares the data
	ThrCommand cmd(&receiver, s);
	ThreadTraits::THREAD thr(boost::bind(_ThrReadWrite, cmd));
	// Start writing the data
	size_t written = 0;
	Timeout& to(_GetTimeout());
	
	while(written != s.size())
	{
		written += sender.write(s.c_str() + written, s.size() - written);
		CPPUNIT_ASSERT(!to.occured());
	}
	
	thr.join();
}

void CoralIOHandlerTestBase::_testReadWriteAll(ICoralIOHandler& sender, ICoralIOHandler& receiver)
{
	// Create long random data to write
	const std::string& s = _GenerateLongRandomData();
	// Create a thread that consumes and compares the data
	ThrCommand cmd(&receiver, s);
	ThreadTraits::THREAD thr(boost::bind(_ThrReadWriteAll, cmd));
	// Start writing the data
	Timeout& to(_GetTimeout());
	sender.writeAll(s.c_str(), s.size(), to);
	CPPUNIT_ASSERT(!to.occured());
	thr.join();
}

void CoralIOHandlerTestBase::_testReadFailure(ICoralIOHandler& sender, ICoralIOHandler& receiver, CoralIOHandlerTestBase::forceFailure_t fcmd)
{
	const std::string s; // Dummy stuff
	// Create a that should detect error on the receiver
	ThrCommand cmd(&receiver, s);
	ThreadTraits::THREAD thr(boost::bind(_ThrReadFailure, cmd));
	CPPUNIT_ASSERT(!sender.fail());
	CPPUNIT_ASSERT(!receiver.fail());
	fcmd(sender);
	CPPUNIT_ASSERT(sender.fail());
	thr.join();
}

void CoralIOHandlerTestBase::_testWriteFailure(ICoralIOHandler& sender, ICoralIOHandler& receiver, CoralIOHandlerTestBase::forceFailure_t fcmd)
{
	const std::string s; // Dummy stuff
	// Create a that should detect error on the receiver
	ThrCommand cmd(&receiver, s);
	ThreadTraits::THREAD thr(boost::bind(_ThrWriteFailure, cmd));
	CPPUNIT_ASSERT(!sender.fail());
	CPPUNIT_ASSERT(!receiver.fail());
	fcmd(sender);
	CPPUNIT_ASSERT(sender.fail());
	thr.join();
}

void CoralIOHandlerTestBase::_testWaitForData(ICoralIOHandler& sender, ICoralIOHandler& receiver)
{
	// Create a thread that writes the data after some wait time
	const std::string s("test");
	ThrCommand cmd(&sender, s);
	ThreadTraits::THREAD thr1(boost::bind(_ThrWait, cmd));
	// Test if we are able to get data in timeout
	CPPUNIT_ASSERT(receiver.waitForData(_GetLongerWaitTimeout()));
	thr1.join();
	// The thread wrote 1 byte. We need to try 2, as the error is detected only when
	// we try to read beyond the stream content
	char buf[2];
	receiver.read(buf, 2);
	// Test if we are able to detect timeout
	ThreadTraits::THREAD thr2(boost::bind(_ThrWait, cmd));
	CPPUNIT_ASSERT(!receiver.waitForData(_GetShortWaitTimeout()));
	thr2.join();
}

END_CORALMESSAGING_NAMESPACE
