/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */

#include <sstream>

#include "CoralMessaging/CTL/CTLPacketReader.h"
#include "CoralMessaging/CAL/CALMessages.h"

#include "CoralMessaging/Stream/DataStringStream.h"
#include "CoralMessaging/CoralSocket.h"
#include "PacketReaderTest.h"

START_CORALMESSAGING_NAMESPACE

using CORALMESSAGING_NAMESPACE_NAME::CoralSocket;

void PacketReaderTest::testOnePacket(void)
{
	uint32_t clientID = 1;
	uint32_t reqID = 2;
	Pointer<Stream::DataStringStream>::Shared str1(new Stream::DataStringStream);
	// Write a message into a coral socket, then read it back
	CAL::MsgObj::Connection mo_conn(true, "oracle://lcg_coral_nightly/CORAL_WRITER");
	CAL::MsgObj::MonitoringInformation mi_conn;
	CAL::Connect msg1(mo_conn, mi_conn);
	CoralSocket csock1(str1, clientID, reqID, msg1);
	msg1.write(csock1);

	Pointer<ICoralIOHandler>::Shared str2(new Stream::DataStringStream(str1->str().str()));
	CTL::PacketReader preader(str2);
	Pointer<CTL::Packet>::Shared packet;
	InfiniteTimeout to;

	while (!packet.get()) {
		packet = preader.pickPacket(to);
		CPPUNIT_ASSERT(!preader.terminated());
	}

	CPPUNIT_ASSERT(packet.get());
	CPPUNIT_ASSERT_EQUAL(clientID, static_cast<uint32_t>(packet->header().clientID()));
	CPPUNIT_ASSERT_EQUAL(reqID, static_cast<uint32_t>(packet->header().requestID()));
	CPPUNIT_ASSERT_EQUAL(msg1.opcode(),packet->header().opcode());
	CPPUNIT_ASSERT_EQUAL(str1->str().str().size(), packet->header().packetSize());
	CPPUNIT_ASSERT_EQUAL(msg1.cacheable(), packet->header().isCacheable());

	IMessageObject::EncodedType s1 = mo_conn.encode();
	IMessageObject::EncodedType s2 = mi_conn.encode();

	uint32_t s = s1->size();
	std::string cmp;
	cmp.append(reinterpret_cast<const char*>(&s), sizeof(uint32_t));
	cmp.append(*s1);
	s = s2->size();
	cmp.append(reinterpret_cast<const char*>(&s), sizeof(uint32_t));
	cmp.append(*s2);
	CPPUNIT_ASSERT_EQUAL(cmp, str1->str().str().substr(20));
}


void PacketReaderTest::testStreamBroken(void)
{
	uint32_t clientID = 1;
	uint32_t reqID = 2;
	Pointer<Stream::DataStringStream>::Shared str1(new Stream::DataStringStream);
	CAL::MsgObj::Connection mo_conn(true, "oracle://lcg_coral_nightly/CORAL_WRITER");
	CAL::MsgObj::MonitoringInformation mi_conn;
	CAL::Connect msg1(mo_conn, mi_conn);
	CoralSocket csock1(str1, clientID, reqID, msg1);
	msg1.write(csock1);

	Pointer<Stream::DataStringStream>::Shared str2(new Stream::DataStringStream(str1->str().str().substr(0, 29)));
	CTL::PacketReader preader(str2);
	Pointer<CTL::Packet>::Shared packet;
	Timeout to(100, 0, 0);
	packet = preader.pickPacket(to);
	CPPUNIT_ASSERT(!packet.get());
	CPPUNIT_ASSERT(preader.terminated());
}

void PacketReaderTest::testGarbage(void)
{
	Pointer<ICoralIOHandler>::Shared str1(new Stream::DataStringStream("garbagegarbagegarbagegarbagegarbage"));
	CTL::PacketReader preader(str1);
	InfiniteTimeout to;
	Pointer<CTL::Packet>::Shared packet(preader.pickPacket(to));
	CPPUNIT_ASSERT(!packet.get());
	CPPUNIT_ASSERT(preader.terminated());
}

void PacketReaderTest::testZeroPayload(void)
{
	uint32_t clientID = 1;
	uint32_t reqID = 2;
	Pointer<Stream::DataStringStream>::Shared str1(new Stream::DataStringStream);
	// Write a message into a coral socket, then read it back
	CAL::OK msg1;
	CoralSocket csock1(str1, clientID, reqID, msg1);
	msg1.write(csock1);

	Pointer<Stream::DataStringStream>::Shared str2(new Stream::DataStringStream(str1->str().str()));
	CTL::PacketReader preader(str2);
	Pointer<CTL::Packet>::Shared packet;
	InfiniteTimeout to;
	packet = preader.pickPacket(to);
	CPPUNIT_ASSERT(packet.get());
	CPPUNIT_ASSERT_EQUAL(size_t(0), packet->header().payloadSize());
}

void PacketReaderTest::testMorePackets(void)
{
	uint32_t clientID = 1;
	uint32_t reqID = 2;
	Pointer<Stream::DataStringStream>::Shared str1(new Stream::DataStringStream);
	// Write a message into a coral socket, then read it back
	CAL::MsgObj::Connection mo_conn(true, "oracle://lcg_coral_nightly/CORAL_WRITER");
	CAL::MsgObj::MonitoringInformation mi_conn;
	CAL::Connect msg1(mo_conn, mi_conn);
	CoralSocket csock1(str1, clientID, reqID, msg1);
	// write 2 messages (packets)
	msg1.write(csock1);
	msg1.write(csock1);

	Pointer<Stream::DataStringStream>::Shared str2(new Stream::DataStringStream(str1->str().str()));
	CTL::PacketReader preader(str2);
	InfiniteTimeout to;
	Pointer<CTL::Packet>::Shared packet = preader.pickPacket(to);
	CPPUNIT_ASSERT(packet.get());
	Pointer<CTL::Packet>::Shared packet2 = preader.pickPacket(to);
	CPPUNIT_ASSERT(packet2.get());
}

END_CORALMESSAGING_NAMESPACE
