/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include <boost/foreach.hpp>
#include <boost/assign/std/vector.hpp>

#include "CoralMessaging/CAL/MsgObj/ParameterCodecs.h"
#include "CoralMessaging/Stream/MessageStreamReader.h"
#include "CoralMessaging/Stream/MessageStreamWriter.h"
#include "CoralMessaging/CAL/MsgObj/AttributeListRepresentation.h"
#include "TestQueryDefinition.h"
#include "QueryRepresentationTest.h"

#include <utility>

START_CORALMESSAGING_NAMESPACE

using CAL::MsgObj::QueryRepresentationTypes;
using CAL::MsgObj::AttributeListRepresentation;
using CAL::MsgObj::Parameter_2;

void QueryRepresentationTest::testEquality()
{
	CPPUNIT_ASSERT(QueryRepresentation() == QueryRepresentation());
	QueryRepresentation q1_1, q1_2;
	_SetupObject(q1_1);
	_SetupObject(q1_2);
	CPPUNIT_ASSERT_EQUAL(q1_1, q1_2);
}

void QueryRepresentationTest::testEncodeDecode()
{
	QueryRepresentation q1_1, q1_2;
	_SetupObject(q1_1);
	IMessageObject::EncodedType s1_1 = q1_1.encode();
	q1_2.decode(*s1_1);
	CPPUNIT_ASSERT_EQUAL(q1_1, q1_2);
}

void QueryRepresentationTest::testReplay()
{
	QueryRepresentation q1, q2;
	_SetupObject(q1);
	IMessageObject::EncodedType s1 = q1.encode();
	q2.decode(*s1);
	TestQueryDefinition qd1(q2.schemaName());
	q2.replay(qd1);
	CPPUNIT_ASSERT_EQUAL(q1, qd1.repr());
}

void QueryRepresentationTest::testAssign()
{
	QueryRepresentation q1;
	_SetupObject(q1);
	QueryRepresentation* q2 = new QueryRepresentation;
	*q2 = q1;
	CPPUNIT_ASSERT_EQUAL(q1, *q2);
	delete q2;
	CPPUNIT_ASSERT_EQUAL(q1, q1);
}

void QueryRepresentationTest::testCopy()
{
	QueryRepresentation q1;
	_SetupObject(q1);
	Pointer<QueryRepresentation>::Scoped q2(dynamic_cast<QueryRepresentation*>(q1.clone()));
	CPPUNIT_ASSERT_EQUAL(q1, *q2);
	QueryRepresentation q3(q1);
	CPPUNIT_ASSERT_EQUAL(q1, q3);
}

void QueryRepresentationTest::testParameterEncoderRW()
{
	typedef QueryRepresentationTypes::groupBy repr_type_1;
	repr_type_1 repr_1, repr_2;
	repr_1.ptr = new repr_type_1::type("groupByValue_1");
	repr_2.ptr = new repr_type_1::type("groupByValue_2");
	typedef std::vector<repr_type_1::type*> VectorType_1;
	VectorType_1 v1;
	boost::assign::push_back(v1)(repr_1.ptr)(repr_2.ptr);
	StrType str(new Stream::DataStringStream);
	Stream::MessageStreamWriter<StrType> msw(str);
	msw.write(v1);
	VectorType_1 v2;
	Stream::MessageStreamReader<StrType> msr(str);
	msr.read(v2);
	CPPUNIT_ASSERT_EQUAL(v1.size(), v2.size());
	CPPUNIT_ASSERT_EQUAL(*v1[0], *v2[0]);
	CPPUNIT_ASSERT(std::equal(v1.begin(), v1.end(), v2.begin(), ctl::_PtrContainerComparator<VectorType_1>()));
}

void QueryRepresentationTest::testSharedAttributeInSetCondition()
{
	using namespace std::rel_ops;

	AttributeList al1;
	al1.extend("id", typeid(double));
	double& al1_val = al1[0].data<double>();
	al1_val = 6.8;

	AttributeList al2(al1);
	double& al2_val = al2[0].data<double>();
	al2_val = 98.4;

	// Check in ParameterCodecs
	AttributeListRepresentation alr1(al1, true);
	CPPUNIT_ASSERT_EQUAL(double(6.8), alr1.attributeList()[0].data<double>());
	al1_val = 10;
	CPPUNIT_ASSERT_EQUAL(double(10), alr1.attributeList()[0].data<double>());
	Parameter_2<std::string, AttributeListRepresentation> p1("", AttributeListRepresentation(al1, true));
	CPPUNIT_ASSERT_EQUAL(double(10), p1.p2().attributeList()[0].data<double>());
	al1_val = 20;
	CPPUNIT_ASSERT_EQUAL(double(20), p1.p2().attributeList()[0].data<double>());

	// Test in QueryRepresentation
	QueryRepresentation qr1, qr2;
	typedef QueryRepresentationTypes::setCondition repr_type;
	repr_type repr1, repr2;
	repr1.ptr = new repr_type::type("atyalapatyala", AttributeListRepresentation(al1, true));
	qr1.replaceMethodCall(repr1);
	repr2.ptr = new repr_type::type("atyalapatyala", AttributeListRepresentation(al2, true));
	qr2.replaceMethodCall(repr2);

	CPPUNIT_ASSERT(qr1 != qr2);
	al1_val = al2_val;
	CPPUNIT_ASSERT_EQUAL(repr1.ptr->p2().attributeList()[0].data<double>(), al2_val);
	CPPUNIT_ASSERT_EQUAL(qr1, qr2);
}

void QueryRepresentationTest::_SetupObject(QueryRepresentation& qr)
{
	qr.setSchemaName("atyalapatyala");
	// A setDistinct call
	typedef QueryRepresentationTypes::setDistinct repr_type_1;
	repr_type_1 repr_1;
	repr_1.ptr = new repr_type_1::type;
	qr.replaceMethodCall(repr_1);

	// groupBy
	typedef QueryRepresentationTypes::groupBy repr_type_2;
	repr_type_2 repr_2;
	repr_2.ptr = new repr_type_2::type("groupby");
	qr.replaceMethodCall(repr_2);

	// addToTableList
	typedef QueryRepresentationTypes::addToTableList repr_type_3;
	repr_type_3 repr_3;
	repr_3.ptr = new repr_type_3::type("table", "alias");
	qr.insertMethodCall(repr_3);
}

END_CORALMESSAGING_NAMESPACE
