#ifndef CORALPROTOCOL_PACKETHEADER_H_
#define CORALPROTOCOL_PACKETHEADER_H_

#include <cstring>
#include <limits>
#include <ostream>

#include "arch.h"
#include "utils.h"
#include "SequenceNumber.h"
#include "Opcode.h"

START_CORALPROTOCOL_NAMESPACE

class PacketHeader {
	friend std::ostream& operator << (PacketHeader& header, std::ostream& os);
	
public:
	// No boundary, etc. check, aBuf must hold properly the header!
	static PacketHeader* create(const std::string& data);

	// Fix the endianness of the new packet header. Cannot be changed in runtime.
    PacketHeader(const t_endianness e = _system_endian);
    
    PacketHeader(const PacketHeader& x) {
    	_header = x._header;
    }
       
    t_endianness endian() const {
        return _header.delim == LITTLE_ENDIAN_MAGIC_WORD() ? e_little : e_big;
    }
    
    static size_t size() {
        return sizeof(_t_header);
    }   
    
    static size_t maxDataSize() {
    	return std::numeric_limits<uint32_t>::max() - size();
    }
    
    uint8_t version() const {
        return _getFirstByte(_header.ver_clientid);
    }

    bool isVersionMismatch() const {
        return version() != _PROTOCOL_VERSION;
    }

    Opcode opcode() const {
        return _getFirstByte(_header.opcode_reqid); 
    }
    
    void setOpcode(const Opcode aOpCode) {
        _header.opcode_reqid = _setFirstByte(_header.opcode_reqid, static_cast<const uint8_t>(aOpCode));
    }

    size_t packetSize() const {
        return endian() == _system_endian ? _header.psize : Utils::swap_word(_header.psize);
    }
    
    size_t payloadSize() const {
        return packetSize() - size();
    }

    void setPacketSize(const uint32_t aSize) {
    	CORALPROTOCOL_DEBUG_ASSERT(aSize <= std::numeric_limits<uint32_t>::max() - size());
    	uint32_t s = aSize + size();
        _header.psize = (endian() == _system_endian ? s : Utils::swap_word(s));
    }
   
    uint24_t clientID() const {
        return _getLast3Byte(_header.ver_clientid); 
    }

    void setClientID(const uint24_t aID) {
        _header.ver_clientid = _setLast3Byte(_header.ver_clientid, aID);
    }

    uint24_t requestID() const {
        return _getLast3Byte(_header.opcode_reqid); 
    }

    void setRequestID(const uint24_t aID) {
        _header.opcode_reqid = _setLast3Byte(_header.opcode_reqid, aID);
    }

	SequenceNumber sequenceNum() const {
        return _getLast3Byte(_header.flags_seqnum); 
    }

    void setSequenceNum(const SequenceNumber aNum) {
        _header.flags_seqnum = _setLast3Byte(_header.flags_seqnum, static_cast<uint24_t>(aNum));
    }

    bool isCacheable() const {   
        return _flags() & (1 << _BIT_CACHEABLE);
    }

    void setCacheable(const bool isCacheable) {
        if (isCacheable) {
            _setFlags(_flags() | (1 << _BIT_CACHEABLE));
        } else {
            _setFlags(_flags() & ~(1 << _BIT_CACHEABLE));
        }
    }
    
    template<class OutputStream>
    bool write(OutputStream& out) const 
    {
    	out.write(reinterpret_cast<const char*>(&_header), size());
       	return !out.fail();
    }
    
    const char* raw() const {
        return (const char*) &_header;
    }    
    
    static const uint32_t LITTLE_ENDIAN_MAGIC_WORD() {
    	return 0x634F7253;   // cOrS
    }
    
    static const uint32_t BIG_ENDIAN_MAGIC_WORD() {
    	return 0x53724F63;   // SrOc
    }

private:
	
    PacketHeader(const std::string& aBuf) {
        memcpy(&_header, aBuf.data(), size());
    }

    uint32_t _setFirstByte(const uint32_t aWord, const uint8_t aByte) const {
        uint32_t orig = (endian() == _system_endian ? aWord : Utils::swap_word(aWord));
        uint32_t val = (static_cast<uint32_t>(aByte) << 24) | (orig & 0x00FFFFFF);
        return (endian() == _system_endian ? val : Utils::swap_word(val));
    }
   
    uint8_t _getFirstByte(const uint32_t aWord) const {
        uint32_t orig = (endian() == _system_endian ? aWord : Utils::swap_word(aWord));
        return (orig & 0xFF000000) >> 24;
    } 

    uint32_t _setLast3Byte(const uint32_t aWord, const uint24_t aVal) const {
        uint32_t orig = (endian() == _system_endian ? aWord : Utils::swap_word(aWord));
        uint32_t val = (static_cast<const uint32_t>(aVal) | (orig & 0xFF000000));
        return (endian() == _system_endian ? val : Utils::swap_word(val));
    }
   
    uint24_t _getLast3Byte(const uint32_t aWord) const {
        uint32_t orig = (endian() == _system_endian ? aWord : Utils::swap_word(aWord));
        return orig & 0x00FFFFFF;
    } 

    
    void _setProtocolVersion(const uint8_t aVer) {
        _header.ver_clientid = _setFirstByte(_header.ver_clientid, aVer);
    }

    uint8_t _flags() const {
        return _getFirstByte(_header.flags_seqnum); 
    }

    void _setFlags(const uint32_t aFlags) {
        _header.flags_seqnum = _setFirstByte(_header.flags_seqnum, aFlags);
    }
    
    struct _t_header {
        uint32_t delim;
        uint32_t ver_clientid;
        uint32_t psize;
        uint32_t opcode_reqid;
        uint32_t flags_seqnum;
    } PACK_STRUCT _header;
    
    static const uint8_t  _PROTOCOL_VERSION = 0;
    static const uint8_t  _BIT_CACHEABLE = 7;
};

std::ostream& operator << (PacketHeader& header, std::ostream& os);

END_CORALPROTOCOL_NAMESPACE

#endif // #ifndef CORALPROTOCOL_PACKETHEADER_H_
