#include <boost/shared_ptr.hpp>

#include "CoralProtocol/PacketHeader.h"
#include "PacketHeaderTest.h"

START_CORALPROTOCOL_NAMESPACE

void PacketHeaderTest::testConstructionFromBuffer()
{
	// big endian
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x03, 0x04, 0x02,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x07, 0x08, 0x09, 0x06,  // uint32_t opcode_reqid;
				  0x80, 0x08, 0x09, 0x0A}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT_EQUAL(e_big, phead->endian());
	CPPUNIT_ASSERT(phead->isCacheable());
	CPPUNIT_ASSERT(!phead->isVersionMismatch());
	CPPUNIT_ASSERT_EQUAL(0x07, (int)phead->opcode());
	CPPUNIT_ASSERT_EQUAL(0x02030405, (int)phead->packetSize());
	CPPUNIT_ASSERT_EQUAL(0x080906, (int)phead->requestID());
	CPPUNIT_ASSERT_EQUAL(0x030402, (int)phead->clientID());
	CPPUNIT_ASSERT_EQUAL(0x00, (int)phead->version());
	CPPUNIT_ASSERT_EQUAL(SequenceNumber(0x08090A), phead->sequenceNum());
	// Test copy
	PacketHeader phead2(*phead);
	CPPUNIT_ASSERT_EQUAL(e_big, phead2.endian());
	CPPUNIT_ASSERT(phead2.isCacheable());
	CPPUNIT_ASSERT(!phead2.isVersionMismatch());
	CPPUNIT_ASSERT_EQUAL(0x07, (int)phead->opcode());
	CPPUNIT_ASSERT_EQUAL(0x02030405, (int)phead->packetSize());
	CPPUNIT_ASSERT_EQUAL(0x080906, (int)phead->requestID());
	CPPUNIT_ASSERT_EQUAL(0x030402, (int)phead->clientID());
	CPPUNIT_ASSERT_EQUAL(0x00, (int)phead->version());
	CPPUNIT_ASSERT_EQUAL(SequenceNumber(0x08090A), phead->sequenceNum());
}

void PacketHeaderTest::testCacheableFlag()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x80, 0x08, 0x09, 0x0A}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT(phead->isCacheable());
	phead->setCacheable(false);
	CPPUNIT_ASSERT(!phead->isCacheable());
	phead->setCacheable(true);
	CPPUNIT_ASSERT(phead->isCacheable());
	// Check construction from "noncacheable" buffer
	str[16] = 0;
	boost::shared_ptr<PacketHeader> phead2(PacketHeader::create(str));
	CPPUNIT_ASSERT(!phead2->isCacheable());
	// Little endianness
	char buf2[] = {0x53, 0x72, 0x4F, 0x63,  // uint32_t delim
		           0x00, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				   0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				   0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				   0x0A, 0x08, 0x09, 0x80}; // uint32_t flags_seqnum;
	
	str.resize(0);
	str.append(buf2, sizeof(buf2));
	boost::shared_ptr<PacketHeader> phead3(PacketHeader::create(str));
	CPPUNIT_ASSERT(phead3->isCacheable());
	phead3->setCacheable(false);
	CPPUNIT_ASSERT(!phead3->isCacheable());
	phead3->setCacheable(true);
	CPPUNIT_ASSERT(phead3->isCacheable());
	// Check construction from "noncacheable" buffer
	str[19] = 0;
	boost::shared_ptr<PacketHeader> phead4(PacketHeader::create(str));
	CPPUNIT_ASSERT(!phead4->isCacheable());
}

void PacketHeaderTest::testVersionMismatch()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x01, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x80, 0x08, 0x09, 0x0A}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT(phead->isVersionMismatch());
}

void PacketHeaderTest::testLittleEndianness()
{
	char buf[] = {0x53, 0x72, 0x4F, 0x63,  // uint32_t delim
		          0x02, 0x03, 0x04, 0x00,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x0A, 0x09, 0x07, 0x80}; // uint32_t flags_seqnum;
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT(!phead->isVersionMismatch());
	CPPUNIT_ASSERT_EQUAL(e_little, phead->endian());
	CPPUNIT_ASSERT(phead->isCacheable());
	CPPUNIT_ASSERT_EQUAL(0x09, (int)phead->opcode());
	CPPUNIT_ASSERT_EQUAL(0x05040302, (int)phead->packetSize());
	CPPUNIT_ASSERT_EQUAL(0x080706, (int)phead->requestID());
	CPPUNIT_ASSERT_EQUAL(0x040302, (int)phead->clientID());
	CPPUNIT_ASSERT_EQUAL(0x00, (int)phead->version());
	CPPUNIT_ASSERT_EQUAL(SequenceNumber(0x07090A), phead->sequenceNum());
}

void PacketHeaderTest::testWrongEndianness()
{
	char buf[] = {0x55, 0x72, 0x4F, 0x63,  // uint32_t delim
		          0x02, 0x03, 0x04, 0x00,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x08, 0x09, 0x0A, 0x00}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT(phead.get() == NULL);
}

void PacketHeaderTest::testOpcode()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x0A, 0x08, 0x09, 0x80}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));

	CPPUNIT_ASSERT_EQUAL(Opcode_gen<0x06>(), phead->opcode());
	phead->setOpcode(Opcode_gen<0x08>());
	CPPUNIT_ASSERT_EQUAL(Opcode_gen<0x08>(), phead->opcode());
}

void PacketHeaderTest::testClientID()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x02, 0x03, 0x00,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x0A, 0x08, 0x09, 0x80}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT_EQUAL(0x020300, (int)phead->clientID());
	phead->setClientID(0xABCDEF);
	CPPUNIT_ASSERT_EQUAL(0xABCDEF, (int)phead->clientID());
}

void PacketHeaderTest::testRequestID()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x0A, 0x08, 0x09, 0x80}; // uint32_t flags_seqnum
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT_EQUAL(0x070809, (int)phead->requestID());
	phead->setRequestID(0xABCDEF);
	CPPUNIT_ASSERT_EQUAL(0xABCDEF, (int)phead->requestID());
}

void PacketHeaderTest::testSequenceNum()
{
	char buf[] = {0x63, 0x4F, 0x72, 0x53,  // uint32_t delim
		          0x00, 0x02, 0x03, 0x04,  // uint32_t ver_clientid
				  0x02, 0x03, 0x04, 0x05,  // uint32_t psize;
				  0x06, 0x07, 0x08, 0x09,  // uint32_t opcode_reqid;
				  0x80, 0x08, 0x09, 0x0A}; // uint32_t flags_seqnum;
	
	std::string str;
	str.append(buf, sizeof(buf));
	boost::shared_ptr<PacketHeader> phead(PacketHeader::create(str));
	CPPUNIT_ASSERT_EQUAL(SequenceNumber(0x08090A), phead->sequenceNum());
	phead->setSequenceNum(SequenceNumber(0xABCDEF));
	CPPUNIT_ASSERT_EQUAL(SequenceNumber(0xABCDEF), phead->sequenceNum());
}

END_CORALPROTOCOL_NAMESPACE
