/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_ACTIVEOBJECT_H_
#define CORALSERVER_ACTIVEOBJECT_H_

#include "CoralServer/base.h"

#include <boost/bind.hpp>

#include "CoralBricks/MonitorObject.h"
#include "utils.h"

START_CORALSERVER_NAMESPACE

template
<
	class ExecutionPolicy,
	class TracingModel = EmptyType
>
class ActiveObject : private boost::noncopyable, public coral::MonitorObject
{
public:
	template<class T>
	ActiveObject(const T& t) : _tracer(t), _runningMethods(0) {}

	template<class T, class U>
	ActiveObject(const T& t, const U& u) : _tracer(t, u), _runningMethods(0) {}

	ActiveObject(const EmptyType&) : _runningMethods(0) {}
	ActiveObject() : _runningMethods(0) {}

	virtual ~ActiveObject() {}

	void waitUntilRunning()
	{
	MONITOR_START_CRITICAL
		while (_runningMethods > 0) _notRunning.wait(MONITOR_LOCK);
	MONITOR_END_CRITICAL
	}

protected:
	typedef ActiveObject<ExecutionPolicy, TracingModel> BaseType;
	typedef TracingModel TracingType;

	template<class OP> class Command
	{
	public:
		Command(ActiveObject* aobj, OP& op)	: _aobj(aobj), _op(op) {}
		void operator () () {_aobj->template _doWork(_op);}
	private:
		ActiveObject* _aobj;
		OP _op;
	};

	template<class OP>
	void _enqueue(OP& op)
	{
	MONITOR_START_CRITICAL
		++_runningMethods;
	MONITOR_END_CRITICAL
		Command<OP> packagedOp(this, op);
		ExecutionPolicy::instance().enqueue(packagedOp);
	}

	template<class OP>
	void _doWork(OP op)
	{
		try
		{
			op();
		}
		catch (...)
		{
		MONITOR_START_CRITICAL
			_decrease();
			throw;
		MONITOR_END_CRITICAL
		}

	MONITOR_START_CRITICAL
		_decrease();
	MONITOR_END_CRITICAL
	}

	TracingModel _tracer;

private:
	void _decrease()
	{
		DEBUG_ASSERT(_runningMethods > 0);
		--_runningMethods;
		_notRunning.notify_all();
	}

	template<class T>
	std::string _desc(const std::string& d, const T& t) {return t.id() + "(" + d + ")";}
	std::string _desc(const std::string& d, const EmptyType&) {return  "(" + d + ")";}

	std::string _description(const std::string& d) {return  _desc(d, _tracer);}

	unsigned long _runningMethods;
	typename synch_traits::CONDITION _notRunning;
};

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_ACTIVEOBJECT_H_*/
