/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include <boost/bind.hpp>

#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "CoralMessaging/CAL/MsgObj/QueryRepresentation.h"

#include "CoralMessageHandler.h"

START_CORALSERVER_NAMESPACE

CoralMessageHandler::~CoralMessageHandler()
{
	waitUntilRunning();
}

bool CoralMessageHandler::_handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	return _handle_p(m, csock);
}

bool CoralMessageHandler::_handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CloseSession>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	return _handle_p(m, csock);
}

bool CoralMessageHandler::_handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	return _handle_p(m, csock);
}

bool CoralMessageHandler::_handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping1>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	return _handle_p(m, csock);
}

bool CoralMessageHandler::_handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping2>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	return _handle_p(m, csock);
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>::Shared msg, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(msg);
	const CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::Connection& m = msg->connection();

	try {
		_session.reset(_GetSessionProxy(m.dbUrl(), m.accessMode()));
	} catch (coral::Exception e) {
		LOGGER_NEWLOG(coral::Info) << "CORAL exception: " << e.what() << commit;
		_sendException(e.what(), *csock);
		throw;
	} catch (...) {
		LOGGER_NEWLOG(coral::Info) << "Unknown exception" << commit;
		_sendException("Unknown error", *csock);
		throw;
	}

	if (_session.get()) {
		CORALMESSAGING_NAMESPACE_NAME::CAL::ConnectReply reply(Config::ServerVersion);
		CoralSocket cs (*csock, reply);
		reply.write(cs);
		LOGGER_NEWLOG(coral::Info) << ":) a client is authenticated." << commit;
		return true;
	} else {
		LOGGER_NEWLOG(coral::Info) << ":( client authentication rejected." << commit;
		_sendException("Authentication unsuccessfull", *csock);
		return false;
	}
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CloseSession>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	_sendOk(*csock);
	LOGGER_NEWLOG(coral::Info) << "Session closed." << commit;
	return true;
}


bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::StartTransaction>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	DEBUG_ASSERT_PTR(csock);
	DEBUG_ASSERT_PTR(_session);
	_session->transaction().start(m->access().access());
	_sendOk(*csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CommitTransaction>::Shared, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT(_session.get());
	_session->transaction().commit();
	_sendOk(*csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::ExecuteQuery>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	DEBUG_ASSERT_PTR(csock);
	DEBUG_ASSERT_PTR(_session);
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::QueryRepresentation;
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::AttributeListRepresentation;
	QueryRepresentation& qr = const_cast<QueryRepresentation&>(m->query());
	ISchema& schema = qr.schemaName().empty() ? _session->nominalSchema() : _session->schema(qr.schemaName());
	Pointer<IQuery>::Shared query(schema.newQuery());
	qr.replay(*query);
	ICursor& cursor = query->execute();
	std::vector<AttributeListRepresentation> resultSet;

	while (cursor.next())
		resultSet.push_back(AttributeListRepresentation(cursor.currentRow()));

	CORALMESSAGING_NAMESPACE_NAME::CAL::ExecuteQueryReply reply(resultSet);
	_sendMessage(reply, *csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::TableSchemaRead>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	DEBUG_ASSERT_PTR(_session);
	ISchema& schema = _session->schema(m->schema().schemaName());
	ITable& table = schema.tableHandle(m->schema().tableName());
	const ITableDescription& desc = table.description();

	// get the primary keys
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::PrimaryKeyDescription;
	PrimaryKeyDescription pkDesc;

	if (desc.hasPrimaryKey())
		pkDesc = PrimaryKeyDescription(desc.primaryKey().columnNames(), desc.primaryKey().tableSpaceName());

	// Fetch the column descriptions
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::ColumnDescription;
	std::vector<ColumnDescription> coldesc;

	for (int i = 0; i < desc.numberOfColumns(); ++i) {
		const IColumn& c = desc.columnDescription(i);
		coldesc.push_back(ColumnDescription(c.name(), c.type(), c.size(), c.isSizeFixed(), c.isNotNull(), c.tableSpaceName()));
	}

	// Fetch the index descriptions
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::IndexDescription;
	std::vector<IndexDescription> indexDesc;

	for (int i = 0; i < desc.numberOfIndices(); ++i) {
		const IIndex& ind = desc.index(i);
		indexDesc.push_back(IndexDescription(ind.name(), ind.columnNames(), ind.isUnique(), ind.tableSpaceName()));
	}

	// Fetch the foreign key descriptions
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::ForeignKeyDescription;
	std::vector<ForeignKeyDescription> fkDesc;

	for (int i = 0; i < desc.numberOfForeignKeys(); ++i) {
		const IForeignKey& fk = desc.foreignKey(i);
		fkDesc.push_back(ForeignKeyDescription(fk.name(), fk.columnNames(), fk.referencedTableName(), fk.referencedColumnNames()));
	}

	// Fetch the unique constraints descriptions
	using CORALMESSAGING_NAMESPACE_NAME::CAL::MsgObj::UniqueConstraintDescription;
	std::vector<UniqueConstraintDescription> ucDesc;

	for (int i = 0; i < desc.numberOfUniqueConstraints(); ++i) {
		const IUniqueConstraint& uc = desc.uniqueConstraint(i);
		ucDesc.push_back(UniqueConstraintDescription(uc.name(), uc.columnNames(), uc.tableSpaceName()));
	}

	CORALMESSAGING_NAMESPACE_NAME::CAL::TableSchemaReadReply reply(coldesc, indexDesc, fkDesc, ucDesc, pkDesc, desc.type(), desc.tableSpaceName());
	_sendMessage(reply, *csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::TableSchemaExists>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	DEBUG_ASSERT_PTR(_session);
	// The method call below throws coral::InvalidSchemaNameException if the schema does not exist
	// We do not need the result
	_session->schema(m->schemaName());
	_sendOk(*csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::ListTables>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	DEBUG_ASSERT_PTR(_session);
	std::set<std::string> tList = (m->schemaName().empty() ? _session->nominalSchema() : _session->schema(m->schemaName())).listTables();
	CORALMESSAGING_NAMESPACE_NAME::CAL::ListTablesReply	reply(std::vector<std::string>(tList.begin(), tList.end()));
	_sendMessage(reply, *csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0 p;
	_sendMessage(p, *csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping1>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	//CORALMESSAGING_NAMESPACE_NAME::CAL::Ping1 p(m->data1());
	CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0 p;
	_sendMessage(p, *csock);
	return true;
}

bool CoralMessageHandler::_handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping2>::Shared m, const Pointer<CoralInputSocket>::Shared csock)
{
	DEBUG_ASSERT_PTR(m);
	//CORALMESSAGING_NAMESPACE_NAME::CAL::Ping2 p(m->data1(), m->data2());
	CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0 p;
	_sendMessage(p, *csock);
	return true;
}

ISessionProxy* CoralMessageHandler::_GetSessionProxy(const std::string& connectionName, bool readOnly)
{
	coral::ConnectionService connService;
	return connService.connect(connectionName, readOnly ? coral::ReadOnly : coral::Update);
}

void CoralMessageHandler::_sendException(const std::string& what, CoralInputSocket& csock)
{
	uint32_t id = CORALMESSAGING_NAMESPACE_NAME::CoralException::getID(typeid(coral::Exception));
	CORALMESSAGING_NAMESPACE_NAME::CAL::Exception err(id, what);
	_sendMessage(err, csock);
}

void CoralMessageHandler::_sendOk(CoralInputSocket& csock)
{
	CORALMESSAGING_NAMESPACE_NAME::CAL::OK ok;
	_sendMessage(ok, csock);
}

void CoralMessageHandler::_sendMessage(const IMessage& msg, const CoralInputSocket& csock)
{
	CoralSocket cs(csock, msg);
	LOGGER_NEWLOG(coral::Info) << "Sending " << msg.name() << " message" << commit;
	msg.write(cs);
}

void CoralMessageHandler::_startReadonlyTransaction()
{
	DEBUG_ASSERT_PTR(_session);
	LOGGER_NEWLOG(coral::Info) << "Transaction not started, opening readonly transaction." << commit;
	_session->transaction().start(true);
}

END_CORALSERVER_NAMESPACE
