/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALMESSAGING_CORALMESSAGEHANDLER_H_
#define CORALMESSAGING_CORALMESSAGEHANDLER_H_

#include <vector>
#include <string>
#include <map>

#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/SessionException.h"
#include "CoralBase/Exception.h"
#include "CoralBase/AttributeList.h"

#include "CoralMessaging/Insulation/TimeStamp.h"
#include "CoralMessaging/Utils/StatisticsCollector.h"
#include "CoralMessaging/CAL/CALMessages.h"
#include "CoralMessaging/CTL/CTLMessages.h"
#include "CoralMessaging/CoralSocket.h"
#include "CoralMessaging/CoralException.h"
#include "CoralMessaging/IMessage.h"

#include "Logger.h"
#include "Pointers.h"
#include "ActiveObject.h"
#include "ThreadPool.h"

START_CORALSERVER_NAMESPACE

using CORALMESSAGING_NAMESPACE_NAME::CoralInputSocket;
using CORALMESSAGING_NAMESPACE_NAME::CoralSocket;
using CORALMESSAGING_NAMESPACE_NAME::IMessage;

/** Handle all the CORAL server request messages
 *
 * This class is used by the MessageHandler state machine. This handler will have one instance
 * per physical connection, and shares the connection and session states between the clients sharing
 * the connection. The only visible function is handle, that creates a Coral reply socket from the
 * input socket and the request message.
 *
 *
 */
class CoralMessageHandler : public ActiveObject<ThreadPool::ThreadPool>
{
public:
	/// Constructor
	CoralMessageHandler() : ActiveObject<ThreadPool::ThreadPool>() {}
	/// destructor
	virtual ~CoralMessageHandler();
 	/** Handle a message. The message type is already known, as the general MessageHandler dispatches
	 * to the right message.
	 *
	 * The template parameter is a pointer to the message.
	 * */
	template<class MESSAGE>
	bool handle(MESSAGE msg, const Pointer<CoralInputSocket>::Shared csock)
	{
		TimeStamp::ValueType start = TimeStamp::now().total_nanoseconds();
		DEBUG_ASSERT_PTR(msg);
		DEBUG_ASSERT_PTR(csock);
		bool res = _handle_a(msg, csock);
		coral::CoralMessaging::StatisticsCollector::Instance().addMessageHandlerTime(*msg, TimeStamp::now().total_nanoseconds() - start);
		return res;
	}

private:
	/** Pack the parameters of a method handler to pass them to a function executed in a different thread.
	 *
	 * Follows the Command design pattern. The MESSAGE parameter must follow pointer semantics. */
	template<class MESSAGE> class Command
	{
	public:
		Command(CoralMessageHandler* mh, MESSAGE m, const Pointer<CoralInputSocket>::Shared csock)
			: _mh(mh), _m(m), _csock(csock)
		{
			// EMPTY
		}

		bool operator () (bool firstPass = true)
		{
			DEBUG_ASSERT_PTR(_mh);
			DEBUG_ASSERT_PTR(_m);
			DEBUG_ASSERT_PTR(_csock);

			bool res = false;
			TimeStamp::ValueType start = TimeStamp::now().total_nanoseconds();

			try
			{
				res = _mh->template _handle_p(_m, _csock);
			}
			catch (coral::Exception& e)
			{
				if (dynamic_cast<coral::TransactionNotActiveException*>(&e) && firstPass)
				{
					_mh->_startReadonlyTransaction();
					operator () (false);
				}
				else
				{
					LOGGER_NEWLOG(coral::Error) << "CORAL Exception when handling " << _m->name()
						<< " (" << *_m << ") : (client " << _csock->clientID() << ", request "
						<< _csock->requestID() << "((" <<	e.what() << "))" << commit;

					uint32_t id = CORALMESSAGING_NAMESPACE_NAME::CoralException::getID(typeid(e));
					CORALMESSAGING_NAMESPACE_NAME::CAL::Exception msg(id, std::string(e.what()));
					CoralSocket cs(*_csock, msg);
					msg.write(cs);
				}
			}
			catch (...)
			{
				throw;
			}

			coral::CoralMessaging::StatisticsCollector::Instance().addMessageHandlerTime(*_m, TimeStamp::now().total_nanoseconds() - start);
			return res;
		}

	private:
		CoralMessageHandler* _mh;
		MESSAGE _m;
		Pointer<CoralInputSocket>::Shared _csock;
	};

	/** Active methods. Enqueue a message handler.
	 *
	 * The template parameter must be a pointer type.
	 *  */
	template<class MESSAGE>
	bool _handle_a(MESSAGE m, const Pointer<CoralInputSocket>::Shared csock)
	{
		Command<MESSAGE> cmd(this, m, csock);
		LOGGER_NEWLOG(coral::Info) << "Handling message " << m->name() << " (" << *m << ") : (client "
				                   << csock->clientID() << ", request " << csock->requestID() << commit;
		_enqueue(cmd);
		return true;
	}

	// Not concurrent (active) methods. They must be executed right away.
	bool _handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CloseSession>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0>::Shared m, const Pointer<CoralInputSocket>::Shared csock);
	bool _handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping1>::Shared m, const Pointer<CoralInputSocket>::Shared csock);
	bool _handle_a(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping2>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	// The concurrent methods
	/** Default handler for the unimplemented messages */
	template<class MESSAGE>
	bool _handle_p(MESSAGE m, const Pointer<CoralInputSocket>::Shared csock)
	{
		LOGGER_NEWLOG(coral::Info) << "Unimplemented handler for message " << m->name() << commit;
		CORALMESSAGING_NAMESPACE_NAME::CAL::Exception err("Unimplemented handler");
		CoralSocket reply(*csock, err);
		err.write(reply);
		return true;
	}

	// The implemented handlers. Template specialization message-by-message
	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CloseSession>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::StartTransaction>::Shared m, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::CommitTransaction>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::ExecuteQuery>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::TableSchemaRead>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::TableSchemaExists>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::ListTables>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping0>::Shared, const Pointer<CoralInputSocket>::Shared csock);
	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping1>::Shared, const Pointer<CoralInputSocket>::Shared csock);
	bool _handle_p(const Pointer<CORALMESSAGING_NAMESPACE_NAME::CAL::Ping2>::Shared, const Pointer<CoralInputSocket>::Shared csock);

	// The helpers
	void _startReadonlyTransaction();

	static ISessionProxy* _GetSessionProxy(const std::string& connectionName, bool readOnly);

	void _sendException(const std::string& what, CoralInputSocket& csock);

	void _sendOk(CoralInputSocket& csock);

	void _sendMessage(const coral::CoralMessaging::IMessage& msg, const CoralInputSocket& csock);

	// The only variable that is shared between the messgae handlers
	// Not protected, we rely on the CORAL locks
	Pointer<ISessionProxy>::Scoped _session;
};

END_CORALSERVER_NAMESPACE

#endif /*CORALMESSAGING_CORALMESSAGEHANDLER_H_*/
