/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string>
#include <boost/lexical_cast.hpp>
#include <boost/assign/list_of.hpp> 

#include "CoralMessaging/Utils/StatisticsCollector.h"
#include "CoralServer/CoralServer.h"
#include "DebugController.h"

START_CORALSERVER_NAMESPACE

DebugController::DebugController()
	: ActiveObject<ThreadPool::ThreadPool>(),
	  _fileName(_FifoPathPrefix() + boost::lexical_cast<std::string>(::getpid()))
{
	_cmdMap = boost::assign::map_list_of
		("stat", &DebugController::_HandleStat)
		("stop", &DebugController::_HandleStop)
		("logoff", &DebugController::_HandleLogOff)
		("logon", &DebugController::_HandleLogOn);
}

DebugController::~DebugController()
{
	// EMPTY
}

void DebugController::stop()
{
	if (_fifo.is_open()) 
	{
		_fifo.close();
		::unlink(_fileName.c_str());
	}
}

void DebugController::start()
{
	boost::function0<void> op = boost::bind(&DebugController::_start_a, this);
	_enqueue(op);
}
    
void DebugController::_start_a()
{
MONITOR_START_CRITICAL
	if (!_createFifo()) return;	
MONITOR_END_CRITICAL
	
	_CmdMap_type::iterator end = _cmdMap.end();
	
	while(1)
	{
		bool e = false;
	MONITOR_START_CRITICAL	
		_fifo.peek();
		e = _fifo.eof();
	MONITOR_END_CRITICAL		
		if (e) 
		{
			_TIMEOUT().actualize();
			ThreadTraits::THREAD::sleep(_TIMEOUT().getXtime());
		}
		else
		{
		MONITOR_START_CRITICAL
			_handleCommand();
		MONITOR_END_CRITICAL
		}	
	}
}

void DebugController::_handleCommand()
{
	std::string cmd;
	_fifo >> cmd;
	_CmdMap_type::iterator i = _cmdMap.find(cmd);
	
	if (i != _cmdMap.end())
	{
		(i->second)();
	}
	else
	{
		LOGGER_NEWLOG(coral::Info) << "Unknown debug command: " << cmd << commit;
	}
}

bool DebugController::_createFifo()
{
	struct stat stats;
	int rc = ::stat(_fileName.c_str(), &stats);
	int mode = S_IRUSR | S_IWUSR | S_IRGRP;

	if (rc == 0) 
	{
		if (!S_ISFIFO(stats.st_mode)) 
		{
			LOGGER_NEWLOG(coral::Error) << "Warning: " << _fileName
				<< " exists but is not a fifo (named pipe). Debug channel may not work." << commit;
		}
	} 
	else 
	{
		rc = ::mkfifo(_fileName.c_str(), mode);

		if (rc != 0) 
		{
			LOGGER_NEWLOG(coral::Error) << "Can't create fifo (named pipe) " << _fileName << addErr << commit;
			return false;
		}
	}
	
	_fifo.open(_fileName.c_str(), std::ios::in|std::ios::out);

	if (!_isValid()) 
	{
		LOGGER_NEWLOG(coral::Error) << "Failed to open " << _fileName << addErr << commit;
		return false;
	}
	
	return true;
}

bool DebugController::_isValid()
{
	return _fifo.is_open() && !_fifo.fail();
}

void DebugController::_HandleStat()
{
	CORALMESSAGING_NAMESPACE_NAME::StatisticsCollector::Instance().print();
}

void DebugController::_HandleLogOn()
{
	theLogger().setLogOn();
	LOGGER_NEWLOG(coral::Info) << "Logging switched on." << commit;
}

void DebugController::_HandleLogOff()
{
	LOGGER_NEWLOG(coral::Info) << "Logging switched off." << commit;
	theLogger().setLogOff();
}

void DebugController::_HandleStop()
{
	LOGGER_NEWLOG(coral::Info) << "Stop command received." << commit;
	theServer().stop();
    ::kill(::getpid(), 9);
    exit(EXIT_SUCCESS);
}

END_CORALSERVER_NAMESPACE
