/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_MESSAGEHANDLER_H_
#define CORALSERVER_MESSAGEHANDLER_H_

#include <boost/statechart/event.hpp>
#include <boost/statechart/state_machine.hpp>
#include <boost/statechart/simple_state.hpp>
#include <boost/statechart/transition.hpp>
#include <boost/statechart/custom_reaction.hpp>

#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/for_each_i.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/seq/size.hpp>
#include <boost/preprocessor/comparison/less.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/arithmetic/dec.hpp>

#include <boost/mpl/list.hpp>
#include <boost/mpl/insert_range.hpp>
#include <boost/mpl/begin_end.hpp>
#include <boost/mpl/transform.hpp>
#include <boost/mpl/placeholders.hpp>

#include "CoralMessaging/CoralSocket.h"
#include "CoralMessaging/CAL/CALMessages.h"

#include "CoralMessageHandler.h"
#include "Pointers.h"

START_CORALSERVER_NAMESPACE

using CORALMESSAGING_NAMESPACE_NAME::IMessage;
using CORALMESSAGING_NAMESPACE_NAME::CoralInputSocket;

namespace sc = boost::statechart;

namespace MessageHandlerSM {

#define ORDINARY_HANDLED_MESSAGES (StartTransaction)(CommitTransaction)(ExecuteQuery)\
	(TableSchemaRead)(TableSchemaExists)(ListTables)(Ping0)(Ping1)(Ping2)

#define STATE_CHANGING_MESSAGES (Connect)(CloseSession)
#define ALL_HANDLED_MESSAGES ORDINARY_HANDLED_MESSAGES STATE_CHANGING_MESSAGES

#define CREATE_LIST_ELEMENT(r, data, i, elem) CORALMESSAGING_NAMESPACE_NAME::CAL::elem BOOST_PP_COMMA_IF(BOOST_PP_LESS(i, data))
#define CREATE_LIST(lst) BOOST_PP_SEQ_FOR_EACH_I(CREATE_LIST_ELEMENT, BOOST_PP_DEC(BOOST_PP_SEQ_SIZE(lst)), lst)

typedef boost::mpl::list<CREATE_LIST(ORDINARY_HANDLED_MESSAGES)> OrdinaryHandledMessages;
typedef boost::mpl::list<CREATE_LIST(STATE_CHANGING_MESSAGES)> StateChangingMessages;
typedef boost::mpl::list<CREATE_LIST(ALL_HANDLED_MESSAGES)> AllHandledMessages;

template<class MESSAGE>
struct Event : public sc::event<Event<MESSAGE> >
{
	explicit Event(typename Pointer<IMessage>::Shared m, typename Pointer<CoralInputSocket>::Shared c)
		: message(m), csock(c) {}

	Event(const Event& x) : sc::event<Event<MESSAGE> >(), message(x.message), csock(x.csock) {}
	Pointer<IMessage>::Shared message;
	Pointer<CoralInputSocket>::Shared csock;
};


struct AllReactions
{
	typedef boost::mpl::transform<AllHandledMessages, sc::custom_reaction< Event<boost::mpl::placeholders::_1> > >::type type;
};

class StUnconnected; // The starting state

struct MessageHandlerStateMachine : public sc::state_machine<MessageHandlerStateMachine, StUnconnected>
{
	Pointer<CoralMessageHandler>::Scoped handler;
};

struct StateGeneral
{
	void sendNotConnected(const Pointer<CoralInputSocket>::Shared in, const std::string& desc = "");
};


struct StConnected;

struct StConnecting : public StateGeneral, public sc::simple_state<StConnecting, MessageHandlerStateMachine >
{
	typedef boost::mpl::list<sc::custom_reaction<Event<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect> > > reactions;

	sc::result react(const Event<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>& ev);
};

struct StUnconnected : public StateGeneral, public sc::simple_state<StUnconnected, MessageHandlerStateMachine>
{
	typedef AllReactions::type reactions;

	template<class MESSAGE>
	sc::result react(const Event<MESSAGE>& ev)
	{
		this->sendNotConnected(ev.csock, "Not connected to a database yet");
		return this->discard_event();
	}

	sc::result react(const Event<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>& ev);
};


struct StConnected : public sc::simple_state<StConnected, MessageHandlerStateMachine>
{
	typedef AllReactions::type reactions;

	template<class MESSAGE>
	sc::result react(const Event<MESSAGE>& ev)
	{
		typename Pointer<MESSAGE>::Shared msg(CORALMESSAGING_NAMESPACE_NAME::message_cast<MESSAGE>(ev.message));
		this->outermost_context().handler->handle(msg, ev.csock);
		return this->discard_event();
	}

	sc::result react(const Event<CORALMESSAGING_NAMESPACE_NAME::CAL::Connect>& ev);

	sc::result react(const Event<CORALMESSAGING_NAMESPACE_NAME::CAL::CloseSession>& ev);
};

} // namespace MessageHandlerSM

class MessageHandler
{
public:
	MessageHandler ();

	virtual ~MessageHandler() {}

	bool handle(const Pointer<IMessage>::Shared message, const Pointer<CoralInputSocket>::Shared csock);

	bool terminated() const;

private:
	MessageHandlerSM::MessageHandlerStateMachine _sm;
};

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_MESSAGEHANDLER_H_*/
