/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_SYNCHRONIZEDQUEUE_H_
#define CORALSERVER_SYNCHRONIZEDQUEUE_H_

#include <limits>
#include <deque>

#include "CoralBricks/MonitorObject.h"

#include "Timeout.h"
#include "Logger.h"
#include "Error.h"
#include "Cloneable.h"

START_CORALSERVER_NAMESPACE

using CORALMESSAGING_NAMESPACE_NAME::Timeout;

/**
 * Ref: An Object Behavioral Pattern for Concurrent Programming, Douglas C. Schmidt
 */
template 
<
	class ELEMENT,
	template <class> class ELEMENT_ACCESS,
	template <class> class STORAGE = std::deque
>
class SynchronizedQueue : public MonitorObject
{
public:
	typedef ELEMENT_ACCESS<ELEMENT> element_type;
	
	SynchronizedQueue(size_t maxSize = std::numeric_limits<size_t>::max()) 
		: _maxSize(maxSize)	{};
	
	virtual ~SynchronizedQueue() {
		_notEmpty.notify_all();
		_notFull.notify_all();
	};
	
	void push(const element_type& e) {
	MONITOR_START_CRITICAL
		// Wait while the queue is full
		while (_full()) _notFull.wait(MONITOR_LOCK);
		_push(e);
		_notEmpty.notify_all();
	MONITOR_END_CRITICAL
	};
	
	element_type pop(const Timeout& tdiff) {
		bool isNotTimeout = true;
	MONITOR_START_CRITICAL 
		// Wait while the queue is empty
		while (_empty() && isNotTimeout) {
			isNotTimeout = _notEmpty.timed_wait(MONITOR_LOCK, tdiff.getXtime());
		}
		element_type e = (isNotTimeout ? _pop() : element_type());
		_notFull.notify_all();
		return e;
	MONITOR_END_CRITICAL
	};
	
	element_type pop() {	
	MONITOR_START_CRITICAL 
		// Wait while the queue is empty
		while (_empty()) _notEmpty.wait(MONITOR_LOCK);
		element_type e = _pop();
		_notFull.notify_all();
		return e;
	MONITOR_END_CRITICAL
	};
	
	bool empty() const {
	MONITOR_START_CRITICAL
		return _empty();
	MONITOR_END_CRITICAL
	};
	
	bool full() const {
	MONITOR_START_CRITICAL
		return _full();
	MONITOR_END_CRITICAL
	}

private:
	
	bool _empty() const {
		return _queue.empty();
	}
	
	bool _full() const {
		return _queue.size() >= _maxSize;
	}
	
	void _push(const element_type& e) {
		DEBUG_ASSERT(!_full());
		_queue.push_back(e);		
	}
	
	element_type _pop() {
		DEBUG_ASSERT(!_empty());
		element_type e = _queue.front();
		_queue.pop_front();		
		return e;
	}
	
	const size_t _maxSize;
	//typedef std::deque<element_type> _queue_t; 
	STORAGE<element_type> _queue;
	
	typename synch_traits::CONDITION _notEmpty;
	typename synch_traits::CONDITION _notFull;
};

END_CORALSERVER_NAMESPACE

#endif /* CORALSERVER_SYNCHRONIZEDQUEUE_H_ */
