/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_TASK_H_
#define CORALSERVER_TASK_H_

#include "CoralBase/Exception.h"

#include <iostream>

#include "Error.h"
#include "Traced.h"

START_CORALSERVER_NAMESPACE

/** Common base of all Task<> classes providing Task-wise object identification for
 * tracing and forces the function operator property */ 
class ITask : public Traced<ITask, void*> {
public:
	ITask() : Traced<ITask, void*>("Task", static_cast<void*>(this)) {};
	virtual ~ITask() {};
	virtual void execute() = 0;
};

/** Task wraps the code that must be executed in a thread. It provides traceability,
 *  catches and logs any unhandled exceptions that are not handled in the operation code.
 * 
 *  The user must implement the thread code as a function object. The function object type
 *  is the template parameter of the class. The object must be:
 * 
 *  - copyable
 *  - must have operator () that executes the task code.
 */
template<class OP_TYPE> class Task : public ITask
{	
public:	
	Task(OP_TYPE& op) : ITask(), _op(op) {};
	virtual ~Task() {};
	
	/** Executes the code, catches and logs all the unhandled exceptions. */
	virtual void execute () {
		try {
			_op();
		} catch (Err& e) {
			LOGGER_NEWLOG(coral::Error) << "CoralServer Exception in " << id() << commit;
		} catch (coral::Exception& e) {
			LOGGER_NEWLOG(coral::Error) << "CORAL Exception in " << id() << e.what() << commit;
		} catch (std::exception& e) {
			LOGGER_NEWLOG(coral::Error) << "C++ Exception in " << id() << e.what() << commit;
		} catch (...) {
			std::cerr << "Unknown exception " << std::endl;
		}
	}
	
private:
	OP_TYPE _op;
};

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_TASK_H_*/
