/*
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 *
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_THREADPOOL_H_
#define CORALSERVER_THREADPOOL_H_

#include <limits>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/shared_ptr.hpp>

#include "CoralBricks/MonitorObject.h"

#include "Logger.h"
#include "Task.h"
#include "ThreadPoolWorker.h"
#include "SynchronizedQueue.h"
#include "Environment.h"

START_CORALSERVER_NAMESPACE

namespace ThreadPool {

class ThreadPool : public Traced<ThreadPool>, public MonitorObject
{
private:
	typedef SynchronizedQueue<ITask, boost::shared_ptr> _queue_t;

public:
	typedef _queue_t::element_type element_type;
	typedef boost::function0<void> op_type;

	ThreadPool(const size_t queueSize, const size_t workerNum);
	virtual ~ThreadPool();

	template<class OP_TYPE>	void enqueue(OP_TYPE& op) {
		element_type tptr(new Task<OP_TYPE>(op));
		LOGGER_NEWLOG(coral::Debug) << tptr->id() << " is waiting for enqueuing" << commit;
		_queue.push(tptr);
		LOGGER_NEWLOG(coral::Debug) << tptr->id() << " is enqueued" << commit;
	}

	element_type pop(const Timeout& td);

	static ThreadPool& instance()
	{
		static ThreadPool tp(std::numeric_limits<size_t>::max(), theEnvironment().get<int>("thread_num"));
		return tp;
	}

	void wait();
	void stop();

private:
	_queue_t _queue;

	typedef boost::ptr_vector<Worker> _workers_t;
	_workers_t _workers;
	ThreadTraits::THREAD_GROUP _thgrp;
};

struct NoThreads {
	static NoThreads& instance()
	{
		static NoThreads nt;
		return nt;
	}

	template<class OP_TYPE>	void enqueue(OP_TYPE& op, const std::string& desc = "")
	{
		LOGGER_NEWLOG(coral::Debug) << "Task (" << desc << ") is executed immediately" << commit;
		op();
		LOGGER_NEWLOG(coral::Debug) << " Task is finished" << commit;
	}
};

} // namespace ThreadPool

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_THREADPOOL_H_*/
