/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_TRACED_H_
#define CORALSERVER_TRACED_H_

#include <sstream>

#include "Logger.h"

START_CORALSERVER_NAMESPACE
/** A Traced class provides traceability and object identification for the class
 * specified in the template type parameter. This class must be inherited in this way:
 * 
 * class ConcreteTracedClass : public Traced<ConcreteTracedClass> ...
 * 
 * Construction and destruction of traced classes are logged. It also provides an
 * object identification object (string) for each traced objects.
 * 
 * The structure of the identification: 
 * 
 * <user-given class-wide prefix><system-generated class-wide unique postfix>
 * 
 * The system does not check if two difefrent classes have different user-given
 * prefixes. To avoid confusion, the user must ensure it.
 *  
 * The class ensures that objects created in different threads have different id-s
 * in the whole application by a global, synchronized postfix generator. The postfixes
 * are simple integer numbers. 
 */
class TracedBase {
public:
	TracedBase() {};
	TracedBase(const TracedBase&) {};
	virtual ~TracedBase() {}; 
	/* type of the object id */
	typedef std::string id_type; 
	virtual const id_type& id() const = 0;
};

template <class TYPE, class POSTFIX_TYPE = std::string>
class Traced : public TracedBase
{
public: 
	/** Create an unique ID for the object. */
	Traced(const char* classPrefix, POSTFIX_TYPE postfix = POSTFIX_TYPE()) {		
		std::stringstream s;
		s << _classPrefix(classPrefix) << postfix;
		_id = s.str();
		LOGGER_NEWLOG(coral::Info) << id() << " created" << commit;
	}
	
	Traced(const Traced& t) : TracedBase(), _id(t._id) {}	
	
	virtual ~Traced() {
		LOGGER_NEWLOG(coral::Info) << id() << " deleted" << commit;
	}
	
	/** Return the object ID. */
	virtual const id_type& id() const {
		return _id;
	}
	
private:
	/** Store the class prefix string */
	static const std::string& _classPrefix (const char* first = "") {
		static std::string cp(first + std::string(":"));
		return cp;
	}
	
	/* The ID */
	id_type _id;
};

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_TRACED_H_*/
