/* 
 * Project: POOL
 * Sub-project: CORAL
 * Web site: http://pool.cern.ch/coral
 * 
 * Author: Zsolt.Molnar@cern.ch (http://www.zsoltmolnar.hu)
 */
#ifndef CORALSERVER_SYNCHRONIZEDQUEUETEST_H_
#define CORALSERVER_SYNCHRONIZEDQUEUETEST_H_

#include "CoralBricks/ThreadTraits.h"
#include "src/SynchronizedQueue.h"
#include "src/Timeout.h"
#include "TestBase.h"
#include "ThreadFunction.h"

START_CORALSERVER_NAMESPACE

class SynchronizedQueueTest : public CPPUNIT_NS :: TestFixture
{
    CPPUNIT_TEST_SUITE (SynchronizedQueueTest);
    /** Test if:
     * - The queue maintains the order of the elements 
     * - properly reports if the queue is full or empty */
    CPPUNIT_TEST (elementOrder);
    /** Test if the the pushung thread is blocked on a full queue */
    CPPUNIT_TEST (blockOnFull);
    /** Test if the popping thread is blocked on an empty queue */
    CPPUNIT_TEST (blockOnEmpty);
    /** Test if a popping blocked thread on an empty queue gets notified when
     * the queue receives an element */
    CPPUNIT_TEST (notifyIfEmptyGotElement);
    /** Test if a pushing blocked thread on a full queue gets notified when
        the queue looses an element (somebody pops en element) */
    CPPUNIT_TEST (notifyIfFullLostElement);
    CPPUNIT_TEST_SUITE_END ();

public:
    void setUp (void) {};
    void tearDown (void) {};

protected:
    void elementOrder (void);
    void blockOnFull (void);
    void blockOnEmpty (void);
    void notifyIfEmptyGotElement (void);
    void notifyIfFullLostElement (void);
    
private:
	/** Create a context for every single test. A test involves:
	 *   - a queue
	 *   - the thread group: a test exits only if all the threads have been 
	 *     finished and joined
	 *  
	 * When creating a test context, it sets up a new queue and a new thread group.
	 * When the test context goes out of scope, its destructor waits while all the 
	 * threads have been joined. Mind, that in case of test failure, cppunit
	 * will generate an assertion causing test context destructor call. 
	 * 
	 * The queue that is being tested contains integer elements. The pop and push
	 * functions of the context are used to push and pop values to/from the internal 
	 * queue. Those operations are executed in separate threads, so the functions
	 * basically create function objects that, upon invocation, pushes/pops the
	 * values in the queue. The function object then passed to a new thread, then
	 * we wait for a condition meaning that the operation successfully executed.
	 * There is a timeout, so we can determine if the operation has been executed 
	 * properly or the thread is blocked on an operation (the queue is full, empty, 
	 * etc.)
	 * 
	 * Remark: we must push only non-negative integers as the negative ones have
	 * internal meanings. 
	 * 
	 * TODO: If a test failed due to a thread got blocked infinitely, the destructor 
	 * (and the whole test execution) will block. Should be optimized.
	 */ 
	class _TestContext {
	public:
		typedef int queue_e_t;
		typedef SynchronizedQueue<queue_e_t, boost::shared_ptr> queue_t;
		typedef boost::shared_ptr<queue_t> queue_p_t;

		_TestContext() : _q(new queue_t(_QUEUE_SIZE)) {};
		
		~_TestContext() {
			_thgrp.join_all();
		}
	
		/** Pop and return a value from the queue. There is a special value
		 * representing timeout. The return value must be checked with isTimeout
		 * to ensure that it represents a valid value */
		queue_e_t pop() {
			_PopFunc pf(_q);
			boost::function0<void> f = pf; 
			_thgrp.create_thread(f);
			// The wait returns false if returned tue to timeout
			_WAIT_TIME().actualize();
			return pf.wait(_WAIT_TIME().getXtime()) ? pf.getVal() : _TIMEOUT_VALUE();
		}
		
		/** Push a value to the queue. 
		 *  return: true: value is pushed, false: timeout occured */
		bool push(const queue_e_t& v) {
			assert(v >= 0);
			_PushFunc pf(_q, v);
			boost::function0<void> f = pf; 
			_thgrp.create_thread(f);
			_WAIT_TIME().actualize();
			return pf.wait(_WAIT_TIME().getXtime());
		}
		
		/** Check if an integer represents timeout */
		static bool isTimeout(const queue_e_t& v) {
			return v == _TIMEOUT_VALUE();
		}
		
		/** Yes, it is not a nice solution, but in this special case we need
		 * direct access to the queue, because we test its full, empty, etc. 
		 * methods as well */
		queue_t& q() {
			return *_q;
		}
		
	private:		
		/** 
		 * This is a code generator macro. It generates function objects to push
		 * and pop a value. Those function objects are executed in the threads.
		 * 
		 * Macro parameters:
		 * 
		 *   __NAME__: the class name
		 *   __CODE__: a piece of code that is executed first when the operator()
		 *             is called. Note that the macro encapsulates nothing, it
		 *             is used because the two wunction objects are really similar,
		 *             and their code cannot be reused with regular C++ constructs.
		 * 
		 * The functions store shared pointers to the queue and the valut that
		 * must be pushed / was popped. It is necessary as the those objects 
		 * must refer to the same entity after copying the functions several times by the
		 * threading subsystem, and they must ensure that there are no scope problems.
		 */
	    #define _FUNCTOR_GENERATOR(_NAME_,_CODE_) \
			class _NAME_ : public ThreadFunction { \
			public: \
				_NAME_(const queue_p_t& q, const queue_e_t& v = _INIT_VALUE()) \
					: ThreadFunction(), _q(q), _v(new queue_e_t(v)) {} \
				_NAME_(const _NAME_& x) : ThreadFunction(x), _q(x._q), _v(x._v) {} \
			\
			virtual void operator() () { \
				_CODE_; \
				_finished(); \
			} \
			\
			queue_e_t getVal() const { \
				return *_v; \
			} \
		\
		private: \
			queue_p_t _q; \
			boost::shared_ptr<queue_e_t> _v; \
		}

		/** Push a value to the queue. The value and the queue is set in the 
		 * constructor. */
		_FUNCTOR_GENERATOR(_PushFunc, _q->push(_v));
		/* Pop a value from the queue. The actual value van be obtained by 
		 * calling getVal() */
		_FUNCTOR_GENERATOR(_PopFunc, *_v = *(_q->pop()));
		#undef _FUNCTOR_GENERATOR

		queue_p_t _q;
		ThreadTraits::THREAD_GROUP _thgrp;
		
		static const size_t _QUEUE_SIZE = 2;
		// Timeout for the blocking call. 
		static Timeout& _WAIT_TIME()
		{
			static Timeout t(1, 0, 0);
			return t;
		}

		static const queue_e_t _INIT_VALUE() { return queue_e_t(-2); };
		static const queue_e_t _TIMEOUT_VALUE() { return queue_e_t(-1); };
	};
};

END_CORALSERVER_NAMESPACE

#endif /*CORALSERVER_SYNCHRONIZEDQUEUETEST_H_*/
