#ifndef CORALSERVERPROXY_PACKET_H
#define CORALSERVERPROXY_PACKET_H

//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: Packet.h,v 1.5 2008/08/28 00:29:02 salnikov Exp $
//
// Description:
//	Class Packet.
//
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <string>
#include <iosfwd>
#include <time.h>
#include <boost/shared_ptr.hpp>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "CoralProtocol/PacketHeader.h"
#include "CoralServerProxy/TimeStamp.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//		---------------------
// 		-- Class Interface --
//		---------------------

/**
 *  Data packet class. Consists of a header and data parts.
 *  Handles memory for the packet in an efficient way. 
 *  Knows how to (de)serialize itself from/to the socket.
 *
 *  @version $Id: Packet.h,v 1.5 2008/08/28 00:29:02 salnikov Exp $
 *
 *  @author Andy Salnikov
 */

namespace coral {
namespace CoralServerProxy {

class NetSocket ;

class Packet ;
typedef boost::shared_ptr<Packet> PacketPtr ;

class Packet  {
public:

  typedef coral::CoralProtocol::PacketHeader HeaderType ;
  
  enum PacketType { Request = 'Q', Reply = 'A', Control = 'C' } ;
  enum ControlOpcode { ServerShutdown, ClientDisconnect } ;
  enum RegularOpcode { ProxyControl = 0xFF } ;

  // constructor from the header and data pointer, data buffer ownership
  // is taken by this object
  Packet ( const coral::CoralProtocol::PacketHeader& header, PacketType type ) ;
  Packet ( const coral::CoralProtocol::PacketHeader& header, PacketType type, const TimeStamp& m_recvTime ) ;

  // Destructor
  ~Packet () ;

  // get packet type
  PacketType type() const { return m_type ; }
  
  // access header 
  CoralProtocol::PacketHeader& header() ;
  
  const CoralProtocol::PacketHeader& header() const ;
  
  // time when the packet was received
  const TimeStamp& recvTime() const { return m_recvTime ; }
  
  // returns pointer to packet data
  char* data() ;
  const char* data() const ;
  
  // get the size of the data besides header
  size_t dataSize() const ;

  // get the size of the whole packet
  size_t packetSize() const ;

  // compare two packets for the purpose of caching. they are equal iff:
  // 1. opcode is the same
  // 2. byte order is the same
  // 3. content of the first array in payload is the same for cacheable packets
  //    or whole payload is the same for non-cacheable packets
  bool operator == ( const Packet& other ) const ;
  
  // compare two packets for the purpose of sorting, 
  // only cache-related things are compared for cacheable packets
  bool operator < ( const Packet& other ) const ;
  
  // Build the packet from scratch.
  static PacketPtr build( PacketType type, 
                          CoralProtocol::Opcode opcode,
                          size_t clientId,
                          size_t requestId,
                          size_t seqNum,
                          size_t dataSize,
                          bool cacheable ) ;

  // Build the reply packet, take clientId, requestId, seqNum from request packet.
  static PacketPtr buildReply ( CoralProtocol::Opcode opcode,
                                const CoralProtocol::PacketHeader& reqHdr,
                                size_t dataSize,
                                bool cacheable ) ;

  // Read the packet from socket, previous content is discarded.
  // Returns number of bytes read, 0 on EOF, negative on error.
  static PacketPtr read( NetSocket& socket, PacketType type ) ;

  // Write the packet to socket.
  // Returns number of bytes written, 0 on EOF, negative on error.
  int write ( NetSocket& socket ) ;

  // Write the packet to socket, but use different header.
  // Returns number of bytes written, 0 on EOF, negative on error.
  int write ( NetSocket& socket, const coral::CoralProtocol::PacketHeader& header ) ;

protected:

private:

  // Data members
  PacketType m_type ;   // packet type
  TimeStamp m_recvTime ;
  char* m_wholePacket ;

};

std::ostream&
operator << ( std::ostream& o, const Packet& p) ;

} // namespace CoralServerProxy
} // namespace coral

#endif // CORALSERVERPROXY_PACKET_H
