//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: Packet.cpp,v 1.9 2008/11/06 22:01:08 salnikov Exp $
//
// Description:
//	Class Packet...
//
// Author List:
//      Andy Salnikov
//
//------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "CoralServerProxy/Packet.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <iostream>
#include <iomanip>
#include <sstream>
#include <memory>
#include <algorithm>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "CoralServerProxy/MsgLogger.h"
#include "CoralServerProxy/NetSocket.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

namespace coral {
namespace CoralServerProxy {

//----------------
// Constructors --
//----------------
Packet::Packet ( const coral::CoralProtocol::PacketHeader& header, PacketType type )
  : m_type(type)
  , m_recvTime()
  , m_wholePacket( new char[header.packetSize()] )
{
  // copy header data to the beginning of the buffer
  new(m_wholePacket) HeaderType( header ) ;
}

Packet::Packet ( const coral::CoralProtocol::PacketHeader& header, PacketType type, const TimeStamp& m_recvTime )
  : m_type(type)
  , m_recvTime(m_recvTime)
  , m_wholePacket( new char[header.packetSize()] )
{
  // copy header data to the beginning of the buffer
  new(m_wholePacket) HeaderType( header ) ;
}

//--------------
// Destructor --
//--------------
Packet::~Packet ()
{
    delete [] m_wholePacket ;
}
// access header
coral::CoralProtocol::PacketHeader&
Packet::header()
{
  return *(reinterpret_cast<HeaderType*>(m_wholePacket)) ;
}

const coral::CoralProtocol::PacketHeader&
Packet::header() const
{
  return *(reinterpret_cast<HeaderType*>(m_wholePacket)) ;
}

// returns pointer to packet data
char*
Packet::data()
{
  return m_wholePacket + sizeof(HeaderType) ;
}

const char*
Packet::data() const
{
  return m_wholePacket + sizeof(HeaderType) ;
}

// get the size of the data besides header
size_t
Packet::dataSize() const
{
  const HeaderType& h = header() ;
  return h.packetSize() - h.size() ;
}

// get the size of the whole packet
size_t
Packet::packetSize() const
{
  return header().packetSize() ;
}

// compare two packets for the purpose of caching. they are equal iff:
// 1. opcode is the same
// 2. byte order is the same
// 3. content of the first array in payload is the same for cacheable packets
//    or whole payload is the same for non-cacheable packets
bool
Packet::operator == ( const Packet& other ) const
{
  // different opcodes means different
  if ( header().opcode() != other.header().opcode() ) return false ;

  // different endiannes means different
  if ( header().endian() != other.header().endian() ) return false ;

  // payload size
  size_t mySize = dataSize() ;
  size_t otherSize = other.dataSize() ;
  const char* myData = data() ;
  const char* otherData = other.data() ;

  if ( header().isCacheable() and other.header().isCacheable() ) {

    // both are cacheable

    // say that empty cacheable packets are equal
    if ( mySize == 0 and otherSize == 0 ) return true ;

    // minimum array size is 4 bytes,
    if ( mySize < 4 or otherSize < 4 ) return false ;

    uint32_t myArraySize, otherArraySize ;
    std::copy ( myData, myData+4, (char*)(&myArraySize) ) ;
    std::copy ( otherData, otherData+4, (char*)(&otherArraySize) ) ;
    if ( header().endian() != CoralProtocol::_system_endian ) {
      // 4-byte swap
      myArraySize = CoralProtocol::Utils::swap_word ( myArraySize ) ;
      otherArraySize = CoralProtocol::Utils::swap_word ( otherArraySize ) ;
    }

    // size must be equal
    if ( myArraySize != otherArraySize ) return false ;

    // compare data
    return std::equal ( myData+4, myData+myArraySize+4, otherData+4 ) ;

  } else {

    // for non-cacheable packets compare whole payload

    // size must be equal
    if ( mySize != otherSize ) return false ;

    // compare data
    return std::equal ( myData, myData+mySize, otherData ) ;


  }
}

// compare two packets for the purpose of sorting,
// only cache-related things are compared for cacheable packets
bool
Packet::operator < ( const Packet& other ) const
{
  const coral::CoralProtocol::PacketHeader& myHeader = header() ;
  const coral::CoralProtocol::PacketHeader& otherHeader = other.header() ;
  // opcodes first
  if ( myHeader.opcode() < otherHeader.opcode() ) return true ;
  if ( myHeader.opcode() > otherHeader.opcode() ) return false ;

  // different endiannes means different
  if ( myHeader.endian() == CoralProtocol::e_little and otherHeader.endian() == CoralProtocol::e_big ) return true ;
  if ( myHeader.endian() == CoralProtocol::e_big and otherHeader.endian() == CoralProtocol::e_little ) return false ;

  // payload size
  size_t mySize = dataSize() ;
  size_t otherSize = other.dataSize() ;
  const char* myData = data() ;
  const char* otherData = other.data() ;


  if ( header().isCacheable() and other.header().isCacheable() ) {

    // both are cacheable

    // minimum array size is 4 bytes, for odd-size packets compare whole payload
    if ( mySize < 4 or otherSize < 4 ) {
      if ( mySize < otherSize ) return true ;
      if ( mySize > otherSize ) return false ;
      return std::equal ( myData, myData+mySize, otherData ) ;
    }

    uint32_t myArraySize, otherArraySize ;
    std::copy ( myData, myData+4, (char*)(&myArraySize) ) ;
    std::copy ( otherData, otherData+4, (char*)(&otherArraySize) ) ;
    if ( header().endian() != CoralProtocol::_system_endian ) {
      // 4-byte swap
      myArraySize = CoralProtocol::Utils::swap_word ( myArraySize ) ;
      otherArraySize = CoralProtocol::Utils::swap_word ( otherArraySize ) ;
    }

    // size must be equal
    if ( myArraySize < otherArraySize ) return true ;
    if ( myArraySize > otherArraySize ) return false ;

    // compare data
    return std::lexicographical_compare( myData+4, myData+myArraySize+4, otherData+4, otherData+otherArraySize+4 ) ;

  } else {

    // shorter packet compare less than longer
    if ( mySize < otherSize ) return true ;
    if ( mySize > otherSize ) return false ;

    // compare data
    return std::lexicographical_compare( myData, myData+mySize, otherData, otherData+otherSize ) ;
  }
}

// Build the packet from scratch.
PacketPtr
Packet::build ( PacketType type,
                CoralProtocol::Opcode opcode,
                size_t clientId,
                size_t requestId,
                size_t seqNum,
                size_t dataSize,
                bool cacheable )
{
  // make a header
  CoralProtocol::PacketHeader header (CoralProtocol::_system_endian) ;
  header.setOpcode ( opcode ) ;
  header.setClientID ( clientId ) ;
  header.setRequestID ( requestId ) ;
  header.setSequenceNum ( CoralProtocol::SequenceNumber(seqNum) ) ;
  header.setPacketSize( dataSize ) ;
  header.setCacheable( cacheable ) ;

  // make a packet
  return PacketPtr ( new Packet ( header, type ) ) ;
}

// Build the reply packet, take clientId, requestId, seqNum from request packet.
PacketPtr
Packet::buildReply ( CoralProtocol::Opcode opcode,
                     const CoralProtocol::PacketHeader& reqHdr,
                     size_t dataSize,
                     bool cacheable )
{
  // make a header
  CoralProtocol::PacketHeader header (CoralProtocol::_system_endian) ;
  header.setOpcode ( opcode ) ;
  header.setClientID ( reqHdr.clientID() ) ;
  header.setRequestID ( reqHdr.requestID() ) ;
  header.setSequenceNum ( reqHdr.sequenceNum() ) ;
  header.setPacketSize( dataSize ) ;
  header.setCacheable( cacheable ) ;

  // make a packet
  return PacketPtr ( new Packet ( header, Packet::Reply ) ) ;
}

// Read the packet from socket, previous content is discarded.
// Returns number of bytes read, 0 on EOF, negative on error.
boost::shared_ptr<Packet>
Packet::read( NetSocket& socket, PacketType type )
{
  const size_t headerSize = sizeof(HeaderType) ;
  char buf[sizeof(HeaderType)] ;

  // get the header first
  int s = socket.read ( buf, headerSize ) ;
  if ( s < 0 ) {
    PXY_ERR ( "Packet::read - failed to read packet header: " << strerror(errno) ) ;
    return PacketPtr() ;
  } else if ( s != int(headerSize) ) {
    PXY_INFO ( "Packet::read - connection closed while reading header: " << strerror(errno) ) ;
    return PacketPtr() ;
  }

  // get the header
  std::auto_ptr<HeaderType> header ( HeaderType::create ( std::string(buf,headerSize) ) ) ;
  if ( not header.get() ) {
    std::ostringstream o ;
    for ( unsigned int i = 0 ; i < headerSize ; ++ i ) {
      o << "\\x" << std::hex << int(buf[i]) ;
    }
    PXY_ERR ( "Packet::read - failed to interpret packet header: " << o.str() ) ;
    return PacketPtr() ;
  }

  // get the length of the remainig stuff
  //PXY_DEBUG ( "Packet::read - packet size: " << header->packetSize() ) ;

  // allocate buffer for remaining packet data
  Packet* p = 0 ;
  try {
    p = new Packet ( *header, type, TimeStamp::now() ) ;
  } catch ( std::exception& e ) {
    PXY_ERR ( "Packet::read - memory allocation failed: packet size = " << header->packetSize() ) ;
    return PacketPtr() ;
  }

  // read this stuff
  s = socket.read ( p->data(), p->dataSize() ) ;
  if ( s < 0 ) {
    PXY_ERR ( "Packet::read - failed to read packet data: " << strerror(errno) ) ;
    delete p ;
    return PacketPtr() ;
  } else if ( s != int(p->dataSize()) ) {
    PXY_INFO ( "Packet::read - connection closed while reading packet data: " << strerror(errno) ) ;
    delete p ;
    return PacketPtr() ;
  }

  return PacketPtr(p) ;
}

// Write the packet to socket.
// Returns number of bytes written, 0 on EOF, negative on error.
int
Packet::write ( NetSocket& socket )
{
  // calculate whole packet size
  size_t fullsize =  packetSize() ;

  // send it down the stream
  int s = socket.write ( m_wholePacket, fullsize ) ;
  //PXY_DEBUG ( "Packet::write: fullsize=" << fullsize << ", s=" << s ) ;
  if ( s > 0 and s < int(fullsize) ) {
    // seems like close happened during transmission
    s = 0 ;
  }

  return s ;
}

// Write the packet to socket, but use different client ID.
// Returns number of bytes written, 0 on EOF, negative on error.
int
Packet::write ( NetSocket& socket, const coral::CoralProtocol::PacketHeader& header )
{
  const size_t datasize = this->dataSize() ;

  // optimization, packets shorter than 16k are sent in one packet
  const unsigned oneBufSize = 1024*16 ;
  if ( datasize <= oneBufSize ) {

    // make buffer big enough
    unsigned char buf[datasize+sizeof(HeaderType)] ;

    // copy the header
    HeaderType* h = new(buf) HeaderType(header) ;
    h->setPacketSize ( dataSize() ) ;

    // copy the payload
    const char* d = data() ;
    std::copy ( d, d+datasize, buf+sizeof(HeaderType) ) ;

    // send it down the stream
    size_t fullsize =  datasize + sizeof(HeaderType) ;
    int s = socket.write ( buf, fullsize ) ;
    //PXY_DEBUG ( "Packet::write: fullsize=" << fullsize << ", s=" << s ) ;
    if ( s > 0 and s < int(fullsize) ) {
      // seems like close happened during transmission
      s = 0 ;
    }

    return s ;

  } else {

    // make sure that data size in header corresponds to real data
    HeaderType h = header ;
    h.setPacketSize ( dataSize() ) ;

    // send the header
    int s = socket.write ( &h, sizeof(HeaderType) ) ;
    //PXY_DEBUG ( "Packet::write: header=" << sizeof(HeaderType) << ", s=" << s ) ;
    if ( s < 0 ) {
      // error happened
      return s ;
    } else if ( s >= 0 and s < int(sizeof(HeaderType)) ) {
      // seems like close happened during transmission
      return 0 ;
    }

    // send data part
    if ( datasize ) {
      int s2 = socket.write ( data(), datasize ) ;
      //PXY_DEBUG ( "Packet::write: dataSize=" << dataSize << ", s2=" << s2 ) ;
      if ( s2 < 0 ) {
        // error happened
        return s2 ;
      } if ( s2 > 0 and s2 < int(datasize) ) {
        // seems like close happened during transmittion
        return 0 ;
      }
      s += s2 ;
    }

    return s ;

  }
}

std::ostream&
operator << ( std::ostream& o, const Packet& p)
{
  const coral::CoralProtocol::PacketHeader& h = p.header() ;
  o << "[Packet type=" << char(p.type()) << " op=" << int(h.opcode())
    << "/" << coral::CoralProtocol::Opcode::name(int(h.opcode()))
    << " size=" << p.dataSize()
    << " ID=<" << (uint32_t)( h.clientID() )
    << ":" << (uint32_t)( h.requestID() )
    << "#" << (uint32_t)( h.sequenceNum().value() & 0x7FFFFF )
    << ( ( h.sequenceNum().value() & 0x800000 ) ? "+" : "" ) << '>' ;
  if ( const_cast<coral::CoralProtocol::PacketHeader&>(h).isCacheable() ) o << " cache" ;

  o << " data=" ;
  o.fill('0') ;
  o.setf ( std::ios::hex, std::ios::basefield ) ;
  const char* b = p.data() ;
  const char* e = b + p.dataSize() ;
  for ( ; b != e ; ++ b ) {
    if ( isprint(*b) ) {
      o << *b ;
    } else {
      o << "\\x" << std::setw(2) << (int(*b) & 0xFF) ;
    }
  }
  o.setf ( std::ios::dec, std::ios::basefield ) ;
  o.fill(' ') ;

  return o << ']' ;
}

} // namespace CoralServerProxy
} // namespace coral
