//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: PacketDispatcher.cpp,v 1.12 2008/11/06 22:01:08 salnikov Exp $
//
// Description:
//	Class PacketDispatcher...
//
// Author List:
//      Andy Salnikov
//
//------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "CoralServerProxy/PacketDispatcher.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <sstream>
#include <algorithm>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "CoralServerProxy/IPacketCache.h"
#include "CoralServerProxy/MsgLogger.h"
#include "CoralServerProxy/Packet.h"
#include "CoralServerProxy/PacketHeaderQueue.h"
#include "CoralServerProxy/PacketQueue.h"
#include "CoralServerProxy/StatCollector.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

namespace {

  bool str2uint ( const std::string& str, unsigned& u )
  {
    char *eptr = 0 ;
    u = strtoul ( str.c_str(), &eptr, 10 ) ;
    return *eptr == '\0' ;
  }


}


//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

namespace coral {
namespace CoralServerProxy {

//----------------
// Constructors --
//----------------
PacketDispatcher::PacketDispatcher ( PacketQueue& inputq,
                                     PacketQueue& serverq,
                                     PacketHeaderQueue& clientq,
                                     IPacketCache& cache,
                                     StatCollector* statCollector )
  : m_inputq ( inputq )
  , m_serverq ( serverq )
  , m_clientq ( clientq )
  , m_queuedPackets()
  , m_multipartCollector()
  , m_cache ( cache )
  , m_statCollector ( statCollector )
{
}

//--------------
// Destructor --
//--------------
PacketDispatcher::~PacketDispatcher ()
{
}

// this is the "run" method used by the Boost.thread
void
PacketDispatcher::operator() ()
{
  // wait for the next packet to come from the queue
  while ( true ) {

    PacketPtr p = m_inputq.pop() ;
    PXY_DEBUG ( "PacketDispatcher: received new packet " << *p ) ;

    // dispatch it to particular processor
    if ( p->type() == Packet::Request ) {
      processRequest ( p ) ;
    } else if ( p->type() == Packet::Reply ) {
      processReply ( p ) ;
    } else if ( p->type() == Packet::Control ) {
      processControl ( p ) ;
    } else {
      // unexpected something, just get rid of it
      PXY_ERR ( "PacketDispatcher: unknown packet type: " << p->type() ) ;
    }

  }
}

void
PacketDispatcher::processRequest ( PacketPtr p )
{
  const CoralProtocol::PacketHeader& header = p->header() ;

  // handle special proxy-control opcode
  if ( header.opcode() == Packet::ProxyControl ) {
    processProxyControl ( p ) ;
    return ;
  }

  // gather a bit of statistics about the packets
  if ( m_statCollector ) {
    m_statCollector->addClientPacket ( p ) ;
  }

  // the transaction requests and CloseSession requests do not pass through,
  // we return OK for read-only transactions, Exception for update transactions
  CoralProtocol::Opcode opcode = header.opcode() ;
  if ( opcode == CoralProtocol::Opcode::E_CloseSession ) {

    PXY_DEBUG ( "PacketDispatcher: received CloseSession" ) ;

    // send OK packet, do not forward it to server
    uint32_t strSize = 0 ;
    PacketPtr p = Packet::buildReply ( CoralProtocol::Opcode::E_OK, header, 4+strSize, false ) ;
    memcpy ( p->data(), &strSize, 4 ) ;

    m_clientq.push ( p, p->header() ) ;

    return ;
  }

  if ( opcode == CoralProtocol::Opcode::E_StartTransaction or
       opcode == CoralProtocol::Opcode::E_CommitTransaction ) {

    PXY_DEBUG ( "PacketDispatcher: received transaction start/commit" ) ;

    // check the packet structure and transaction mode
    const char* err = 0 ;
    if ( opcode == CoralProtocol::Opcode::E_StartTransaction ) {
      // check the size of the packet, must be 5
      if ( header.payloadSize() != 5 ) {
        PXY_DEBUG ( "PacketDispatcher: Unexpected transaction packet size: " << header.payloadSize() ) ;
        err = "Unexpected packet size" ;
      } else {
        uint8_t mode = p->data()[4] ;
        if ( mode == 0 ) {
          PXY_DEBUG ( "PacketDispatcher: Update transactions are not supported with proxy" ) ;
          err = "Update transactions are not supported with proxy" ;
        } else if ( mode > 1 ) {
          PXY_DEBUG ( "PacketDispatcher: Unexpected transaction mode: " << mode ) ;
          err = "Unexpected transaction mode (>1)" ;
        }
      }
    }

    if ( err ) {

      // send error packet
      uint32_t strSize = strlen ( err ) ;

      PacketPtr p = Packet::buildReply ( CoralProtocol::Opcode::E_CTL, header, 1 + 4+strSize, false ) ;
      char* data = p->data() ;
      *data ++ = 0x7 ;
      memcpy ( data, &strSize, 4 ) ;
      data += 4 ;
      memcpy ( data, err, strSize ) ;

      m_clientq.push ( p, p->header() ) ;

    } else {

      // send OK packet
      uint32_t strSize = 0 ;

      PacketPtr p = Packet::buildReply ( CoralProtocol::Opcode::E_OK, header, 4+strSize, false ) ;
      memcpy ( p->data(), &strSize, 4 ) ;

      m_clientq.push ( p, p->header() ) ;

    }

    return ;
  }

  // analyze packet header
  if ( header.sequenceNum().value() != 0 ) {

    // multipart requests are not handled now, just send it upstream
    PXY_DEBUG ( "PacketDispatcher: multipart request, moving it to server queue" ) ;
    m_serverq.push ( p ) ;

  } else if ( header.isCacheable() ) {

    // packet (or rather response to it) can be cached

    const IPacketCache::PacketPtrList* r = m_cache.find( p ) ;
    if ( r ) {

      // cached already, just send it back to the client
      PXY_DEBUG ( "PacketDispatcher: reply found in cache" ) ;
      for ( IPacketCache::PacketPtrList::const_iterator it = r->begin() ; it != r->end() ; ++ it ) {
        m_clientq.push ( *it, header ) ;
        if ( m_statCollector ) {
          // gather a bit of statistics about the packets
          m_statCollector->addServerPacket ( *it, header, true ) ;
        }
      }

    } else {

      // It's not in the cache yet, try to see if the identical request was
      // already sent to the server
      PacketPtr firstRequest = m_queuedPackets.findIdenticalRequest ( p ) ;
      if ( firstRequest ) {

        // Not in cache yet, but identical request is already processing.
        PXY_DEBUG ( "PacketDispatcher: identical packet is queued already" ) ;
        m_queuedPackets.add ( p ) ;

        // See if there are any parts of the multi-part reply arrived
        MultipartCollector::iterator it = m_multipartCollector.find ( firstRequest->header() ) ;
        if ( it != m_multipartCollector.end() ) {

          PXY_DEBUG ( "PacketDispatcher: partial reply received from server" ) ;

          // send all already received parts to client parts
          const MultipartCollector::PacketPtrList& replies = it->second ;
          for ( MultipartCollector::PacketPtrList::const_iterator it = replies.begin() ; it != replies.end() ; ++ it ) {
            m_clientq.push ( *it, header ) ;
            if ( m_statCollector ) {
              // gather a bit of statistics about the packets
              m_statCollector->addServerPacket ( *it, header, true ) ;
            }
          }

        }

      } else {

        // no trace yet of the similar packets
        PXY_DEBUG ( "PacketDispatcher: packet not in cache or queue" ) ;
        m_queuedPackets.add ( p ) ;
        m_serverq.push ( p ) ;

      }

    }


  } else {

    // non-cacheable packet, just send it upstream
    PXY_DEBUG ( "PacketDispatcher: packet non-cacheable, moving it to server queue" ) ;
    m_serverq.push ( p ) ;

  }
}

void
PacketDispatcher::processReply ( PacketPtr p )
{
  // See if the request for this reply is in the queued packet list.
  // Queued packet list only contains packets that client wants to be
  // cached. Server may override the cacheable flag of the packets,
  // so even if the server reply is not cacheable we still have to
  // check client requests queue

  if ( m_statCollector ) {
    // gather a bit of statistics about the packets
    m_statCollector->addServerPacket ( p, p->header(), false ) ;
  }

  PacketPtr req = m_queuedPackets.findRequestForReply ( p->header() ) ;
  if ( not req ) {

    // nobody expects us here, the request may have been non-cacheable, or maybe multi-part
    PXY_DEBUG ( "PacketDispatcher: no request in the queue, send reply to client" ) ;

    // we don't have to be cached, just send me to the client
    m_clientq.push ( p, p->header() ) ;

  } else {

    // if this is the last packet in the sequence then we'll clear the queue
    // from the requests awaiting it
    bool remove = not ( p->header().sequenceNum().value() & 0x800000 ) ;

    // get the list of all requests for whom I'm a good reply
    std::list<PacketPtr> requests ;
    m_queuedPackets.getRequests ( req, requests, remove ) ;
    PXY_DEBUG ( "PacketDispatcher: got " << requests.size() << " requests from queued packets list" ) ;

    // send this reply to every request
    for ( std::list<PacketPtr>::const_iterator it = requests.begin() ; it != requests.end() ; ++ it ) {
      PXY_DEBUG ( "PacketDispatcher: send reply to client, original request: " << *(*it) ) ;
      m_clientq.push ( p, (*it)->header() ) ;
    }

    // now if the reply is multi-part add it to the parts collector
    if ( p->header().sequenceNum().value() ) {

      //
      std::pair<MultipartCollector::iterator,bool> res = m_multipartCollector.addPacket ( p ) ;
      if ( not res.second ) {

        PXY_ERR ( "PacketDispatcher: failed to add reply part to collector: " << *p ) ;

      } else {

        PXY_DEBUG ( "PacketDispatcher: added reply part to collector: " << *p ) ;

        // if we got the last part them move result to the cache
        if ( remove ) {

          // copy to cache
          const MultipartCollector::PacketPtrList& parts = res.first->second ;
          m_cache.insert ( req, parts ) ;

          // and remove it from collector
          m_multipartCollector.erase( res.first ) ;

        }

      }

    } else {

      // for single-part packet if the reply is cacheable send it to the cache
      if ( p->header().isCacheable() ) {
        PXY_DEBUG ( "PacketDispatcher: insert packet into cache: " << *req << " -> " << *p ) ;
        IPacketCache::PacketPtrList replies ( 1, p ) ;
        m_cache.insert ( req, replies ) ;
      }

    }
  }

}

void
PacketDispatcher::processControl ( PacketPtr p )
{
  unsigned opcode = p->header().opcode() ;

  if ( opcode == Packet::ServerShutdown ) {

    // server closed connection, will shutdown all clients too
    PXY_INFO ( "PacketDispatcher : got shutdown packet, clear queued packets list" ) ;
    m_queuedPackets.clear() ;
    m_multipartCollector.clear() ;

    // we also want to make sure that on the next connection to a server
    // a connect packet is sent to the server, so we have to remove all
    // connect packets from a cache
    m_cache.clearConnect() ;

  } else if ( opcode == Packet::ClientDisconnect ) {

    // remove all memory about queued packets coming from this client
    m_queuedPackets.deleteClientId ( p->header().clientID() ) ;

  }

  // send it to both client and server queues
  m_serverq.push ( p ) ;
  m_clientq.push ( p, p->header() ) ;
}

void
PacketDispatcher::processProxyControl( PacketPtr p )
{
  std::string cmd ( std::string ( p->data(), p->dataSize() ) ) ;

  PXY_DEBUG ( "PacketDispatcher - got ProxyControl command: \"" << cmd << "\"" ) ;

  if ( cmd.compare( 0, 5, "stat " ) == 0 && not m_statCollector ) {
    // send reply packet
    sendReply ( "FAIL: statistics is not collected", p->header() ) ;
    return ;
  }


  if ( cmd == "stat clear" ) {

    // reset statistics
    m_statCollector->clear() ;

    // send reply packet
    sendReply ( "OK", p->header() ) ;

  } else if ( cmd.compare( 0, 14, "stat set-size " ) == 0 ) {

    // get the number for the new stat size
    unsigned s ;
    if ( ::str2uint ( std::string(cmd,14), s ) ) {

      m_statCollector->setStatSize ( s ) ;
      // send reply packet
      sendReply ( "OK", p->header() ) ;

    } else {

      PXY_ERR ( "PacketDispatcher - invalid number in stat-size: \"" << cmd << "\"" ) ;
      // send reply packet
      sendReply ( "FAIL: invalid number", p->header() ) ;

    }

  } else if ( cmd.compare( 0, 20, "stat auto-reset-sec " ) == 0 ) {

    // get the number for the new auto-reset interval
    unsigned s ;
    if ( ::str2uint ( std::string(cmd,20), s ) ) {

      m_statCollector->setAutoResetSec ( s ) ;
      // send reply packet
      sendReply ( "OK", p->header() ) ;

    } else {

      PXY_ERR ( "PacketDispatcher - invalid number in auto-reset-sec: \"" << cmd << "\"" ) ;

      // send reply packet
      sendReply ( "FAIL: invalid number", p->header() ) ;

    }

  } else if ( cmd == "stat report" ) {

    // dump the stat
    std::ostringstream stream ;
    m_statCollector->printJSON ( stream ) ;
    const std::string res = stream.str() ;

    // send reply packet
    sendReply ( stream.str(), p->header() ) ;

  } else {

    PXY_ERR ( "PacketDispatcher - invalid command: \"" << cmd << "\"" ) ;

    // send reply packet
    sendReply ( "invalid command", p->header() ) ;
  }

}

void
PacketDispatcher::sendReply( const std::string& res, const coral::CoralProtocol::PacketHeader& reqHeader )
{
  coral::CoralProtocol::PacketHeader header ( reqHeader ) ;

  header.setPacketSize ( res.size() ) ;
  PacketPtr reply = Packet::buildReply ( header.opcode(), header, res.size(), false ) ;
  std::copy ( res.begin(), res.end(), reply->data() ) ;

  // send the reply to the client queue
  m_clientq.push ( reply, reply->header() ) ;
}

} // namespace CoralServerProxy
} // namespace coral
