// $Id: Schema.cpp,v 1.9 2008/04/03 21:42:02 rado Exp $
#include "Schema.h"
#include "ISessionProperties.h"
#include "DomainProperties.h"
#include "Table.h"
#include "Statement.h"
#include "ErrorHandler.h"
#include "OracleTableBuilder.h"
#include "Query.h"
#include "ViewFactory.h"
#include "View.h"

#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/ITableSchemaEditor.h"
#include "RelationalAccess/RelationalServiceException.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"

#include "CoralKernel/Service.h"

coral::FrontierAccess::Schema::Schema( const coral::FrontierAccess::ISessionProperties& sessionProperties )
  : m_sessionProperties( sessionProperties ), m_tables(), m_tablesReadFromDataDictionary( false ), m_views(), m_viewsReadFromDataDictionary( false )
{
}

coral::FrontierAccess::Schema::~Schema()
{
  this->reactOnEndOfTransaction();
}

std::string coral::FrontierAccess::Schema::schemaName() const
{
  return m_sessionProperties.schemaName();
}

std::set<std::string> coral::FrontierAccess::Schema::listTables() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::listTables" );
  
  std::set<std::string> result;

  if ( ! m_tablesReadFromDataDictionary )
    this->readTablesFromDataDictionary();

  for ( std::map< std::string, coral::FrontierAccess::Table*>::const_iterator iTable = m_tables.begin(); iTable != m_tables.end(); ++iTable )
  {
    result.insert( iTable->first );
  }

  return result;
}

bool coral::FrontierAccess::Schema::existsTable( const std::string& tableName ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::existsTable" );

  // Check if the table is already in the local cache.
  if ( m_tables.find( tableName ) == m_tables.end() )
  {
    // Check if the data dictionary was read
    if ( m_tablesReadFromDataDictionary )
    {
      return false;
    }
    else
    {
      // Check inside the database.
      bool status = false;

      coral::FrontierAccess::Statement statement( m_sessionProperties, "SELECT TABLE_NAME FROM ALL_TABLES WHERE OWNER=:\"owner\" AND TABLE_NAME=:\"table\"" );
      
      statement.setNumberOfPrefetchedRows( 100 );
      coral::AttributeList bindData;
      bindData.extend<std::string>( "owner" );
      bindData.extend<std::string>( "table" );
      bindData["owner"].data<std::string>() = m_sessionProperties.schemaName();
      bindData["table"].data<std::string>() = tableName;
      statement.execute( bindData, true /*force reload*/ );
      coral::AttributeList output;
      output.extend<std::string>( "TABLE_NAME" );
      const std::string& tName = output.begin()->data<std::string>();
      statement.defineOutput( output );
      while( statement.fetchNext() )
      {
        if( tName == tableName )
        {
          status = true;
          break;
        }
      }

      if ( status == false )
      {
        return false;
      }
      else {
        // The table exists. Create the table object and insert it into the map
        coral::FrontierAccess::Table* table = new coral::FrontierAccess::Table( m_sessionProperties, tableName );
        m_tables.insert( std::make_pair( tableName, table ) );
        return true;
      }
    }
  }

  return true;
}

void coral::FrontierAccess::Schema::dropTable( const std::string& /*tableName*/ )
{
  throw coral::InvalidOperationException( m_sessionProperties.domainProperties().service()->name(),
                                          "Frontier does not support DDL operations", "coral::FrontierAccess::Schema::dropTable" );
}

void coral::FrontierAccess::Schema::dropIfExistsTable( const std::string& /*tableName*/ )
{
  throw coral::InvalidOperationException( m_sessionProperties.domainProperties().service()->name(),
                                          "Frontier does not support DDL operations", "coral::FrontierAccess::Schema::dropIfExistsTable" );
}

coral::ITable& coral::FrontierAccess::Schema::createTable( const coral::ITableDescription& /*description*/ )
{
  throw coral::InvalidOperationException( m_sessionProperties.domainProperties().service()->name(),
                                          "Frontier does not support DDL operations", "coral::FrontierAccess::Schema::createTable" );
  coral::ITable* table = 0;
  return *table;
}

coral::ITable& coral::FrontierAccess::Schema::tableHandle( const std::string& tableName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "coral::FrontierAccess::Schema::tableHandle" );
  
  // The fully qualified table name
  std::string fullTableName = m_sessionProperties.schemaName() + ".\"" + tableName + "\"";

  if( ! this->existsTable( tableName ) )
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(), fullTableName );
  
  // The table exists and is in the cache
  std::map< std::string, Table* >::iterator iTable = m_tables.find( tableName );
  
  // Paranoia check
  if( iTable == m_tables.end() )
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(), fullTableName );

  if( iTable->second == 0 )
  {
    coral::FrontierAccess::Table* table = new coral::FrontierAccess::Table( m_sessionProperties, tableName );
    iTable->second = table;
  }
    
  return *( iTable->second );
}
    
void coral::FrontierAccess::Schema::truncateTable( const std::string& )
{
  throw coral::InvalidOperationException( m_sessionProperties.domainProperties().service()->name(),
                                          "Frontier does not support DDL operations", "coral::FrontierAccess::Schema::truncateTable" );
}

void coral::FrontierAccess::Schema::callProcedure( const std::string& , const coral::AttributeList& )
{
    throw coral::InvalidOperationException( m_sessionProperties.domainProperties().service()->name(),
                                            "Frontier does not support DDL operations", "coral::FrontierAccess::Schema::callProcedure" );
}

coral::IQuery* coral::FrontierAccess::Schema::newQuery() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "coral::FrontierAccess::Schema::newQuery" );
  return new coral::FrontierAccess::Query( m_sessionProperties );
}

coral::IViewFactory* coral::FrontierAccess::Schema::viewFactory()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "coral::FrontierAccess::Schema::viewFactory" );
  return new coral::FrontierAccess::ViewFactory( m_sessionProperties );
}

bool coral::FrontierAccess::Schema::existsView( const std::string& viewName ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "coral::FrontierAccess::Schema::existsView" );

  // Check if the view is already in the local cache.
  if ( m_views.find( viewName ) == m_views.end() )
  {
    // Check if the data dictionary was read
    if ( m_viewsReadFromDataDictionary )
    {
      return false;
    }
    else
    {
      // Check inside the database.

      // The fully qualified view name
      std::string fullViewName = m_sessionProperties.schemaName() + ".\"" + viewName + "\"";

      coral::FrontierAccess::Statement
        statement( m_sessionProperties,
                   "SELECT AO.OBJECT_NAME FROM ALL_OBJECTS AO, ALL_VIEWS AV WHERE AO.OWNER=:\"owner\" AND AO.OBJECT_NAME=:\"view\" AND AO.OBJECT_NAME = AV.VIEW_NAME" );

      bool status = false;
      statement.setNumberOfPrefetchedRows( 100 );
      coral::AttributeList bindData;
      bindData.extend<std::string>( "owner" );
      bindData.extend<std::string>( "view" );
      bindData["owner"].data<std::string>() = m_sessionProperties.schemaName();
      bindData["view"].data<std::string>() = viewName;
      statement.execute( bindData, true /*force reload*/ );
      coral::AttributeList output;
      output.extend<std::string>( "OBJECT_NAME" );
      const std::string& vName = output.begin()->data<std::string>();
      statement.defineOutput( output );
      while( statement.fetchNext() )
      {
        if( vName == viewName )
        {
          status = true;
          break;
        }
      }

      if ( status == false )
      {
        return false;
      }
      else {
        // The view exists. Create the view object and insert it into the map
        coral::FrontierAccess::View* view = new coral::FrontierAccess::View( m_sessionProperties, viewName );
        m_views.insert( std::make_pair( viewName, view ) );
        return true;
      }
    }
  }

  return true;
}

void coral::FrontierAccess::Schema::dropView( const std::string& /* viewName */ )
{
  // FIXME - throw invalid operation exception
}

void coral::FrontierAccess::Schema::dropIfExistsView( const std::string& /* viewName */)
{
  // FIXME - throw invalid operation exception
}

coral::IView& coral::FrontierAccess::Schema::viewHandle( const std::string& viewName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::viewHandle" );

  // Check if the view is already in the local cache.
  std::map< std::string, coral::FrontierAccess::View* >::iterator iView = m_views.find( viewName );

  // The fully qualified view name
  const std::string fullViewName = m_sessionProperties.schemaName() + ".\"" + viewName + "\"";
  
  if ( iView == m_views.end() )
  {
    // Check if the data dictionary was read
    if ( m_viewsReadFromDataDictionary )
      throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(), fullViewName );


// FIXME - getting view description
    // Try to get the descriptor handle
//     // The view exists. Create the view object and insert it into the map
//     coral::FrontierAccess::View* view = new coral::FrontierAccess::View( m_sessionProperties, viewName, describeHandle );
//     iView = m_views.insert( std::make_pair( viewName, view ) ).first;
  }
  else
  {
    throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(),fullViewName );
  }

  return *( iView->second );
}

std::set<std::string> coral::FrontierAccess::Schema::listViews() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::listViews" );
  std::set<std::string> result;
  if ( ! m_viewsReadFromDataDictionary )
    this->readViewsFromDataDictionary();

  for ( std::map< std::string, coral::FrontierAccess::View*>::const_iterator iView = m_views.begin(); iView != m_views.end(); ++iView )
  {
    result.insert( iView->first );
  }

  return result;
}

void coral::FrontierAccess::Schema::reactOnEndOfTransaction()
{
  for ( std::map< std::string, coral::FrontierAccess::Table* >::iterator iTable = m_tables.begin(); iTable != m_tables.end(); ++iTable )
    if ( iTable->second )
      delete iTable->second;

  m_tables.clear();

  m_tablesReadFromDataDictionary = false;

  for ( std::map< std::string, coral::FrontierAccess::View* >::iterator iView = m_views.begin(); iView != m_views.end(); ++iView )
    if ( iView->second )
      delete iView->second;

  m_views.clear();

  m_viewsReadFromDataDictionary = false;
}

void coral::FrontierAccess::Schema::readTablesFromDataDictionary() const
{
  coral::FrontierAccess::Statement statement( m_sessionProperties, "SELECT TABLE_NAME FROM ALL_TABLES WHERE OWNER=:\"owner\"" );
  statement.setNumberOfPrefetchedRows( 100 );
  coral::AttributeList bindData;
  bindData.extend<std::string>( "owner" );
  bindData.begin()->data<std::string>() = m_sessionProperties.schemaName();
  statement.execute( bindData, true /*force reload*/ );
  coral::AttributeList output;
  output.extend<std::string>( "TABLE_NAME" );
  const std::string& tableName = output.begin()->data<std::string>();
  statement.defineOutput( output );
  while( statement.fetchNext() )
  {
    if ( m_tables.find( tableName ) == m_tables.end() )
    {
      m_tables[tableName] = 0;
    }
  }
  m_tablesReadFromDataDictionary = true;
}

void coral::FrontierAccess::Schema::readViewsFromDataDictionary() const
{
  coral::FrontierAccess::Statement statement( m_sessionProperties, "SELECT VIEW_NAME FROM ALL_VIEWS WHERE OWNER=:\"owner\"" );
  statement.setNumberOfPrefetchedRows( 100 );
  coral::AttributeList bindData;
  bindData.extend<std::string>( "owner" );
  bindData.begin()->data<std::string>() = m_sessionProperties.schemaName();
  statement.execute( bindData, true /*force reload*/ );
  coral::AttributeList output;
  output.extend<std::string>( "VIEW_NAME" );
  const std::string& viewName = output.begin()->data<std::string>();
  statement.defineOutput( output );
  while( statement.fetchNext() )
  {
    if ( m_views.find( viewName ) == m_views.end() )
    {
      m_views[viewName] = 0;
    }
  }
  m_viewsReadFromDataDictionary = true;
}
