#include "CoralBase/MessageStream.h"
#include "RelationalAccess/AuthenticationServiceException.h"
#include "CoralCommon/AuthenticationCredentials.h"
#include "AuthenticationCredentialSet.h"
#include "ReplicaSet.h"
#include "LFCReplicaService.h"
#include "LFCDatabaseServiceSet.h"
#include "LFCFile.h"
#include "LFCReplica.h"
#include "LFCNameServer.h"
#include "LFCLogicalConnectionString.h"
#include "LFCCredentialString.h"
#include "LFCDbReplicaMapping.h"
#include "LFCException.h"
#include "LFCReplicaServiceException.h"

coral::LFCReplicaService::LFCReplicaService::LFCReplicaService( const std::string& componentName ):coral::Service( componentName ),
                                                                                                   m_credentialMap(),
                                                                                                   m_nameServer(new LFCNameServer){}
/// Standard Destructor
coral::LFCReplicaService::LFCReplicaService::~LFCReplicaService(){
  delete m_nameServer;
  for (std::map< std::string, coral::AuthenticationCredentialSet* >::iterator iAuthCred = m_credentialMap.begin();
       iAuthCred != m_credentialMap.end();++iAuthCred ){
    delete iAuthCred->second;
  }
  m_credentialMap.clear();
}

/* Performs a lookup for a given logical connection string.  */
coral::IDatabaseServiceSet*
coral::LFCReplicaService::LFCReplicaService::lookup( const std::string& logicalName,
                                                     coral::AccessMode accessMode,
                                                     std::string authenticationMechanism) const {
  
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Looking up logical string \"" << logicalName <<"\" with access mode="<<
    ((accessMode==coral::ReadOnly)? std::string("Read-Only") : std::string("Update") )<<coral::MessageStream::endmsg;
  
  ReplicaFilter filter;
  if(accessMode == coral::Update) filter.setAccessMode(accessMode);
  if(!authenticationMechanism.empty()) filter.setAuthenticationMechanism(authenticationMechanism);
  // only replica declared On are considered
  filter.setStatus(coral::LFCReplicaService::On);
  
  LFCDatabaseServiceSet* dbServiceSet = new LFCDatabaseServiceSet;
  
  try {
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    ::sleep(1);
    m_nameServer->startTransaction("Listing available replicas.");
    LFCLogicalConnectionString logicalConn(logicalName);
    LFCFile lfnFolder = logicalConn.lfcFile();
    lfnFolder.lookup();
    if(!lfnFolder.exists()){
      m_nameServer->commitTransaction();
      delete dbServiceSet;
      dbServiceSet = 0;
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" not found. Closing transaction."<< coral::MessageStream::endmsg;
      throw LFCReplicaServiceException("Logical connection string is unknown.","LFCReplicaService::listReplicas");
    }
    if(!lfnFolder.isDirectory()){
      m_nameServer->commitTransaction();
      delete dbServiceSet;
      dbServiceSet = 0;
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" does not describe a folder in LFC name server. Closing transaction."<< coral::MessageStream::endmsg;
      throw LFCReplicaServiceException("Logical connection string is not mapped to a folder.","LFCReplicaService::listReplicas");
    }
    m_credentialMap.clear();
    const std::vector<std::string>& files = lfnFolder.listFiles();
    for(std::vector<std::string>::const_iterator iF=files.begin();iF!=files.end();++iF){
      LFCFile file(lfnFolder.path(),*iF);
      file.lookup();
      if(!file.isDirectory()){
        if(file.isReadable()) {
           LFCReplicaSet* lfcReplicas = m_nameServer->listReplicas( file.path() );
           for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
                iRd != lfcReplicas->replicas().end();
                ++iRd){
             LFCDbReplicaMapping mapping(*(*iRd));
             ReplicaDescription* replicaDescr = 0;
             try {
               replicaDescr = mapping.dbReplica();
               if(filter.select(*replicaDescr)) {
                 if(!replicaDescr->authenticationMechanism().empty()){
                   // first treating authentication data
                   std::map<std::string,coral::AuthenticationCredentialSet*>::iterator iCm =
                       m_credentialMap.find(replicaDescr->physicalConnectionString());
                   if(iCm==m_credentialMap.end()){
                     coral::AuthenticationCredentialSet* credForReplica = new coral::AuthenticationCredentialSet(this->name(),
                                                                                                                 replicaDescr->physicalConnectionString());
                     iCm = m_credentialMap.insert(std::make_pair(replicaDescr->physicalConnectionString(),credForReplica)).first;
                   }
                   if(*iF==coral::IAuthenticationCredentials::defaultRole()){
                       iCm->second->appendCredentialItem(coral::IAuthenticationCredentials::userItem(),replicaDescr->userName());
                       iCm->second->appendCredentialItem(coral::IAuthenticationCredentials::passwordItem(),replicaDescr->password());
                   } else {
                       iCm->second->appendCredentialItemForRole(coral::IAuthenticationCredentials::userItem(),replicaDescr->userName(),*iF);
                       iCm->second->appendCredentialItemForRole(coral::IAuthenticationCredentials::passwordItem(),replicaDescr->password(),*iF);
                   }
                 }
                 // now lookup data...
                 if(!dbServiceSet->appendReplica(replicaDescr)){
                   delete replicaDescr;
                 }
               } else {
                 delete replicaDescr;
               }
             } catch (const LFCDbReplicaMappingException& exc){
               log << coral::Error << "Failed to process replica for \""<<file.path()<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
               if(replicaDescr) delete replicaDescr;
             }
           }
           delete lfcReplicas;
        }
      }
    }
    log << coral::Debug << "Lookup completed. Closing transaction."<< coral::MessageStream::endmsg;
    m_nameServer->commitTransaction();
  } catch (const LFCException& exc){
    delete dbServiceSet;
    dbServiceSet = 0;
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::listReplicas");
  }
  return dbServiceSet;          
}

/**
 * Returns a reference to the credentials object for a given connection string.
 * If the connection string is not known to the service an UnknownConnectionException is thrown.
 */
const coral::IAuthenticationCredentials&
coral::LFCReplicaService::LFCReplicaService::credentials( const std::string& connectionString ) const {
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Searching connection \""<<connectionString<<"\" in the credential map. Total entries="<<m_credentialMap.size()<<coral::MessageStream::endmsg;
  std::map<std::string,coral::AuthenticationCredentialSet*>::const_iterator iCred = m_credentialMap.find(connectionString);
  if(iCred==m_credentialMap.end()){
    throw coral::UnknownConnectionException("LFCReplicaService",connectionString);
  }
  return iCred->second->credentials();
}

/**
 * Returns a reference to the credentials object for a given connection string and a "role".
 * If the connection string is not known to the service an UnknownConnectionException is thrown.
 * If the role specified does not exist an UnknownRoleException is thrown.
 */
const coral::IAuthenticationCredentials&
coral::LFCReplicaService::LFCReplicaService::credentials( const std::string& connectionString,
                                                          const std::string& role ) const {
  
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Searching connection \""<<connectionString<<"\" in the credential map. Total entries="<<m_credentialMap.size()<<coral::MessageStream::endmsg;
  std::map<std::string,coral::AuthenticationCredentialSet*>::const_iterator iCred = m_credentialMap.find(connectionString);
  if(iCred==m_credentialMap.end()){
    throw coral::UnknownConnectionException("LFCReplicaService",connectionString);
  }
  return iCred->second->credentials(role);
}

void
coral::LFCReplicaService::LFCReplicaService::addReplica( const std::string& logicalName,
                                                         const std::string& theRole,
                                                         const ReplicaDescription& replica ){
  
  coral::MessageStream log( this->name() );
  std::string role = theRole;
  if(role.empty()) role=coral::IAuthenticationCredentials::defaultRole();
  log << coral::Debug << "Adding replica for logical string=\"" << logicalName <<"\", role=\""<< role<<"\": pcs=\""<<replica.physicalConnectionString()<<"\"."<<coral::MessageStream::endmsg;
  
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Creating a new logical connection string.");
    LFCLogicalConnectionString logicalConn(logicalName);
    LFCFile lfnFolder = logicalConn.lfcFile();
    LFCFile roleFile(lfnFolder.path(),role);
    roleFile.lookup();
    std::string lfnGuid;
    if(!roleFile.exists()){
      lfnFolder.lookup();
      if(!lfnFolder.exists()){
        m_nameServer->makeDir(lfnFolder.path());
      } else {
        if(!lfnFolder.isDirectory()){
          log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" does not describe a folder in LFC name server. Aborting transaction."<< coral::MessageStream::endmsg;
          m_nameServer->abortTransaction();
          throw LFCReplicaServiceException("Cannot create folder \""+lfnFolder.path()+"\", a file with this name already exists.",
                                    "LFCReplicaService::addReplica");
        }
      }
      lfnGuid = m_nameServer->makeFile(roleFile.path());
    } else {
      lfnGuid = roleFile.guid();
    }
    // using the same guid for the uniqueness of the pcs
    LFCDbReplicaMapping mapping(replica,lfnGuid);
    m_nameServer->addReplica(lfnGuid,
                             mapping.lfcReplica());
    log << coral::Debug << "Committing transaction."<< coral::MessageStream::endmsg;
    m_nameServer->commitTransaction();
    log << coral::Info << "Add Replica operation completed."<< coral::MessageStream::endmsg;
  } catch (const LFCException& ex) {
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(ex.what(),"LFCReplicaService::addReplica");
  }
}

void coral::LFCReplicaService::LFCReplicaService::importReplicas( const coral::LFCReplicaService::ReplicaSet& replicas ){
  for(std::map<std::string,std::map<std::string,std::multimap<std::string,ReplicaDescription*> > >::const_iterator iLcs=replicas.replicaMap().begin();
      iLcs!=replicas.replicaMap().end();++iLcs){
    std::string lcs = iLcs->first;
    for(std::map<std::string,std::multimap<std::string,ReplicaDescription*> >::const_iterator iRole=iLcs->second.begin();
        iRole!=iLcs->second.end();++iRole){
      std::string role = iRole->first;
      for(std::multimap<std::string,ReplicaDescription*>::const_iterator iReplica=iRole->second.begin();
          iReplica!=iRole->second.end();++iReplica){
        this->addReplica(lcs,role,*(iReplica->second));
      }
    }
  }
}

coral::LFCReplicaService::ReplicaSet*
coral::LFCReplicaService::LFCReplicaService::listReplicas( const std::string& logicalName,
                                                           const std::string& role,
                                                           const ReplicaFilter& filter,
                                                           bool scanChildFolders) const {
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Listing replicas for logical string=\"" << logicalName <<"\", role=\""<< role<<"\""<<coral::MessageStream::endmsg;
  ReplicaSet* replicaSet = 0;
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Listing available replicas.");
    LFCLogicalConnectionString logicalConn(logicalName);
    LFCFile lfnFolder = logicalConn.lfcFile();
    lfnFolder.lookup();
    if(!lfnFolder.exists()){
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" not found. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->commitTransaction();
      throw LFCReplicaServiceException("Logical connection string is unknown.","LFCReplicaService::listReplicas");
    }
    if(!lfnFolder.isDirectory()){
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" does not describe a folder in LFC name server. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->commitTransaction();
      throw LFCReplicaServiceException("Logical connection string is not mapped to a folder.","LFCReplicaService::listReplicas");
    }
    replicaSet = new ReplicaSet;
    if(!role.empty()){
      LFCFile roleFile(lfnFolder.path(),role);
      roleFile.lookup();
      if(!roleFile.exists()){
        log << coral::Error << "Role \""<<role<<"\" does not exists for logical connection \""<<lfnFolder.path()<<"\"."<< coral::MessageStream::endmsg;
        log << coral::Error << "Nothing to do . Closing transaction."<< coral::MessageStream::endmsg;
        delete replicaSet;
        replicaSet = 0;
        m_nameServer->commitTransaction();
        throw LFCReplicaServiceException("Role "+role+" does not exists for logical connection "+lfnFolder.path(),"LFCReplicaService::listReplicas");
      }
      if(roleFile.isDirectory()){
        delete replicaSet;
        replicaSet = 0;
        log << coral::Error << "Role entry \""<<lfnFolder.path()<<"\" describes a folder in LFC name server."<< coral::MessageStream::endmsg;
        log << coral::Debug << "Nothing to do. Closing transaction."<< coral::MessageStream::endmsg;
        m_nameServer->commitTransaction();
        throw LFCReplicaServiceException("Logical connection string is not mapped to a folder.","LFCReplicaService::listReplicas");
      }
      if(roleFile.isReadable()) {
        getReplicaEntries(lfnFolder.path(),role,roleFile.path(),filter,replicaSet );
      }
    } else {
      scanFolderTree( lfnFolder, filter, replicaSet, scanChildFolders );
    }
    log << coral::Debug << "Closing transaction."<< coral::MessageStream::endmsg;
    m_nameServer->commitTransaction();
    log << coral::Info << "Search completed."<< coral::MessageStream::endmsg;
  } catch (const LFCException& exc){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::listReplicas");
  }
  return replicaSet;
}

size_t coral::LFCReplicaService::LFCReplicaService::deleteReplicas( const std::string& logicalString,
                                                                    const std::string& role,
                                                                    const ReplicaFilter& filter ){
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Deleting replicas for logical string=\"" << logicalString <<"\", role=\""<< role<<"\""<<coral::MessageStream::endmsg;
  size_t deleted = 0;
  bool keepFolder = false;
  
  LFCReplicaSet* lfcReplicas = 0;
  ReplicaDescription* replicaDescr = 0;
  
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Deleting replicas for "+logicalString);
    std::vector<std::string> lfns;
    LFCLogicalConnectionString logicalConn(logicalString);
    LFCFile lfnFolder = logicalConn.lfcFile();
    lfnFolder.lookup();
    if(!lfnFolder.exists()){
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" not found. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      throw LFCReplicaServiceException("Logical connection string is unknown.","LFCReplicaService::deleteReplicas");
    }
    if(!lfnFolder.isDirectory()){
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" does not describe a folder in LFC name server. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      throw LFCReplicaServiceException("Logical connection string is not mapped to a folder.","LFCReplicaService::deleteReplicas");
    }
    if(role.empty()){
      const std::vector<std::string>& files =  lfnFolder.listFiles();
      if(files.size()==0){
        log << coral::Error << "No roles are associated to the path \""<<lfnFolder.path()<<"\". Closing transaction."<< coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();
        throw LFCReplicaServiceException("No roles are associated to the path "+lfnFolder.path(),"LFCReplicaService::deleteReplicas");
      }
      for(std::vector<std::string>::const_iterator iF=files.begin();iF!=files.end();++iF){
        LFCFile file(lfnFolder.path(),*iF);
        file.lookup();
        if(!file.isDirectory()){
          lfns.push_back(file.path());
        } else {
          keepFolder = true;
        }
      }
    } else {
      keepFolder = true;
      LFCFile roleFile(lfnFolder.path(),role);
      roleFile.lookup();
      if(roleFile.exists()){
        if(roleFile.isDirectory()){
          log << coral::Error << "Role file \""<<roleFile.path()<<"\" is a folder. Closing transaction."<< coral::MessageStream::endmsg;
          m_nameServer->abortTransaction();
          throw LFCReplicaServiceException(roleFile.path()+" is a Folder.","LFCReplicaService::deleteReplicas");
        }
        lfns.push_back(roleFile.path());
      } else {
        log << coral::Error << "Role file \""<<roleFile.path()<<"\" does not exist. Closing transaction."<< coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();
        throw LFCReplicaServiceException(roleFile.path()+" file does not exists.","LFCReplicaService::deleteReplicas");        
      }
    }
    for(std::vector<std::string>::iterator iR = lfns.begin();iR!=lfns.end();++iR){
      lfcReplicas = m_nameServer->listReplicas( *iR );
      log << coral::Debug << "Deleting replicas for file "<<*iR <<coral::MessageStream::endmsg;
      size_t replicasForThisRole = lfcReplicas->replicas().size();
      size_t deletedForThisRole = 0;
      for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
           iRd != lfcReplicas->replicas().end();
           ++iRd){
        LFCDbReplicaMapping mapping(*(*iRd));
        try {
          replicaDescr = mapping.dbReplica();
          if(filter.select(*replicaDescr)) {
            deletedForThisRole += m_nameServer->deleteReplica(*iR,(*iRd)->pfn);
          }
          delete replicaDescr;
          replicaDescr = 0;
        } catch (const LFCDbReplicaMappingException& exc){
          log << coral::Error << "Failed to process replica for \""<<*iR<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
          if(replicaDescr) delete replicaDescr;
          replicaDescr = 0;
        }
      }
      delete lfcReplicas;
      deleted += deletedForThisRole;
      lfcReplicas = 0;
      if(deletedForThisRole==replicasForThisRole){
        log << coral::Debug << "Deleting file "<<*iR <<" from LFC name server."<<coral::MessageStream::endmsg;
        m_nameServer->rmFile(*iR);
      }
    }
    log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
    m_nameServer->commitTransaction();
    log << coral::Info << "Deleted "<<deleted<<" replicas." << coral::MessageStream::endmsg;
    if(!keepFolder){
      log << coral::Debug << "Starting transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
      m_nameServer->startTransaction("Deleting folder "+logicalString);
      lfnFolder.lookup();
      if(lfnFolder.listFiles().empty()) {
        log << coral::Debug << "Removing folder "<<lfnFolder.path()<<" from LFC name server."<<coral::MessageStream::endmsg;
        m_nameServer->rmDir(lfnFolder.path());
      } 
      log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
      m_nameServer->commitTransaction();
      log << coral::Info << "Delete operation completed." << coral::MessageStream::endmsg;
    }
  } catch (const LFCException& ex){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    if(lfcReplicas) delete lfcReplicas;
    if(replicaDescr) delete replicaDescr;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(ex.what(),"LFCReplicaService::deleteReplicas");    
  }
  return deleted;
}

void coral::LFCReplicaService::LFCReplicaService::setReplicaPassword( const std::string& dbHostName,
                                                                      const std::string& userName,
                                                                      const std::string& newPassword ){
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Setting new password for db account with username=\"" << userName <<"\" on host=\""<< dbHostName<<"\"" <<coral::MessageStream::endmsg;
  LFCReplicaSet* lfcReplicas = 0;
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Changing password for user "+userName+"@"+dbHostName);
    lfcReplicas = m_nameServer->listReplicas("",dbHostName,"");
    if(lfcReplicas->replicas().empty()){
      log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      log << coral::Info << "No replica found for database host \"" << dbHostName << "\"."<<coral::MessageStream::endmsg;
    } else {
      bool userFound = false;
      for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
           iRd != lfcReplicas->replicas().end(); ++iRd){
        LFCDbReplicaMapping mapping(*(*iRd));
        ReplicaDescription* replicaDescr = 0;
        try {
          replicaDescr = mapping.dbReplica();
          if(replicaDescr->userName()==userName){
            userFound = true;
            log << coral::Debug << "Changing password: deleting replica "<<(*iRd)->pfn <<" with username=\""<<userName<<"\""<<coral::MessageStream::endmsg;
            m_nameServer->deleteReplicaByFileId( (*iRd)->fileid, (*iRd)->pfn);
            LFCReplica replica(*(*iRd));
            replica.fsParameter = LFCCredentialString::credentials(newPassword,replicaDescr->guid());
            log << coral::Debug << "Changing password: adding replica "<<(*iRd)->pfn <<" with username=\""<<userName<<"\""<<coral::MessageStream::endmsg;
            m_nameServer->addReplica( replica );
          }
          delete replicaDescr;
          replicaDescr = 0;
        } catch (const LFCDbReplicaMappingException& exc){
          log << coral::Error << "Failed to process replica with host=\""<<dbHostName<<"\", userName=\""<<userName<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
          if(replicaDescr) delete replicaDescr;
          replicaDescr = 0;
        }
        
      }
      if(userFound){
        log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
        m_nameServer->commitTransaction();        
        log << coral::Info << "Change password operation completed." << coral::MessageStream::endmsg;
      } else {
        log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();
        log << coral::Info << "No replica found for database user \"" << userName << "\"." <<coral::MessageStream::endmsg;
      }
    }
    delete lfcReplicas;
  }catch (const LFCException& exc){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    if(lfcReplicas) delete lfcReplicas;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::changePassword");
  }
}

bool coral::LFCReplicaService::LFCReplicaService::verifyReplicaPassword( const std::string& dbHostName,
                                                                         const std::string& userName,
                                                                         const std::string& currentPassword ) const {
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Verifying password for db account with username=\"" << userName <<"\" on host=\""<< dbHostName<<"\"" <<coral::MessageStream::endmsg;
  int notSamePassword = 0;
  LFCReplicaSet* lfcReplicas = 0;
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Verifying password for user "+userName+"@"+dbHostName);
    lfcReplicas = m_nameServer->listReplicas("",dbHostName,"");
    bool userFound = false;
    if(lfcReplicas->replicas().empty()){
      log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      log << coral::Info << "No replica found for database host \"" << dbHostName << "\"."<<coral::MessageStream::endmsg;
    } else {
      for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
           iRd != lfcReplicas->replicas().end(); ++iRd){
        LFCDbReplicaMapping mapping(*(*iRd));
        ReplicaDescription* replicaDescr = 0;
        try {
          replicaDescr = mapping.dbReplica();
          if(replicaDescr->userName()==userName){
            userFound = true;
            if(replicaDescr->password()!=currentPassword){
              ++notSamePassword;
              log << coral::Debug << "On replica \""<<(*iRd)->pfn<<"\" the password specified for db_server=\""<<dbHostName<<" and user=\""<<userName<<"\" is INVALID."<< coral::MessageStream::endmsg;
            }
          }
          delete replicaDescr;
          replicaDescr = 0;
        } catch (const LFCDbReplicaMappingException& exc){
          log << coral::Error << "Failed to process replica with host=\""<<dbHostName<<"\", userName=\""<<userName<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
          if(replicaDescr) delete replicaDescr;
          replicaDescr = 0;
        }
      }
      if(userFound){
        log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
        m_nameServer->commitTransaction();
        
        if(notSamePassword){
          log << coral::Info << "The password provided is INVALID for "<<notSamePassword<<" entries with db_server=\""<<dbHostName<<"\" and user=\""<<userName<<"\"."<<coral::MessageStream::endmsg;          
        } else {
          log << coral::Info << "The password provided is VALID for all of the entries with db_server=\""<<dbHostName<<"\" and user=\""<<userName<<"\"."<<coral::MessageStream::endmsg;
        }
      } else {
        log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();
        log << coral::Info << "No replica found for database user \"" << userName << "\"." <<coral::MessageStream::endmsg;
      }
    }
    delete lfcReplicas;
  }catch (const LFCException& exc){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    if(lfcReplicas) delete lfcReplicas;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::changePassword");
  }
  return notSamePassword==0;
}

void coral::LFCReplicaService::LFCReplicaService::setReplicaStatus( const std::string& dbHostName,
                                                                    const std::string& userName,
                                                                    ReplicaStatus newStatus){
  coral::MessageStream log( this->name() );
  LFCReplicaSet* lfcReplicas = 0;
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Changing status for replicas on dbhost="+dbHostName);
    lfcReplicas = m_nameServer->listReplicas("",dbHostName,"");
    bool userFound = false;
    if(lfcReplicas->replicas().empty()){
      log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      log << coral::Info << "No replica found for database host \"" << dbHostName << "\"." <<coral::MessageStream::endmsg;
    } else {
      for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
           iRd != lfcReplicas->replicas().end(); ++iRd){
        LFCDbReplicaMapping mapping(*(*iRd));
        ReplicaDescription* replicaDescr = 0;
        try {
          replicaDescr = mapping.dbReplica();
          if(userName.empty() || (replicaDescr->userName()==userName)){
            userFound = true;
            std::string userLog("");
            if(!userName.empty()) userLog = " and userName=\""+userName+"\"";
            log << coral::Debug << "Setting status="<<LFCDbReplicaMapping::statusCode(newStatus)<<" for db host=\""<< dbHostName <<userLog<<coral::MessageStream::endmsg;
            m_nameServer->setReplicaStatus((*iRd)->pfn,LFCDbReplicaMapping::statusCode(newStatus));
          }
          delete replicaDescr;
          replicaDescr = 0;
        } catch (const LFCDbReplicaMappingException& exc){
          log << coral::Error << "Failed to process replica with host=\""<<dbHostName<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
          if(replicaDescr) delete replicaDescr;
          replicaDescr = 0;
        }          
      }
      if(userFound){
        log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
        m_nameServer->commitTransaction();
        log << coral::Info << "Set status operation completed." << coral::MessageStream::endmsg;
      } else {
        log << coral::Debug << "Nothing to do. Closing transaction." <<coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();        
        log << coral::Info << "No replica found for database user \"" << userName << "\"." <<coral::MessageStream::endmsg;
      }
    }
    delete lfcReplicas;
  } catch (const LFCException& exc){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    if(lfcReplicas) delete lfcReplicas;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::changePassword");
  }
}

void coral::LFCReplicaService::LFCReplicaService::setReplicaAccessPermission( const std::string& logicalName,
                                                                              const std::string& role,
                                                                              const std::string& groupName,
                                                                              ReplicaPermission permission){
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Set replica access permission for logical string=\"" << logicalName <<"\", role=\""<< role<<"\", "
      <<"group \""<<groupName<<"\""<< coral::MessageStream::endmsg;
  
  try {  
    log << coral::Debug << "Opening transaction on LFC host \""<<m_nameServer->lfcHost()<<"\""<<coral::MessageStream::endmsg;
    m_nameServer->startTransaction("Deleting replicas for "+logicalName);
    LFCLogicalConnectionString logicalConn(logicalName);
    LFCFile lfnFolder = logicalConn.lfcFile();
    lfnFolder.lookup();
    if(!lfnFolder.exists()){
      log << coral::Debug << "Nothing to do. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" not found."<< coral::MessageStream::endmsg;
      throw LFCReplicaServiceException("Logical connection string is unknown.","LFCReplicaService::setReplicaAccessPermission");
    }
    if(!lfnFolder.isDirectory()){
      log << coral::Debug << "Nothing to do. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      log << coral::Error << "Logical connection string \""<<lfnFolder.path()<<"\" does not describe a folder in LFC name server."<< coral::MessageStream::endmsg;
      throw LFCReplicaServiceException("Logical connection string is unknown.","LFCReplicaService::setReplicaAccessPermission");
    }
    int oldMode = m_nameServer->getPermissionToGroup(lfnFolder.path(),groupName);    
    m_nameServer->setPermissionToGroup(lfnFolder.path(),groupName,LFCDbReplicaMapping::permissionMode(permission)|1|oldMode);
    LFCFile roleFile(lfnFolder.path(),role);
    roleFile.lookup();
    if(roleFile.exists()){
      if(roleFile.isDirectory()){
        log << coral::Error << "Role file \""<<roleFile.path()<<"\" is a folder. Closing transaction."<< coral::MessageStream::endmsg;
        m_nameServer->abortTransaction();
        throw LFCReplicaServiceException(roleFile.path()+" is a Folder.","LFCReplicaService::setReplicaAccessPermission");
      }
    } else {
      log << coral::Error << "Role file \""<<roleFile.path()<<"\" does not exist. Closing transaction."<< coral::MessageStream::endmsg;
      m_nameServer->abortTransaction();
      throw LFCReplicaServiceException(roleFile.path()+" file does not exists.","LFCReplicaService::setReplicaAccessPermission");
    }
    m_nameServer->setPermissionToGroup(roleFile.path(),groupName,LFCDbReplicaMapping::permissionMode(permission));
    log << coral::Debug << "Committing transaction." << coral::MessageStream::endmsg;
    m_nameServer->commitTransaction();
    log << coral::Info << "Set Permission operation completed." << coral::MessageStream::endmsg;
  } catch (const LFCException& exc){
    log << coral::Error << "Access to LFC server \""<<m_nameServer->lfcHost()<<"\" failed. Aborting transaction."<< coral::MessageStream::endmsg;
    m_nameServer->abortTransaction();
    throw LFCReplicaServiceException(exc.what(),"LFCReplicaService::setReplicaAccessPermission");
  }
}

void coral::LFCReplicaService::LFCReplicaService::scanFolderTree( const LFCFile& folder,
                                                                  const ReplicaFilter& filter,
                                                                  ReplicaSet* replicaSet,
                                                                  bool scanChildFolders ) const {
  const std::vector<std::string>& files = folder.listFiles();
  for(std::vector<std::string>::const_iterator iF=files.begin();iF!=files.end();iF++){
    LFCFile child(folder.path(),*iF);
    child.lookup();
    if(child.isDirectory()){
      if(scanChildFolders) scanFolderTree(child,filter,replicaSet,true);
    } else {
      if(child.isReadable()){
        getReplicaEntries(folder.path(),*iF, child.path(),filter,replicaSet );
      }
    }
  }
}

void coral::LFCReplicaService::LFCReplicaService::getReplicaEntries( const std::string& folder,
                                                                     const std::string& fileName,
                                                                     const std::string& filePath,
                                                                     const ReplicaFilter& filter,
                                                                     ReplicaSet* replicaSet ) const {
  LFCReplicaSet* lfcReplicas = m_nameServer->listReplicas( filePath );
  for (std::vector<LFCReplica*>::const_iterator iRd = lfcReplicas->replicas().begin();
       iRd != lfcReplicas->replicas().end();
       ++iRd){
    LFCDbReplicaMapping mapping(*(*iRd));
    ReplicaDescription* replicaDescr = 0;
    try{
      replicaDescr = mapping.dbReplica();
      if(filter.select(*replicaDescr)) {
        replicaSet->appendReplica(folder,fileName,replicaDescr);
      } else {
        delete replicaDescr;
      }
    } catch (const LFCDbReplicaMappingException& exc){
      coral::MessageStream log( this->name() );
      log << coral::Error << "Failed to process replica for \""<<filePath<<"\": "<< exc.what()<< coral::MessageStream::endmsg;
      if(replicaDescr) delete replicaDescr;
    } 
  }
  delete lfcReplicas;
}

