#ifndef ORACLE_CONNECTION_PROPERTIES_H
#define ORACLE_CONNECTION_PROPERTIES_H

#include <string>
#include "CoralBricks/ThreadTraits.h"

struct OCIEnv;
struct OCIError;
struct OCIServer;

namespace coral {

  class ITypeConverter;

  namespace OracleAccess {

    class TypeConverter;
    class DomainProperties;

    /**
     * Class ConnectionProperties
     *
     * A class holding the running parameters of a database connection
     */

    class ConnectionProperties
    {
    public:
      /// Constructor
      ConnectionProperties( const DomainProperties& domainProperties,
                            const std::string& connectionString );

      /// Destructor
      virtual ~ConnectionProperties();

      /// Sets the OCI handles and resets the type converter
      void setHandles( OCIEnv* environmentHandle,
                       OCIError* errorHandle,
                       OCIServer* serverHandle,
                       int majorServerVersion );

      /// Returns the domain properties
      const DomainProperties& domainProperties() const;

      /// Returns the connection string
      std::string connectionString() const;

      /// Returns the type converter
      coral::ITypeConverter& typeConverter();
      const coral::ITypeConverter& typeConverter() const;

      /// Returns the OCI environment handle
      OCIEnv* environmentHandle() const;

      /// Returns the OCI error handle
      OCIError* errorHandle() const;

      /// Returns the OCI server handle
      OCIServer* serverHandle() const;

      /// Returns the server version
      int serverVersion() const;

    private:
      /// The domain properties
      const DomainProperties& m_domainProperties;

      /// The connection string
      std::string m_connectionString;

      /// The type converter
      TypeConverter* m_typeConverter;

      /// The OCI environment handle
      OCIEnv* m_environmentHandle;

      /// The OCI error handle
      OCIError* m_errorHandle;

      /// The OCI server handle
      OCIServer* m_serverHandle;

      /// The server major version
      int m_serverVersion;

      /// The mutex lock
      mutable boost::mutex m_mutex;
    };

  }

}


// Inline methods

inline const coral::OracleAccess::DomainProperties&
coral::OracleAccess::ConnectionProperties::domainProperties() const
{
  return m_domainProperties;
}


inline std::string
coral::OracleAccess::ConnectionProperties::connectionString() const
{
  return m_connectionString;
}


inline OCIEnv*
coral::OracleAccess::ConnectionProperties::environmentHandle() const
{
  boost::mutex::scoped_lock lock( m_mutex );
  return m_environmentHandle;
}


inline OCIError*
coral::OracleAccess::ConnectionProperties::errorHandle() const
{
  boost::mutex::scoped_lock lock( m_mutex );
  return m_errorHandle;
}


inline OCIServer*
coral::OracleAccess::ConnectionProperties::serverHandle() const
{
  boost::mutex::scoped_lock lock( m_mutex );
  return m_serverHandle;
}


inline int
coral::OracleAccess::ConnectionProperties::serverVersion() const
{
  boost::mutex::scoped_lock lock( m_mutex );
  return m_serverVersion;
}

#endif
