#ifndef ORACLEACCESS_ORACLESTATEMENT_H
#define ORACLEACCESS_ORACLESTATEMENT_H

#ifdef WIN32
#include <WTypes.h> // AK 04.12.2008 include this first on Win (bug #35683)
#endif

#include "oratypes.h"
#include <string>
#include <vector>
#include <typeinfo>

struct OCIStmt;
struct OCINumber;
struct OCIDate;
struct OCIDateTime;
struct OCILobLocator;

namespace coral {

  class AttributeList;
  class Date;
  class TimeStamp;

  namespace OracleAccess {

    class ISessionProperties;
    class IOutputVariableHolder;
    class StatementStatistics;

    /**
     * @class OracleStatement OracleStatement.h
     *
     * Class responsible for the execution of SQL statements
     * and for the retrieval of its results.
     *
     */

    class OracleStatement {
    public:
      /**
       * Constructor
       *
       */
      OracleStatement( const ISessionProperties& properties,
                       const std::string& sqlStatement  );

      /// Destructor
      ~OracleStatement();

      /**
       * Prepares and executes the statement.
       * @param inputData The input bind data.
       * Returns false in case of any error, true otherwise.
       *
       */
      bool execute( const coral::AttributeList& inputData );

      /**
       * Returns the C++ type for an output column [0,..]
       *
       * @param columnId The column identifier
       *
       * @param typeConverter The SQL to C++ type converter
       *
       */
      const std::type_info* typeForOutputColumn( int columnId ) const;

      /**
       * Defines the output. It assumes that the privided output AttributeList matches exactly
       * the actual output from the statement.
       * Returns false in case of any error, true otherwise.
       *
       * @param outputData The buffers where the output data should be delivered.
       *
       */
      bool defineOutput( coral::AttributeList& outputData );

      /**
       * Returns the number of rows processed.
       *
       */
      long numberOfRowsProcessed() const;

      /**
       * Sets the number of prefetched rows
       *
       * @param numberOfRows The number of rows to be prefetched.
       *
       */
      bool setNumberOfPrefetchedRows( unsigned int numberOfRows );

      /**
       * Sets the size of the cache for prefetched rows
       *
       * @param sizeInMB The size of the cache in MBytes.
       *
       */
      bool setCacheSize( unsigned int sizeInMB );

      /**
       * Fetches the next row. Returns false in case no more rows are there.
       *
       */
      bool fetchNext();

      /**
       * Retrieves the current row number
       *
       */
      unsigned int currentRowNumber() const;

    private:
      /// Resets the internal structures.
      void reset();

      /// Copies the data from the buffers to the defined output
      bool copyOutput();

    private:
      /// The session properties
      const ISessionProperties& m_sessionProperties;

      /// The OCI statement handle
      OCIStmt*                 m_statementHandle;

      /// Indicator array
      std::vector< sb2 >       m_indicators;

      /// The output data
      coral::AttributeList*    m_outputData;

      /// The buffer holding the output variables
      std::vector< IOutputVariableHolder* > m_outputVariables;

      // The statistics for the statement
      StatementStatistics* m_statementStatistics;
    };

  }

}

#endif
