#include "Query.h"
#include "ISessionProperties.h"
#include "DomainProperties.h"
#include "Cursor.h"
#include "OracleStatement.h"

#include "RelationalAccess/SchemaException.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include "CoralKernel/Service.h"

coral::OracleAccess::Query::Query( const coral::OracleAccess::ISessionProperties& properties,
                                   const std::string& tableName ):
  coral::OracleAccess::QueryDefinition( properties, tableName ),
  m_cursor( 0 ),
  m_forUpdate( false ),
  m_cacheSize( 0 ),
  m_outputBuffer( 0 ),
  m_outputTypes()
{}


coral::OracleAccess::Query::Query( const coral::OracleAccess::ISessionProperties& properties ):
  coral::OracleAccess::QueryDefinition( properties ),
  m_cursor( 0 ),
  m_forUpdate( false ),
  m_cacheSize( 0 ),
  m_outputBuffer( 0 ),
  m_outputTypes()
{}


coral::OracleAccess::Query::~Query()
{
  if ( m_outputBuffer ) delete m_outputBuffer;
  if ( m_cursor ) delete m_cursor;
}

void
coral::OracleAccess::Query::setForUpdate()
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::setForUpdate" );

  if ( this->sessionProperties().isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( this->sessionProperties().domainProperties().service()->name(),
                                                          "IQuery::setForUpdate" );

  m_forUpdate = true;
}


void
coral::OracleAccess::Query::setRowCacheSize( int numberOfCachedRows )
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::setRowCacheSize" );
  m_cacheSize = numberOfCachedRows;
}


void
coral::OracleAccess::Query::setMemoryCacheSize( int sizeInMB )
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::setMemoryCacheSize" );
  m_cacheSize = -sizeInMB;
}


void
coral::OracleAccess::Query::defineOutputType( const std::string& outputIdentifier,
                                              const std::string& cppTypeName )
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::defineOutputTypes" );
  if ( m_outputBuffer ) {
    delete m_outputBuffer;
    m_outputBuffer = 0;
  }
  m_outputTypes.insert( std::make_pair( outputIdentifier, cppTypeName ) );
}


void
coral::OracleAccess::Query::defineOutput( coral::AttributeList& outputDataBuffer )
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::defineOutput" );

  if ( m_outputBuffer ) delete m_outputBuffer;
  m_outputBuffer = new coral::AttributeList( outputDataBuffer );
  const unsigned int numberOfVariables = outputDataBuffer.size();
  for ( unsigned int i = 0; i < numberOfVariables; ++i ) {
    (*m_outputBuffer)[i].shareData( outputDataBuffer[i] );
  }
}


coral::ICursor&
coral::OracleAccess::Query::execute()
{
  if ( m_cursor )
    throw coral::QueryExecutedException( this->sessionProperties().domainProperties().service()->name(),
                                         "IQuery::execute" );

  // Get the sql fragment from the definition
  std::string sqlStatement = this->sqlFragment();

  // Lock for update
  if ( m_forUpdate ) sqlStatement += " FOR UPDATE";

  // Prepare the statement.
  coral::OracleAccess::OracleStatement* statement = new coral::OracleAccess::OracleStatement( this->sessionProperties(),
                                                                                              sqlStatement );

  // Define the cache size.
  if ( m_cacheSize < 0 ) {
    statement->setCacheSize( -m_cacheSize );
  }
  else if ( m_cacheSize > 0 ) {
    statement->setNumberOfPrefetchedRows( m_cacheSize );
  }

  // Bind and execute.
  bool result = statement->execute( this->bindData() );
  if ( ! result )
    throw coral::QueryException( this->sessionProperties().domainProperties().service()->name(),
                                 "Could not execute a query",
                                 "IQuery::execute" );

  const std::vector< std::string >& outputNames = this->outputVariables();

  // Define the output.
  if ( ! m_outputBuffer ) {
    m_outputBuffer = new coral::AttributeList;
    int position = 0;
    for ( std::vector<std::string>::const_iterator iOutput = outputNames.begin();
          iOutput != outputNames.end(); ++iOutput, ++position ) {
      // Check if the type is known.
      std::map< std::string, std::string >::const_iterator iType = m_outputTypes.find( *iOutput );
      if ( iType != m_outputTypes.end() ) {
        m_outputBuffer->extend( iType->first, iType->second );
      }
      else {
        const std::type_info* typeId = statement->typeForOutputColumn( position );
        if ( typeId == 0 )
          throw coral::QueryException( this->sessionProperties().domainProperties().service()->name(),
                                       "Could not identify output type for \"" + *iOutput + "\"",
                                       "IQuery::execute" );
        m_outputBuffer->extend( *iOutput,*typeId );
      }
    }
  }

  statement->defineOutput( *m_outputBuffer );

  // Return the cursor
  m_cursor = new coral::OracleAccess::Cursor( statement, *m_outputBuffer );
  return *m_cursor;
}
