#include "Schema.h"
#include "ISessionProperties.h"
#include "DomainProperties.h"
#include "Table.h"
#include "OracleStatement.h"
#include "OracleErrorHandler.h"
#include "OracleTableBuilder.h"
#include "Query.h"
#include "ViewFactory.h"
#include "View.h"

#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/ITableSchemaEditor.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "CoralKernel/Service.h"

#include "oci.h"

#include <sstream>

coral::OracleAccess::Schema::Schema( const coral::OracleAccess::ISessionProperties& sessionProperties ):
  m_sessionProperties( sessionProperties ),
  m_tables(),
  m_tablesReadFromDataDictionary( false ),
  m_views(),
  m_viewsReadFromDataDictionary( false ),
  m_tableMutex(),
  m_viewMutex()
{}


coral::OracleAccess::Schema::~Schema()
{
  this->reactOnEndOfTransaction();
}

std::string coral::OracleAccess::Schema::schemaName() const
{
  return m_sessionProperties.schemaName();
}

std::set<std::string>
coral::OracleAccess::Schema::listTables() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::listTables" );
  std::set<std::string> result;
  if ( ! m_tablesReadFromDataDictionary )
    this->readTablesFromDataDictionary();

  boost::mutex::scoped_lock lock( m_tableMutex );
  for ( std::map< std::string, coral::OracleAccess::Table*>::const_iterator iTable = m_tables.begin();
        iTable != m_tables.end(); ++iTable ) {
    result.insert( iTable->first );
  }

  return result;
}


bool
coral::OracleAccess::Schema::existsTable( const std::string& tableName ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::existsTable" );

  // Check if the table is already in the local cache.
  bool isInCache = false;
  {
    boost::mutex::scoped_lock lock( m_tableMutex );
    isInCache = ( m_tables.find( tableName ) == m_tables.end() ? false : true );
  }

  if ( ! isInCache ) {

    // Check inside the database.

    // The fully qualified table name
    std::string fullTableName = m_sessionProperties.schemaName() + ".\"" + tableName + "\"";
    
    // Try to get the descriptor handle
    void* temporaryPointer = 0;
    sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                   &temporaryPointer,
                                   OCI_HTYPE_DESCRIBE, 0, 0 );
    if ( status != OCI_SUCCESS ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not allocate a describe handle",
                                    "ISchema::existsTable" );
    }

    OCIDescribe* describeHandle = static_cast< OCIDescribe* >( temporaryPointer );

    status = OCIDescribeAny( m_sessionProperties.serviceContextHandle(),
                             m_sessionProperties.errorHandle(),
                             const_cast<char *>( fullTableName.c_str() ),
                             ::strlen( fullTableName.c_str() ),
                             OCI_OTYPE_NAME, OCI_DEFAULT,
                             OCI_PTYPE_TABLE,
                             describeHandle );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "retrieving the describe handle of " + fullTableName );
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
      if ( errorHandler.lastErrorCode() != 4043 ) {
        coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
        log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
        throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                      "Could not retrieve a describe handle",
                                      "ISchema::existsTable" );
      }


      // Check if it is a synonym
      std::string tableForSynonym = this->synonymForTable( tableName );
      if ( tableForSynonym.empty() )
        return false;
      else
        return this->existsTable( tableForSynonym );
    }
    else {
      boost::mutex::scoped_lock lock( m_tableMutex );

      if ( m_tables.find( tableName ) != m_tables.end() ) { // Another thread was faster...
        OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
        return true;
      }

      // The table exists. Create the table object and insert it into the map
      coral::OracleAccess::Table* table = new coral::OracleAccess::Table( m_sessionProperties,
                                                                          tableName,
                                                                          describeHandle );
      m_tables.insert( std::make_pair( tableName, table ) );
      return true;
    }
  }
  else
    return true;
}


std::string
coral::OracleAccess::Schema::synonymForTable( const std::string& tableName ) const
{
  // The fully qualified table name
  std::string fullTableName = m_sessionProperties.schemaName() + ".\"" + tableName + "\"";
    
  // Try to get the descriptor handle
  void* temporaryPointer = 0;
  sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                 &temporaryPointer,
                                 OCI_HTYPE_DESCRIBE, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not allocate a describe handle",
                                  "OracleAccess::Schema::synonymForTable" );
  }

  OCIDescribe* describeHandle = static_cast< OCIDescribe* >( temporaryPointer );

  status = OCIDescribeAny( m_sessionProperties.serviceContextHandle(),
                           m_sessionProperties.errorHandle(),
                           const_cast<char *>( fullTableName.c_str() ),
                           ::strlen( fullTableName.c_str() ),
                           OCI_OTYPE_NAME, OCI_DEFAULT,
                           OCI_PTYPE_SYN,
                           describeHandle );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "retrieving the describe handle of " + fullTableName );
    OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
    if ( errorHandler.lastErrorCode() != 4043 ) {
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve a describe handle",
                                    "OracleAccess::Schema::synonymForTable" );
    }
    else {
      return "";
    }
  }

  // Retrieve the corresponding table name.
  OCIParam* parameterHandle = 0;
  status = OCIAttrGet( describeHandle, OCI_HTYPE_DESCRIBE,
                       &parameterHandle, 0, OCI_ATTR_PARAM,
                       m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Retrieving the describe parameter for a synonym" );
    coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
    log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
    OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the OCI describe parameter for a synonym",
                                  "OracleAccess::Schema::synonymForTable" );
  }

  // Retrieving the table name
  text* textPlaceHolder = 0;
  ub4 textSize = 0;
  status = OCIAttrGet( parameterHandle, OCI_DTYPE_PARAM,
                       &textPlaceHolder, &textSize,
                       OCI_ATTR_NAME,
                       m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Retrieving the name of the object referenced by a synonym" );
    coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
    log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
    OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the name of the object referenced by a synonym",
                                  "OracleAccess::Schema::synonymForTable" );
  }

  std::ostringstream osName;
  osName << textPlaceHolder;
  std::string sName = osName.str();

  OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
  return sName.substr( 0, textSize );
}


void
coral::OracleAccess::Schema::dropTable( const std::string& tableName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::dropTable" );

  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::dropTable" );

  if ( ! this->existsTable( tableName ) ) {
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),
                                            tableName,
                                            "ISchema::dropTable" );
  }
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "DROP TABLE " + m_sessionProperties.schemaName() + ".\"" + tableName + "\"" );

  if ( ! statement.execute( coral::AttributeList() ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop a table",
                                  "ISchema::dropTable" );

  boost::mutex::scoped_lock lock( m_tableMutex );
  std::map< std::string, coral::OracleAccess::Table* >::iterator iTable = m_tables.find( tableName );
  if ( iTable->second ) delete iTable->second;
  m_tables.erase( iTable );
}


void
coral::OracleAccess::Schema::dropIfExistsTable( const std::string& tableName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::dropIfExistsTable" );
  if ( this->existsTable( tableName ) )
    this->dropTable( tableName );
}


coral::ITable&
coral::OracleAccess::Schema::createTable( const coral::ITableDescription& description )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::createTable" );

  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::createTable" );

  // Get the name of the table
  std::string tableName = description.name();

  // Check if a table with this name already exists
  if ( this->existsTable( tableName ) )
    throw coral::TableAlreadyExistingException( m_sessionProperties.domainProperties().service()->name(),
                                                tableName );

  // Construct and execute the statement creating the table
  coral::OracleAccess::OracleTableBuilder builder( description,
                                                   m_sessionProperties );
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  builder.statement() );
  if ( ! statement.execute( coral::AttributeList() ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not create a new table",
                                  "ISchema::createTable" );

  // Get the table handle
  coral::ITable& table = this->tableHandle( tableName );

  // Use now the schema editor to add the indices
  coral::ITableSchemaEditor& editor = table.schemaEditor();
  for ( int i = 0; i < description.numberOfIndices(); ++i ) {
    const coral::IIndex& index = description.index( i );
    editor.createIndex( index.name(), index.columnNames(), index.isUnique(), index.tableSpaceName() );
  }

  return table;
}


coral::ITable&
coral::OracleAccess::Schema::tableHandle( const std::string& tableName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::tableHandle" );

  // Check if the table is already in the local cache.
  std::map< std::string, coral::OracleAccess::Table* >::iterator iTable;
  bool needDescription = false;
  {
    boost::mutex::scoped_lock lock( m_tableMutex );
    iTable = m_tables.find( tableName );
    needDescription = ( iTable == m_tables.end() || iTable->second == 0 );
  }

  if ( needDescription ) {

    // The fully qualified table name
    const std::string fullTableName = m_sessionProperties.schemaName() + ".\"" + tableName + "\"";
    
    // Try to get the descriptor handle
    void* temporaryPointer = 0;
    sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                   &temporaryPointer,
                                   OCI_HTYPE_DESCRIBE, 0, 0 );
    if ( status != OCI_SUCCESS ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not allocate a describe handle",
                                    "ISchema::tableHandle" );
    }

    OCIDescribe* describeHandle = static_cast< OCIDescribe* >( temporaryPointer );

    status = OCIDescribeAny( m_sessionProperties.serviceContextHandle(),
                             m_sessionProperties.errorHandle(),
                             const_cast<char *>( fullTableName.c_str() ),
                             ::strlen( fullTableName.c_str() ),
                             OCI_OTYPE_NAME, OCI_DEFAULT,
                             OCI_PTYPE_TABLE,
                             describeHandle );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "retrieving the describe handle of " + fullTableName );
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
      if ( errorHandler.lastErrorCode() != 4043 ) {
        coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
        log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
        throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                      "Could not retrieve a describe handle",
                                      "ISchema::tableHandle" );
      }
      else {
        // Check if it is a synonym
        std::string tableForSynonym = this->synonymForTable( tableName );
        if ( tableForSynonym.empty() )
          throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),
                                                  fullTableName );
        else
          return this->tableHandle( tableForSynonym );
      }
    }

    boost::mutex::scoped_lock lock( m_tableMutex );
    iTable = m_tables.find( tableName );
    if ( iTable != m_tables.end() && iTable->second != 0 ) { // Another thread was faster...
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
    }
    else {
      // The table exists. Create the table object and insert it into the map
      coral::OracleAccess::Table* table = new coral::OracleAccess::Table( m_sessionProperties,
                                                                          tableName,
                                                                          describeHandle );
      if ( iTable == m_tables.end() )
        iTable = m_tables.insert( std::make_pair( tableName, table ) ).first;
      else
        iTable->second = table;
    }
    return *( iTable->second );
  }

  return *( iTable->second );
}


void
coral::OracleAccess::Schema::truncateTable( const std::string& tableName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::truncateTable" );
  if ( ! this->existsTable( tableName ) )
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),
                                            tableName, "truncateTable" );

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "TRUNCATE TABLE " + m_sessionProperties.schemaName() + ".\"" + tableName + "\"" );

  if ( ! statement.execute( coral::AttributeList() ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not truncate table \"" + tableName + "\"",
                                  "ISchema::truncateTable" );
}


void
coral::OracleAccess::Schema::callProcedure( const std::string& procedureName,
                                            const coral::AttributeList& inputArguments )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::callProcedure" );
  std::ostringstream os;
  std::string sqlProcName = procedureName;
  size_t dotPos = sqlProcName.find(".");
  
  if (dotPos != std::string::npos) sqlProcName.replace(dotPos, 1, "\".\"");
 
  os << "CALL " << m_sessionProperties.schemaName() << ".\"" << sqlProcName << "\" ( ";

  for ( coral::AttributeList::const_iterator iAttribute = inputArguments.begin();
        iAttribute != inputArguments.end(); ++iAttribute ) {
    if ( iAttribute != inputArguments.begin() ) os << ", ";
    os << ":\"" << iAttribute->specification().name() << "\"";
  }

  os << " )";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );

  if ( ! statement.execute( inputArguments ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Error calling \"" + sqlProcName + "\"",
                                  "ISchema::callProcedure" );
}


coral::IQuery*
coral::OracleAccess::Schema::newQuery() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::newQuery" );
  return new coral::OracleAccess::Query( m_sessionProperties );
}


coral::IViewFactory*
coral::OracleAccess::Schema::viewFactory()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::viewFactory" );

  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::viewFactory" );

  return new coral::OracleAccess::ViewFactory( m_sessionProperties );
}


bool
coral::OracleAccess::Schema::existsView( const std::string& viewName ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::existsView" );

  // Check if the view is already in the local cache.
  bool isInCache = false;
  {
    boost::mutex::scoped_lock lock( m_tableMutex );
    isInCache = ( m_views.find( viewName ) != m_views.end() );
  }

  if ( ! isInCache ) {

    // Check inside the database.

    // The fully qualified view name
    std::string fullViewName = m_sessionProperties.schemaName() + ".\"" + viewName + "\"";

    // Try to get the descriptor handle
    void* temporaryPointer = 0;
    sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                   &temporaryPointer,
                                   OCI_HTYPE_DESCRIBE, 0, 0 );
    if ( status != OCI_SUCCESS ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not allocate a describe handle",
                                    "ISchema::existsView" );
    }

    OCIDescribe* describeHandle = static_cast< OCIDescribe* >( temporaryPointer );

    status = OCIDescribeAny( m_sessionProperties.serviceContextHandle(),
                             m_sessionProperties.errorHandle(),
                             const_cast<char *>( fullViewName.c_str() ),
                             ::strlen( fullViewName.c_str() ),
                             OCI_OTYPE_NAME, OCI_DEFAULT,
                             OCI_PTYPE_VIEW,
                             describeHandle );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "retrieving the describe handle of " + fullViewName );
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
      if ( errorHandler.lastErrorCode() != 4043 ) {
        coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
        log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
        throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                      "Could not retrieve a describe handle",
                                      "ISchema::existsView" );
      }
      return false;
    }
    else {
      boost::mutex::scoped_lock lock( m_viewMutex );
      if ( m_views.find( viewName ) != m_views.end() ) { // Another thread was faster...
        OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
        return true;
      }

      // The view exists. Create the view object and insert it into the map
      coral::OracleAccess::View* view = new coral::OracleAccess::View( m_sessionProperties,
                                                                       viewName,
                                                                       describeHandle );
      m_views.insert( std::make_pair( viewName, view ) );
      return true;
    }
  }
  else
    return true;
}


void
coral::OracleAccess::Schema::dropView( const std::string& viewName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::dropView" );
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::dropView" );

  if ( ! this->existsView( viewName ) ) {
    throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(),
                                            viewName,
                                            "ISchema::dropView" );
  }
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "DROP VIEW " + m_sessionProperties.schemaName() + ".\"" + viewName + "\"" );

  if ( ! statement.execute( coral::AttributeList() ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop a view",
                                  "ISchema::dropView" );

  boost::mutex::scoped_lock lock( m_viewMutex );
  std::map< std::string, coral::OracleAccess::View* >::iterator iView = m_views.find( viewName );
  if ( iView->second ) delete iView->second;
  m_views.erase( iView );
}


void
coral::OracleAccess::Schema::dropIfExistsView( const std::string& viewName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::dropIfExistsView" );
  if ( this->existsView( viewName ) )
    this->dropView( viewName );
}


coral::IView&
coral::OracleAccess::Schema::viewHandle( const std::string& viewName )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::viewHandle" );

  // Check if the view is already in the local cache.
  std::map< std::string, coral::OracleAccess::View* >::iterator iView;
  bool needDescription = false;
  {
    boost::mutex::scoped_lock lock( m_viewMutex );
    iView = m_views.find( viewName );
    needDescription = ( iView == m_views.end() || iView->second == 0 );
  }

  if ( needDescription ) {

    // The fully qualified view name
    const std::string fullViewName = m_sessionProperties.schemaName() + ".\"" + viewName + "\"";

    // Try to get the descriptor handle
    void* temporaryPointer = 0;
    sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                   &temporaryPointer,
                                   OCI_HTYPE_DESCRIBE, 0, 0 );
    if ( status != OCI_SUCCESS ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not allocate a describe handle",
                                    "ISchema::viewHandle" );
    }

    OCIDescribe* describeHandle = static_cast< OCIDescribe* >( temporaryPointer );

    status = OCIDescribeAny( m_sessionProperties.serviceContextHandle(),
                             m_sessionProperties.errorHandle(),
                             const_cast<char *>( fullViewName.c_str() ),
                             ::strlen( fullViewName.c_str() ),
                             OCI_OTYPE_NAME, OCI_DEFAULT,
                             OCI_PTYPE_VIEW,
                             describeHandle );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "retrieving the describe handle of " + fullViewName );
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
      if ( errorHandler.lastErrorCode() != 4043 ) {
        coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
        log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
        throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                      "Could not retrieve a describe handle",
                                      "ISchema::viewHandle" );
      }
      else {
        throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(),
                                                fullViewName );
      }
    }

    boost::mutex::scoped_lock lock( m_viewMutex );
    iView = m_views.find( viewName );
    if ( iView != m_views.end() && iView->second != 0 ) { // Another thread was faster...
      OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
    }
    else {
      // The view exists. Create the view object and insert it into the map
      coral::OracleAccess::View* view = new coral::OracleAccess::View( m_sessionProperties,
                                                                       viewName,
                                                                       describeHandle );
      if ( iView == m_views.end() )
        iView = m_views.insert( std::make_pair( viewName, view ) ).first;
      else
        iView->second = view;
    }
    return *( iView->second );
  }

  return *( iView->second );
}



std::set<std::string>
coral::OracleAccess::Schema::listViews() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ISchema::listViews" );
  std::set<std::string> result;
  if ( ! m_viewsReadFromDataDictionary )
    this->readViewsFromDataDictionary();

  boost::mutex::scoped_lock lock( m_viewMutex );
  for ( std::map< std::string, coral::OracleAccess::View*>::const_iterator iView = m_views.begin();
        iView != m_views.end(); ++iView ) {
    result.insert( iView->first );
  }

  return result;
}


void
coral::OracleAccess::Schema::reactOnEndOfTransaction()
{
  boost::mutex::scoped_lock lock( m_tableMutex );
  for ( std::map< std::string, coral::OracleAccess::Table* >::iterator iTable = m_tables.begin();
        iTable != m_tables.end(); ++iTable )
    if ( iTable->second )
      delete iTable->second;

  m_tables.clear();

  m_tablesReadFromDataDictionary = false;

  boost::mutex::scoped_lock lockViews( m_viewMutex );
  for ( std::map< std::string, coral::OracleAccess::View* >::iterator iView = m_views.begin();
        iView != m_views.end(); ++iView )
    if ( iView->second )
      delete iView->second;

  m_views.clear();

  m_viewsReadFromDataDictionary = false;
}


void
coral::OracleAccess::Schema::readTablesFromDataDictionary() const
{
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "SELECT TABLE_NAME FROM ALL_TABLES WHERE OWNER=:\"owner\"" );
  statement.setNumberOfPrefetchedRows( 100 );
  coral::AttributeList bindData;
  bindData.extend<std::string>( "owner" );
  bindData.begin()->data<std::string>() = m_sessionProperties.schemaName();
  statement.execute( bindData );
  coral::AttributeList output;
  output.extend<std::string>( "TABLE_NAME" );
  const std::string& tableName = output.begin()->data<std::string>();
  statement.defineOutput( output );

  boost::mutex::scoped_lock lock( m_tableMutex );

  while( statement.fetchNext() ) {
    if ( m_tables.find( tableName ) == m_tables.end() ) {
      m_tables[tableName] = 0;
    }
  }
  m_tablesReadFromDataDictionary = true;
}


void
coral::OracleAccess::Schema::readViewsFromDataDictionary() const
{
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "SELECT VIEW_NAME FROM ALL_VIEWS WHERE OWNER=:\"owner\"" );
  statement.setNumberOfPrefetchedRows( 100 );
  coral::AttributeList bindData;
  bindData.extend<std::string>( "owner" );
  bindData.begin()->data<std::string>() = m_sessionProperties.schemaName();
  statement.execute( bindData );
  coral::AttributeList output;
  output.extend<std::string>( "VIEW_NAME" );
  const std::string& viewName = output.begin()->data<std::string>();
  statement.defineOutput( output );

  boost::mutex::scoped_lock lock( m_viewMutex );

  while( statement.fetchNext() ) {
    if ( m_views.find( viewName ) == m_views.end() ) {
      m_views[viewName] = 0;
    }
  }
  m_viewsReadFromDataDictionary = true;
}
