// Workaround for Windows compilation error (vc71 and vc9).
// If the WTypes.h header is not included at the beginning, another header 
// defines symbol BOOLEAN which is needed by RpcNdr.h (included by WTypes.h).
#ifdef WIN32
#include <WTypes.h>
#endif

#include "Session.h"
#include "SessionPropertiesProxy.h"
#include "ConnectionProperties.h"
#include "DomainProperties.h"
#include "MonitorController.h"
#include "OracleErrorHandler.h"
#include "Transaction.h"
#include "Schema.h"
#include "OracleStatement.h"

#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/IMonitoringService.h"

#include "CoralCommon/MonitoringEventDescription.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/MessageStream.h"

#include "CoralKernel/Service.h"
#include "CoralKernel/Context.h"

#include "oci.h"

#include <locale>

coral::OracleAccess::Session::Session(  coral::IDevConnection& connection,
                                        coral::OracleAccess::ConnectionProperties& connectionProperties,
                                        const std::string& schemaName,
                                        coral::AccessMode mode )
  : coral::IDevSession( connection )
  , m_sessionHandle( 0 )
  , m_schema( 0 )
  , m_transaction( 0 )
  , m_mutex()
  , m_mutexForSchemas()
  , m_schemas()
{
  m_properties = new coral::OracleAccess::SessionProperties( connectionProperties, schemaName, static_cast<coral::ISession&>( *this ), mode == coral::ReadOnly);
  m_monitorController = new coral::OracleAccess::MonitorController( *m_properties );
}


coral::OracleAccess::Session::~Session()
{
  if ( this->isUserSessionActive() ) this->endUserSession();
  delete m_monitorController;
  delete m_properties;
}


coral::IMonitoring&
coral::OracleAccess::Session::monitoring()
{
  return *m_monitorController;
}


void
coral::OracleAccess::Session::startUserSession( const std::string& userName,
                                                const std::string& password )
{
  if ( ! this->isValid() ) return;

  boost::mutex::scoped_lock lock( m_mutex );

  // Creating an error handler to be used in the subsequent calls.
  coral::OracleAccess::OracleErrorHandler errorHandler( m_properties->errorHandle() );

  // Creating a service context handle
  void* temporaryPointer = 0;
  sword status = OCIHandleAlloc( m_properties->environmentHandle(),
                                 &temporaryPointer,
                                 OCI_HTYPE_SVCCTX, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::SessionException( "Could not allocate an OCI service context handle",
                                   "ISession::connectAsUser",
                                   m_properties->domainProperties().service()->name() );
  }
  OCISvcCtx* serviceContextHandle = static_cast< OCISvcCtx* >( temporaryPointer );

  // Setting the server to the service context
  status = OCIAttrSet( serviceContextHandle, OCI_HTYPE_SVCCTX, 
                       m_properties->serverHandle(), 0, OCI_ATTR_SERVER,
                       m_properties->errorHandle() );
  if ( status != OCI_SUCCESS ) {
    errorHandler.handleCase( status, "setting the server to the service context" );
    coral::MessageStream log( m_properties->domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      OCIHandleFree( serviceContextHandle, OCI_HTYPE_SVCCTX );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
      throw coral::SessionException( "Could not set the server to the service context",
                                     "ISession::connectAsUser",
                                     m_properties->domainProperties().service()->name() );
    }
    else {
      log << coral::Warning << errorHandler.message() << coral::MessageStream::endmsg;
    }
  }

  // Updating the context handle
  m_properties->setServiceContextHandle( serviceContextHandle );

  // Allocating the user session handle
  temporaryPointer = 0;
  status = OCIHandleAlloc( m_properties->environmentHandle(), &temporaryPointer,
                           OCI_HTYPE_SESSION, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::SessionException( "Could not allocate a user session handle",
                                   "ISession::startUserSession",
                                   m_properties->domainProperties().service()->name() );
  }
  m_sessionHandle = static_cast< OCISession* >( temporaryPointer );

  // Setting the user session to the service context
  status = OCIAttrSet( m_properties->serviceContextHandle(), OCI_HTYPE_SVCCTX, 
                       m_sessionHandle, 0, OCI_ATTR_SESSION, m_properties->errorHandle() );
  if ( status != OCI_SUCCESS ) {
    errorHandler.handleCase( status, "setting the user session to the service context" );
    coral::MessageStream log( m_properties->domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
      OCIHandleFree( m_sessionHandle, OCI_HTYPE_SESSION );
      m_sessionHandle = 0;
      throw coral::SessionException( "Could not allocate a user session handle",
                                     "ISession::startUserSession",
                                     m_properties->domainProperties().service()->name() );
    }
    else {
      log << coral::Warning << errorHandler.message() << coral::MessageStream::endmsg;
    }
  }

  // Setting the user name and the password
  status = OCIAttrSet( m_sessionHandle, OCI_HTYPE_SESSION,
                       const_cast< char *>( userName.c_str() ), ::strlen( userName.c_str() ),
                       OCI_ATTR_USERNAME, m_properties->errorHandle() );
  if ( status == OCI_SUCCESS ) {
    status = OCIAttrSet( m_sessionHandle, OCI_HTYPE_SESSION,
                         const_cast< char *>( password.c_str() ), ::strlen( password.c_str() ),
                         OCI_ATTR_PASSWORD, m_properties->errorHandle() );
  }
  if ( status != OCI_SUCCESS ) {
    errorHandler.handleCase( status, "setting the user name and the password" );
    coral::MessageStream log( m_properties->domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
      OCIHandleFree( m_sessionHandle, OCI_HTYPE_SESSION );
      m_sessionHandle = 0;
      throw coral::SessionException( "Could not set user name and password",
                                     "ISession::startUserSession",
                                     m_properties->domainProperties().service()->name() );
    }
    else {
      log << coral::Warning << errorHandler.message() << coral::MessageStream::endmsg;
    }
  }

  // Authenticating
  status = OCISessionBegin( m_properties->serviceContextHandle(),
                            m_properties->errorHandle(),
                            m_sessionHandle,
                            OCI_CRED_RDBMS, OCI_DEFAULT);
  if ( status != OCI_SUCCESS ) {
    errorHandler.handleCase( status, "authenticating" );
    coral::MessageStream log( m_properties->domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
      OCIHandleFree( m_sessionHandle, OCI_HTYPE_SESSION );
      m_sessionHandle = 0;
      if ( errorHandler.lastErrorCode() == 1017 ) { // Failed to authenticate
        throw coral::AuthenticationFailureException( m_properties->domainProperties().service()->name() );
      }
      else if ( true ) { // Temporarily unavailable. Retry!
        throw coral::StartSessionException( m_properties->domainProperties().service()->name(),
                                            "ISession::startUserSession" );
      }
      else { // all other failures
        throw coral::SessionException( errorHandler.message(),
                                       "ISession::startUserSession",
                                       m_properties->domainProperties().service()->name() );
      }
    }
    else {
      log << coral::Warning << errorHandler.message() << coral::MessageStream::endmsg;
    }
  }

  // Enable sql trace
  char* cerntrace_on = ::getenv( "CORAL_ORA_CERN_TRACE_ON" );
  if ( cerntrace_on ) {
    std::string sqlStatement = "BEGIN CERN_TRACE.CSTART_TRACE( 2 ); END;";
    coral::OracleAccess::OracleStatement statement( *m_properties, sqlStatement );
    if ( ! statement.execute( coral::AttributeList() ) ) {
      coral::MessageStream log( m_properties->domainProperties().service()->name() );
      log << coral::Warning << "Could not enable SQL TRACE " << coral::MessageStream::endmsg;
    }
  }
  else {
    char* sqltrace_on = ::getenv( "CORAL_ORA_SQL_TRACE_ON" );
    if ( sqltrace_on ) {
      std::string sqlStatement;
      // SQL trace identifier
      char* sqltrace_id = ::getenv( "CORAL_ORA_SQL_TRACE_IDENTIFIER" );
      if ( sqltrace_id ) 
      {
        sqlStatement = "ALTER SESSION SET tracefile_identifier='";
        sqlStatement += ::getenv( "CORAL_ORA_SQL_TRACE_IDENTIFIER" );
        sqlStatement += "'";
        coral::OracleAccess::OracleStatement statement( *m_properties, sqlStatement );
        if ( ! statement.execute( coral::AttributeList() ) ) {
          coral::MessageStream log( m_properties->domainProperties().service()->name() );
          log << coral::Warning << "Could not set the SQL TRACE IDENTIFIER " << coral::MessageStream::endmsg;
        }
      }
      sqlStatement = "ALTER SESSION SET SQL_TRACE=TRUE";
      if ( std::string( sqltrace_on ) == "3106" )
        sqlStatement = "ALTER SESSION SET EVENTS '3106 TRACE NAME ERRORSTACK LEVEL 10'";
      if ( std::string( sqltrace_on ) == "10046" )
        sqlStatement = "ALTER SESSION SET EVENTS '10046 TRACE NAME CONTEXT FOREVER, LEVEL 12'";
      if ( std::string( sqltrace_on ) == "10053" )
        sqlStatement = "ALTER SESSION SET EVENTS '10053 TRACE NAME CONTEXT FOREVER, LEVEL 1'";
      if ( std::string( sqltrace_on ) == "10132" )
        sqlStatement = "ALTER SESSION SET EVENTS '10132 TRACE NAME CONTEXT FOREVER, LEVEL 12'";
      coral::OracleAccess::OracleStatement statement( *m_properties, sqlStatement );
      if ( ! statement.execute( coral::AttributeList() ) ) {
        coral::MessageStream log( m_properties->domainProperties().service()->name() );
        log << coral::Warning << "Could not enable SQL TRACE " << coral::MessageStream::endmsg;
      }
    }
  }

  // Create new schema and transaction objects.
  m_schema = new coral::OracleAccess::Schema( *m_properties );
  m_transaction = new coral::OracleAccess::Transaction( *m_properties, *m_schema );

  // Record the beginning of the session
  if ( m_properties->monitoringService() ) {
    m_properties->monitoringService()->record( "oracle://" + m_properties->connectionString() + "/" + m_properties->schemaName(),
                                               coral::monitor::Session,
                                               coral::monitor::Info,
                                               monitoringEventDescription.sessionBegin() );
  }
}


bool
coral::OracleAccess::Session::isUserSessionActive() const
{
  if ( ! this->isValid() ) return false;
  boost::mutex::scoped_lock lock( m_mutex );
  return ( m_sessionHandle != 0 );
}


void
coral::OracleAccess::Session::endUserSession()
{
  if ( ! this->isUserSessionActive() ) return;

  // Abort any active transaction
  if ( this->transaction().isActive() ) this->transaction().rollback();

  boost::mutex::scoped_lock lock( m_mutex );

  // Stop the tracing
  char* cerntrace_on = ::getenv( "CORAL_ORA_CERN_TRACE_ON" );
  if ( cerntrace_on ) {
    this->transaction().start();
    std::string sqlStatement = "BEGIN CERN_TRACE.CSTOP_TRACE('" + std::string( cerntrace_on ) + "'); END;";
    coral::OracleAccess::OracleStatement statement( *m_properties, sqlStatement );
    if ( ! statement.execute( coral::AttributeList() ) ) {
      coral::MessageStream log( m_properties->domainProperties().service()->name() );
      log << coral::Warning << "Could not enable SQL TRACE" << coral::MessageStream::endmsg;
    }
    this->transaction().commit();
  }


  if ( m_transaction ) {
    delete m_transaction;
    m_transaction = 0;
  }

  boost::mutex::scoped_lock lockS( m_mutexForSchemas );
  for ( std::map< std::string,
          std::pair< coral::OracleAccess::SessionPropertiesProxy*,
          coral::OracleAccess::Schema* > >::iterator iSchema = m_schemas.begin();
        iSchema != m_schemas.end(); ++iSchema ) {
    delete iSchema->second.second;
    delete iSchema->second.first;
  }
  m_schemas.clear();
  if ( m_schema ) {
    delete m_schema;
    m_schema = 0;
  }

  // Close the user session
  OCISessionEnd( m_properties->serviceContextHandle(),
                 m_properties->errorHandle(),
                 m_sessionHandle, OCI_DEFAULT );
  OCIHandleFree( m_sessionHandle, OCI_HTYPE_SESSION );
  m_sessionHandle = 0;

  // Record the ending of the session
  if ( m_properties->monitoringService() ) {
    m_properties->monitoringService()->record( "oracle://" + m_properties->connectionString() + "/" + m_properties->schemaName(),
                                               coral::monitor::Session,
                                               coral::monitor::Info,
                                               monitoringEventDescription.sessionEnd() );
  }

  OCISvcCtx* serviceContextHandle = m_properties->serviceContextHandle();
  OCIHandleFree( serviceContextHandle, OCI_HTYPE_SVCCTX );
    
  m_properties->setServiceContextHandle( 0 );
}


coral::ITransaction&
coral::OracleAccess::Session::transaction()
{
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "ISession::transaction" );
  boost::mutex::scoped_lock lock( m_mutex );
  return *m_transaction;
}


coral::ISchema&
coral::OracleAccess::Session::nominalSchema()
{
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );
  boost::mutex::scoped_lock lock( m_mutex );
  return *m_schema;
}


coral::ISchema&
coral::OracleAccess::Session::schema( const std::string& _schemaName )
{
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );

  // Form the schema name correctly by setting everything to upper case
  std::string schemaName = _schemaName;
  for ( std::string::size_type i = 0; i < schemaName.size(); ++i ) {
    schemaName[i] = std::toupper( schemaName[i], std::locale::classic() );
  }

  boost::mutex::scoped_lock lock( m_mutexForSchemas );

  // Check first is the schema with the corresponding name exists in the map
  std::map< std::string, std::pair< coral::OracleAccess::SessionPropertiesProxy*, coral::OracleAccess::Schema* > >::iterator iSchema =
    m_schemas.find( schemaName );
  if ( iSchema != m_schemas.end() ) return *( iSchema->second.second );

  // Check in the database if a schema with such a name exists
  void* temporaryPointer = 0;
  sword status = OCIHandleAlloc( m_properties->environmentHandle(),
                                 &temporaryPointer,
                                 OCI_HTYPE_DESCRIBE, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::SessionException( m_properties->domainProperties().service()->name(),
                                  "Could not allocate a describe handle",
                                  "ISession::schema" );
  }
  OCIDescribe* describeHandle = static_cast<OCIDescribe*>(temporaryPointer);

  OCIDescribeAny( m_properties->serviceContextHandle(),
                  m_properties->errorHandle(),
                  const_cast<char *>( schemaName.c_str() ),
                  ::strlen( schemaName.c_str() ),
                  OCI_OTYPE_NAME,
                  OCI_DEFAULT,
                  OCI_PTYPE_SCHEMA,
                  describeHandle );
  coral::OracleAccess::OracleErrorHandler errorHandler( m_properties->errorHandle() );
  errorHandler.handleCase( status, "retrieving the describe handle of schema " + schemaName );
  OCIHandleFree( describeHandle, OCI_HTYPE_DESCRIBE );
  if ( errorHandler.lastErrorCode() == 4043 ) {
    coral::MessageStream log( m_properties->domainProperties().service()->name() );
    log << coral::Debug << errorHandler.message() << coral::MessageStream::endmsg;
    throw coral::InvalidSchemaNameException( m_properties->domainProperties().service()->name() );
  }

  // The schema exists. Insert it into the map of the known ones
  coral::OracleAccess::SessionPropertiesProxy* 
      properties = new coral::OracleAccess::SessionPropertiesProxy( *m_properties, schemaName );
  
  coral::OracleAccess::Schema* schema = new coral::OracleAccess::Schema( *properties );

  m_schemas.insert( std::make_pair( schemaName, std::make_pair( properties, schema ) ) );

  return *schema;
}
