#ifndef ORACLEACCESS_SESSION_H
#define ORACLEACCESS_SESSION_H

#include "RelationalAccess/AccessMode.h"
#include "CoralCommon/IDevSession.h"
#include <map>
#include "CoralBricks/ThreadTraits.h"

struct OCIServer;
struct OCISession;

namespace coral {

  namespace OracleAccess {

    class ConnectionProperties;
    class SessionProperties;
    class SessionPropertiesProxy;
    class MonitorController;
    class Transaction;
    class Schema;

    /**
     * Class Session
     *
     * Implementation of the ISession interface for the OracleAccess module
     */
    class Session : virtual public coral::IDevSession
    {
    public:
      /// Constructor
      Session( coral::IDevConnection& connection,
               ConnectionProperties& connectionProperties,
               const std::string& schemaName,
               coral::AccessMode mode );

      /// Destructor
      ~Session();

      /**
       * Returns the reference to the underlying IMonitoring object.
       */
      coral::IMonitoring& monitoring();
  
      /**
       * Authenticates with the database server using a user/password pair.
       * If the authentication fails an AuthenticationFailureException is thrown.
       */
      void startUserSession( const std::string& userName,
                             const std::string& password );

      /**
       * Terminates a user session without dropping the connection to the database server
       */
      void endUserSession();

      /**
       * Returns the status of a user session.
       */
      bool isUserSessionActive() const;
  
      /**
       * Returns the corresponding ITransaction object.
       * If a connection is not yet established, a ConnectionNotActiveException is thrown.
       */
      coral::ITransaction& transaction();
  
      /**
       * Returns a reference to the working ISchema object.
       * If a connection is not yet established, a ConnectionNotActiveException is thrown.
       */
      coral::ISchema& nominalSchema();

      /**
       * Returns a reference to the ISchema object corresponding to the specified name.
       * If a connection is not yet established, a ConnectionNotActiveException is thrown.
       * If no schema exists corresponding to the specified name an InvalidSchemaNameException is thrown.
       */
      coral::ISchema& schema( const std::string& schemaName );

      /**
       * Returns the technology name for the remote session.
       * For plugins establishing a database connection through a middle-tier 
       * (e.g. CoralAccess), this is discovered when establishing the remote session.
       */
      std::string remoteTechnologyName() const
      {
        return "Oracle"; // same as Domain.flavorName()
      }

    private:
      /// The properties of the session
      SessionProperties* m_properties;

      /// The monitoring controller
      MonitorController* m_monitorController;

      /// The session handle
      OCISession*        m_sessionHandle;

      /// The schema object
      Schema*            m_schema;

      /// The transaction object
      Transaction*       m_transaction;

      /// Mutex for the handles
      mutable boost::mutex m_mutex;

      /// Mutex for the schemas
      mutable boost::mutex m_mutexForSchemas;

      /// A map of the other schemas
      std::map< std::string, std::pair< SessionPropertiesProxy*, Schema* > > m_schemas;
    };

  }

}

#endif
