#ifndef ORACLE_SESSION_PROPERTIES_H
#define ORACLE_SESSION_PROPERTIES_H

#include "ISessionProperties.h"
#include "CoralBricks/ThreadTraits.h"

struct OCIEnv;
struct OCIError;
struct OCIServer;
struct OCISvcCtx;

namespace coral {

  class ISession;

  namespace OracleAccess {

    class ConnectionProperties;

    /**
     * Class SessionProperties
     *
     * An implementation of the ISessionProperties interface.
     */

    class SessionProperties : virtual public ISessionProperties
    {
    public:
      /// Constructor
      SessionProperties( ConnectionProperties& connectionProperties,
                         const std::string& schemaName,
                         coral::ISession& session,
                         bool readOnly );

      /// Destructor
      virtual ~SessionProperties() {}

      /// Sets the service context handle
      void setServiceContextHandle( OCISvcCtx* serviceContextHandle );

      /// Sets the monitoring service
      void setMonitoringService( coral::monitor::IMonitoringService* monitoringService );

      /// Returns the domain properties
      const DomainProperties& domainProperties() const;

      /// Returns the connection string
      std::string connectionString() const;

      /// Returns the session object
      coral::ISession& session();

      /// Returns the type converter
      coral::ITypeConverter& typeConverter();
      const coral::ITypeConverter& typeConverter() const;

      /// Returns the OCI environment handle
      OCIEnv* environmentHandle() const;

      /// Returns the OCI error handle
      OCIError* errorHandle() const;

      /// Returns the OCI server handle
      OCIServer* serverHandle() const;

      /// Returns the OCI service context handle
      OCISvcCtx* serviceContextHandle() const;

      /// Returns the monitoring service
      coral::monitor::IMonitoringService* monitoringService() const;

      /// Returns the transaction state
      bool isTransactionActive() const;

      /// Returns the server version
      int serverVersion() const;

      /// Returns the corresponding schema
      coral::ISchema& schema() const;

      /// Returns the schema name
      std::string schemaName() const;

      /// Returns the readOnly flag
      bool isReadOnly() const;

    private:
      /// The domain properties
      const DomainProperties& m_domainProperties;

      /// The connection string
      std::string       m_connectionString;

      /// The type converter
      coral::ITypeConverter& m_typeConverter;

      /// The environment handle
      OCIEnv*           m_environmentHandle;

      /// The error handle
      OCIError*         m_errorHandle;

      /// The server handle
      OCIServer*        m_serverHandle;

      /// The server major version
      int               m_serverVersion;

      /// The session reference
      coral::ISession&  m_session;

      /// The OCI service context handle
      OCISvcCtx*        m_serviceContextHandle;

      /// The monitoring service
      coral::monitor::IMonitoringService* m_monitoringService;

      /// The schema name
      std::string       m_schemaName;

      /// The read-only flag
      bool              m_isReadOnly;

      /// The mutex lock
      mutable boost::mutex  m_mutex;
    };

  }

}


// Inline methods
inline const coral::OracleAccess::DomainProperties&
coral::OracleAccess::SessionProperties::domainProperties() const
{
  return m_domainProperties;
}


inline std::string
coral::OracleAccess::SessionProperties::connectionString() const
{
  return m_connectionString;
}


inline coral::ITypeConverter&
coral::OracleAccess::SessionProperties::typeConverter()
{
  return m_typeConverter;
}


inline const coral::ITypeConverter&
coral::OracleAccess::SessionProperties::typeConverter() const
{
  return m_typeConverter;
}


inline OCIEnv*
coral::OracleAccess::SessionProperties::environmentHandle() const
{
  return m_environmentHandle;
}


inline OCIError*
coral::OracleAccess::SessionProperties::errorHandle() const
{
  return m_errorHandle;
}


inline OCIServer*
coral::OracleAccess::SessionProperties::serverHandle() const
{
  return m_serverHandle;
}


inline int
coral::OracleAccess::SessionProperties::serverVersion() const
{
  return m_serverVersion;
}


inline void
coral::OracleAccess::SessionProperties::setServiceContextHandle( OCISvcCtx* serviceContextHandle )
{
  boost::mutex::scoped_lock lock(m_mutex);
  m_serviceContextHandle = serviceContextHandle;
}


inline OCISvcCtx*
coral::OracleAccess::SessionProperties::serviceContextHandle() const
{
  boost::mutex::scoped_lock lock(m_mutex);
  return m_serviceContextHandle;
}


inline void
coral::OracleAccess::SessionProperties::setMonitoringService( coral::monitor::IMonitoringService* monitoringService )
{
  boost::mutex::scoped_lock lock(m_mutex);
  m_monitoringService = monitoringService;
}


inline coral::monitor::IMonitoringService*
coral::OracleAccess::SessionProperties::monitoringService() const
{
  boost::mutex::scoped_lock lock(m_mutex);
  return m_monitoringService;
}


inline coral::ISession&
coral::OracleAccess::SessionProperties::session()
{
  return m_session;
}


inline bool
coral::OracleAccess::SessionProperties::isReadOnly() const
{
  return m_isReadOnly;
}


inline std::string
coral::OracleAccess::SessionProperties::schemaName() const
{
  return m_schemaName;
}

#endif
