#include "Table.h"
#include "ISessionProperties.h"
#include "DomainProperties.h"
#include "TableDescriptionProxy.h"
#include "DataEditor.h"
#include "PrivilegeManager.h"
#include "Query.h"

#include "RelationalAccess/SessionException.h"

#include "CoralKernel/Service.h"

coral::OracleAccess::Table::Table( const coral::OracleAccess::ISessionProperties& sessionProperties,
                                   const std::string& tableName,
                                   OCIDescribe* descriptor ):
  m_sessionProperties( sessionProperties ),
  m_descriptionProxy( new coral::OracleAccess::TableDescriptionProxy( sessionProperties,
                                                                      descriptor,
                                                                      tableName ) ),
  m_privilegeManager( new coral::OracleAccess::PrivilegeManager( sessionProperties,
                                                                 sessionProperties.schemaName() + ".\"" + tableName + "\"" ) ),
  m_dataEditor( new coral::OracleAccess::DataEditor( sessionProperties, *m_descriptionProxy ) )
{
}


coral::OracleAccess::Table::~Table()
{
  delete m_dataEditor;
  delete m_privilegeManager;
  delete m_descriptionProxy;
}


const coral::ITableDescription&
coral::OracleAccess::Table::description() const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ITable::description" );
  return static_cast< const coral::ITableDescription& >( *m_descriptionProxy );
}


coral::ITableSchemaEditor&
coral::OracleAccess::Table::schemaEditor()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ITable::schemaEditor" );

  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ITable::schemaEditor" );

  return static_cast< coral::ITableSchemaEditor& >( *m_descriptionProxy );
}


coral::ITableDataEditor&
coral::OracleAccess::Table::dataEditor()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ITable::dataEditor" );
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ITable::dataEditor" );

  return *m_dataEditor;
}


coral::ITablePrivilegeManager&
coral::OracleAccess::Table::privilegeManager()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),
                                                "ITable::privilegeManager" );
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ITable::privilegeManager" );

  return *m_privilegeManager;
}


coral::IQuery*
coral::OracleAccess::Table::newQuery() const
{
  return new coral::OracleAccess::Query( m_sessionProperties,
                                         m_descriptionProxy->name() );
}
