#ifdef WIN32
#include <WTypes.h> // AV 17.04.2008 include this first on Win (bug #35683)
#endif

#include "TableDescriptionProxy.h"
#include "ISessionProperties.h"
#include "DomainProperties.h"
#include "OracleErrorHandler.h"
#include "OracleStatement.h"
#include "OracleTableBuilder.h"
#include "ColumnProxy.h"

#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/ITypeConverter.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "CoralBase/Date.h"
#include "CoralBase/TimeStamp.h"
#include "CoralBase/MessageStream.h"

#include "CoralKernel/Service.h"

#include "oci.h"

coral::OracleAccess::TableDescriptionProxy::TableDescriptionProxy( const coral::OracleAccess::ISessionProperties& properties,
                                                                   OCIDescribe* describeHandle,
                                                                   const std::string& tableName ):
  m_sessionProperties( properties ),
  m_describeHandle( describeHandle ),
  m_description( new coral::TableDescription( properties.domainProperties().service()->name() ) ),
  m_columnDescriptionRead( false ),
  m_columns(),
  m_uniqueConstraintsRead( false ),
  m_primaryKeyRead( false ),
  m_indicesRead( false ),
  m_foreignKeysRead( false ),
  m_mutex()
{
  m_description->setName( tableName );
}


coral::OracleAccess::TableDescriptionProxy::~TableDescriptionProxy()
{
  for ( std::vector< coral::OracleAccess::ColumnProxy* >::iterator iColumn = m_columns.begin();
        iColumn != m_columns.end(); ++iColumn ) delete *iColumn;

  delete m_description;

  OCIHandleFree( m_describeHandle, OCI_HTYPE_DESCRIBE );
}


std::string
coral::OracleAccess::TableDescriptionProxy::name() const
{
  return m_description->name();
}


std::string
coral::OracleAccess::TableDescriptionProxy::type() const
{
  return m_description->type();
}


void
coral::OracleAccess::TableDescriptionProxy::readColumnDescription()
{
  boost::mutex::scoped_lock lock( m_mutex );
  if ( m_columnDescriptionRead ) return;

  // Retrieving the describe parameter handle
  OCIParam* parameterHandle = 0;
  sword status = OCIAttrGet( m_describeHandle, OCI_HTYPE_DESCRIBE,
                             &parameterHandle, 0, OCI_ATTR_PARAM,
                             m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Retrieving the describe parameter for a table" );
    coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
    log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the OCI describe parameter for a table",
                                  "TableDescriptionProxy::readColumnDescription" );
  }

  // Retrieving the column list
  OCIParam* columnList = 0;
  status = OCIAttrGet( parameterHandle, OCI_DTYPE_PARAM,
                       &columnList, 0, OCI_ATTR_LIST_COLUMNS,
                       m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Retrieving the column list for a table" );
    coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
    log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the column list for a table",
                                  "TableDescriptionProxy::readColumnDescription" );
  }


  // Retrieving the number of columns
  ub2 numberOfColumns = 0;
  status = OCIAttrGet( parameterHandle, OCI_DTYPE_PARAM,
                       &numberOfColumns, 0, OCI_ATTR_NUM_COLS,
                       m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Retrieving the number of columns in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
    coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
    log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the number of columns in a table",
                                  "TableDescriptionProxy::readColumnDescription" );
  }



  for ( ub4 iColumn = 1; iColumn <= numberOfColumns; ++iColumn )
  {
    // Retrieving the column parameter
    void* temporaryPointer = 0;
    status = OCIParamGet( columnList, OCI_DTYPE_PARAM, m_sessionProperties.errorHandle(),
                          &temporaryPointer, iColumn );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the parameter of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the parameter of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }
    OCIParam* columnParameter = static_cast< OCIParam* >( temporaryPointer );

    // Retrieving the column name
    text* textPlaceHolder = 0;
    ub4 textSize = 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &textPlaceHolder, &textSize,
                         OCI_ATTR_NAME,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the name of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the name of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }
    std::ostringstream osColumnName;
    osColumnName << textPlaceHolder;
    std::string columnName = osColumnName.str();
    columnName = columnName.substr( 0, textSize );

    // Retrieving the column type
    ub2 columnType = 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &columnType, 0,
                         OCI_ATTR_DATA_TYPE,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the type of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the type of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }

    // Retrieving the precision of the column
    ub1 precision = 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &precision, 0,
                         OCI_ATTR_PRECISION,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the precision of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the precision of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }
    
    // Retrieving the scale of the column
    sb1 scale = 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &scale, 0,
                         OCI_ATTR_SCALE,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the scale of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the scale of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }

    // Retrieving the data size of the column
    ub4 columnSize = 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &columnSize, 0,
                         OCI_ATTR_DATA_SIZE,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the size of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the size of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }

    // Retrieving the nullness of the column
    ub1 isNull= 0;
    status = OCIAttrGet( columnParameter, OCI_DTYPE_PARAM,
                         &isNull, 0,
                         OCI_ATTR_IS_NULL,
                         m_sessionProperties.errorHandle() );
    if ( status != OCI_SUCCESS ) {
      coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
      errorHandler.handleCase( status, "Retrieving the null-ness of a column in table " + m_sessionProperties.schemaName() + "." + m_description->name() );
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << errorHandler.message() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the null-ness of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }

    std::string columnTypeName = "";
    if ( columnType == OCI_TYPECODE_VARCHAR2 ||
         columnType == OCI_TYPECODE_VARCHAR ||
         columnType == OCI_TYPECODE_CLOB ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(std::string) );
    }
    else if ( columnType == OCI_TYPECODE_DATE ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(coral::Date) );
    }
    else if ( columnType == OCI_TYPECODE_TIMESTAMP ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(coral::TimeStamp) );
    }
    else if ( columnType == OCI_TYPECODE_TIMESTAMP_TZ ) {
      columnTypeName = "TIMESTAMP WITH TIME ZONE";
    }
    else if ( columnType == OCI_TYPECODE_CHAR ) {
      if ( columnSize == 1 ) {
        columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(char) );
      }
      else {
        columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(std::string) );
      }
    }
    else if ( columnType == OCI_TYPECODE_NUMBER ) {
      std::ostringstream os;
      if ( scale == -127 ) { // This is a float
        os << "FLOAT(" << static_cast<int>(precision) << ")";
      }
      else {
        os << "NUMBER";
        if ( precision != 0 ) {
          os << "(" << static_cast<int>(precision);
          if ( scale != 0 ) {
            os << "," << static_cast<int>(scale);
          }
          os << ")";
        }
      }
      std::string sqlType = os.str();
      columnTypeName = m_sessionProperties.typeConverter().cppTypeForSqlType( sqlType );
    }
    else if ( columnType == OCI_TYPECODE_BFLOAT ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(float) );
    }
    else if ( columnType == OCI_TYPECODE_BDOUBLE ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(double) );
    }
    else if ( columnType == OCI_TYPECODE_BLOB ) {
      columnTypeName = coral::AttributeSpecification::typeNameForId( typeid(coral::Blob) );
    }
    else {
      coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );
      log << coral::Error << "Cannot handle the type of column " << columnName << " in table "
          << m_sessionProperties.schemaName() << "." << m_description->name() << coral::MessageStream::endmsg;    
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not handle the type of a column in a table",
                                    "TableDescriptionProxy::readColumnDescription" );
    }

    bool fixedSize = true;
    int size = 0;
    if ( columnTypeName == coral::AttributeSpecification::typeNameForId( typeid(std::string) ) ) {
      size = columnSize;
      if ( columnType == OCI_TYPECODE_VARCHAR2 ||
           columnType == OCI_TYPECODE_VARCHAR ||
           columnType == OCI_TYPECODE_CLOB ) fixedSize = false;
      if ( columnType == OCI_TYPECODE_CLOB ) size = 100000;
    }
    if ( columnTypeName == coral::AttributeSpecification::typeNameForId( typeid(coral::TimeStamp) ) ) {
      size = scale;
    }
    m_description->insertColumn( columnName, columnTypeName, size, fixedSize );

    if ( isNull == 0 ) {
      m_description->setNotNullConstraint( columnName );
    }
  }

  /// Construct the column proxies.
  for ( int i = 0; i < numberOfColumns; ++i )
    m_columns.push_back( new coral::OracleAccess::ColumnProxy( m_description->columnDescription( i ),
                                                               *this ) );

  m_columnDescriptionRead = true;
}


void
coral::OracleAccess::TableDescriptionProxy::refreshConstraints() const
{
  const_cast< coral::OracleAccess::TableDescriptionProxy *>( this )->readColumnDescription();

  {
    boost::mutex::scoped_lock lock( m_mutex );
    if ( m_uniqueConstraintsRead ) return;

    std::ostringstream os;
    os << "SELECT ALL_CONSTRAINTS.CONSTRAINT_NAME AS CONSTRAINT_NAME, "
       << "ALL_CONS_COLUMNS.COLUMN_NAME AS COLUMN_NAME, "
       << "ALL_INDEXES.TABLESPACE_NAME AS TABLESPACE_NAME "
       << "FROM ALL_CONSTRAINTS, ALL_CONS_COLUMNS, ALL_INDEXES "
       << "WHERE ALL_CONSTRAINTS.OWNER = :\"OWNER\" "
       << "AND ALL_CONSTRAINTS.TABLE_NAME = :\"TABLE_NAME\" "
       << "AND ALL_CONSTRAINTS.CONSTRAINT_TYPE = 'U' "
       << "AND ALL_CONS_COLUMNS.OWNER = ALL_CONSTRAINTS.OWNER "
       << "AND ALL_CONS_COLUMNS.TABLE_NAME = ALL_CONSTRAINTS.TABLE_NAME "
       << "AND ALL_INDEXES.OWNER = ALL_CONSTRAINTS.OWNER "
       << "AND ALL_INDEXES.TABLE_NAME = ALL_CONSTRAINTS.TABLE_NAME "
       << "AND ALL_CONS_COLUMNS.CONSTRAINT_NAME = ALL_CONSTRAINTS.CONSTRAINT_NAME " // The two
       << "AND ALL_INDEXES.INDEX_NAME = ALL_CONS_COLUMNS.CONSTRAINT_NAME " // join conditions
       << "ORDER BY ALL_CONSTRAINTS.CONSTRAINT_NAME, ALL_CONS_COLUMNS.POSITION";
    coral::OracleAccess::OracleStatement statement( m_sessionProperties, os.str() );
    coral::AttributeList* bindData = new coral::AttributeList;
    bindData->extend<std::string>( "OWNER" );
    bindData->extend<std::string>( "TABLE_NAME" );
    (*bindData)[0].data<std::string>() = m_sessionProperties.schemaName();
    (*bindData)[1].data<std::string>() = m_description->name();
    if ( ! statement.execute( *bindData ) ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the unique constraints defined in a table",
                                    "TableDescriptionProxy::refreshConstraints" );
    }
    delete bindData; // I have no idea why the static crashes...
    
    coral::AttributeList output;
    output.extend<std::string>( "CONSTRAINT_NAME" );
    output.extend<std::string>( "COLUMN_NAME" );
    output.extend<std::string>( "TABLESPACE_NAME" );
    statement.defineOutput( output );
    const std::string& constraintName = output[0].data<std::string>();
    const std::string& columnName = output[1].data<std::string>();
    const std::string& tableSpaceName = output[2].data<std::string>();
    std::map< std::string, std::pair< std::string, std::vector< std::string > > > constraints;
    while( statement.fetchNext() ) {
      std::map< std::string, std::pair< std::string, std::vector< std::string > > >::iterator iConstraint = constraints.find( constraintName );
      if ( iConstraint == constraints.end() ) {
        iConstraint = constraints.insert( std::make_pair( constraintName, std::make_pair( tableSpaceName, std::vector< std::string >() ) ) ).first;
      }
      iConstraint->second.second.push_back( columnName );
    }
    
    for ( std::map< std::string, std::pair< std::string, std::vector< std::string > > >::const_iterator iConstraint = constraints.begin();
          iConstraint != constraints.end(); ++iConstraint ) {
      m_description->setUniqueConstraint( iConstraint->second.second,
                                          iConstraint->first,
                                          true,
                                          iConstraint->second.first );
    }
    
    m_uniqueConstraintsRead = true;
  }

  // The indices should be refreshed as well!!
  this->refreshIndexInfo();
  this->refreshPrimaryKeyInfo();
}


void
coral::OracleAccess::TableDescriptionProxy::refreshPrimaryKeyInfo() const
{
  const_cast< coral::OracleAccess::TableDescriptionProxy *>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );
  if ( m_primaryKeyRead ) return;

  std::ostringstream os;
  os << "SELECT "
     << "B.COLUMN_NAME AS COLUMN_NAME, "
     << "C.TABLESPACE_NAME AS TABLESPACE_NAME "
     << "FROM "
     << "ALL_CONSTRAINTS A, "
     << "ALL_CONS_COLUMNS B, "
     << "ALL_INDEXES C "
     << "WHERE "
     << "A.OWNER=:\"OWNER\" "
     << "AND A.TABLE_NAME=:\"TABLE_NAME\" "
     << "AND A.CONSTRAINT_TYPE = 'P' "
     << "AND B.OWNER = A.OWNER "
     << "AND B.TABLE_NAME = A.TABLE_NAME "
     << "AND B.CONSTRAINT_NAME = A.CONSTRAINT_NAME "
     << "AND B.CONSTRAINT_NAME = C.INDEX_NAME "
     << "AND C.OWNER = B.OWNER "
     << "AND C.TABLE_NAME = B.TABLE_NAME "
     << "ORDER BY B.POSITION";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties, os.str() );
  coral::AttributeList* bindData = new coral::AttributeList;
  bindData->extend<std::string>( "OWNER" );
  bindData->extend<std::string>( "TABLE_NAME" );
  (*bindData)[0].data<std::string>() = m_sessionProperties.schemaName();
  (*bindData)[1].data<std::string>() = m_description->name();
  if ( ! statement.execute( *bindData ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the primary key information defined in a table",
                                  "TableDescriptionProxy::refreshPrimaryKeyInfo" );
  }
  delete bindData; // I have no idea why the static crashes...

  coral::AttributeList output;
  output.extend<std::string>( "COLUMN_NAME" );
  output.extend<std::string>( "TABLESPACE_NAME" );
  statement.defineOutput( output );
  const std::string& columnName = output[0].data<std::string>();
  const std::string& tableSpaceName = output[1].data<std::string>();
  std::pair<std::string, std::vector<std::string> > pk = std::make_pair( std::string(""),
                                                                         std::vector<std::string>() );
  while( statement.fetchNext() ) {
    pk.first = tableSpaceName;
    pk.second.push_back( columnName );
  }

  if ( ! pk.first.empty() )
    m_description->setPrimaryKey( pk.second, pk.first );

  m_primaryKeyRead = true;
}


// A simple structure to hold the index information
namespace coral {
  namespace OracleAccess {
    class IndexInfo {
    public:
      IndexInfo( bool _unique = false,
                 std::string tsName = "" ): unique(_unique),
                                            tableSpaceName( tsName ),
                                            columns()
      {}
      bool unique;
      std::string tableSpaceName;
      std::vector<std::string> columns;
    };
  }
}


void
coral::OracleAccess::TableDescriptionProxy::refreshIndexInfo() const
{
  const_cast< coral::OracleAccess::TableDescriptionProxy *>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );
  if ( m_indicesRead ) return;

  std::ostringstream os;
  os << "SELECT "
     << "A.INDEX_NAME AS INDEX_NAME, "
     << "A.UNIQUENESS AS UNIQUENESS, "
     << "A.TABLESPACE_NAME AS TABLESPACE_NAME, "
     << " B.COLUMN_NAME AS COLUMN_NAME "
     << "FROM "
     << "ALL_INDEXES A, "
     << "ALL_IND_COLUMNS B "
     << "WHERE "
     << "A.OWNER =:\"OWNER\" "
     << "AND A.TABLE_NAME =:\"TABLE_NAME\" "
     << "AND A.INDEX_NAME = B.INDEX_NAME "
     << "AND B.INDEX_OWNER = A.OWNER "
     << "AND B.TABLE_NAME = A.TABLE_NAME "
     << "AND A.INDEX_NAME NOT IN "
     << "( SELECT CONSTRAINT_NAME "
     << "FROM ALL_CONSTRAINTS "
     << "WHERE "
     << "OWNER = A.OWNER "
     << "AND TABLE_NAME = A.TABLE_NAME "
     << "AND ( CONSTRAINT_TYPE = 'P' OR CONSTRAINT_TYPE = 'U' ) ) "
     << "ORDER BY A.INDEX_NAME, B.COLUMN_POSITION";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties, os.str() );
  coral::AttributeList* bindData = new coral::AttributeList;
  bindData->extend<std::string>( "OWNER" );
  bindData->extend<std::string>( "TABLE_NAME" );
  (*bindData)[0].data<std::string>() = m_sessionProperties.schemaName();
  (*bindData)[1].data<std::string>() = m_description->name();
  if ( ! statement.execute( *bindData ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the index information defined in a table",
                                  "TableDescriptionProxy::refreshPrimaryKeyInfo" );
  }
  delete bindData; // I have no idea why the static crashes...

  coral::AttributeList output;
  output.extend<std::string>( "INDEX_NAME" );
  output.extend<std::string>( "UNIQUENESS" );
  output.extend<std::string>( "TABLESPACE_NAME" );
  output.extend<std::string>( "COLUMN_NAME" );
  statement.defineOutput( output );
  const std::string& indexName = output[0].data<std::string>();
  const std::string& uniqueness = output[1].data<std::string>();
  const std::string& tableSpaceName = output[2].data<std::string>();
  const std::string& columnName = output[3].data<std::string>();

  // Read the index information
  std::map< std::string, coral::OracleAccess::IndexInfo > indices;
  while( statement.fetchNext() ) {
    std::map< std::string, IndexInfo >::iterator index = indices.find( indexName );
    if ( index == indices.end() ) {
      index = indices.insert( std::make_pair( indexName,
                                              coral::OracleAccess::IndexInfo( ( uniqueness == "UNIQUE") ? true : false,
                                                                              tableSpaceName ) ) ).first;
    }
    index->second.columns.push_back( columnName );
  }

  for ( std::map< std::string, coral::OracleAccess::IndexInfo >::const_iterator index = indices.begin();
        index != indices.end(); ++index ) {
    m_description->createIndex( index->first,
                                index->second.columns,
                                index->second.unique,
                                index->second.tableSpaceName );
  }

  m_indicesRead = true;
}



// A simple structure to hold the foreign key information
namespace coral {
  namespace OracleAccess {
    class ForeignKeyInfo {
    public:
      ForeignKeyInfo( std::string _tableName = "" ): tableName(_tableName),
                                                     columns(),
                                                     refColumns()
      {}
      std::string tableName;
      std::vector<std::string> columns;
      std::vector<std::string> refColumns;
    };
  }
}




void
coral::OracleAccess::TableDescriptionProxy::refreshForeignKeyInfo() const
{
  const_cast<coral::OracleAccess::TableDescriptionProxy *>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  std::ostringstream os;
  os << "SELECT "
     << "A.CONSTRAINT_NAME AS FK_NAME, "
     << "B.TABLE_NAME AS REF_TABLE, "
     << "C.COLUMN_NAME AS FK_COL, "
     << "D.COLUMN_NAME AS REF_COL "
     << "FROM "
     << "(SELECT CONSTRAINT_NAME, R_CONSTRAINT_NAME FROM ALL_CONSTRAINTS WHERE OWNER=:\"OWNER\" AND TABLE_NAME =:\"TABLE_NAME\" AND CONSTRAINT_TYPE='R') A, "
     << "(SELECT TABLE_NAME, CONSTRAINT_NAME FROM ALL_CONSTRAINTS WHERE OWNER=:\"OWNER\") B, "
     << "(SELECT COLUMN_NAME, CONSTRAINT_NAME, POSITION FROM ALL_CONS_COLUMNS WHERE OWNER=:\"OWNER\" AND TABLE_NAME=:\"TABLE_NAME\" ORDER BY POSITION) C, "
     << "(SELECT COLUMN_NAME, CONSTRAINT_NAME, TABLE_NAME, POSITION FROM ALL_CONS_COLUMNS WHERE OWNER=:\"OWNER\") D "
     << "WHERE "
     << "A.R_CONSTRAINT_NAME = B.CONSTRAINT_NAME "
     << "AND A.CONSTRAINT_NAME = C.CONSTRAINT_NAME "
     << "AND B.CONSTRAINT_NAME = D.CONSTRAINT_NAME "
     << "AND C.POSITION = D.POSITION "
     << "ORDER BY FK_NAME, D.POSITION";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties, os.str() );
  coral::AttributeList* bindData = new coral::AttributeList;
  bindData->extend<std::string>( "OWNER" );
  bindData->extend<std::string>( "TABLE_NAME" );
  (*bindData)[0].data<std::string>() = m_sessionProperties.schemaName();
  (*bindData)[1].data<std::string>() = m_description->name();
  if ( ! statement.execute( *bindData ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not retrieve the foreign key information defined in a table",
                                  "TableDescriptionProxy::refreshForeignKeyInfo" );
  }
  delete bindData; // I have no idea why the static crashes...

  coral::AttributeList output;
  output.extend<std::string>( "FK_NAME" );
  output.extend<std::string>( "REF_TABLE" );
  output.extend<std::string>( "FK_COL" );
  output.extend<std::string>( "REF_COL" );
  statement.defineOutput( output );
  const std::string& foreignKeyName = output[0].data<std::string>();
  const std::string& referencedTableName = output[1].data<std::string>();
  const std::string& foreignKeyColumnName = output[2].data<std::string>();
  const std::string& referencedColumnName= output[3].data<std::string>();

  // Read the foreign key information
  std::map< std::string, coral::OracleAccess::ForeignKeyInfo > fkeys;
  while( statement.fetchNext() ) {
    std::map< std::string, ForeignKeyInfo >::iterator fkey = fkeys.find( foreignKeyName );
    if ( fkey == fkeys.end() ) {
      fkey = fkeys.insert( std::make_pair( foreignKeyName,
                                           coral::OracleAccess::ForeignKeyInfo( referencedTableName ) ) ).first;
    }
    fkey->second.columns.push_back( foreignKeyColumnName );
    fkey->second.refColumns.push_back( referencedColumnName );
  }

  for ( std::map< std::string, coral::OracleAccess::ForeignKeyInfo >::const_iterator iKey = fkeys.begin();
        iKey != fkeys.end(); ++iKey ) {
    m_description->createForeignKey( iKey->first,
                                     iKey->second.columns,
                                     iKey->second.tableName,
                                     iKey->second.refColumns );
  }

  m_foreignKeysRead = true;
}


std::string
coral::OracleAccess::TableDescriptionProxy::tableSpaceName() const
{
  std::string tableSpaceName = m_description->tableSpaceName();
  if ( tableSpaceName.empty() ) {
    boost::mutex::scoped_lock lock( m_mutex );

    coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                    "SELECT TABLESPACE_NAME FROM ALL_TABLES WHERE OWNER=:\"owner\" AND TABLE_NAME=:\"table_name\"" );
    coral::AttributeList bindData;
    bindData.extend<std::string>( "owner" );
    bindData.extend<std::string>( "table_name" );
    bindData[0].data<std::string>() = m_sessionProperties.schemaName();
    bindData[1].data<std::string>() = m_description->name();
    if ( ! statement.execute( bindData ) ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not retrieve the tablespace for a table",
                                    "ITableDescription::tableSpaceName" );
    }
    coral::AttributeList outputData;
    outputData.extend<std::string>( "TABLESPACE_NAME" );
    statement.defineOutput( outputData );
    statement.fetchNext();
    tableSpaceName = outputData.begin()->data<std::string>();
    const_cast< coral::TableDescription* >( m_description )->setTableSpaceName( tableSpaceName );
  }
  return tableSpaceName;
}


int
coral::OracleAccess::TableDescriptionProxy::numberOfColumns() const
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  return m_description->numberOfColumns();
}


const coral::IColumn&
coral::OracleAccess::TableDescriptionProxy::columnDescription( int columnIndex ) const
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( columnIndex < 0 || columnIndex >= static_cast<int>( m_columns.size() ) )
    throw coral::InvalidColumnIndexException( m_sessionProperties.domainProperties().service()->name() );
  return *( m_columns[ columnIndex ] );
}


const coral::IColumn&
coral::OracleAccess::TableDescriptionProxy::columnDescription( const std::string& columnName ) const
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  const coral::IColumn* pColumn = 0;
  for ( std::vector< coral::OracleAccess::ColumnProxy* >::const_iterator iColumn = m_columns.begin();
        iColumn != m_columns.end(); ++iColumn )
    if ( ( *iColumn )->name() == columnName ) {
      pColumn = *iColumn;
      break;
    }

  if ( pColumn == 0 )
    throw coral::InvalidColumnNameException( m_sessionProperties.domainProperties().service()->name() );
  return *pColumn;
}


bool
coral::OracleAccess::TableDescriptionProxy::hasPrimaryKey() const
{
  if ( ! m_primaryKeyRead ) this->refreshPrimaryKeyInfo();
  return m_description->hasPrimaryKey();
}


const coral::IPrimaryKey&
coral::OracleAccess::TableDescriptionProxy::primaryKey() const
{
  if ( ! m_primaryKeyRead ) this->refreshPrimaryKeyInfo();
  return m_description->primaryKey();
}


int
coral::OracleAccess::TableDescriptionProxy::numberOfIndices() const
{
  if ( ! m_indicesRead ) this->refreshIndexInfo();
  return m_description->numberOfIndices();
}


const coral::IIndex&
coral::OracleAccess::TableDescriptionProxy::index( int indexId ) const
{
  if ( ! m_indicesRead ) this->refreshIndexInfo();
  return m_description->index( indexId );
}


int
coral::OracleAccess::TableDescriptionProxy::numberOfForeignKeys() const
{
  if ( ! m_foreignKeysRead ) this->refreshForeignKeyInfo();
  return m_description->numberOfForeignKeys();
}


const coral::IForeignKey&
coral::OracleAccess::TableDescriptionProxy::foreignKey( int foreignKeyIdentifier ) const
{
  if ( ! m_foreignKeysRead ) this->refreshForeignKeyInfo();
  return m_description->foreignKey( foreignKeyIdentifier );
}


int
coral::OracleAccess::TableDescriptionProxy::numberOfUniqueConstraints() const
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  this->refreshConstraints();
  boost::mutex::scoped_lock lock( m_mutex );
  return m_description->numberOfUniqueConstraints();
}


const coral::IUniqueConstraint&
coral::OracleAccess::TableDescriptionProxy::uniqueConstraint( int uniqueConstraintIdentifier ) const
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  this->refreshConstraints();
  boost::mutex::scoped_lock lock( m_mutex );
  return m_description->uniqueConstraint( uniqueConstraintIdentifier );
}


void
coral::OracleAccess::TableDescriptionProxy::insertColumn( const std::string& name,
                                                          const std::string& type,
                                                          int size,
                                                          bool fixedSize,
                                                          std::string tableSpaceName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->insertColumn( name,
                               type,
                               size,
                               fixedSize,
                               tableSpaceName );

  std::string sqltype = coral::OracleAccess::OracleTableBuilder::sqlType( m_sessionProperties,
                                                                          type,
                                                                          size,
                                                                          fixedSize );
  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" << m_description->name() << "\" ADD ( \"" << name << "\" " << sqltype << " )";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->dropColumn( name );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not add a new column to a table",
                                  "ITableSchemaEditor::insertColumn" );
  }

  m_columns.push_back( new coral::OracleAccess::ColumnProxy( m_description->columnDescription( m_columns.size() ),
                                                             *this ) );
}


void
coral::OracleAccess::TableDescriptionProxy::dropColumn( const std::string& name )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->dropColumn( name );

  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" << m_description->name() << "\" DROP COLUMN \"" << name << "\"";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop a column from a table",
                                  "ITableSchemaEditor::dropColumn" );
  }

  // Reconstruct the column proxies
  for ( std::vector< coral::OracleAccess::ColumnProxy* >::iterator iColumn = m_columns.begin();
        iColumn != m_columns.end(); ++iColumn ) delete *iColumn;
  m_columns.clear();
  int numberOfColumns = m_description->numberOfColumns();
  for ( int i = 0; i < numberOfColumns; ++i )
    m_columns.push_back( new coral::OracleAccess::ColumnProxy( m_description->columnDescription( i ),
                                                               *this ) );
}


void
coral::OracleAccess::TableDescriptionProxy::renameColumn( const std::string& originalName,
                                                          const std::string& newName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->renameColumn( originalName,
                               newName );
  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" << m_description->name()
     << "\" RENAME COLUMN \"" << originalName << "\" TO \"" << newName << "\"";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->renameColumn( newName,
                                 originalName );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not rename a column in a table",
                                  "ITableSchemaEditor::dropColumn" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::changeColumnType( const std::string& columnName,
                                                              const std::string& typeName,
                                                              int size,
                                                              bool fixedSize )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->changeColumnType( columnName,
                                   typeName,
                                   size,
                                   fixedSize );

  std::string sqltype = coral::OracleAccess::OracleTableBuilder::sqlType( m_sessionProperties,
                                                                          typeName,
                                                                          size,
                                                                          fixedSize );
  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" << m_description->name() << "\" MODIFY ( \"" << columnName << "\" " << sqltype << " )";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not change the type of a column in a table",
                                  "ITableSchemaEditor::changeColumnType" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::setNotNullConstraint( const std::string& columnName,
                                                                  bool isNotNull )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->setNotNullConstraint( columnName,
                                       isNotNull );

  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" << m_description->name() << "\" MODIFY ( \"" << columnName << "\" ";
  if ( isNotNull ) os << "NOT ";
  os << "NULL )";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->setNotNullConstraint( columnName,
                                         ! isNotNull );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not change the NOT NULL constraint of a column in a table",
                                  "ITableSchemaEditor::setNotNullConstraint" );
  }

}


void
coral::OracleAccess::TableDescriptionProxy::setUniqueConstraint( const std::string& columnName,
                                                                 std::string name,
                                                                 bool isUnique,
                                                                 std::string tableSpaceName )
{
  this->setUniqueConstraint( std::vector<std::string>( 1, columnName ),
                             name,
                             isUnique,
                             tableSpaceName );
}


void
coral::OracleAccess::TableDescriptionProxy::setUniqueConstraint( const std::vector<std::string>& columnNames,
                                                                 std::string name,
                                                                 bool isUnique,
                                                                 std::string tableSpaceName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  this->refreshConstraints();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->setUniqueConstraint( columnNames,
                                      name,
                                      isUnique,
                                      tableSpaceName );

  if ( isUnique ) {
    std::ostringstream os;
    os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" + m_description->name() << "\" ADD ( ";
    if ( ! name.empty() ) {
      os << "CONSTRAINT \"" << name << "\" ";
    }
    os << "UNIQUE ( ";
    for ( std::vector<std::string>::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn ) {
      if ( iColumn != columnNames.begin() ) os << ", ";
      os << "\"" << *iColumn << "\"";
    }
    os << " ) )";
    std::string tableSpace = tableSpaceName;
    if ( tableSpace.empty() )
      tableSpace = m_sessionProperties.domainProperties().tableSpaceForIndices();
    if ( ! tableSpace.empty() )
      os << " USING INDEX TABLESPACE \"" << tableSpace << "\"";
    coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                    os.str() );
    if ( ! statement.execute( coral::AttributeList() ) ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not create a unique constraint in the table",
                                      "ITableSchemaEditor::setUniqueConstraint" );
    }
  }
  else {
    std::ostringstream os;
    os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" + m_description->name() << "\" DROP ";
    if ( name.empty() ) {
      os << "UNIQUE ( \"" << columnNames[0] << "\" )";
    }
    else {
      os << "CONSTRAINT \"" << name << "\"";
    }
    coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                    os.str() );
    if ( ! statement.execute( coral::AttributeList() ) ) {
      throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                    "Could not drop a unique constraint from the table",
                                      "ITableSchemaEditor::setUniqueConstraint" );
    }
  }
}


void
coral::OracleAccess::TableDescriptionProxy::setPrimaryKey( const std::string& columnName,
                                                           std::string tableSpaceName )
{
  this->setPrimaryKey( std::vector< std::string >( 1, columnName ), tableSpaceName );
}


void
coral::OracleAccess::TableDescriptionProxy::setPrimaryKey( const std::vector<std::string>& columnNames,
                                                           std::string tableSpaceName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_primaryKeyRead ) this->refreshPrimaryKeyInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->setPrimaryKey( columnNames,
                                tableSpaceName );
  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" + m_description->name() << "\" ADD ( CONSTRAINT \""
     << m_description->name() << "_PK\" PRIMARY KEY ( ";
  for ( std::vector<std::string>::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    if ( iColumn != columnNames.begin() ) os << ", ";
    os << "\"" << *iColumn << "\"";
  }
  os << " )";
  std::string tableSpace = tableSpaceName;
  if ( tableSpace.empty() )
    tableSpace = m_sessionProperties.domainProperties().tableSpaceForIndices();
  if ( ! tableSpace.empty() )
    os << " USING INDEX TABLESPACE \"" << tableSpace << "\"";
  os << " )";
  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->dropPrimaryKey();
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not add a primary key constraint on a table",
                                  "ITableSchemaEditor::setPrimaryKey" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::dropPrimaryKey()
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_primaryKeyRead ) this->refreshPrimaryKeyInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->dropPrimaryKey();

  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" + m_description->name() << "\" DROP PRIMARY KEY";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop a primary key constraint from the table",
                                  "ITableSchemaEditor::dropPrimaryKey" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::createIndex( const std::string& indexName,
                                                         const std::string& columnName,
                                                         bool isUnique,
                                                         std::string tableSpaceName )
{
  this->createIndex( indexName, std::vector<std::string>( 1, columnName ), isUnique, tableSpaceName );
}


void
coral::OracleAccess::TableDescriptionProxy::createIndex( const std::string& name,
                                                         const std::vector<std::string>& columnNames,
                                                         bool isUnique,
                                                         std::string tableSpaceName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_indicesRead ) this->refreshIndexInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  // Create the transient one
  m_description->createIndex( name,
                              columnNames,
                              isUnique,
                              tableSpaceName );

  std::ostringstream os;
  os << "CREATE ";
  if ( isUnique ) os << "UNIQUE ";
  os << "INDEX " << m_sessionProperties.schemaName() << ".\"" << name << "\" ON "
     << m_sessionProperties.schemaName() << ".\"" << m_description->name() << "\" ( ";
  for ( std::vector<std::string>::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    if ( iColumn != columnNames.begin() ) os << ", ";
    os << "\"" << *iColumn << "\"";
  }
  os << " )";

  std::string tableSpace = tableSpaceName;
  if ( tableSpace.empty() )
    tableSpace = m_sessionProperties.domainProperties().tableSpaceForIndices();
  if ( ! tableSpace.empty() )
    os << " TABLESPACE \"" << tableSpace << "\"";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->dropIndex( name );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not create an index on the table",
                                  "ITableSchemaEditor::createIndex" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::dropIndex( const std::string& indexName )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_indicesRead ) this->refreshIndexInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->dropIndex( indexName );

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  "DROP INDEX " + m_sessionProperties.schemaName() + "\"" + indexName + "\"" );
  if ( ! statement.execute( coral::AttributeList() ) )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop an index from the table",
                                  "ITableSchemaEditor::dropIndex" );
}


void
coral::OracleAccess::TableDescriptionProxy::createForeignKey( const std::string& name,
                                                              const std::string& columnName,
                                                              const std::string& referencedTableName,
                                                              const std::string& referencedColumnName )
{
  this->createForeignKey( name,
                          std::vector< std::string >( 1, columnName ),
                          referencedTableName,
                          std::vector< std::string >( 1, referencedColumnName ) );
}


void
coral::OracleAccess::TableDescriptionProxy::createForeignKey( const std::string& name,
                                                              const std::vector<std::string>& columnNames,
                                                              const std::string& referencedTableName,
                                                              const std::vector<std::string>& referencedColumnNames )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_foreignKeysRead ) this->refreshForeignKeyInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->createForeignKey( name,
                                   columnNames,
                                   referencedTableName,
                                   referencedColumnNames );

  std::ostringstream os;
  os << "ALTER TABLE  " << m_sessionProperties.schemaName() << ".\"" + m_description->name()
     << "\" ADD ( CONSTRAINT \"" << name << "\" FOREIGN KEY ( ";
  for ( std::vector<std::string>::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    if ( iColumn != columnNames.begin() ) os << ", ";
    os << "\"" << *iColumn << "\"";
  }
  os << " ) REFERENCES " << m_sessionProperties.schemaName() << ".\"" << referencedTableName << "\" ( ";
  for ( std::vector<std::string>::const_iterator iColumn = referencedColumnNames.begin();
        iColumn != referencedColumnNames.end(); ++iColumn ) {
    if ( iColumn != referencedColumnNames.begin() ) os << ", ";
    os << "\"" << *iColumn << "\"";
  }
  os << " ) )";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    m_description->dropForeignKey( name );
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not create a foreign key constraint on the table",
                                  "ITableSchemaEditor::createForeignKey" );
  }
}


void
coral::OracleAccess::TableDescriptionProxy::dropForeignKey( const std::string& name )
{
  if ( ! m_columnDescriptionRead )
    const_cast<coral::OracleAccess::TableDescriptionProxy*>( this )->readColumnDescription();
  if ( ! m_foreignKeysRead ) this->refreshForeignKeyInfo();

  boost::mutex::scoped_lock lock( m_mutex );

  m_description->dropForeignKey( name );

  std::ostringstream os;
  os << "ALTER TABLE " << m_sessionProperties.schemaName() << ".\"" + m_description->name() << "\" DROP CONSTRAINT \"" << name << "\"";

  coral::OracleAccess::OracleStatement statement( m_sessionProperties,
                                                  os.str() );
  if ( ! statement.execute( coral::AttributeList() ) ) {
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),
                                  "Could not drop a foreign key constraint from the table",
                                  "ITableSchemaEditor::dropForeignKey" );
  }
}
