#include "DmlApp.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"

#include <iostream>
#include <stdexcept>
#include <sstream>

DmlApp::DmlApp( const std::string& connectionString,
                const std::string& userName,
                const std::string& password ):
  TestBase(),
  m_connectionString( connectionString ),
  m_userName( userName ),
  m_password( password )
{}


DmlApp::~DmlApp()
{}


void
DmlApp::run()
{
  coral::ISession* session = this->connect( m_connectionString,
                                            m_userName, m_password );
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();

  std::cout << "About to drop previous table" << std::endl;
  schema.dropIfExistsTable( "T_1" );

  std::cout << "Describing new tables" << std::endl;
  coral::TableDescription description;
  description.setName( "T_1" );
  description.insertColumn( "ID",
                            coral::AttributeSpecification::typeNameForId( typeid(int) ) );
  description.setPrimaryKey( "ID" );
  description.insertColumn( "x",
                            coral::AttributeSpecification::typeNameForId( typeid(float) ) );
  description.setNotNullConstraint( "x" );
  description.insertColumn( "Y",
                            coral::AttributeSpecification::typeNameForId( typeid(double) ) );
  description.insertColumn( "Z",
                            coral::AttributeSpecification::typeNameForId( typeid(double) ) );
  description.insertColumn( "data0",
                            coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
  description.insertColumn( "data1",
                            coral::AttributeSpecification::typeNameForId( typeid(std::string) ),
                            2,
                            true );
  description.insertColumn( "data2",
                            coral::AttributeSpecification::typeNameForId( typeid(std::string) ),
                            321,
                            false );

  std::cout << "About to create the tables" << std::endl;
  coral::ITableDataEditor& editor = schema.createTable( description ).dataEditor();

  coral::AttributeList rowBuffer;
  editor.rowBuffer( rowBuffer );

  for ( int i = 0; i < 10; ++i ) {
    rowBuffer["ID"].data<int>() = i;
    rowBuffer["x"].data<float>() = i + 0.01*i;
    rowBuffer["Y"].data<double>() = i + 0.001*i;
    rowBuffer["Z"].data<double>() = i + 0.0001*i;
    std::ostringstream os0;
    os0 << "The data0 is " << i;
    rowBuffer["data0"].data<std::string>() = os0.str();
    std::ostringstream os1;
    os1 << i;
    rowBuffer["data1"].data<std::string>() = os1.str();
    std::ostringstream os2;
    os2 << "Hello " << 10*(i-1) + i;
    rowBuffer["data2"].data<std::string>() = os2.str();

    if ( i%3 == 1 ) {
      rowBuffer["Y"].setNull();
    }
    else {
      rowBuffer["Y"].setNull( false );
    }

    editor.insertRow( rowBuffer );
  }

  // Now try to insert a duplicate entry
  bool exceptionCaught = false;
  try {
    editor.insertRow( rowBuffer );
  }
  catch ( coral::DuplicateEntryInUniqueKeyException& ) {
    exceptionCaught = true;
  }

  if ( ! exceptionCaught )
    throw std::runtime_error( "DuplicateEntryInUniqueKeyException not thrown !!!" );


  // Remove a few entries...
  coral::AttributeList inputData;
  inputData.extend<int>( "id" );
  inputData.begin()->data<int>() = 5;
  int rowsDeleted = editor.deleteRows( "ID<:id", inputData );
  std::cout << "Removed " << rowsDeleted << " rows" << std::endl;

  // Udpate the entries;
  inputData.extend<double>("increment");
  inputData[1].data<double>() = 1.111;
  inputData[0].data<int>() += 2;
  int rowsUpdated = editor.updateRows( "Y   = Y+ :increment", "ID>:id", inputData );
  std::cout << "Updated " << rowsUpdated << " rows" << std::endl;


  session->transaction().commit();
  delete session;
}
