#include "QueriesApp.h"

#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Date.h"
#include "CoralBase/TimeStamp.h"

#include "CoralCommon/Sleep.h"

#include <iostream>
#include <stdexcept>


QueriesApp::QueriesApp( const std::string& connectionString,
                        const std::string& userName,
                        const std::string& password ):
  TestBase(),
  m_connectionString( connectionString ),
  m_userName( userName ),
  m_password( password )
{}


QueriesApp::~QueriesApp()
{}


void
QueriesApp::run()
{
  this->fillData();
  coral::sys::sleep(1);
  this->readData();
}



void
QueriesApp::fillData()
{
  coral::ISession* session = this->connect( m_connectionString,
                                            m_userName, m_password );
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();

  std::cout << "About to drop previous table" << std::endl;
  schema.dropIfExistsTable( "T" );

  std::cout << "Describing new table" << std::endl;
  coral::TableDescription description;
  description.setName( "T" );
  description.insertColumn( "ID",
                            coral::AttributeSpecification::typeNameForId( typeid(int) ) );
  description.setPrimaryKey( "ID" );
  description.insertColumn( "TheDate",
                            coral::AttributeSpecification::typeNameForId( typeid(coral::Date) ) );
  description.insertColumn( "TheTime",
                            coral::AttributeSpecification::typeNameForId( typeid(coral::TimeStamp) ),
                            6 );

  std::cout << "About to create the table" << std::endl;
  coral::ITable& table = schema.createTable( description );

  coral::AttributeList rowBuffer;
  table.dataEditor().rowBuffer( rowBuffer );

  for ( int i = 0; i < 5; ++i )
  {
    rowBuffer["ID"].data<int>() = i;
    rowBuffer["TheDate"].data<coral::Date>() = coral::Date( 2005, 11, i+1 );
    rowBuffer["TheTime"].data<coral::TimeStamp>() = coral::TimeStamp::now();
    table.dataEditor().insertRow( rowBuffer );
  }

  coral::IBulkOperation* bulkInserter = table.dataEditor().bulkInsert( rowBuffer, 3 );

  long fraction = 111111111;
  for ( int i = 5; i < 10; ++i, fraction /= 10 )
  {
    rowBuffer["ID"].data<int>() = i;
    rowBuffer["TheDate"].data<coral::Date>() = coral::Date( 2005, 11, i+1 );
    coral::TimeStamp thisMoment = coral::TimeStamp::now();
    rowBuffer["TheTime"].data<coral::TimeStamp>() = coral::TimeStamp( thisMoment.year(),
                                                                      thisMoment.month(),
                                                                      thisMoment.day(),
                                                                      thisMoment.hour(),
                                                                      thisMoment.minute(),
                                                                      thisMoment.second(),
                                                                      (i-4) * fraction );
    bulkInserter->processNextIteration();
  }
  bulkInserter->flush();
  delete bulkInserter;

  session->transaction().commit();
  delete session;
}



void
QueriesApp::readData()
{
  coral::ISession* session = this->connect( m_connectionString,
                                            m_userName, m_password );
  session->transaction().start( true );

  coral::ISchema& schema = session->nominalSchema();

  coral::IQuery* query = schema.tableHandle("T").newQuery();
  query->setRowCacheSize( 20 );
  coral::ICursor& cursor = query->execute();
  int i = 0;
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    currentRow.toOutputStream( std::cout ) << std::endl;
    ++i;
  }

  if ( i != 10 )
    throw std::runtime_error( "Unexpected number of rows" );

  delete query;

  session->transaction().commit();
  delete session;
}
