#include "ThreadBody.h"

#include "../Common/Reporter.h"

#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include "CoralCommon/Sleep.h"

#include <sstream>
#include <cmath>
#include <memory>

ThreadBody::ThreadBody( coral::IConnection& connection,
                        int threadID,
                        Reporter& reporter,
                        const std::string& schemaName,
                        const std::string& userName,
                        const std::string& password ):
  m_threadID( threadID ),
  m_reporter( reporter ),
  m_connection( connection ),
  m_schemaName( schemaName ),
  m_userName( userName ),
  m_password( password )
{}


ThreadBody::~ThreadBody()
{}

void
ThreadBody::operator()()
{
  try {
    std::ostringstream os;
    os << "Thread no " << m_threadID << " running...";
    m_reporter.reportToSTDOUT( os.str() );

    std::auto_ptr<coral::ISession> session( m_connection.newSession( m_schemaName ) );

    session->startUserSession( m_userName, m_password );
    session->transaction().start();

    coral::ISchema& schema = session->nominalSchema();

    std::ostringstream os1;
    os1 << "Thread no " << m_threadID << " about to create and fill a table";
    m_reporter.reportToSTDOUT( os1.str() );

    // The table name
    std::ostringstream osTableName;
    osTableName << "ORA_UT_MS_T_" << m_threadID;
    const std::string tableName = osTableName.str();

    schema.dropIfExistsTable( tableName );

    coral::TableDescription description;
    description.setName( tableName );
    description.insertColumn( "I", "int" );
    description.insertColumn( "F", "float" );
    description.insertColumn( "D", "double" );
    description.setPrimaryKey( "I" );
    description.setNotNullConstraint( "F" );

    coral::ITable& table = schema.createTable( description );
    coral::AttributeList rowBuffer;
    rowBuffer.extend<int>( "I" );
    rowBuffer.extend<float>( "F" );
    rowBuffer.extend<double>( "D" );

    int& i = rowBuffer[0].data<int>();
    float& f = rowBuffer[1].data<float>();
    double& d = rowBuffer[2].data<double>();

    for ( int row01 = 0; row01 < 100; ++row01 ) {
      i = row01;
      f = row01 + 0.001 * m_threadID;
      d = row01 + 0.000001 * m_threadID;
      table.dataEditor().insertRow( rowBuffer );
    }
    session->transaction().commit();

    coral::sys::sleep(1);

    session->transaction().start( true );
    std::ostringstream os2;
    os2 << "Thread no " << m_threadID << " about to perform a query";
    m_reporter.reportToSTDOUT( os2.str() );

    std::auto_ptr<coral::IQuery> query( schema.tableHandle( tableName ).newQuery() );

    coral::AttributeList outputBuffer;
    outputBuffer.extend<double>( "RES" );

    query->addToOutputList( "F+D", "RES" );
    query->addToOrderList( "I" );
    query->defineOutput( outputBuffer );

    coral::ICursor& cursor = query->execute();
    int row02 = 0;
    double& res = outputBuffer[0].data<double>();
    while ( cursor.next() ) {
      if ( ::fabs( res - ( 2* row02 + 0.001001 * m_threadID ) ) > 0.00001 )
        throw std::runtime_error( "Unexpected data" );
      ++row02;
    }

    if ( row02 != 100 )
      throw std::runtime_error( "Unexpected number of rows" );

    session->transaction().commit();
  }
  catch ( std::exception& e ) {
    std::ostringstream os;
    os << "[OVAL] : Standard C++ exception raised in Thread no " << m_threadID << " : " << e.what();
    m_reporter.reportToSTDERR( os.str() );
  }
  catch (...) {
    std::ostringstream os;
    os << "[OVAL] : Unknonwn exception ... raised in Thread no " << m_threadID;
    m_reporter.reportToSTDERR( os.str() );
  }

  std::ostringstream os;
  os << "Thread no " << m_threadID << " ended gracefully";
  m_reporter.reportToSTDOUT( os.str() );
}
