#include "RelationalAuthenticationService.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"
#include "CoralCommon/AuthenticationCredentials.h"
#include "AuthenticationCredentialSet.h"
#include "CoralCommon/Cipher.h"
#include "XMLAuthenticationFileParser.h"
#include "XMLAuthenticationFileContent.h"
#include "RelationalAccess/AuthenticationServiceException.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/IAuthenticationCredentials.h"
#include "RelationalAccess/IConnectionService.h"
#include "RelationalAccess/IConnectionServiceConfiguration.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralKernel/Context.h"
#include "CoralKernel/IPropertyManager.h"
#include <cstdlib>
#include <boost/bind.hpp>
#include <sstream>
#include <fstream>

#define VERIFICATION_KEY TABLE_NAME

static std::string CREDENTIAL_DATABASE("sqlite_file:authentication.db");
static std::string TABLE_NAME("CORAL_CREDENTIALS");
static std::string CONNECTION_COLUMN("CRED0");
static std::string ROLE_COLUMN("CRED1");
static std::string USERNAME_COLUMN("CRED2");
static std::string PASSWORD_COLUMN("CRED3");
static std::string VERIFICATION_KEY_COLUMN("CRED4");

coral::RelationalAuthenticationService::RelationalAuthenticationService::RelationalAuthenticationService( const std::string& componentName )
  : coral::Service( componentName ),
    m_isLoaded( false ),
    m_isLoading( false ),
    m_connectionString( "" ),
    m_key( ""),
    m_data()
{
	coral::Property* pm = dynamic_cast<coral::Property*>(coral::Context::instance().PropertyManager().property("CredentialDatabase"));
  if(pm){
    m_connectionString = pm->get();
    boost::function1<void, std::string> cb0(boost::bind(&coral::RelationalAuthenticationService::RelationalAuthenticationService::setDataSource, this, _1));
    m_callback0_ID = pm->registerCallback(cb0);
  }
	pm = dynamic_cast<coral::Property*>(coral::Context::instance().PropertyManager().property("EncryptionKey"));
  if(pm){
    m_key = pm->get();
    boost::function1<void, std::string> cb1(boost::bind(&coral::RelationalAuthenticationService::RelationalAuthenticationService::setEncryptionKey, this, _1));
    m_callback1_ID = pm->registerCallback(cb1);
  }
}


coral::RelationalAuthenticationService::RelationalAuthenticationService::~RelationalAuthenticationService()
{
	coral::Property* pm = dynamic_cast<coral::Property*>(coral::Context::instance().PropertyManager().property("CredentialDatabase"));
  if(pm){
    pm->unregisterCallback(m_callback0_ID);
  }
	pm = dynamic_cast<coral::Property*>(coral::Context::instance().PropertyManager().property("EncryptionKey"));
  if(pm){
    pm->unregisterCallback(m_callback1_ID);
  }
	for ( std::map< std::string, coral::AuthenticationCredentialSet* >::iterator iConnection = m_data.begin();
		iConnection != m_data.end(); ++iConnection ) delete iConnection->second;
}

void
coral::RelationalAuthenticationService::RelationalAuthenticationService::connectToDatabase(  const std::string& connectionString,
                                                                                             const std::string& encryptionKey)
{
  m_connectionString = connectionString;
  m_key = encryptionKey;
}

/// Sets the input connection string
void
coral::RelationalAuthenticationService::RelationalAuthenticationService::setDataSource(  const std::string& connectionString )
{
  m_connectionString = connectionString;
}

/// Sets the input connection string
void
coral::RelationalAuthenticationService::RelationalAuthenticationService::setEncryptionKey(  const std::string& key ){
  m_key = key;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::createCredentialDatabase()
{
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::Update ));
  session->transaction().start();
  coral::ISchema& schema = session->nominalSchema();
  if(schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    std::stringstream msg;
    msg <<"Cannot create credential database, a table named \"" << TABLE_NAME << "\" already exists.";
    throw coral::AuthenticationServiceException(msg.str(),
                                                "RelationalAuthenticationService::dropCredentialDatabase",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::TableDescription descr;
  descr.setName(TABLE_NAME);
  int columnSize = 2000;
  descr.insertColumn( CONNECTION_COLUMN, coral::AttributeSpecification::typeNameForType<std::string>(),columnSize,false);
  descr.insertColumn( ROLE_COLUMN, coral::AttributeSpecification::typeNameForType<std::string>(),columnSize,false);
  descr.insertColumn( USERNAME_COLUMN, coral::AttributeSpecification::typeNameForType<std::string>(),columnSize,false);
  descr.insertColumn( PASSWORD_COLUMN, coral::AttributeSpecification::typeNameForType<std::string>(),columnSize,false);
  descr.insertColumn( VERIFICATION_KEY_COLUMN, coral::AttributeSpecification::typeNameForType<std::string>(),columnSize,false);
  descr.setNotNullConstraint( CONNECTION_COLUMN );
  descr.setNotNullConstraint( ROLE_COLUMN );
  descr.setNotNullConstraint( USERNAME_COLUMN );
  descr.setNotNullConstraint( PASSWORD_COLUMN );
  descr.setNotNullConstraint( VERIFICATION_KEY_COLUMN );
  std::vector<std::string> columnsForIndex;
  columnsForIndex.push_back(CONNECTION_COLUMN);
  columnsForIndex.push_back(ROLE_COLUMN);
  descr.setPrimaryKey( columnsForIndex );
  coral::MessageStream log( this->name() );
  log << coral::Info << "Creating table \"" << TABLE_NAME << "\"."<< coral::MessageStream::endmsg;
  schema.createTable( descr );
  session->transaction().commit();
  return true;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::dropCredentialDatabase()
{
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::Update ));
  session->transaction().start();
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    std::stringstream msg;
    msg <<"Cannot drop credential database, a table named \"" << TABLE_NAME << "\" does not exists.";
    throw coral::AuthenticationServiceException(msg.str(),
                                                "RelationalAuthenticationService::dropCredentialDatabase",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::MessageStream log( this->name() );
  log << coral::Info << "Dropping table \"" << TABLE_NAME << "\"."<< coral::MessageStream::endmsg;
  schema.dropTable( TABLE_NAME );
  session->transaction().commit();
  return true;
}


bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::addCredentialEntry( const std::string& userName,
                                                                                             const std::string& password,
                                                                                             const std::string& serviceConnectionString,
                                                                                             std::string theRole)
{
  std::string role = theRole;
  if(role.empty()) role = coral::IAuthenticationCredentials::defaultRole();
  std::string connString = serviceConnectionString;
  std::string user = userName;
  std::string passwd = password;
  std::string verifKey = VERIFICATION_KEY;
  if(!m_key.empty()){
    coral::MessageStream log( this->name() );
    log << coral::Debug << "Encoding credential data."<< coral::MessageStream::endmsg;
    connString = coral::Cipher::encode(serviceConnectionString,m_key);
    role = coral::Cipher::encode(role, m_key);
    user = coral::Cipher::encode(userName, m_key);
    passwd = coral::Cipher::encode(password, m_key);
    verifKey =  coral::Cipher::encode(TABLE_NAME, m_key);
  }
  
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::Update ));
  session->transaction().start();
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::addCredentialEntry",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::AttributeList readBuff;
  readBuff.extend<std::string>(CONNECTION_COLUMN);
  readBuff.extend<std::string>(ROLE_COLUMN);
  std::auto_ptr<coral::IQuery> query(schema.tableHandle(TABLE_NAME).newQuery());
  coral::AttributeList whereData;
  whereData.extend<std::string>(CONNECTION_COLUMN);
  whereData.extend<std::string>(ROLE_COLUMN);
  whereData[ CONNECTION_COLUMN ].data<std::string>() = connString;
  whereData[ ROLE_COLUMN ].data<std::string>() = role;
  std::stringstream whereClause;
  whereClause << CONNECTION_COLUMN << " = :"<<CONNECTION_COLUMN;
  whereClause << " AND ";
  whereClause << ROLE_COLUMN << " = :"<<ROLE_COLUMN;
  query->defineOutput(readBuff);
  query->setCondition( whereClause.str(), whereData );
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) {
    coral::MessageStream log( this->name() );
    if( theRole == coral::IAuthenticationCredentials::defaultRole() || theRole.empty()){
      log << coral::Warning << "Credential parameters for connection string \""
          << serviceConnectionString << "\" have already been defined. Only new elements are appended, while existing will be ignored."
          << coral::MessageStream::endmsg;
    } else {
      log << coral::Warning << "Credential parameters for connection string \""
          << serviceConnectionString << "\" and role \""<< theRole
          << "\" have already been defined. Only new elements are appended, while existing will be ignored."
          << coral::MessageStream::endmsg;      
    }
    session->transaction().commit();
    return false;
  }

  coral::ITableDataEditor& inserter = schema.tableHandle(TABLE_NAME).dataEditor();
  coral::AttributeList dataBuffer;
  inserter.rowBuffer(dataBuffer);
  dataBuffer[ CONNECTION_COLUMN ].data<std::string>() = connString;
  dataBuffer[ ROLE_COLUMN ].data<std::string>() = role;
  dataBuffer[ USERNAME_COLUMN ].data<std::string>() = user;
  dataBuffer[ PASSWORD_COLUMN ].data<std::string>() = passwd;
  dataBuffer[ VERIFICATION_KEY_COLUMN ].data<std::string>() = verifKey;
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Inserting credential data for connection \""<<serviceConnectionString<<"\"."<< coral::MessageStream::endmsg;
  inserter.insertRow( dataBuffer );
  session->transaction().commit();
  m_isLoaded = false;
  return true;
}

bool coral::RelationalAuthenticationService::RelationalAuthenticationService::removeCredentialEntry( const std::string& serviceConnectionString,
                                                                                                    std::string role){
  if(role.empty()) role = coral::IAuthenticationCredentials::defaultRole();
  std::string connectionS = serviceConnectionString;
  if(!m_key.empty()){
    coral::MessageStream log( this->name() );
    log << coral::Debug << "Encoding credential data."<< coral::MessageStream::endmsg;
    connectionS = coral::Cipher::encode(serviceConnectionString,m_key);
    role = coral::Cipher::encode(role, m_key);
  }
  
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::Update ));
  session->transaction().start();
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::removeCredentialentry",
                                                "coral::RelationalAuthenticationService");    
  }
  
  coral::AttributeList dataBuff;
  std::string condition = CONNECTION_COLUMN + " = :" + CONNECTION_COLUMN + " AND " +
                          ROLE_COLUMN + " = :" + ROLE_COLUMN;
  dataBuff.extend<std::string>(CONNECTION_COLUMN);
  dataBuff.extend<std::string>(ROLE_COLUMN);
  dataBuff[CONNECTION_COLUMN].data<std::string>() = connectionS;
  dataBuff[ROLE_COLUMN].data<std::string>() = role;
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Deleting credential data for connection \""<<serviceConnectionString<<"\"."<< coral::MessageStream::endmsg;
  schema.tableHandle(TABLE_NAME).dataEditor().deleteRows(condition,dataBuff);
  session->transaction().commit();
  m_isLoaded = false;
  return true;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::importCredentials( const std::string& xmlAuthenticationFile ){
  bool inserted = false;
  coral::XMLAuthenticationFileParser fileParser(this->name());
  std::map< std::string, AuthenticationCredentialSet* > data;
  fileParser.parse( xmlAuthenticationFile,data );
  if(data.empty()) return false;
  std::string verifKey = VERIFICATION_KEY;
  if(!m_key.empty())  verifKey = coral::Cipher::encode(verifKey,m_key);
  
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::Update ));
  coral::ISchema& schema = session->nominalSchema();
  session->transaction().start();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::importCredentials",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::AttributeList readBuff;
  readBuff.extend<std::string>(CONNECTION_COLUMN);
  readBuff.extend<std::string>(ROLE_COLUMN);
  std::auto_ptr<coral::IQuery> query(schema.tableHandle(TABLE_NAME).newQuery());
  query->defineOutput(readBuff);
  coral::ICursor& cursor = query->execute();
  std::set<std::string> pkData;
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    std::string connectionString = currentRow[CONNECTION_COLUMN].data<std::string>();
    std::string role = currentRow[ROLE_COLUMN].data<std::string>();
    std::string pkEntry("");
    pkEntry.append(connectionString).append(role);
    pkData.insert(pkEntry);
  }

  coral::AttributeList dataBuffer;
  schema.tableHandle(TABLE_NAME).dataEditor().rowBuffer( dataBuffer );
  coral::IBulkOperation* rowInserter = schema.tableHandle( TABLE_NAME ).dataEditor().bulkInsert(dataBuffer, 1000);
  for( std::map< std::string, AuthenticationCredentialSet* >::const_iterator iEntry = data.begin();
       iEntry != data.end();
       ++iEntry){
    std::string service = iEntry->first;
    std::string srv = service;
    if(!m_key.empty()) srv = coral::Cipher::encode(service,m_key);
    if( iEntry->second->hasDefault()){
      std::string role = coral::IAuthenticationCredentials::defaultRole();
      std::string user = iEntry->second->credentials().valueForItem(coral::IAuthenticationCredentials::userItem());
      std::string passwd = iEntry->second->credentials().valueForItem(coral::IAuthenticationCredentials::passwordItem());
      if(!m_key.empty()){
        role = coral::Cipher::encode(role,m_key);
        user = coral::Cipher::encode(user,m_key);
        passwd = coral::Cipher::encode(passwd,m_key);
      }
      std::string pkEntry("");
      pkEntry.append(srv).append(role);
      std::set<std::string>::const_iterator iE = pkData.find(pkEntry);
      if(iE!=pkData.end()){
        coral::MessageStream log( this->name() );
        log << coral::Warning << "Credential parameters for connection string \""
            << service << "\" have already been defined. Only new elements are appended, while existing will be ignored."
            << coral::MessageStream::endmsg;
      } else {
        coral::MessageStream log( this->name() );
        log << coral::Debug << "Importing credential parameters for connection string \""
            << service << "\"." << coral::MessageStream::endmsg;
        dataBuffer[ CONNECTION_COLUMN ].data<std::string>() = srv;
        dataBuffer[ ROLE_COLUMN ].data<std::string>() = role;
        dataBuffer[ USERNAME_COLUMN ].data<std::string>() = user;
        dataBuffer[ PASSWORD_COLUMN ].data<std::string>() = passwd;
        dataBuffer[ VERIFICATION_KEY_COLUMN ].data<std::string>() = verifKey;
        rowInserter->processNextIteration();
        inserted = true;
      } 
    }
    
    for(std::map< std::string, AuthenticationCredentials* >::const_iterator iRole = iEntry->second->roleMap().begin();
        iRole != iEntry->second->roleMap().end();
        ++iRole){
      std::string role = iRole->first;
      std::string rl = role;
      std::string user = iRole->second->valueForItem(coral::IAuthenticationCredentials::userItem());
      std::string passwd = iRole->second->valueForItem(coral::IAuthenticationCredentials::passwordItem());
      if(!m_key.empty()){
        rl =  coral::Cipher::encode(role,m_key);
        user = coral::Cipher::encode(user,m_key);
        passwd = coral::Cipher::encode(passwd,m_key);
      }
      std::string pkEntry("");
      pkEntry.append(srv).append(role);
      std::set<std::string>::const_iterator iE = pkData.find(pkEntry);
      if(iE!=pkData.end()){
        coral::MessageStream log( this->name() );
        log << coral::Warning << "Credential parameters for connection string \""
            << service << "\" and role \""<< role 
          << "\" have already been defined. Only new elements are appended, while existing will be ignored."
          << coral::MessageStream::endmsg;
      } else {
        coral::MessageStream log( this->name() );
        log << coral::Debug << "Importing credential parameters for connection string \""
            << service << "\" and role \""<< role << "\"." << coral::MessageStream::endmsg;
        dataBuffer[ CONNECTION_COLUMN ].data<std::string>() = srv;
        dataBuffer[ ROLE_COLUMN ].data<std::string>() = rl;
        dataBuffer[ USERNAME_COLUMN ].data<std::string>() = user;
        dataBuffer[ PASSWORD_COLUMN ].data<std::string>() = passwd;
        dataBuffer[ VERIFICATION_KEY_COLUMN ].data<std::string>() = verifKey;
        rowInserter->processNextIteration();
        inserted = true;
      }
    }
  }
  rowInserter->flush();
  delete rowInserter;
  session->transaction().commit();
  if(inserted) m_isLoaded = false;
  return inserted;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::exportCredentials( const std::string& xmlAuthenticationFile ){
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::ReadOnly ));
  session->transaction().start(true);
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::exportCredentials",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::AttributeList dataBuff;
  dataBuff.extend<std::string>(CONNECTION_COLUMN);
  dataBuff.extend<std::string>(ROLE_COLUMN);
  dataBuff.extend<std::string>(USERNAME_COLUMN);
  dataBuff.extend<std::string>(PASSWORD_COLUMN);
  dataBuff.extend<std::string>(VERIFICATION_KEY_COLUMN);
  std::auto_ptr<coral::IQuery> query(schema.tableHandle(TABLE_NAME).newQuery());
  query->defineOutput(dataBuff);
  query->addToOrderList(CONNECTION_COLUMN);
  query->addToOrderList(ROLE_COLUMN);
  bool foundValid = false;
  coral::ICursor& cursor = query->execute();
  coral::XMLAuthenticationFileContent* outFile = 0;
  std::ofstream*  authenticationFileStream = 0;
  std::string prevConn("");
  while ( cursor.next() ) {
    if(!outFile){
      authenticationFileStream = new std::ofstream(xmlAuthenticationFile.c_str(),std::ios_base::out);
      outFile = new coral::XMLAuthenticationFileContent(*authenticationFileStream);
    }
    
    const coral::AttributeList& currentRow = cursor.currentRow();
    std::string connectionString = currentRow[CONNECTION_COLUMN].data<std::string>();
    std::string role = currentRow[ROLE_COLUMN].data<std::string>();
    std::string userName =  currentRow[USERNAME_COLUMN].data<std::string>();   
    std::string password =  currentRow[PASSWORD_COLUMN].data<std::string>();
    std::string verifKey = currentRow[VERIFICATION_KEY_COLUMN].data<std::string>();
    if(!m_key.empty()){
      coral::MessageStream log( this->name() );
      log << coral::Debug << "Decoding credential data."<< coral::MessageStream::endmsg;
      connectionString = coral::Cipher::decode(connectionString,m_key);
      role = coral::Cipher::decode(role,m_key);
      userName =  coral::Cipher::decode(userName,m_key);  
      password =  coral::Cipher::decode(password,m_key);
      verifKey = coral::Cipher::decode(verifKey,m_key);
    }
    // only consider entries where the verification key is matching...
    if( VERIFICATION_KEY == verifKey ){
      foundValid = true;

      if(prevConn!=connectionString){
        if(!prevConn.empty()) outFile->closeConnectionEntry();
        outFile->openConnectionEntry(connectionString);
        prevConn = connectionString;
      }
      if(role == coral::IAuthenticationCredentials::defaultRole()){
        outFile->addCredentialEntry(userName,password );
      } else{
        outFile->openRoleEntry( role );
        outFile->addCredentialEntry(userName,password );
        outFile->closeRoleEntry();
      }
      coral::MessageStream log( this->name() );
      log << coral::Debug << "Exporting credential data for connection \""<<connectionString<<"\""<<coral::MessageStream::endmsg;
    } else {
      coral::MessageStream log( this->name() );
      if(m_key.empty()){
        log << coral::Debug << "Skipping credential item, required key has not been provided."<<coral::MessageStream::endmsg;
      } else {
        log << coral::Debug << "Skipping credential item, provided key is invalid."<<coral::MessageStream::endmsg;
      }
      
    }
    
  }
  session->transaction().commit();
  if(outFile) {
    outFile->close();
    authenticationFileStream->close();
    delete outFile;
    delete authenticationFileStream;
  }
  return foundValid;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::listCredentials( std::ostream& out ){

  std::string serviceTag("service");
  std::string roleTag("role");
  std::string userTag("username");
  size_t serviceMaxW = serviceTag.length();
  size_t roleMaxW = roleTag.length();
  size_t userMaxW = userTag.length();

  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::ReadOnly ));
  session->transaction().start(true);
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::listEntries",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::AttributeList dataBuff;
  dataBuff.extend<std::string>(CONNECTION_COLUMN);
  dataBuff.extend<std::string>(ROLE_COLUMN);
  dataBuff.extend<std::string>(USERNAME_COLUMN);
  dataBuff.extend<std::string>(VERIFICATION_KEY_COLUMN);
  std::auto_ptr<coral::IQuery> query(schema.tableHandle(TABLE_NAME).newQuery());
  query->addToOutputList(CONNECTION_COLUMN);
  query->addToOutputList(ROLE_COLUMN);
  query->addToOutputList(USERNAME_COLUMN);
  query->addToOutputList(VERIFICATION_KEY_COLUMN);
  query->defineOutput(dataBuff);
  bool foundValid = false;
  bool anyEntry = false;
  coral::ICursor& cursor = query->execute();
  std::map<std::pair<std::string,std::string>,std::string> credentials;
  while ( cursor.next() ) {
    anyEntry = true;
    const coral::AttributeList& currentRow = cursor.currentRow();
    std::string connectionString = currentRow[CONNECTION_COLUMN].data<std::string>();
    std::string role = currentRow[ROLE_COLUMN].data<std::string>();
    std::string userName =  currentRow[USERNAME_COLUMN].data<std::string>();   
    std::string verifKey = currentRow[VERIFICATION_KEY_COLUMN].data<std::string>();
    if(!m_key.empty()){
      connectionString = coral::Cipher::decode(connectionString,m_key);
      role = coral::Cipher::decode(role,m_key);
      userName =  coral::Cipher::decode(userName,m_key);  
      verifKey = coral::Cipher::decode(verifKey,m_key);
    }
    // only consider entries where the verification key is matching...
    if( VERIFICATION_KEY == verifKey ){
      foundValid = true;
      credentials.insert(std::make_pair(std::make_pair(connectionString,role),userName));
      if(connectionString.length()>serviceMaxW) serviceMaxW = connectionString.length();
      if(role!=coral::IAuthenticationCredentials::defaultRole()){
        if(role.length()>roleMaxW) roleMaxW = role.length();
      }
      if(userName.length()>userMaxW) userMaxW = userName.length();
    }
    
  }
  session->transaction().commit();
  
  if(foundValid){
    out << "Found "<<credentials.size()<<" entries."<<std::endl;
    std::string space("  ");
    out << std::setiosflags(std::ios_base::left);
    out <<space<<std::setw(serviceMaxW)<<serviceTag;
    out <<space<<std::setw(roleMaxW)<<roleTag;
    out <<space<<std::setw(userMaxW)<<userTag;
    out <<std::endl;
    
    out <<space<<std::setfill('-')<<std::setw(serviceMaxW)<<"";
    out <<space<<std::setw(roleMaxW)<<"";
    out <<space<<std::setw(userMaxW)<<"";
    out <<std::endl;
    out << std::setfill(' ');
    for(std::map<std::pair<std::string,std::string>,std::string>::const_iterator iE=credentials.begin();
        iE!=credentials.end();++iE){
      std::string roleV = iE->first.second;
      if(roleV == coral::IAuthenticationCredentials::defaultRole()) roleV = "";
      out <<space<<std::setw(serviceMaxW)<<iE->first.first;
      out <<space<<std::setw(roleMaxW)<<roleV;
      out <<space<<std::setw(userMaxW)<<iE->second;
      out << std::endl;
    }
    out <<space<<std::setfill('-')<<std::setw(serviceMaxW)<<"";
    out <<space<<std::setw(roleMaxW)<<"";
    out <<space<<std::setw(userMaxW)<<"";
    out <<std::endl;
    out <<std::endl;
  } else {
    if(!anyEntry) {
      out << "No entries found in the provided credential repository." << std::endl;
    } else {
      out << "No credential item found for the provided key." << std::endl;
    }
  }
  
  return foundValid;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::initCredentialAccess(){
  std::string initAuthenticationService("");
  char* authPath = ::getenv ( "CORAL_AUTH_PATH" );
  if ( authPath ) {
    initAuthenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  else {
    if ( ::getenv ( "CORAL_AUTH_USER" ) && ::getenv ( "CORAL_AUTH_PASSWORD" ) )
      initAuthenticationService = "CORAL/Services/EnvironmentAuthenticationService";
    else
      initAuthenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  coral::IHandle<coral::IAuthenticationService> initServ =
    coral::Context::instance().query<coral::IAuthenticationService>(initAuthenticationService);
  if(!initServ.isValid()){
    coral::Context::instance().loadComponent( initAuthenticationService );
    initServ = coral::Context::instance().query<coral::IAuthenticationService>( initAuthenticationService );
  }
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Retriving initializing credential data from service \""<< (dynamic_cast<coral::Service*>(initServ.get()))->name()<<"\""<<coral::MessageStream::endmsg;
  try {
    std::string initUserName = initServ->credentials( m_connectionString ).valueForItem( IAuthenticationCredentials::userItem() );
    std::string initPassword = initServ->credentials( m_connectionString ).valueForItem( IAuthenticationCredentials::passwordItem() );
    coral::AuthenticationCredentialSet* initialSet = new coral::AuthenticationCredentialSet( this->name(),m_connectionString );
    initialSet->appendCredentialItem( coral::IAuthenticationCredentials::userItem(), initUserName );
    initialSet->appendCredentialItem( coral::IAuthenticationCredentials::passwordItem(), initPassword );
    log << coral::Debug << "Loading initial credentials for service \""<< m_connectionString <<"\""<<coral::MessageStream::endmsg;
    m_data.insert( std::make_pair( m_connectionString, initialSet ) );
  } catch (const coral::UnknownConnectionException& ex){
    throw coral::AuthenticationServiceException("Cannot access credential database, initial credentials have not been provided.",
                                                "RelationalAuthenticationService::initCredentialAcces",
                                                "coral::RelationalAuthenticationService");    
  }
  return true;
}

bool
coral::RelationalAuthenticationService::RelationalAuthenticationService::loadCredentialData()
{
  if(m_connectionString.empty()){
    throw coral::AuthenticationServiceException("Credential database has not been provided.",
                                                "RelationalAuthenticationService::initialize",
                                                "coral::RelationalAuthenticationService");    
  }
  m_isLoading = true;
    
  coral::IHandle<coral::IConnectionService> connServ = coral::Context::instance().query<coral::IConnectionService>();
  if(!connServ.isValid()){
    coral::Context::instance().loadComponent("CORAL/Services/ConnectionService");
    connServ = coral::Context::instance().query<coral::IConnectionService>();
  }
  
  std::auto_ptr<coral::ISessionProxy> session(connServ->connect(m_connectionString, coral::ReadOnly ));
  session->transaction().start(true);
  coral::ISchema& schema = session->nominalSchema();
  if(!schema.existsTable(TABLE_NAME)){
    session->transaction().commit();
    throw coral::AuthenticationServiceException("Cannot find credential database.",
                                                "RelationalAuthenticationService::loadCredentialData",
                                                "coral::RelationalAuthenticationService");    
  }
  coral::AttributeList dataBuff;
  dataBuff.extend<std::string>(CONNECTION_COLUMN);
  dataBuff.extend<std::string>(ROLE_COLUMN);
  dataBuff.extend<std::string>(USERNAME_COLUMN);
  dataBuff.extend<std::string>(PASSWORD_COLUMN);
  dataBuff.extend<std::string>(VERIFICATION_KEY_COLUMN);
  std::auto_ptr<coral::IQuery> query(schema.tableHandle(TABLE_NAME).newQuery());
  query->defineOutput(dataBuff);
  bool foundValid = false;
  bool anyRow = false;
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    anyRow = true;
    const coral::AttributeList& currentRow = cursor.currentRow();
    std::string connectionString = currentRow[CONNECTION_COLUMN].data<std::string>();
    std::string role = currentRow[ROLE_COLUMN].data<std::string>();
    std::string userName =  currentRow[USERNAME_COLUMN].data<std::string>();   
    std::string password =  currentRow[PASSWORD_COLUMN].data<std::string>();
    std::string verifKey = currentRow[VERIFICATION_KEY_COLUMN].data<std::string>();
    if(!m_key.empty()){
      connectionString = coral::Cipher::decode(connectionString,m_key);
      role = coral::Cipher::decode(role,m_key);
      userName =  coral::Cipher::decode(userName,m_key);  
      password =  coral::Cipher::decode(password,m_key);
      verifKey = coral::Cipher::decode(verifKey,m_key);
    }
    // only consider entries where the verification key is matching...
    if( VERIFICATION_KEY == verifKey ){
      foundValid = true;
      coral::AuthenticationCredentialSet* credential = 0;
      std::map< std::string, coral::AuthenticationCredentialSet* >::iterator iConnection = m_data.find( connectionString );
      if ( iConnection != m_data.end() ) {
        credential = iConnection->second;
      } else {
        credential = new coral::AuthenticationCredentialSet( this->name(), connectionString );
        m_data.insert( std::make_pair( connectionString, credential ) );
      }
      coral::MessageStream log( this->name() );
      log << coral::Debug << "Loading credential data for connection \""<<connectionString<<"\""<<coral::MessageStream::endmsg;
      if(role==coral::IAuthenticationCredentials::defaultRole()){
        credential->appendCredentialItem( coral::IAuthenticationCredentials::userItem(), userName );
        credential->appendCredentialItem( coral::IAuthenticationCredentials::passwordItem(), password );
      } else {
        credential->appendCredentialItemForRole( coral::IAuthenticationCredentials::userItem(), userName, role );
        credential->appendCredentialItemForRole( coral::IAuthenticationCredentials::passwordItem(), password, role );
      }
    } else {
      coral::MessageStream log( this->name() );
      if(m_key.empty()){
        log << coral::Debug << "Skipping credential item, required key has not been provided."<<coral::MessageStream::endmsg;
      } else {
        log << coral::Debug << "Skipping credential item, provided key is invalid."<<coral::MessageStream::endmsg;
      }
      
    }
    
  }
  session->transaction().commit();
  if(!foundValid){
    coral::MessageStream log( this->name() );
    if(anyRow){
      if(m_key.empty()){
        log << coral::Error << "No credential item found, required key has not been provided."<<coral::MessageStream::endmsg;
      } else {
        log << coral::Error << "No credential item found for the provided key."<<coral::MessageStream::endmsg;
      } 
    } else {
      log << coral::Debug << "No credential item found."<<coral::MessageStream::endmsg;        
    }
 }
  
  m_isLoading = false;
  m_isLoaded = true;
  return true;
}

void coral::RelationalAuthenticationService::RelationalAuthenticationService::reset(){
  for ( std::map< std::string, coral::AuthenticationCredentialSet* >::iterator iConnection = m_data.begin();
        iConnection != m_data.end(); ++iConnection ) delete iConnection->second;
  m_data.clear();
  m_isLoaded = false;  
}



const coral::IAuthenticationCredentials&
coral::RelationalAuthenticationService::RelationalAuthenticationService::credentials( const std::string& connectionString ) const
{
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Looking up credential data for connection \""<<connectionString<<"\""<<coral::MessageStream::endmsg;
  //boost::mutex::scoped_lock lock(m_mutexLock);
  bool lookUpOk = m_isLoaded;
  if(m_isLoading){
    const_cast< coral::RelationalAuthenticationService::RelationalAuthenticationService* >( this )->initCredentialAccess();
    lookUpOk = true;
  }
  if ( ! lookUpOk ) {
    const_cast< coral::RelationalAuthenticationService::RelationalAuthenticationService* >( this )->loadCredentialData();
  }
  std::map< std::string, coral::AuthenticationCredentialSet* >::const_iterator iConnection = m_data.find( connectionString );
  if ( iConnection == m_data.end() )
    throw coral::UnknownConnectionException( this->name(), connectionString );
  return iConnection->second->credentials();
}


const coral::IAuthenticationCredentials&
coral::RelationalAuthenticationService::RelationalAuthenticationService::credentials( const std::string& connectionString,
                                                                                      const std::string& role ) const
{
  coral::MessageStream log( this->name() );
  log << coral::Debug << "Looking up credential data for connection \""<<connectionString<<"\" and role \""<<role<<"\"."<<coral::MessageStream::endmsg;
  //boost::mutex::scoped_lock lock(m_mutexLock);
  bool lookUpOk = m_isLoaded;
  if(m_isLoading){
    const_cast< coral::RelationalAuthenticationService::RelationalAuthenticationService* >( this )->initCredentialAccess();
    lookUpOk = true;
  }
  if ( ! lookUpOk ) {
    const_cast< coral::RelationalAuthenticationService::RelationalAuthenticationService* >( this )->loadCredentialData();
  }
  std::map< std::string, coral::AuthenticationCredentialSet* >::const_iterator iConnection = m_data.find( connectionString );
  if ( iConnection == m_data.end() )
    throw coral::UnknownConnectionException( this->name(), connectionString );
  return iConnection->second->credentials( role );
}

