#include "Schema.h"
#include "SessionProperties.h"
#include "SQLiteTableBuilder.h"
#include "DomainProperties.h"
#include "SQLiteStatement.h"
#include "Table.h"
#include "Query.h"
#include "ViewFactory.h"
#include "View.h"

#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/ITableSchemaEditor.h"
#include "RelationalAccess/SchemaException.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "CoralKernel/Service.h"

#include <string>

coral::SQLiteAccess::Schema::Schema(coral::SQLiteAccess::SessionProperties& sessionProperties):
  m_sessionProperties(sessionProperties),
  m_tables(),
  m_tablesReadFromDataDictionary( false ),
  m_views(),
  m_viewsReadFromDataDictionary( false )
{}

coral::SQLiteAccess::Schema::~Schema()
{
  this->reactOnEndOfTransaction();
}

std::string coral::SQLiteAccess::Schema::schemaName() const
{
  return "";
}

std::set<std::string>
coral::SQLiteAccess::Schema::listTables() const{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::listTables" );
  std::set<std::string> result;
  if ( ! m_tablesReadFromDataDictionary ){
    this->readTablesFromDataDictionary();
  }
  for ( std::map< std::string, coral::SQLiteAccess::Table* >::const_iterator iTable = m_tables.begin(); iTable != m_tables.end(); ++iTable ){
    result.insert( iTable->first );
  }
  return result;
}

bool
coral::SQLiteAccess::Schema::existsTable( const std::string& name ) const
{
  if ( ! m_sessionProperties.isTransactionActive() ){
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::existsTable" );
  }
  // Check if the table is already in the local cache.
  if ( m_tables.find( name ) == m_tables.end() ) {
    // Check if the data dictionary was read
    if ( m_tablesReadFromDataDictionary ) {
      return false;
    }else{
      // Check from the data dictionary
      readTablesFromDataDictionary();
      if ( m_tables.find( name ) == m_tables.end() ) return false;
      //coral::SQLiteAccess::Table* table = new coral::SQLiteAccess::Table( m_sessionProperties,name);
      //m_tables.insert( std::make_pair( name, table ) );
      return true;
    }
  }
  return true;
}

coral::ITable&
coral::SQLiteAccess::Schema::createTable( const coral::ITableDescription& description )
{
  //error if table exists.
  if ( ! m_sessionProperties.isTransactionActive() ){
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::createTable" );    
  }  // Get the name of the table
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::createTable" );
  std::string name = description.name();
  // Check if a table with this name already exists
  if ( this->existsTable( name ) ) {
    throw coral::TableAlreadyExistingException( m_sessionProperties.domainProperties().service()->name(),name);
  }
  
  // Construct and execute the statement creating the table
  coral::SQLiteAccess::SQLiteTableBuilder builder( description, m_sessionProperties);
  coral::SQLiteAccess::SQLiteStatement statement( m_sessionProperties);
  statement.prepare( builder.statement() );
  if ( ! statement.execute() ){
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),"Could not create a new table","ISchema::createTable" );
  }
  //register the new table
  coral::SQLiteAccess::Table* t=new coral::SQLiteAccess::Table(m_sessionProperties,name);
  m_tables.insert( std::make_pair(name,t) );
  // Get the table handle
  coral::ITable& table = this->tableHandle(name);
  // Use now the schema editor to add the indices
  coral::ITableSchemaEditor& editor = table.schemaEditor();
  for ( int i = 0; i < description.numberOfIndices(); ++i ) {
    const coral::IIndex& index = description.index( i );
    editor.createIndex( index.name(), index.columnNames(), index.isUnique(), index.tableSpaceName() );
  }
  return table;
}

void
coral::SQLiteAccess::Schema::dropTable( const std::string& name )
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );  
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::dropTable "<<name<<coral::MessageStream::endmsg;
#endif
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::dropTable" );
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::dropTable" );
  if ( ! this->existsTable( name ) ) {
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),name,"ISchema::dropTable" );
  }
  std::string query("DROP TABLE \"");
  query+=name;
  query+="\"";
  SQLiteStatement statement(m_sessionProperties);
  statement.prepare(query);
  if( !statement.execute() ){
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(), "Could not drop a table", "ISchema::dropTable" );
  }
  std::map< std::string, coral::SQLiteAccess::Table* >::iterator iTable = m_tables.find( name );
  if ( iTable->second ) delete iTable->second;
  m_tables.erase( iTable );
  
  statement.reset();
  query="DELETE FROM coral_sqlite_fk WHERE tablename=\'";
  query+=name;
  query+="\'";
  if(!statement.prepare(query) ){
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),"Could not clean table coral_sqlite_fk","Schema::dropTable" );
  }
  statement.execute();
  statement.reset();
}
  
void
coral::SQLiteAccess::Schema::dropIfExistsTable( const std::string& name )
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(), "ISchema::dropIfExistsTable" );
  if ( this->existsTable( name ) ) {
    this->dropTable(name);
  }
}

coral::ITable&
coral::SQLiteAccess::Schema::tableHandle( const std::string& name )
{
  /*#ifdef _DEBUG
  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
                           m_sessionProperties.domainProperties().service()->name(),
                           coral::Nil );  
  log<<coral::Debug<<"Schema::tableHandle "<<name<<coral::MessageStream::endmsg;
#endif
  */
  if ( ! m_sessionProperties.isTransactionActive() ){
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::tableHandle" );
  }
  // Check if the table is already in the local cache
  std::map< std::string, coral::SQLiteAccess::Table* >::iterator iTable = m_tables.find( name );
  if ( iTable != m_tables.end() ){ //found in local cache
    if( !iTable->second ){
      iTable->second=new coral::SQLiteAccess::Table(m_sessionProperties,name);
    }
    return *( iTable->second );
  }
  // Check if the data dictionary was read
  if ( m_tablesReadFromDataDictionary ){
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(), name,"ISchema::tableHandle" );
  }
  readTablesFromDataDictionary();
  iTable = m_tables.find( name );
  if ( iTable == m_tables.end() ) {
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),name,"ISchema::tableHandle" );
  }
  coral::SQLiteAccess::Table* table = new coral::SQLiteAccess::Table( m_sessionProperties,name);
  iTable->second = table;
  return *( iTable->second );
}


void
coral::SQLiteAccess::Schema::truncateTable( const std::string& tableName ){
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::truncateTable" );
  if ( ! this->existsTable( tableName ) )
    throw coral::TableNotExistingException( m_sessionProperties.domainProperties().service()->name(),tableName, "truncateTable" );
  std::string query( "DELETE FROM \"" );
  query+=tableName;
  query+="\"";  
  coral::SQLiteAccess::SQLiteStatement statement( m_sessionProperties );
  statement.prepare(query);
  if ( ! statement.execute() )
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(),"Could not truncate table \"" + tableName + "\"","ISchema::truncateTable" );
}

void 
coral::SQLiteAccess::Schema::callProcedure( const std::string& procedureName,
					    const coral::AttributeList& /* inputArguments */ ){
  throw coral::SchemaException(m_sessionProperties.domainProperties().service()->name(),"Operation not supported, could not call prodecure \"" + procedureName + "\"","ISchema::callProcedure");
}

coral::IQuery*
coral::SQLiteAccess::Schema::newQuery() const 
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::newQuery" );
  return new coral::SQLiteAccess::Query( m_sessionProperties );
}

coral::IViewFactory*
coral::SQLiteAccess::Schema::viewFactory()
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::viewFactory" );
   if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ISchema::viewFactory" );
  return new coral::SQLiteAccess::ViewFactory( m_sessionProperties );
}

bool
coral::SQLiteAccess::Schema::existsView( const std::string& name ) const
{
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::existsView" );
  // Check if the view is already in the local cache.
  if ( m_views.find( name ) == m_views.end() ) {
    // Check if the data dictionary was read
    if ( m_viewsReadFromDataDictionary ) {
      return false;
    }else{
      // Check inside the database.
      readViewsFromDataDictionary();
      if ( m_views.find( name ) == m_views.end() ) return false;
      // The view exists. Create the view object and insert it into the map
      coral::SQLiteAccess::View* view = new coral::SQLiteAccess::View( m_sessionProperties, name);
      m_views.insert( std::make_pair( name, view ) );
      return true;
    }
  }
  return true;
}


void
coral::SQLiteAccess::Schema::dropView( const std::string& name )
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );  
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::dropView "<<name<<coral::MessageStream::endmsg;
#endif
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::dropView" );
  if ( m_sessionProperties.isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),"ISchema::dropView" );
  if ( ! this->existsView( name ) ) {
    throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(),name,"ISchema::dropView" );
  }
  std::string query("DROP VIEW \"");
  query+=name;
  query+="\"";
  SQLiteStatement statement(m_sessionProperties);
  statement.prepare(query);
  if( !statement.execute() ){
    throw coral::SchemaException( m_sessionProperties.domainProperties().service()->name(), "Could not drop a view", "ISchema::dropView" );
  }
  std::map< std::string, coral::SQLiteAccess::View* >::iterator iView = m_views.find( name );
  if ( iView->second ) delete iView->second;
  m_views.erase( iView);
}


void
coral::SQLiteAccess::Schema::dropIfExistsView( const std::string& name )
{
  
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::dropIfExistsView" );
  if ( this->existsView( name ) ) {
    this->dropView(name);
  }
}


std::set<std::string>
coral::SQLiteAccess::Schema::listViews() const
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );  
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::listViews "<<coral::MessageStream::endmsg;
#endif
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::listViews" );
  std::set<std::string> result;
  if ( ! m_viewsReadFromDataDictionary ){
    this->readViewsFromDataDictionary();
  }
  for ( std::map< std::string, coral::SQLiteAccess::View* >::const_iterator iView = m_views.begin(); iView != m_views.end(); ++iView ){
    result.insert( iView->first );
  }
  return result;
}  


coral::IView&
coral::SQLiteAccess::Schema::viewHandle( const std::string& name )
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );  
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::viewHandle "<<name<<coral::MessageStream::endmsg;
#endif
  if ( ! m_sessionProperties.isTransactionActive() )
    throw coral::TransactionNotActiveException( m_sessionProperties.domainProperties().service()->name(),"ISchema::viewHandle" );
  // Check if the view is already in the local cache
  std::map< std::string, coral::SQLiteAccess::View* >::iterator iView = m_views.find( name );
  if ( iView != m_views.end() ){ //found in local cache
    if( !iView->second ){
      iView->second=new coral::SQLiteAccess::View(m_sessionProperties,name);
    }
    return *( iView->second );
  }
  // Check if the data dictionary was read
  readViewsFromDataDictionary();
  m_viewsReadFromDataDictionary=true;
  iView = m_views.find( name );
  if ( iView == m_views.end() ) {
    throw coral::ViewNotExistingException( m_sessionProperties.domainProperties().service()->name(),name,"ISchema::viewHandle" );
  }
  coral::SQLiteAccess::View* view = new coral::SQLiteAccess::View( m_sessionProperties,name);
  iView->second = view;
  return *( iView->second );
}

void
coral::SQLiteAccess::Schema::reactOnEndOfTransaction()
{
  for ( std::map< std::string, coral::SQLiteAccess::Table* >::iterator iTable = m_tables.begin(); iTable != m_tables.end(); ++iTable )
    if ( iTable->second ){
      delete iTable->second;
    }
  for ( std::map< std::string, coral::SQLiteAccess::View* >::iterator iView = m_views.begin(); iView != m_views.end(); ++iView )
    if ( iView->second ){
      delete iView->second;
    }
  m_tables.clear();
  m_tablesReadFromDataDictionary = false;
  m_views.clear();
  m_viewsReadFromDataDictionary = false;
}

void
coral::SQLiteAccess::Schema::readTablesFromDataDictionary() const
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );  
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::readTablesFromDataDictionary"<<coral::MessageStream::endmsg;
#endif
  coral::SQLiteAccess::SQLiteStatement stmt( m_sessionProperties );
  stmt.prepare("SELECT name FROM sqlite_master WHERE type='table' UNION ALL SELECT name FROM sqlite_temp_master WHERE type='table' ORDER BY 1");
  coral::AttributeList output;
  output.extend<std::string>( "name" );
  while( stmt.fetchNext() ){
    if( stmt.defineOutput( output ) ){
      const std::string& tableName = output.begin()->data<std::string>();
      if ( m_tables.find( tableName ) == m_tables.end() ) {
        m_tables[tableName]=0; //create 0 handle
      }
    }
  }
  m_tablesReadFromDataDictionary = true;
}

void
coral::SQLiteAccess::Schema::readViewsFromDataDictionary() const
{
#ifdef _DEBUG
//  coral::MessageStream log( m_sessionProperties.domainProperties().service(),
//                           m_sessionProperties.domainProperties().service()->name(),
//                           coral::Nil );
  coral::MessageStream log( m_sessionProperties.domainProperties().service()->name() );  
  log<<coral::Debug<<"Schema::readViewsFromDataDictionary"<<coral::MessageStream::endmsg;
#endif
  coral::SQLiteAccess::SQLiteStatement stmt( m_sessionProperties );
  stmt.prepare("SELECT name FROM sqlite_master WHERE type='view' UNION ALL SELECT name FROM sqlite_temp_master WHERE type='view' ORDER BY 1");
  coral::AttributeList output;
  output.extend<std::string>( "name" );
  while( stmt.fetchNext() ){
    if( stmt.defineOutput( output ) ){
      if( m_views.find( output["name"].data<std::string>() ) == m_views.end() ) {
        m_views[output["name"].data<std::string>()]=0; 
      }
    }
  }
  m_viewsReadFromDataDictionary = true;
}
