#include "Session.h"
#include "SessionProperties.h"
#include "ConnectionProperties.h"
#include "DomainProperties.h"
#include "MonitorController.h"
#include "Transaction.h"
#include "Schema.h"

#include "RelationalAccess/MonitoringException.h"
#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/IAuthenticationCredentials.h"
#include "RelationalAccess/IMonitoringService.h"

#include "CoralBase/SearchPath.h"
#include "CoralBase/MessageStream.h"

#include "CoralCommon/URIParser.h"
#include "CoralCommon/MonitoringEventDescription.h"

#include "CoralKernel/Service.h"

#include "sqlite3.h"

#include <stdlib.h>

#include <string>

#ifdef WIN32
#include <io.h>
namespace coral
{
  bool isReadonly( const std::string& fName )
  {
    return( ::_access( fName.c_str(), 02 ) != -1 );
  }
}
#else // U**x like
#include <unistd.h>
namespace coral
{
  bool isReadonly( const std::string& fName )
  {
    return( ::access( fName.c_str(), R_OK ) != -1 );
  }
}
#endif
//#include <iostream>
coral::SQLiteAccess::Session::Session(coral::IDevConnection& connection, coral::SQLiteAccess::ConnectionProperties& connectionProperties, const std::string& /* schemaName */, coral::AccessMode mode)
  : coral::IDevSession( connection )
  , m_transaction(0)
  , m_schema(0)
  , m_dbHandle(0)
{
  m_properties = new coral::SQLiteAccess::SessionProperties(connectionProperties,static_cast<coral::ISession&>( *this ), mode == coral::ReadOnly);
  m_monitorController = new coral::SQLiteAccess::MonitorController( *m_properties );
}
coral::SQLiteAccess::Session::~Session()
{
  if ( this->isUserSessionActive() ) this->endUserSession();  
  delete m_monitorController;
  delete m_properties;
}
coral::IMonitoring&
coral::SQLiteAccess::Session::monitoring()
{
  return *m_monitorController;
}

void
coral::SQLiteAccess::Session::startUserSession( const std::string& /* userName */ ,	const std::string&  /* password */)
{
  if ( this->isUserSessionActive() ) return;
  
//  if ( m_properties->isReadOnly() && !m_properties->connectionProperties().databaseFileName().isReadable() )
  if ( m_properties->isReadOnly() && !coral::isReadonly( m_properties->connectionProperties().databaseFileName().string() ) )
  {
    throw coral::AuthenticationFailureException( m_properties->domainProperties().service()->name() );
  }

int rstat=-1;
#if SQLITE_VERSION_NUMBER < 3006000
 rstat=sqlite3_open(m_properties->connectionProperties().databaseFileName().string().c_str(), &m_dbHandle);
#else
 if( m_properties->isReadOnly() ){
   rstat=sqlite3_open_v2(m_properties->connectionProperties().databaseFileName().string().c_str(),&m_dbHandle,SQLITE_OPEN_READONLY,NULL);
 }else{
   rstat=sqlite3_open_v2(m_properties->connectionProperties().databaseFileName().string().c_str(),&m_dbHandle,SQLITE_OPEN_READWRITE|SQLITE_OPEN_CREATE,NULL);
 }
#endif
 if(rstat!=SQLITE_OK){
    this->endUserSession();
    throw coral::SessionException( "Could not open file "+std::string(m_properties->connectionProperties().databaseFileName().string()), "Session::startUserSession", m_properties->domainProperties().service()->name() );
 }
  m_properties->setdbHandle(m_dbHandle);
  if ( getenv ( "CORAL_SQLITE_TEMP_STORE_MEMORY" ) ){
    std::stringstream cmd;
    cmd << "PRAGMA temp_store=MEMORY";
    //std::cout << cmd.str() << std::endl;
    sqlite3_exec(m_dbHandle, cmd.str().c_str(), 0, 0, 0 );
  }else if ( getenv ( "CORAL_SQLITE_TEMP_STORE_DIRECTORY" ) ) {
    std::stringstream cmd;
    cmd << "PRAGMA temp_store_directory='"
        << getenv ( "CORAL_SQLITE_TEMP_STORE_DIRECTORY" ) << "'";
    //std::cout << cmd.str() << std::endl;
    sqlite3_exec(m_dbHandle, cmd.str().c_str(), 0, 0, 0 );
  }

  // Create new schema and transaction objects.
  m_schema = new coral::SQLiteAccess::Schema( *m_properties );
  m_transaction = new coral::SQLiteAccess::Transaction( *m_properties );
  // Record the beginning of the session
  if ( m_properties->monitoringService() ) {
    try{
      m_properties->monitoringService()->record( m_properties->connectionString(),
                                                 coral::monitor::Session,
                                                 coral::monitor::Info,
                                                 monitoringEventDescription.sessionBegin() );
    }catch(const coral::MonitoringException& e){
      coral::MessageStream log( m_properties->domainProperties().service()->name() );  
      log<<coral::Error<<"caught  coral::MonitoringException in Session::startUserSession"<<e.what()<<coral::MessageStream::endmsg;
    }
  }
  return;
}

void
coral::SQLiteAccess::Session::endUserSession()
{
  if ( this->isUserSessionActive() ) {
    // Abort any active transaction
    if ( this->transaction().isActive() ) this->transaction().rollback();
    if ( m_transaction ) {
      delete m_transaction;
      m_transaction = 0;
    }
    if ( m_schema ) {
      delete m_schema;
      m_schema = 0;
    }
    if ( m_properties->monitoringService() ) {
      m_properties->monitoringService()->record( m_properties->connectionString(),
						 coral::monitor::Session,
						 coral::monitor::Info,
						 monitoringEventDescription.sessionEnd() );
    }
    sqlite3_close( m_dbHandle );
    m_dbHandle=0;
    m_properties->setdbHandle(0);
  }
}


bool
coral::SQLiteAccess::Session::isUserSessionActive() const
{
  return ( m_dbHandle!=0 );
}

coral::ITransaction&
coral::SQLiteAccess::Session::transaction()
{
  if ( ! this->isUserSessionActive() ) {
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(), "ISession::transaction" );
  }
  return *m_transaction;
}
coral::ISchema& 
coral::SQLiteAccess::Session::nominalSchema(){
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );
  return *m_schema;
}
coral::ISchema&
coral::SQLiteAccess::Session::schema( const std::string& /* schemaName */)
{
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );
  return *m_schema;
}

sqlite3* 
coral::SQLiteAccess::Session::getdbHandle() const
{
  return m_dbHandle;
}
