#include <iostream>
#include <stdexcept>

#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/RelationalServiceException.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/Exception.h"
#include "CoralBase/AttributeSpecification.h"

#include "CoralKernel/Context.h"

#include <string>

int main( int, char** )
{
  std::string con = "sqlite_file:zhen.db";
  
  try
  {
    coral::Context& ctx = coral::Context::instance();

    ctx.loadComponent( "CORAL/RelationalPlugins/sqlite" );
    
    coral::IHandle<coral::IRelationalDomain> iHandle=ctx.query<coral::IRelationalDomain>( "CORAL/RelationalPlugins/sqlite" );
    
    if ( ! iHandle.isValid() )
    {
      throw coral::NonExistingDomainException( "sqlite" );
    }
    
    std::pair<std::string, std::string> connectionAndSchema = iHandle->decodeUserConnectionString( con );
    coral::IConnection*  connection=iHandle->newConnection( connectionAndSchema.first );
    connection->connect();
    coral::ISession* session=connection->newSession(connectionAndSchema.second);  
    std::string userName, password;
    session->startUserSession( userName, password );
    //
    // create test table x1
    //
    session->transaction().start(false);
    session->nominalSchema().dropIfExistsTable( "x1" );
    std::cout<<"create table x1 (id INT PRIMARY KEY, name STRING NOT NUUL UNIQUE, address STRING UNIQUE, order INT)"<<std::endl;
    coral::TableDescription description1;
    description1.setName("x1");
    description1.insertColumn("id", coral::AttributeSpecification::typeNameForId( typeid(int) ) );
    description1.insertColumn("name", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    description1.insertColumn("address", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    description1.insertColumn("order", coral::AttributeSpecification::typeNameForId( typeid(int) ) );
    description1.setPrimaryKey( "id" );
    description1.setNotNullConstraint( "name" );
    description1.setUniqueConstraint( "name" );
    description1.setUniqueConstraint( "address" );
    session->nominalSchema().createTable(description1 );
    session->transaction().commit();
    
    //
    // create test table x2 (combined PK and indices)
    //
    session->transaction().start(false);
    session->nominalSchema().dropIfExistsTable( "x2" );
    std::cout<<"create table x2 (id INT, name STRING NOT NUUL, address STRING) primary key(id,name)"<<std::endl;
    coral::TableDescription description2;
    description2.setName("x2");
    description2.insertColumn("id", coral::AttributeSpecification::typeNameForId( typeid(int) ) );
    description2.insertColumn("name", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    description2.insertColumn("address", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    std::vector<std::string> pks;
    pks.push_back("id");
    pks.push_back("name");
    description2.setPrimaryKey( pks, "" );
    description2.setNotNullConstraint( "name" );
    session->nominalSchema().createTable(description2 );
    session->transaction().commit();

    //
    // create test table x3 (foreign keys)
    //
    session->transaction().start(false);
    session->nominalSchema().dropIfExistsTable( "x3" );
    std::cout<<"create table x3 (id INT primary key, name STRING NOT NUUL, address STRING) foreign key(id) references x2(id)"<<std::endl;
    coral::TableDescription description3;
    description3.setName("x3");
    description3.insertColumn("id", coral::AttributeSpecification::typeNameForId( typeid(int) ) );
    description3.insertColumn("name", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    description3.insertColumn("address", coral::AttributeSpecification::typeNameForId( typeid(std::string) ) );
    description3.setPrimaryKey( "id", "" );
    description3.setNotNullConstraint( "name" );
    description3.createForeignKey("","id","x2","id");
    session->nominalSchema().createTable(description3 );
    session->transaction().commit();

    //
    // describe tables
    //
    session->transaction().start();
    coral::ISchema& schema= session->nominalSchema();
    std::set<std::string> tables=schema.listTables();
    for( std::set<std::string>::iterator it= tables.begin(); it!=tables.end(); ++it){
      
      coral::ITable& tab=schema.tableHandle(*it);
      const coral::ITableDescription& desc=tab.description();
      std::cout<<"Table name: "<<desc.name()<<std::endl;
      std::cout<<"Number of columns: "<<desc.numberOfColumns()<<std::endl;
      for( int i=0; i<desc.numberOfColumns(); ++i ){
	std::cout<<"   columnName  "<<desc.columnDescription(i).name()<<std::endl;
      }
      std::cout<<"Has primary key "<<desc.hasPrimaryKey()<<std::endl;
      const coral::IPrimaryKey& mypk=desc.primaryKey();
      const std::vector<std::string>& colNames=mypk.columnNames();
      for( std::vector<std::string>::const_iterator it=colNames.begin(); it!=colNames.end(); ++it ){
	std::cout<<"PK columnName  "<<*it<<std::endl;
      }
      std::cout<<"Number of indices: "<<desc.numberOfIndices()<<std::endl;
      std::cout<<"Number of foreign keys: "<<desc.numberOfForeignKeys()<<std::endl;
      try{
	const coral::IForeignKey& fk=desc.foreignKey(0);
	std::cout<<" foreign key 0 name "<<fk.name()<<std::endl;
	std::cout<<" foreign key referenced table name "<<fk.referencedTableName()<<std::endl;     
      }catch(const coral::InvalidForeignKeyIdentifierException& ){
	std::cout<<"invalid foreign key"<<std::endl;
      } 
      std::cout<<"Number of unique constraints: "<<desc.numberOfUniqueConstraints()<<std::endl;
    }
    std::cout << "Committing..." << std::endl;
    session->transaction().commit();
    session->endUserSession();
    delete session;
    session=0;
    // Disconnecting
    std::cout << "Disconnecting..." << std::endl;
    connection->disconnect();
    delete connection;
  }
  catch ( coral::Exception& se ) {
    std::cerr << "CORAL Exception : " << se.what() << std::endl;
    return 1;
  }
  catch ( std::exception& e ) {
    std::cerr << "Standard C++ exception : " << e.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Exception caught (...)" << std::endl;
    return 1;
  }
  return 0;
}
