#include "Queries.h"
#include "TestEnv/TestingConfig.h"

Queries::Queries(TestEnv * env) : Testing(env)
{
}


Queries::~Queries()
{
}


void
Queries::run()
{
    session->transaction().start( true );

    coral::ISchema& workingSchema = session->nominalSchema();

    // Run a query on a table
    coral::IQuery* query0 = workingSchema.tableHandle(T1).newQuery();
    query0->addToOutputList( "Data" );
    query0->addToOutputList( "Z" );
    query0->addToOrderList( "id" );
    coral::ICursor& cursor0 = query0->execute();
    int i = 0;
    while ( cursor0.next() ) {
      std::cout << "Row " << i + 1 << std::endl;
      const coral::Blob& blob = cursor0.currentRow()[0].data<coral::Blob>();
      int expectedBlobSize = 1000 * ( i + 1 );
      if ( blob.size() != expectedBlobSize ) {
        std::cerr << "Error : returned blob size = " << blob.size() << " while expected is " << expectedBlobSize << std::endl;
        delete query0;
        throw std::runtime_error( "Unexpected blob data" );
      }
      const unsigned char* p = static_cast<const unsigned char*>( blob.startingAddress() );
      for ( int j = 0; j < expectedBlobSize; ++j, ++p ) {
        if ( *p != ( i + j )%256 ) {
          delete query0;
          throw std::runtime_error( "Unexpected blob data" );
        }
      }

      bool nullNess = ( i%4 == 2 ) ? true : false;
      if ( cursor0.currentRow()[1].isNull() != nullNess ) {
        delete query0;
        throw std::runtime_error( "Unexpected null-ness" );
      }
      ++i;
    }
    delete query0;
    if ( i != 7 ) throw std::runtime_error( "Unexpected number of returned rows" );

    // Run a wildcard query on the table
    coral::IQuery* query1 = workingSchema.newQuery();
    query1->addToTableList(T3);
    coral::ICursor& cursor1 = query1->execute();
    while ( cursor1.next() ) {
      cursor1.currentRow().toOutputStream( std::cout ) << std::endl;
    }
    delete query1;

    // Run a wildcard query on the view
    coral::IQuery* query2 = workingSchema.newQuery();
    query2->addToTableList(V0);
    coral::ICursor& cursor2 = query2->execute();
    while ( cursor2.next() ) {
      cursor2.currentRow().toOutputStream( std::cout ) << std::endl;
    }
    delete query2;

    // Run a query on a table
    coral::IQuery* query3 = workingSchema.tableHandle(T3).newQuery();
    query3->addToOutputList( "id" );
    query3->addToOrderList( "Qn" );
    query3->setRowCacheSize( 100 );
    query3->limitReturnedRows( 2, 2 );
    coral::ICursor& cursor3 = query3->execute();
    int nRows = 0;
    while ( cursor3.next() ) {
      cursor3.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query3;
    if ( nRows != 2 )
      throw std::runtime_error( "Unexpected number of rows returned #1" );

    // Perform a query involving two tables.
    coral::IQuery* query4 = workingSchema.newQuery();
    query4->addToOutputList(T1 + ".X", "x" );
    query4->addToOutputList(T1 + ".t", "t" );
    query4->addToOutputList(T3 + ".Qn", "Qn" );
    query4->addToTableList(T1);
    query4->addToTableList(T3);
    std::string condition = T1 + ".id = " + T3 + ".fk1 AND " + T1 + ".Z<:zmax";
    coral::AttributeList conditionData;
    conditionData.extend<float>( "zmax" );
    query4->setCondition( condition, conditionData );
    query4->setMemoryCacheSize( 5 );
    conditionData[0].data<float>() = 3;
    coral::ICursor& cursor4 = query4->execute();
    nRows = 0;
    while ( cursor4.next() ) {
      cursor4.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query4;
    if ( nRows != 2 )
      throw std::runtime_error( "Unexpected number of rows returned #2" );

    // Perform a query involving a subquery.
    coral::IQuery* query5 = workingSchema.newQuery();
    query5->addToOutputList( "A.t", "t" );
    query5->addToOutputList( "B.tx", "tx" );

    coral::IQueryDefinition& subQuery1 = query5->defineSubQuery( "A" );
    subQuery1.addToOutputList( "t" );
    subQuery1.addToOutputList( "id" );
    subQuery1.addToTableList(T1);
    coral::AttributeList cond1;
    cond1.extend<float>( "xmin" );
    subQuery1.setCondition( "X < :xmin", cond1 );
    query5->addToTableList( "A" );

    coral::IQueryDefinition& subQuery2 = query5->defineSubQuery( "B" );
    subQuery2.addToOutputList( "id" );
    subQuery2.addToOutputList( "TX", "tx" );
    subQuery2.addToTableList(V0);
    coral::AttributeList cond2;
    cond2.extend<float>( "tymax" );
    subQuery2.setCondition( "TX < :tymax", cond2 );
    query5->addToTableList( "B" );

    query5->setCondition( "A.id = B.id", coral::AttributeList() );

    
    //for mysql exact float values doesn't work propper
    //it could fluctuate between +0.1 and -0.1
    //cond1[0].data<float>() = 1.005;
    
    cond1[0].data<float>() = 1.0049;
    cond2[0].data<float>() = 210;
    nRows = 0;
    coral::ICursor& cursor5 = query5->execute();
    while ( cursor5.next() ) {
      cursor5.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query5;
    if ( nRows != 3 )
      throw std::runtime_error( "Unexpected number of rows returned #3" );

    session->transaction().commit();
 
}
