#include "TestDriver.h"

#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IViewFactory.h"
#include "RelationalAccess/IView.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"


#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include "CoralBase/Blob.h"
#include "CoralBase/TimeStamp.h"
#include "CoralBase/Date.h"

#include "CoralCommon/Utilities.h"

static const int rowsWithSimpleInsert = 10;
static const int rowsWithBulkInsert = 55;

InputOutput::InputOutput( TestEnv * env ) : Testing(env) {
}

void InputOutput::prepareSchema(){    
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  schema.dropIfExistsTable(T1);
  
  if ( m_rowBuffer ) delete m_rowBuffer;
  m_rowBuffer = new coral::AttributeList;
  coral::AttributeList& rowBuffer = *m_rowBuffer;
  
  coral::TableDescription description;
  description.setName(T1);
  description.insertColumn( "ID",
                            coral::AttributeSpecification::typeNameForId( typeid( int ) ) );
  description.setNotNullConstraint( "ID" );
  description.setPrimaryKey( "ID" );
  rowBuffer.extend<int>( "ID" );

  description.insertColumn( "V_B",
                            coral::AttributeSpecification::typeNameForId( typeid( bool ) ) );
  rowBuffer.extend<bool>( "V_B" );

  description.insertColumn( "V_UC",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned char ) ) );
  rowBuffer.extend<unsigned char>( "V_UC" );

  description.insertColumn( "V_SC",
                            coral::AttributeSpecification::typeNameForId( typeid( char ) ) );
  rowBuffer.extend<char>( "V_SC" );

  description.insertColumn( "V_US",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned short ) ) );
  rowBuffer.extend<unsigned short>( "V_US" );

  description.insertColumn( "V_SS",
                            coral::AttributeSpecification::typeNameForId( typeid( short ) ) );
  rowBuffer.extend<short>( "V_SS" );

  description.insertColumn( "V_UI",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned int ) ) );
  rowBuffer.extend<unsigned int>( "V_UI" );

  description.insertColumn( "V_SI",
                            coral::AttributeSpecification::typeNameForId( typeid( int ) ) );
  rowBuffer.extend<int>( "V_SI" );

  description.insertColumn( "V_UL",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned long ) ) );
  rowBuffer.extend<unsigned long>( "V_UL" );

  description.insertColumn( "V_SL",
                            coral::AttributeSpecification::typeNameForId( typeid( long ) ) );
  rowBuffer.extend<long>( "V_SL" );

  description.insertColumn( "V_ULL",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned long long ) ) );
  rowBuffer.extend<unsigned long long>( "V_ULL" );

  description.insertColumn( "V_SLL",
                            coral::AttributeSpecification::typeNameForId( typeid( long long ) ) );
  rowBuffer.extend<long long>( "V_SLL" );

  description.insertColumn( "V_F",
                            coral::AttributeSpecification::typeNameForId( typeid( float ) ) );
  rowBuffer.extend<float>( "V_F" );

  description.insertColumn( "V_D",
                            coral::AttributeSpecification::typeNameForId( typeid( double ) ) );
  rowBuffer.extend<double>( "V_D" );

  description.insertColumn( "V_LD",
                            coral::AttributeSpecification::typeNameForId( typeid( double ) ) );
  rowBuffer.extend<double>( "V_LD" );

  description.insertColumn( "V_BLOB",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::Blob ) ) );
  rowBuffer.extend<coral::Blob>( "V_BLOB" );

  description.insertColumn( "V_DATE",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::Date ) ) );
  rowBuffer.extend<coral::Date>( "V_DATE" );

  description.insertColumn( "V_TIMESTAMP",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::TimeStamp ) ) );
  rowBuffer.extend<coral::TimeStamp>( "V_TIMESTAMP" );

  description.insertColumn( "V_SG",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ) );
  rowBuffer.extend<std::string>( "V_SG" );

  description.insertColumn( "V_SF",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ),
                            5 );
  rowBuffer.extend<std::string>( "V_SF" );

  description.insertColumn( "V_SV",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ),
                            50,
                            false );
  rowBuffer.extend<std::string>( "V_SV" );

  schema.createTable( description );
  
  session->transaction().commit();

}

void InputOutput::writeSimple(){ 
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(T1);
  coral::ITableDataEditor& dataEditor = table.dataEditor();
  coral::AttributeList& rowBuffer = *m_rowBuffer;

  for ( int i = 0; i < rowsWithSimpleInsert; ++i ) {
    this->fillData( i );
    dataEditor.insertRow( rowBuffer );
  }

  session->transaction().commit();
}

void InputOutput::writeBulk(){
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(T1);
  coral::ITableDataEditor& dataEditor = table.dataEditor();
  coral::AttributeList& rowBuffer = *m_rowBuffer;

  coral::IBulkOperation* bulkOperation = dataEditor.bulkInsert( rowBuffer,
                                                                rowsWithSimpleInsert );

  for ( int i = rowsWithSimpleInsert; i < rowsWithSimpleInsert + rowsWithBulkInsert; ++i ) {
    this->fillData( i );
    bulkOperation->processNextIteration();
  }
  bulkOperation->flush();

  delete bulkOperation;

  session->transaction().commit();

}

void InputOutput::read(){
  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(T1);
  coral::IQuery* query = table.newQuery();
  query->addToOrderList( "ID" );
  query->defineOutput( *m_rowBuffer );
  query->setRowCacheSize( rowsWithSimpleInsert );

  coral::ICursor& cursor = query->execute();
  for ( int i = 0; i < rowsWithSimpleInsert + rowsWithBulkInsert; ++i ) {
    if ( ! cursor.next() )
      throw std::runtime_error( "Unexpected number of returned rows" );
    this->checkData( i );
  }
  if ( cursor.next() )
    throw std::runtime_error( "Unexpected number of returned rows" );

  delete query;

  session->transaction().commit();

}

