#include "RandomQuery.h"
#include "TestEnv/TestingConfig.h"

#include "RelationalAccess/ISessionProperties.h"
#include "RelationalAccess/ITypeConverter.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/IColumn.h"

namespace cool {
	
  /** @file types.h
   *
   * Type definitions of the basic C++ types that can be stored in COOL.
   * Also includes the definitions of types that are not yet supported.
   *
   * @author Andrea Valassi, Marco Clemencic and Sven A. Schmidt
   * @date 2004-11-05 
   */

  // -------------------------------------------------------------------------
  /// Type definitions for all basic C++ types known by COOL (both those
  /// already supported and those candidate to be supported in the future).
  // -------------------------------------------------------------------------

  typedef bool                      Bool;

  typedef signed char               SChar;
  typedef unsigned char             UChar;

  typedef short                     Int16;
  typedef unsigned short            UInt16;

  typedef int                       Int32;
  typedef unsigned int              UInt32;

  typedef long long                 Int64;
  typedef unsigned long long        UInt64;

  typedef UInt64                    UInt63;

  typedef float                     Float;
  typedef double                    Double;

  typedef std::string               String255;
  typedef std::string               String4k;
  typedef std::string               String64k;
  typedef std::string               String16M;  

  typedef coral::Blob               Blob64k;
  typedef coral::Blob               Blob16M;

}

RandomQuery::RandomQuery(TestEnv * env)
  : Testing(env)
{
  _tables.push_back(T1);
  _tables.push_back(T2);
  _tables.push_back(T3);
}

std::string
RandomQuery::createFullString(size_t maxc, size_t offset)
{
  std::string h;
  
  size_t pos = 0;
  while(pos < maxc){
    char c = ((pos + offset) % 256);
    h = h + c;  
    pos++;
  }
  return h;
}

void
RandomQuery::fillBlob(coral::Blob& blob, size_t maxc)
{
  blob.resize( maxc + 1 );
  unsigned char* p = static_cast<unsigned char*>( blob.startingAddress() );
  for ( size_t j = 0; j < maxc; ++j, ++p ) *p = rand() % 256;
} 

void
RandomQuery::setup()
{  
  session->transaction().start(false);

  //create the tables and sets up the values for the tests
  coral::TableDescription desc01;
  desc01.setName(T1);
  
  desc01.insertColumn( "id", coral::AttributeSpecification::typeNameForId( typeid(long long) ) );
  desc01.insertColumn( "name", "string", 255, false );      
  
  desc01.setPrimaryKey( "id" );
    
  session->nominalSchema().dropIfExistsTable( desc01.name() );
  
  coral::ITable& table01 = session->nominalSchema().createTable( desc01 );
  coral::AttributeList rowBuffer01;
  table01.dataEditor().rowBuffer( rowBuffer01 );
  for(size_t i = 1; i < 10; i++ ){
    rowBuffer01["id"].data<long long>() = i;
    rowBuffer01["name"].data<std::string>() = "row#1";
    table01.dataEditor().insertRow( rowBuffer01 );
  }
//  table.privilegeManager().grantToUser( "PUBLIC", coral::ITablePrivilegeManager::Select );

  coral::TableDescription desc02;
  desc02.setName(T2);
  
  desc02.insertColumn( "id", coral::AttributeSpecification::typeNameForId( typeid(long long) ) );
  desc02.insertColumn( "s01", coral::AttributeSpecification::typeNameForId(typeid(std::string)), 255, false );
  desc02.insertColumn( "s02", coral::AttributeSpecification::typeNameForId(typeid(std::string)), 255, false );
  desc02.insertColumn( "d01", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  
  coral::ITypeConverter& typec = session->typeConverter();
  
  std::set<std::string> types = typec.supportedSqlTypes();
  std::set<std::string>::iterator i;
  for(i = types.begin();i != types.end(); i++ ){
    std::cout << *i << std::endl;
  }
  if(types.find("CLOB") != types.end()){
    //we can use a clob here
  }
  
  desc02.insertColumn( "d02", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  
  desc02.setPrimaryKey( "id" );
    
  session->nominalSchema().dropIfExistsTable( desc02.name() );
  
  coral::ITable& table02 = session->nominalSchema().createTable( desc02 );
  coral::AttributeList rowBuffer02;
  table02.dataEditor().rowBuffer( rowBuffer02 );
  
  for(size_t i = 1; i < 10; i++ ){
    rowBuffer02["id"].data<long long>() = i;
    //all characters
    rowBuffer02["s01"].data<std::string>() = "hello world";
    //all viewable characters
    rowBuffer02["s02"].data<std::string>() = createFullString(94, 32);
    //fill a blob with random data
    fillBlob(rowBuffer02["d01"].data<coral::Blob>(), 5632);
    //fill a blob with random data
    fillBlob(rowBuffer02["d02"].data<coral::Blob>(), 5631);
    
    table02.dataEditor().insertRow( rowBuffer02 );
  }
  
  coral::MsgLevel oldLevel = coral::MessageStream::msgVerbosity();
  coral::MessageStream::setMsgVerbosity( coral::Verbose );  
  
  coral::TableDescription desc03;
  desc03.setName(T1);
  
  desc03.insertColumn( "NODE_ID", coral::AttributeSpecification::typeNameForId(typeid( long long )) );
  std::cout << "NODE_ID" << "\t" << coral::AttributeSpecification::typeNameForId(typeid( cool::UInt32 )) << std::endl;
  desc03.insertColumn( "NODE_PARENTID", coral::AttributeSpecification::typeNameForId(typeid(long long)));
  desc03.insertColumn( "NODE_NAME", coral::AttributeSpecification::typeNameForId(typeid(std::string)));
  desc03.insertColumn( "NODE_FULLPATH", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "NODE_DESCRIPTION", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "NODE_ISLEAF", coral::AttributeSpecification::typeNameForId(typeid(long long)) );
  desc03.insertColumn( "NODE_SCHEMA_VERSION", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "NODE_INSTIME", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "LASTMOD_DATE", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "FOLDER_VERSIONING", coral::AttributeSpecification::typeNameForId(typeid(long long)) );
  desc03.insertColumn( "FOLDER_PAYLOADSPEC", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  desc03.insertColumn( "FOLDER_PAYLOAD_INLINE", coral::AttributeSpecification::typeNameForId(typeid(long long)) );
  desc03.insertColumn( "FOLDER_PAYLOAD_EXTREF", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  desc03.insertColumn( "FOLDER_CHANNELSPEC", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  desc03.insertColumn( "FOLDER_CHANNEL_EXTREF", coral::AttributeSpecification::typeNameForId(typeid(coral::Blob)) );
  desc03.insertColumn( "FOLDER_IOVTABLENAME", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "FOLDER_TAGTABLENAME", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "FOLDER_IOV2TAGTABLENAME", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );
  desc03.insertColumn( "FOLDER_CHANNELTABLENAME", coral::AttributeSpecification::typeNameForId(typeid(std::string)) );

  desc03.setPrimaryKey( "NODE_ID" );
    
  session->nominalSchema().dropIfExistsTable( desc03.name() );
  coral::ITable& table03 = session->nominalSchema().createTable( desc03 );
  coral::AttributeList rowBuffer03;
  table03.dataEditor().rowBuffer( rowBuffer03 );
  
  for(size_t i = 1; i < 10; i++ ){
    rowBuffer03[0].data<long long>() = i;
    rowBuffer03[1].data<long long>() = i * 2;
    rowBuffer03[2].data<std::string>() = "My Node";
    rowBuffer03[3].data<std::string>() = createFullString(94, 32);
    rowBuffer03[4].data<std::string>() = createFullString(94, 32);
    rowBuffer03[5].data<long long>() = 0;
    rowBuffer03[6].data<std::string>() = createFullString(94, 32);
    rowBuffer03[7].data<std::string>() = createFullString(94, 32);
    rowBuffer03[8].data<std::string>() = createFullString(94, 32);
    rowBuffer03[9].data<long long>() = i * 4;
    fillBlob(rowBuffer03[10].data<coral::Blob>(), 5632);
    rowBuffer03[11].data<long long>() = i * 4;
    fillBlob(rowBuffer03[12].data<coral::Blob>(), 5631);
    fillBlob(rowBuffer03[13].data<coral::Blob>(), 5631);
    fillBlob(rowBuffer03[14].data<coral::Blob>(), 5631);
    rowBuffer03[15].data<std::string>() = createFullString(94, 32);
    rowBuffer03[16].data<std::string>() = createFullString(94, 32);
    rowBuffer03[17].data<std::string>() = createFullString(94, 32);
    rowBuffer03[18].data<std::string>() = "MyTable";
    
    table03.dataEditor().insertRow( rowBuffer03 );
  }

  
  session->transaction().commit();
     
  coral::MessageStream::setMsgVerbosity( oldLevel );
   
  sleep(4); // Avoid ORA-01466
  
}

typedef void (*rquery_t)(coral::ISessionProxy * session, const std::vector<std::string>& tables);

void
rquery01(coral::ISessionProxy * session, const std::vector<std::string>& tables)
{
  std::cout << "Start Test01" << std::endl;
  session->transaction().start(true);
  
  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(tables[1] /* T2 */);
  coral::IQuery * query = table.newQuery();

  query->addToOutputList( "s02" );
  query->addToOutputList( "id" );
  
  coral::AttributeList mylist;
  mylist.extend( "s02", coral::AttributeSpecification::typeNameForId(typeid(std::string)));
  mylist.extend( "id", coral::AttributeSpecification::typeNameForId( typeid(long long) ));
  
  query->defineOutput( mylist );
 
  coral::ICursor& cursor = query->execute();
  while(cursor.next()){
    const coral::AttributeList& row = cursor.currentRow();
    std::cout << row[0].data<std::string>() << '\t' << row[1].data<long long>() << std::endl;
  }
    
  delete query;
  session->transaction().commit();
}

void
rquery02(coral::ISessionProxy * session, const std::vector<std::string>& tables)
{
  std::cout << "Start Test02" << std::endl;
  session->transaction().start(true);
  
  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(tables[1] /* T2 */);
  coral::IQuery * query = table.newQuery();

//  coral::MsgLevel oldLevel = coral::MessageStream::msgVerbosity();
//  coral::MessageStream::setMsgVerbosity( coral::Verbose );
  
  query->addToOutputList( "id" );
  query->addToOutputList( "s02" );
  
  coral::AttributeList mylist;
  mylist.extend( "id", coral::AttributeSpecification::typeNameForId( typeid(long long) ));
  mylist.extend( "s02", coral::AttributeSpecification::typeNameForId(typeid(std::string)));
  
  query->defineOutput( mylist );
  
  coral::ICursor& cursor = query->execute();
  while(cursor.next()){
    const coral::AttributeList& row = cursor.currentRow();
    std::cout << row[0].data<long long>() << '\t' << row[1].data<std::string>() << std::endl;
  }
  
//  coral::MessageStream::setMsgVerbosity( oldLevel );
  
  delete query;
  session->transaction().commit();
}

void
rquery03(coral::ISessionProxy * session, const std::vector<std::string>& tables)
{
  std::cout << "Start Test03" << std::endl;
  session->transaction().start(true);
  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(tables[0] /* T2 */);
  coral::IQuery * query = table.newQuery();
//  coral::IQuery * query = schema.newQuery();
  
  ///prepare the query the same as it happens in cool
  
  //add our table T1
//  query->addToTableList( tables[0], "T1" );

  query->addToOutputList( "FOLDER_CHANNELSPEC" );
  query->addToOutputList( "FOLDER_CHANNELTABLENAME" );
  query->addToOutputList( "FOLDER_CHANNEL_EXTREF" );
  query->addToOutputList( "FOLDER_IOV2TAGTABLENAME" );
  query->addToOutputList( "FOLDER_IOVTABLENAME" );
  query->addToOutputList( "FOLDER_PAYLOADSPEC" );
  query->addToOutputList( "FOLDER_PAYLOAD_EXTREF" );
  query->addToOutputList( "FOLDER_PAYLOAD_INLINE" );
  query->addToOutputList( "FOLDER_TAGTABLENAME" );
  query->addToOutputList( "FOLDER_VERSIONING" );
  query->addToOutputList( "LASTMOD_DATE" );
  query->addToOutputList( "NODE_DESCRIPTION" );
  query->addToOutputList( "NODE_FULLPATH" );
  query->addToOutputList( "NODE_ID" );
  query->addToOutputList( "NODE_INSTIME" );
  query->addToOutputList( "NODE_ISLEAF" );  
  query->addToOutputList( "NODE_NAME" );
  query->addToOutputList( "NODE_PARENTID" );
  query->addToOutputList( "NODE_SCHEMA_VERSION" );

  coral::AttributeList attrl;

  ///original order from cool     
  attrl.extend("NODE_ID", typeid( cool::UInt32 ));
  attrl.extend("NODE_PARENTID", typeid(long long) /*typeid( cool::UInt32 )*/);
  attrl.extend("NODE_NAME", typeid(std::string) /*typeid( cool::String255 )*/ );
  attrl.extend("NODE_FULLPATH", typeid( cool::String255 ));
  attrl.extend("NODE_DESCRIPTION", typeid( cool::String255 ));
  attrl.extend("NODE_ISLEAF", typeid( cool::Bool ));
  attrl.extend("NODE_SCHEMA_VERSION", typeid( cool::String255 ));
  attrl.extend("NODE_INSTIME", typeid( cool::String255 ));
  attrl.extend("LASTMOD_DATE", typeid( cool::String255 ));
  attrl.extend("FOLDER_PAYLOADSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_PAYLOAD_INLINE", typeid( cool::UInt16 ));
  attrl.extend("FOLDER_PAYLOAD_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_CHANNELSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_CHANNEL_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_VERSIONING", typeid( cool::Int32 ));
  attrl.extend("FOLDER_IOVTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_IOV2TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_CHANNELTABLENAME", typeid( cool::String255 ));
  
  
  ///sorted order here
  ///crashes with segfault in defineOutput
  /**
  attrl.extend("FOLDER_CHANNELSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_CHANNELTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_CHANNEL_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_IOV2TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_IOVTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_PAYLOADSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_PAYLOAD_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_PAYLOAD_INLINE", typeid( cool::UInt16 ));
  attrl.extend("FOLDER_TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_VERSIONING", typeid( cool::Int32 ));
  attrl.extend("LASTMOD_DATE", typeid( cool::String255 ));
  attrl.extend("NODE_DESCRIPTION", typeid( cool::String255 ));
  attrl.extend("NODE_FULLPATH", typeid( cool::String255 ));
  attrl.extend("NODE_ID", typeid( cool::UInt32 ));
  attrl.extend("NODE_INSTIME", typeid( cool::String255 ));
  attrl.extend("NODE_ISLEAF", typeid( cool::Bool ));
  attrl.extend("NODE_NAME", typeid( cool::String255 ));
  attrl.extend("NODE_PARENTID", typeid( cool::UInt32 ));
  attrl.extend("NODE_SCHEMA_VERSION", typeid( cool::String255 ));
  */  
    
  boost::shared_ptr<coral::AttributeList> dataBuffer;
  
  dataBuffer.reset(new coral::AttributeList(attrl));
  
  query->defineOutput( *dataBuffer );
 
  coral::MsgLevel oldLevel = coral::MessageStream::msgVerbosity();
  coral::MessageStream::setMsgVerbosity( coral::Verbose );
   
  size_t rows = 0;
  
  coral::ICursor& cursor = query->execute();
  while(cursor.next()){
    rows++;
  }

  std::cout << "rows fetched : " << rows << std::endl;
    
  coral::MessageStream::setMsgVerbosity( oldLevel );
  
  delete query;
  session->transaction().commit();
}

void
rquery04(coral::ISessionProxy * session, const std::vector<std::string>& tables)
{
  std::cout << "Start Test03" << std::endl;
  session->transaction().start(true);
  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle(tables[0] /* T2 */);
  coral::IQuery * query = table.newQuery();
//  coral::IQuery * query = schema.newQuery();
  
  ///prepare the query the same as it happens in cool
  
  //add our table T1
//  query->addToTableList( tables[0], "T1" );

  query->addToOutputList( "FOLDER_CHANNELSPEC" );
  query->addToOutputList( "FOLDER_CHANNELTABLENAME" );
  query->addToOutputList( "FOLDER_CHANNEL_EXTREF" );
  query->addToOutputList( "FOLDER_IOV2TAGTABLENAME" );
  query->addToOutputList( "FOLDER_IOVTABLENAME" );
  query->addToOutputList( "FOLDER_PAYLOADSPEC" );
  query->addToOutputList( "FOLDER_PAYLOAD_EXTREF" );
  query->addToOutputList( "FOLDER_PAYLOAD_INLINE" );
  query->addToOutputList( "FOLDER_TAGTABLENAME" );
  query->addToOutputList( "FOLDER_VERSIONING" );
  query->addToOutputList( "LASTMOD_DATE" );
  query->addToOutputList( "NODE_DESCRIPTION" );
  query->addToOutputList( "NODE_FULLPATH" );
  query->addToOutputList( "NODE_ID" );
  query->addToOutputList( "NODE_INSTIME" );
  query->addToOutputList( "NODE_ISLEAF" );  
  query->addToOutputList( "NODE_NAME" );
  query->addToOutputList( "NODE_PARENTID" );
  query->addToOutputList( "NODE_SCHEMA_VERSION" );

  coral::AttributeList attrl;
 
  ///sorted order here
  ///crashes with segfault in defineOutput

  attrl.extend("FOLDER_CHANNELSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_CHANNELTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_CHANNEL_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_IOV2TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_IOVTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_PAYLOADSPEC", typeid( cool::String64k ));
  attrl.extend("FOLDER_PAYLOAD_EXTREF", typeid( cool::String64k ));
  attrl.extend("FOLDER_PAYLOAD_INLINE", typeid( cool::UInt16 ));
  attrl.extend("FOLDER_TAGTABLENAME", typeid( cool::String255 ));
  attrl.extend("FOLDER_VERSIONING", typeid( cool::Int32 ));
  attrl.extend("LASTMOD_DATE", typeid( cool::String255 ));
  attrl.extend("NODE_DESCRIPTION", typeid( cool::String255 ));
  attrl.extend("NODE_FULLPATH", typeid( cool::String255 ));
  attrl.extend("NODE_ID", typeid( cool::UInt32 ));
  attrl.extend("NODE_INSTIME", typeid( cool::String255 ));
  attrl.extend("NODE_ISLEAF", typeid( cool::Bool ));
  attrl.extend("NODE_NAME", typeid( cool::String255 ));
  attrl.extend("NODE_PARENTID", typeid( cool::UInt32 ));
  attrl.extend("NODE_SCHEMA_VERSION", typeid( cool::String255 ));  
    
  boost::shared_ptr<coral::AttributeList> dataBuffer;
  
  dataBuffer.reset(new coral::AttributeList(attrl));
  
  query->defineOutput( *dataBuffer );
 
  coral::MsgLevel oldLevel = coral::MessageStream::msgVerbosity();
  coral::MessageStream::setMsgVerbosity( coral::Verbose );
   
  size_t rows = 0;
  
  coral::ICursor& cursor = query->execute();
  while(cursor.next()){
    rows++;
  }

  std::cout << "rows fetched : " << rows << std::endl;
    
  coral::MessageStream::setMsgVerbosity( oldLevel );
  
  delete query;
  session->transaction().commit();
}

const size_t querym_max = 4;
rquery_t query_matrix[querym_max] = {&rquery02, &rquery02, &rquery03, &rquery04};

void
RandomQuery::execute()
{
//  session->transaction().start(true);
//  size_t rnumber = (rand() % 3);
  size_t rnumber = 2;
  if(rnumber < querym_max) query_matrix[rnumber](session, _tables);
     
//  session->transaction().commit();
}
