#include "TestEnv/SealSwitch.h"

#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/IConnectionService.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IDatabaseServiceSet.h"
#include "RelationalAccess/IDatabaseServiceDescription.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/IMonitoringReporter.h"

namespace coral {

#ifdef CORAL_SEALED

#include "SealKernel/Context.h"

SealSwitcher::SealSwitcher()
  : m_context( new seal::Context )
{

}

SealSwitcher::~SealSwitcher()
{

}

void
SealSwitcher::loadComponent(std::string component)
{
  seal::PluginManager* pm = seal::PluginManager::get();
  pm->initialise();
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_context.get() );

  seal::IHandle<seal::IMessageService> msg =
    m_context->query<seal::IMessageService>( "SEAL/Services/MessageService" );
  if( !msg ){
    loader->load( "SEAL/Services/MessageService" );
    std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
    m_context->query( v_msgSvc );
    if ( ! v_msgSvc.empty() ) {
      seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
      msgSvc->setOutputLevel( seal::Msg::Info);
    }
  }
  
  loader->load( component );
}

coral::IConnectionService&
SealSwitcher::getConnectionService()
{
  seal::IHandle<coral::IConnectionService> handle(m_context->query<coral::IConnectionService>( "CORAL/Services/ConnectionService" ));
  if ( ! handle ) loadComponent("CORAL/Services/ConnectionService");
  handle = m_context->query<coral::IConnectionService>( "CORAL/Services/ConnectionService" );
  if ( ! handle ) throw std::runtime_error( "Could not load the CORAL connection service" );
  return *handle;
}

coral::IRelationalDomain&
SealSwitcher::getDomain(const std::string& connectionString)
{
  std::vector< seal::IHandle<coral::IRelationalService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) throw std::runtime_error( "Could not locate the relational service" );
  seal::IHandle<coral::IRelationalService>& relationalService = v_svc.front();
  return relationalService->domainForConnection( connectionString );
}

const coral::IAuthenticationCredentials&
SealSwitcher::getCredentials(const std::string& connectionString)
{
  seal::IHandle<coral::IAuthenticationService> authsvc = m_context->component( "CORAL/Services/XMLAuthenticationService" );
  if( ! authsvc ) throw std::runtime_error( "Could not locate authentication service!" );
  return authsvc->credentials( connectionString );
}

std::string
SealSwitcher::getConnectionString(const std::string serviceName)
{
  seal::IHandle<coral::ILookupService> lookup = m_context->component("CORAL/Services/XMLLookupService");
  if ( ! lookup ) {
    loadComponent("CORAL/Services/XMLLookupService");
  }
  lookup = m_context->query<coral::ILookupService>("CORAL/Services/XMLLookupService");
  if(! lookup ) throw std::runtime_error( "Could not locate the lookup service" );
  
  coral::IDatabaseServiceSet * serviceSet = lookup->lookup(serviceName);
  if(!(serviceSet->numberOfReplicas() > 0)){
    throw std::runtime_error( "Could not find the Service Name" );
  }
  //extract the connection string
  return serviceSet->replica(0).connectionString();  
}

const coral::IMonitoringReporter&
SealSwitcher::getMonitoringReporter()
{
  std::vector< seal::IHandle<coral::monitor::IMonitoringService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) throw std::runtime_error( "Could not locate the monitoring service" );
  seal::IHandle<coral::monitor::IMonitoringService>& monitoringService = v_svc.front();
  return monitoringService->reporter();
}

seal::Handle<seal::Context>&
SealSwitcher::getSealContext()
{
  return m_context;
}

  
#else

SealSwitcher::SealSwitcher()
{

}

SealSwitcher::~SealSwitcher()
{

}

void
SealSwitcher::loadComponent(std::string component)
{
  coral::Context& context = coral::Context::instance();
  context.loadComponent( component );
}

coral::IConnectionService&
SealSwitcher::getConnectionService()
{
  coral::IHandle<coral::IConnectionService> handle(coral::Context::instance().query<coral::IConnectionService>( "CORAL/Services/ConnectionService" ));
  if( !handle.isValid() ) loadComponent("CORAL/Services/ConnectionService");
  handle = coral::Context::instance().query<coral::IConnectionService>( "CORAL/Services/ConnectionService" );  
  if( !handle.isValid() ) throw std::runtime_error( "Could not load the CORAL connection service" );
  return *(handle.get());
}

coral::IRelationalDomain&
SealSwitcher::getDomain(const std::string& connectionString)
{
  coral::IHandle<coral::IRelationalService> svc = coral::Context::instance().query<coral::IRelationalService>();  
  if ( ! svc.isValid() ) throw std::runtime_error( "Could not locate the relational service" );
  return svc->domainForConnection( connectionString );
}

const coral::IAuthenticationCredentials&
SealSwitcher::getCredentials(const std::string& connectionString)
{
  coral::IHandle<coral::IAuthenticationService>
      authsvc = coral::Context::instance().query<coral::IAuthenticationService>( "CORAL/Services/XMLAuthenticationService" );

  if( ! authsvc.isValid() )
    throw std::runtime_error( "Could not locate authentication service!" );
    
  return authsvc->credentials( connectionString );
}

std::string
SealSwitcher::getConnectionString(const std::string serviceName)
{
  coral::Context& ctx = coral::Context::instance();
  coral::IHandle<coral::ILookupService> lookupService = ctx.query<coral::ILookupService>();
  if ( ! lookupService.isValid() ) {
    ctx.loadComponent("CORAL/Services/XMLLookupService");
  }
  lookupService = ctx.query<coral::ILookupService>();
  if(! lookupService.isValid() ){
    throw std::runtime_error( "Could not locate the lookup service" );
  }
  
  coral::IDatabaseServiceSet * serviceSet = lookupService->lookup(serviceName);
  if(!(serviceSet->numberOfReplicas() > 0)){
    throw std::runtime_error( "Could not find the Service Name" );
  }
  //extract the connection string
  return serviceSet->replica(0).connectionString();  
}

const coral::IMonitoringReporter&
SealSwitcher::getMonitoringReporter()
{
  coral::IHandle<coral::monitor::IMonitoringService>
      monitoringService = coral::Context::instance().query<coral::monitor::IMonitoringService>();
  if( ! monitoringService.isValid() ) throw std::runtime_error( "Could not locate monitoring service!" );

  return monitoringService->reporter();
}

#endif

}
