#include <iostream>
#include <sstream>
#include <stdexcept>
#include <time.h>
#include "PluginManager/PluginManager.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/Context.h"
#include "SealKernel/Service.h"
#include "SealKernel/PropertyManager.h"
#include "SealKernel/IMessageService.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IDatabaseServiceSet.h"
#include "RelationalAccess/IDatabaseServiceDescription.h"


void reportReplica( const coral::IDatabaseServiceDescription& replica,
                    const std::string& logicalName )
{
  std::cout << logicalName << " : "
            << replica.connectionString() << ", "
            << replica.authenticationMechanism() << ", "
            << ( ( replica.accessMode() == coral::ReadOnly ) ? std::string("Read-Only") : std::string("Update") )
            << std::endl;
}


int main( int, char** )
{
	clock_t startClock, endClock;
	double elapsedTime;
 try {
	int numberOfReplicas=0;	
    seal::PluginManager* pm = seal::PluginManager::get();
    pm->initialise();

    seal::Handle< seal::Context > context = new seal::Context();
    seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( context.get() );

    loader->load( "CORAL/Services/XMLLookupService" );

    seal::IHandle<seal::Service> svc = context->query<seal::Service>( "CORAL/Services/XMLLookupService" );

    seal::Handle<seal::PropertyManager>& pmgr = svc->propertyManager();
    seal::PropertyBase* property = pmgr->property( "LookupFile" );
    property->set( std::string( "dblookup.xml" ) );

    std::vector< seal::IHandle<coral::ILookupService> > v_lookupSvc;
    context->query( v_lookupSvc );

    if ( ! v_lookupSvc.empty() )
    {
      const coral::ILookupService& lookupSvc = *( v_lookupSvc.front() );

	std::cout<<"Performs the XMLLookupService test"<<std::endl;

	startClock = clock();
      coral::IDatabaseServiceSet* dbSet = lookupSvc.lookup( "myservice1" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice1" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;


	endClock=clock();
        elapsedTime=((double) (endClock - startClock)) / CLOCKS_PER_SEC;
        std::cout<<"Time taken in micro seconds : "<<elapsedTime<<std::endl;
	std::cout<<"Ends the XMLLookupService test"<<std::endl;

    }
    else {
      throw std::runtime_error("Could not retrieve an lookup service.");
    }
 }
  catch ( std::exception& e ) {
    std::cerr << "Standard C++ exception : " << e.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Exception caught (...)" << std::endl;
    return 1;
  }

  try {

    int numberOfReplicas = 0;
    seal::PluginManager* pm = seal::PluginManager::get();
    pm->initialise();

    seal::Handle< seal::Context > context = new seal::Context();
    seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( context.get() );

// Code for Seal Message Service
   // Load the SEAL Message service and set the output threshold to Debug
   loader->load( "SEAL/Services/MessageService" );
   std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
   context->query( v_msgSvc );
   if ( ! v_msgSvc.empty() ) {
     seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
     msgSvc->setOutputLevel( seal::Msg::Debug);
   }
	
	
// Code for Seal Message Service
    loader->load( "CORAL/Services/LFCLookupService" );

    seal::IHandle<seal::Service> svc = context->query<seal::Service>( "CORAL/Services/LFCLookupService" );

    std::vector< seal::IHandle<coral::ILookupService> > v_lookupSvc;
    context->query( v_lookupSvc );
    if ( ! v_lookupSvc.empty() ) {
    	const coral::ILookupService& lookupSvc = *( v_lookupSvc.front() );

	std::cout<<"Performs the LFCLookupService test"<<std::endl;

	startClock = clock();
       coral::IDatabaseServiceSet* dbSet = lookupSvc.lookup( "myservice1");
       numberOfReplicas = dbSet->numberOfReplicas();

      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice1");
      numberOfReplicas = dbSet->numberOfReplicas();

      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

	endClock=clock();
	elapsedTime=((double) (endClock - startClock)) / CLOCKS_PER_SEC;

        std::cout<<"Time taken in micro seconds : "<<elapsedTime<<std::endl;
	std::cout<<"Ends the LFCLookupService test"<<std::endl;
    }
    else {
      throw std::runtime_error("Could not retrieve an lookup service.");
    }
    std::cout << "Exiting..." << std::endl;
  }
  catch ( std::exception& e ) {
    std::cerr << "Standard C++ exception : " << e.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Exception caught (...)" << std::endl;
    return 1;
  }
  std::cout << "[OVAL] Success" << std::endl;
  return 0;
}
